/*
 ******************************************************************************
 ******************************************************************************
 *****                             PEXtk                                  *****
 *****									  *****
 *****		    An Immediate Mode Interface to PEX			  *****
 *****									  *****
 *****				Example Programs			  *****
 *****									  *****
 *                                                                            *
 *	Copyright 1991, 1993. All Rights Reserved                             *
 *			SHOgraphics					      *
 *			321 Soquel Way           			      *
 *			Sunnyvale, CA      94086  			      *
 *                                                                            *
 *****									  *****

    Permission to use, copy, modify, and distribute this software and
 its documentation for any purpose and without fee is hereby granted,
 provided that the above copyright notice appear in all copies and that
 both that copyright notice and this permission notice appear in
 supporting documentation, and that the name of SHOgraphics not be
 used in advertising or publicity pertaining to distribution of the 
 software without specific, written prior permission.

 SHOGRAPHICS DISCLAIMS ANY AND ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 INCLUDING ALL EXPRESSED AND IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS. IN NO EVENT SHALL SHOGRAPHICS BE LIABLE FOR ANY DAMAGES OF ANY
 KIND, INCLUDING BUT NOT LIMITED TO SPECIAL, INDIRECT OR CONSEQUENTIAL 
 DAMAGES RESULTING FROM LOSS OF USE, DATA OR PROSPECTIVE PROFITS, WHETHER 
 IN AN ACTION OF CONTRACT, NEGLIENCE OR OTHER TORTIOUS ACTION, ARISING OUT 
 OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

 ************************************************************************ 
 *		PEXtk Example Program for index color routines
 */
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <stdio.h>
#include <PEXtk.h>

#include "read_demo_obj.c"

    Fcoord3D	 	ctr;
    int			iter, verbose_flag, ModelType;
    int			pause_flag;
    float		near, far;
    float		left, right, bottom, top, fov, aspect;
    int			win_width, win_height;


main(argc, argv)
int argc; char **argv;
{
    int 		i,j,k,m,n;
    Display 		*dpy;
    Window 		window;
    Fcoord3D	 	vt, vn, vtt[255];
    PCOLOR		colorObj;
    int		 	*infoData;

    ModelType = 0; pause_flag = verbose_flag = 0;
    iter = 100;
    win_width=500; win_height=500;
    left = -5.0; right = 5.0; bottom = -5.0; top = 5.0;
    near = 1.0; far = 1000.0; aspect = 1.0; fov = 90.0;

    parse_line(argc,argv);

    if ((dpy = XOpenDisplay ("")) == NULL) {
	printf(" Error opening display\n");
	return 1;
    }

    if(Pinit(dpy)){
	printf("ERROR... initializing PEX server\n");
	exit(1);
    }

    if( (window = Pcreate_window(dpy, win_width, win_height)) == NULL){
	printf("ERROR... creating window\n");
	exit(1);
    }

    Pset_drawable( window );
    Pzbuffer(1);
    Pbuffer_mode(  PEXtk_DOUBLE_BUFFER );
    Pviewport( 0, win_width-1, 0, win_height-1);
    Pbackface( PEXtk_CULL_BACKFACE);

    Pset_color_mode( PEXtk_COLOR_INDEXED);
    Pset_color_table_rgb( 0, 1.0, 0.0, 0.0);	/* RED */
    Pset_color_table_rgb( 1, 0.0, 1.0, 0.0);	/* GREEN */
    Pset_color_table_rgb( 2, 0.0, 0.0, 1.0);	/* BLUE */
    Pset_color_table_rgb( 3, 1.0, 0.0, 1.0);	/* MAGENTA */
    Pset_color_table_rgb( 4, 1.0, 1.0, 0.0);	/* YELLOW */
    Pset_color_table_rgb( 5, 0.0, 1.0, 1.0);	/* CYAN */

	Set_view();

	Pcolor_ind( 1);
	Pline_color_ind( 1);
	Ptranslate( -ctr.x, -ctr.y, 0.0, 1);

    i = 1; j = 0;
	Create_str();
    while( i < iter ){
	Pclear_rgb( 0.0, 0.0, 0.0);
	Protate_z( 3.0, 1 );
	Pcolor_ind( j);
	if( i % 10 == 0 )	/* cycle through colors */
	    j = (j+1) % 6;
	Pcall_struct( 1);
	Pswap_buffers(); 
	i++;
    }
    if( pause_flag ){
	printf("enter a char ");
	fscanf( stdin, "%c",&i);
	}
      Delete_str();
    Pexit();
    XDestroyWindow( dpy, window);
    XCloseDisplay( dpy);
    exit(0);
}

Set_view()
{
    int i, j;
    switch( ModelType ){
	case 0:
	    Ppersp_view( 0, fov, aspect, near, far);
	    break;
	case 1:
	    Portho_view( 0, left, right, bottom, top, near, far);
	    break;
	case 2:
	    Portho_2d_view( 0, left, right, bottom, top );
	    break;
	case 3:
	    Pwindow_view( 0, left, right, bottom, top, near, far );
	    break;
	default:
	    Ppersp_view( 0, fov, aspect, near, far);
	    break;
    }
}

Create_str()
{
    int i, j, k, m, n;
    Fcoord3D	 vn, vt[255];
	Pcreate_struct(1);
	for(i=0,j=0; i<poly_cnt; i++)
	{
           k = vert_indices[j];
           for( m=0;m<facet_Count[i]; m++){
             n = vert_indices[j++];
             vt[m].x = verts[n].x; vt[m].y = verts[n].y; vt[m].z = verts[n].z;
	   }
	Ppoly_fill( m, vt);
	}
	Pclose_struct();
}

Delete_str()
{
    if(Pdelete_struct(1))
	printf("ERROR... deleting structure 1\n");
}

Exec_str()
{
    Pcall_struct( 1);
}

/*
 *                      +---------------------------+
 * ---------------------+         parse_line        +------------------------
 *                      +---------------------------+
 *  FUNCTION:
 *      Parses the command line.
 *
 *  USAGE:
 *      parse_line(argc, argv,options)
 *	int	   argc 	   (in)  - number of arg's
 *	char   	   argv 	   (in)  - pointer to arg list
 *
 *  DESCRIPTION:
 *	This routine will parse the command line for the following options:
 *	   opt  argument	type	  comment
 *	   ---  ------------    ----      ----------------------------------
 *	    -f filename.obj	string	  .obj file to read
 *	    -i iter		int	  number of iterations
 *	    -p 			void	  sets pause flag
 *	    -v 			void	  sets verbose flag
 *	    -m mode		int	  mode: 0=persp,1=ortho,2=ortho2,
 *					  3=window
 *	    -w size		int	  window size
 *	    -z trans		float	  z translation value
 *	    -h near		float	  near (hither) plane
 *	    -y far		float	  far (yon) plane
 *	    -a aspect		float	  aspect ratio
 *	    -l left		float	  left
 *	    -r right		float	  right
 *	    -t top		float	  top
 *	    -b bottom		float	  bottom
 *                                
*/
int parse_line(argc,argv)
int argc;
char **argv;
{
     char *temp,ffile[32];
     extern char *optarg;
     extern int optind;
     int file_err;
     int c;

    if( argc < 2 ){
	Usage();
	exit(1);
    }

    file_err = 1;

    while((c=getopt(argc,argv,"a:i:h:y:l:r:b:t:m:f:t:w:pvz:")) != EOF)
    {
	switch(c){
	    case 'f':
		  read_demo_obj(optarg, &ctr, 1.0);
		  file_err = 0;
		break;
	    case 'i':
		  sscanf(optarg,"%d",&iter);
		break;
	    case 'm':
		  sscanf(optarg,"%d",&ModelType);
		break;
	    case 'a':
		  sscanf(optarg,"%f",&aspect);
		break;
	    case 'p':
		  pause_flag = 1;
		break;
	    case 'v':
		  verbose_flag = 1;
		break;
	    case 'w':
		  sscanf(optarg,"%d",&win_width);
		  win_height = win_width;
		break;
	    case 'z':
		  sscanf(optarg,"%f",&z_trans_factor);
		break;
	    case 'h':
		  sscanf(optarg,"%f",&near);
		break;
	    case 'y':
		  sscanf(optarg,"%f",&far);
		break;
	    case 'l':
		  sscanf(optarg,"%f",&left);
		break;
	    case 'r':
		  sscanf(optarg,"%f",&right);
		break;
	    case 'b':
		  sscanf(optarg,"%f",&bottom);
		break;
	    case 't':
		  sscanf(optarg,"%f",&top);
		break;
	}
    }

    for(; optind < argc;optind++ ){
		  read_demo_obj(argv[optind], &ctr, 1.0);
		  file_err = 0;
    }
    if( file_err ){
	Usage();
	exit(1);
    }
}

Usage()
{
    printf("Error...  Usage is:\n");
    printf(" index_clr [options] file.obj\n");
    printf("\t -i iter\t(number of iterations)\n");
    printf("\t -m mode {0-3}\t(view type,0=persp,1=ortho)\n");
    printf("\t\t\t(2=ortho2,3=window)\n");
    printf("\t -l left\t(left coord, float)\n");
    printf("\t -r right\t(right coord, float)\n");
    printf("\t -t top\t\t(top coord, float)\n");
    printf("\t -b bottom\t(bottom coord, float)\n");
    printf("\t -p\t\t(pause flag)\n");
    printf("\t -w size\t(set window width and height)\n");
    printf("\t -a aspect\t(aspect ratio, float)\n");
    printf("\t -h hither\t(set hither plane)\n");
    printf("\t -y yon\t\t(set yon plane)\n");
    printf("\t -v\t\t(verbose flag)\n");
    printf("\t -z z_value\n");
    printf("\t -f file.obj\n");
}
