/*
 *  $Header: /auto/nfs/sequoia/u/an/work/rws/pextk/src/pextk.h,v 1.3 1994/01/04 00:41:51 rws Exp $
 ******************************************************************************
 ******************************************************************************
 *****                             PEXtk                                  *****
 *****									  *****
 *****		    An Immediate Mode Interface to PEX			  *****
 *****									  *****
 *****									  *****
 *****    This module contains definitions used internally to pextk       *****
 *****									  *****
 *                                                                            *
 *	Copyright 1992, All Rights Reserved                                   *
 *			SHOgraphics					      *
 *			1890 North Shoreline Blvd			      *
 *			Mountain View, CA    94043			      *
 *                                                                            *
 *   written by: Bob Schulte		date: February 1992                   *
 *                                                                            *
 *****									  *****

    Permission to use, copy, modify, and distribute this software and
 its documentation for any purpose and without fee is hereby granted,
 provided that the above copyright notice appear in all copies and that
 both that copyright notice and this permission notice appear in
 supporting documentation, and that the names of SHOgraphics,
 the X Consortium, and MIT not be used in advertising or publicity
 pertaining to distribution of the software without specific, written
 prior permission.

 SHOGRAPHICS DISCLAIMS ANY AND ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 INCLUDING ALL EXPRESSED AND IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS. IN NO EVENT SHALL SHOGRAPHICS BE LIABLE FOR ANY DAMAGES OF ANY
 KIND, INCLUDING BUT NOT LIMITED TO SPECIAL, INDIRECT OR CONSEQUENTIAL 
 DAMAGES RESULTING FROM LOSS OF USE, DATA OR PROSPECTIVE PROFITS, WHETHER 
 IN AN ACTION OF CONTRACT, NEGLIENCE OR OTHER TORTIOUS ACTION, ARISING OUT 
 OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

 ************************************************************************ */
/*
 *  Module Description
 *	This module contains the structure and type definitions that are 
 *	used locally in PEXtk.
 *
 *  Dependencies
 *	X window ID's are unsigned longs.
 *
 */

/* Need multibuf.h for struct PEXtk_WINDOW */
#include <X11/extensions/multibuf.h>

/* 
 * Size to use for and dls block of data, such as the allocated node
 * pointer blocks, and allocated name blocks.
 */
#define	DLS_SIZE		500
#define	LABEL_SIZE		20
#define	DLS_NODE_BLOCK_SIZE	500
#define	CTABLE_SIZE		512

/* 
 * Number of linked list dpy list nodes to allocate each time
 */
#define	DLS_NODE_LL_SIZE	1024

typedef char	*PTR;

typedef struct {
	float	r,g,b;
	} L_COLOR;

typedef struct {
	int	type;
	float	intensity;
	float	angle;
	float	dirx,diry,dirz;
	float	xpos, ypos, zpos;
	float	conc;
	float	atten1, atten2;
	} LIGHT;

	/*
	 * Note:
	 *	Although each 'drawable' contains this data, some of
	 *	the ID's may be the same (copied from one drawable
	 *	to the next). This structure allows each drawable to
	 *	take on different attributes.
	 */
typedef struct PEXtk_WINDOW {
	U_LONG	w_id;			/* the drawable */
	U_LONG	back_w_id;		/* back w_id for a drawable */
	Multibuffer 	mbuf[2];		/* back mbuf id's */
	int		width;		/* number of pixels wide */
	int		height;		/* number of pixels high */
	int		depth;		/* number of bit planes */
	PEXRenderer	Rdr_ID;		/* renderer id for a w_id */
	PEXLookupTable	View_ID;	/* PEXViewLUT tbl id for a w_id */
	PEXLookupTable	Light_ID;	/* lgt tbl id for a w_id */
	PEXLookupTable	Light_ID0;	/* null lgt id */
	PEXLookupTable	Text_ID;	/* text font id */
	PEXLookupTable	Depthcue_ID;	/* depthcue tbl id */
	unsigned long   dyn_state;	/* mask for renderer dynamics */
	} PEXtk_WINDOW;

typedef struct {
	long	id;
	int	type;
	int	width;
	int	height;
	float	left;		/* window lower left x coord */
	float	right;		/* window lower left y coord */
	float	bottom;		/* window upper right x coord */
	float	top;		/* window upper right y coord */
	float	near;
	float	far;
	float	aspect;
	float	fov;
	} PEXtk_VIEW;

typedef	struct DPY_LIST_NODE {
	int		opcode;		/* The opcode */
	int		name;		/* The dls name */
	struct DPY_LIST_NODE	*next;	/* The next node in list */
	    union {
		char 	c;
		int 	i;
		long 	l;
		float	f;
		PTR	ptr;
		PTR	*pptr;
	    } arg1;
	    union {
		char 	c;
		int 	i;
		long 	l;
		float	f;
		PTR	ptr;
		PTR	*pptr;
	    } arg2;
	    union {
		char 	c;
		int 	i;
		long 	l;
		float	f;
		PTR	ptr;
		PTR	*pptr;
	    } arg3;
	    union {
		char 	c;
		int 	i;
		long 	l;
		float	f;
		PTR	ptr;
		PTR	*pptr;
	    } arg4;
	    union {
		char 	c;
		int 	i;
		long 	l;
		float	f;
		PTR	ptr;
		PTR	*pptr;
	    } arg5;
	    union {
		char 	c;
		int 	i;
		long 	l;
		float	f;
		PTR	ptr;
		PTR	*pptr;
	    } arg6;
	    union {
		char 	c;
		int 	i;
		long 	l;
		float	f;
		PTR	ptr;
		PTR	*pptr;
	    } arg7;
	} DPY_LIST_NODE;

typedef struct PTR_DLS_NODE {
	DPY_LIST_NODE	    *ptr;
	struct PTR_DLS_NODE *next;
	} PTR_DLS_NODE;

typedef	struct PLABEL {
	int		label[LABEL_SIZE];	/* The PEX Label id */
	int		attr[LABEL_SIZE];	/* The PEX InteriorStyle */
	struct PLABEL	*next;		/* The next label structure */
	int		cnt;		/* Number of items in the label list */
	} PLABEL;

	/*
	 * Each 'struct name' has the following data associated with it
	 */
typedef	struct {
	int		id;		/* The PEX structure id */
	int		name;		/* The PEXtk structure name */
	int		empty;		/* True if structure is empty */
	long		attr;		/* Current i_style, edgeFlag, interp */
	PLABEL 		*lbl_list; 	/* Label list */
	PLABEL 		*cur_lbl; 	/* current label list ptr */
	PLABEL 		*last_lbl; 	/* last label list ptr */
	int		lbl_size;	/* total number of label slots used */
	int		lbl_blocks;	/* total number of label blocks */
	} STRUCT;

typedef	struct DLS_NAME {
					/* This structure keeps track of
					 * the dpy list 'names' used, and
					 * their associated starting
					 * dls node pointer
					 */
	int		name[DLS_SIZE];	/* The PEXtk Display List names */
	DPY_LIST_NODE	*node[DLS_SIZE];/* The Display List node ptrs */
					/* associates the start node ptr */
					/* in 'node' for each 'name' */
	struct DLS_NAME *next;		/* The next label structure */
	int		cnt;		/* Number of items in this block */
	} DLS_NAME;

typedef	struct DLS_ALLOC_NODE {
	DPY_LIST_NODE	*node[DLS_SIZE];/* The Display List alloc'd ptrs */
	struct DLS_ALLOC_NODE *next;	/* The next alloc_node block in list */
	int		cnt;		/* Number of items in this block */
	} DLS_ALLOC_NODE;

typedef	struct DPY_POINT3_NODE {
	PEXCoord	pts[3];
	struct DPY_POINT3_NODE	*next;	/* The next node in list */
	} DPY_POINT3_NODE;

typedef	struct PTR_POINT3_NODE {
	DPY_POINT3_NODE	*ptr;
	struct PTR_POINT3_NODE	*next;	/* The next node in list */
	} PTR_POINT3_NODE;

typedef	struct DPY_POINT4_NODE {
	PEXCoord	pts[4];
	struct DPY_POINT4_NODE	*next;	/* The next node in list */
	} DPY_POINT4_NODE;

typedef	struct PTR_POINT4_NODE {
	DPY_POINT4_NODE	*ptr;
	struct PTR_POINT4_NODE	*next;	/* The next node in list */
	} PTR_POINT4_NODE;

typedef struct PCOLOR_LIST {
	int	index;
	PCOLOR  col;
	} PCOLOR_LIST;

typedef	struct CTABLE_LIST {
	PCOLOR_LIST	tbl[CTABLE_SIZE];/* The Color Table values */
	struct CTABLE_LIST *next;	/* The next Color Table block in list*/
	int		cnt;		/* Number of items in this block */
	} CTABLE_LIST;

typedef	struct {
	PEXCoord	pts;	/* points x,y,z */
	} BLD_VERT;

typedef	struct {
	PEXColor	color;	/* color r,g,b */
	} BLD_COLOR;

typedef	struct {
	PEXCoord	norm;	/* normal x,y,z */
	} BLD_NORMAL;

typedef	struct {
	PEXCoord	texture; /* texture x,y,z */
	} BLD_TEXTURE;

typedef struct {
	PEXColor	  colorLine;
	int	  	  colorLineInd;	/* for line_clr_ind */
	PEXColor	  colorPoly;
	int		  colorPolyInd;	/* for poly_clr_ind */
	PEXColor	  colorMrkr;
	int		  colorMrkrInd;	/* for mrkr_clr_ind */
	PEXColor	  colorText;
	int		  colorTextInd;	/* for text_clr_ind */
	PEXReflectionAttributes poly_mtl;	/* PEX Refl. Attrib */
	PEXColor	  mtlColor;
	PEXColor	  colorBg;	/* background color (Pclear) */
	int		  colorBgInd;	/* background color (Pclear_ind) */
	PEXColorSpecifier colorLight;
	int		  colorLightInd;

	int		  edgeFlag;	/* PEX Surface Edge Flag, PEXOn/PEXOff*/
	int		  edgeFlag_str;	/* PEX SurfaceEdgeFlag, struct mode */
	int		  edgeType;	/* PEX Surface Edge Type, Solid,Dash,*/
	int		  border;	/* Border Type */
	int		  backcull;	/* Backcull mode */
	int		  refl_model;	/* PEX Reflection Model */
	int		  line_interp;	/* PEX Polyline Interp Method */
	/*
	 * Note that the interior style and the surface interpolation
	 * method is an 'implicit attribute' because some routines, 
	 * such as poly_fill, poly_shade, have an implicit style of 
	 * Solid, while poly has a style of Hollow; and the Ppoly_shade
	 * routine has an 'implicit attribute' of surface interpolation
	 * color.
	 * Since PEXtk supports the poly_fill_area and poly_index,
	 * which have explicit interior styles, the mixing of the
	 * two can cause problems if the explicit setting routine,
	 * shade_mode, is not used correctly. To account for this,
	 * and immediate mode interior style and an interior style
	 * for the currently opened structure is maintained.
	 * The same is true for the surface interpolation method.
	 */
	int		  interp_meth;	/* PEX Surface Interp Method */
	int		  interp_meth_str;/* PEX Structure Surface Interp */
	int		  i_style;	/* Current Immed mode Interior Style */
	int		  i_style_str;	/* Current Structure Interior Style */
	float		  line_width;	/* Current line width */
	int		  line_type;	/* Current line type */
	float		  mrkr_scale;	/* Current marker scale */
	int		  mrkr_type;	/* Current marker type */
	PEXVector2D	  text_vec;	/* Text string up vector orientation */
	PEXFont		  font;		/* Text font type */
	PEXTextFontEntry  text_entry;	/* Text font table entry */
	int		  path;		/* Text font path direction */
	float		  text_xsize;	/* Text char x size */
	float		  text_ysize;	/* Text char y size */
	float		  texta_xsize;	/* Text Annotate char x size */
	float		  texta_ysize;	/* Text Annotate char y size */

	float		  curv_umin;	/* nurb curve umin */
	float		  curv_umax;	/* nurb curve umax */
	float		  surf_umin;	/* nurb surface umin */
	float		  surf_umax;	/* nurb surface umax */
	float		  surf_vmin;	/* nurb surface vmin */
	float		  surf_vmax;	/* nurb surface vmax */

	int		  u_prec;	/* utility precision in u dir*/
	int		  v_prec;	/* utility precision in v dir*/
	int		  curv_meth;	/* nurb curve approx method */
	int		  curv_utol;	/* nurb curve tolerance */
	int		  surf_meth;	/* nurb surface approx method */
	int		  surf_utol;	/* nurb surface u-dir tolerance */
	int		  surf_vtol;	/* nurb surface v-dir tolerance */

	} PEXtk_ATTRIB;

	/*
	 * PEXtk state data
	 */
typedef struct  PEXTK_STATE {

	int	color_type;	/* curr. color type: 
				  PEXRgbFloatColour, PEXIndexedColour, etc */
	int	color_mode;	/* curr. color mode, Pset_color_mode       */
	int	buffer_mode;	/* buffer mode, double or single buffer    */
				/* dependencies are cur_draw 		   */
	int	shade_mode;	/* bitmask for current shade mode 	   */
	int	mat_mode;	/* matrix mode, model/viewing or proj      */
	int	num_rdr;
	int	shape;		/* shape flag 		*/
	int	req_type;	/* PEX request type     */
	int	maj_ver;	/* PEX major version    */
	int	min_ver;	/* PEX minor version    */
	int	fp_format;	/* Supported PEX floating point format   */
	int	debug_mode;	/* PEXtk_DEBUG flag, true turns on debug */

	PEXPipelineContext  pc_ID;  /* Pipeline context for color approx/ind */
	unsigned long	pc_mask[3]; /* Pipeline context mask bits 	    */
	PEXPCAttributes	pcAttrib; /* Renderer Attributes 		    */

	PEXLookupTable	lut_approx;	/* Lookup table for color approx    */
	PEXLookupTable	lut_index;	/* Lookup table for color indexing  */
	PEXRendererAttributes 	rdrAttrib; /* Renderer Attributes 	    */
	PEXtk_WINDOW	win[PEXtk_MAX_BUFFERS];	/* array of window id's     */
	int	dirty_mask;		/* mask indicating an item is dirty */

	int	sys_mode;		/* Current PEXtk state mode      */
	int	edit_mode;		/* Current PEXtk editing mode    */

	int	error;			/* Current error code 		 */

	int	cur_w;			/* the current window index in w_id */
	int	num_w_id;		/* number of drawables 		 */
	int	upd_act;		/* the mbuf update action 	 */

	long	cur_draw;		/* current rendering window      */	
	long	cur_disp_buf;		/* current display window        */	
					/*   only used with the          */
					/*   Pset_display_buffer routine */
	long	cur_rdr;		/* current renderer 		 */	
	long	cur_id;			/* current id: rdr or struct     */	
	int	cur_view;		/* current viewing id 		 */	

	int	mtx_loc;		/* matrix stack location 	 */
	int	vpt_loc;		/* viewport stack location 	 */
	int	view_loc;		/* viewing stack location 	 */
	int	attr_loc;		/* attribute stack location 	 */
	int	name_loc;		/* name stack location 		 */

	PEXtk_VIEW	*view_stk;	/* pointer to view stack */
	PEXViewEntry    *vue_tbl;     	/* pointer to view table */

	/* pointers used to access items that are pushed/popped 	 */
	PEXtk_ATTRIB	*attr;		/* pointer to current attribute  */
	PEXtk_VIEW	*view;		/* pointer to current view 	 */
	PEXtk_VIEWPORT	*vpt;		/* pointer to current viewport   */
	PEXDeviceRect	*clip;		/* pointer to current clip rect  */
	Matrix		*mat_stk;	/* pointer to matrix stack 	 */
	PEXName 	*names;		/* Name Stack ptr used for picking */
	PEXNameSet 	nameSet;	/* Name Set to be used for picking */
	PEXViewEntry    *vue;     	/* pointer to current view table */

	int	zbuffer;		/* zbuffer mode 		 */
	int	antialias;		/* antialias mode 		 */
	int	backcull;		/* current backcull mode 	 */
	int	dither;			/* True if dither is on 	 */
	int	dither_pattern;		/* dithering pattern to use 	 */
	int	transparency;		/* True if transparency is on 	 */

	int	num_lgts;		/* number of lights in buffer 	 */
	int	active_lgts;		/* currently active light 	 */
	LIGHT	lgt[32];

	/*
	 * The create_struct "name to PEX structure ID" list is
	 * a contiguous list
	 */
	STRUCT	*str_list;		/* pointer to structure id,name */
	int	str_size;		/* maximum number of structures */
	int	str_cnt;		/* current number of structures */
	long	cur_struct;	/* PEX structure id of current open struct */
	int	cur_str_name;	/* 'name' of current open struct */
	int	cur_struct_loc;	/* loc in str_list of current open struct */
	int	cur_label;	/* Current label: make_label, replace_struct */
	int	cur_label_block;/* block in label_list of current label */
	int	cur_label_loc;	/* loc in label_list of current label */
	int	cur_dls_name;		/* name of cur. open dls */

	} PEXTK_STATE;

	/*
	 * PEXtk X data
	 */
typedef struct PEXTK_X {
	Display *dpy;
	GC	gc;
	/* int	win_width, win_height; */
	int	win_screen;
	int	win_depth;
	Visual	*win_visual;
	int	c_class;     /* visual class */
	Colormap cmap;
	U_LONG  red_shift;
	U_LONG  green_shift;
	U_LONG  blue_shift;
	} PEXTK_X;

	/*
	 * PEXtk geometry building data
	 */
typedef struct  PEXTK_BUILD {

	int	bld_mask;		/* state of building process */
					/* pts, colors, normals */
	BLD_NORMAL *bld_polynorm;	/* Normals set with Ppoly_normal */
	int	polynorm_size;		/* size of build polynorm list */
	int	polynorm_cnt;		/* number of items in polynorm list */

	int	bld_mode;	/* current primitive building mode */
	int	bld_style;	/* interior style for polygon/tmesh */

	BLD_VERT *bld_vert;	/* Pvertex vertex list */
	int	vert_size;	/* size of bld vert list */
	int	vert_cnt;	/* number of items in bld list */

	BLD_COLOR *bld_color;	/* Pvertex_color... color list */
	int	color_size;	/* size of bld color list */
	int	color_cnt;	/* number of items in bld color list */

	BLD_NORMAL *bld_normal;	/* Pvertex_normal... normal list */
	int	normal_size;	/* size of bld normal list */
	int	normal_cnt;	/* number of items in bld normal list */

	BLD_TEXTURE *bld_texture;	/* Pvertex_texture... texture list */
	int	texture_size;		/* size of bld texture list */
	int	texture_cnt;		/* number of items in bld tex list */

	char	*buffer;		/* buffer for pts/clr/normal data */
	int	buffer_size;		/* size of buffer */

	PEXCoord cursor_pos;	/* pos of cursor for move/draw */	

	PEXListOfVertex plines; /* PEXPolylineSet buffer for Pbegin_line */
	PEXListOfVertex ppoly;  /* PEXFillAreaWithData buffer for Pbegin_poly*/

	int	pt2d_cnt;	/* size of Ppoly_point_2d list */
	int	pt3d_cnt;	/* size of Ppoly_point_3d list */
	int	nv_cnt;		/* size of Ppoly_point_nv list */
	int	uv_cnt;		/* size of Ppoly_point_uv list */
	int	nvuv_cnt;	/* size of Ppoly_point_nv_uv list */

	double	*cos_buffer_u;		/* buffer circle cosine values */
	double	*sin_buffer_u;		/* buffer circle sine values */
	int	trig_buf_u_size;	/* size of buffer */
	double	*cos_buffer_v;		/* buffer sphere cosine values */
	double	*sin_buffer_v;		/* buffer sphere sine values */
	int	trig_buf_v_size;	/* size of buffer */

} PEXTK_BUILD;

	/*
	 * PEXtk geometry buffer 
	 */
typedef struct  PEXTK_BUFFER {

	PEXListOfVertex *plines;
	int	plines_size;
	int	plines_cnt;

	/* pxlPolygonData *ppoly; */
	/* PEXListOfVertex *ppoly;	/* not used */
	/* PEXFacetData *pfacet;	/* not used */
	int	ppoly_size;
	int	ppoly_cnt;

	unsigned short	*indicies;	/* index buf for build sphere */
	int	indicies_size;		/* size of indicies buf */

	int	cdata_size;		/* size of connectivitydata buf */
	PEXConnectivityData *cdata;
	PEXListOfUShort  *shorts_list;

#ifdef PEX4_0
	short	*vert_index;		/* pointer to index vertices */
	int	index_size;		/* max num verts in a polygon */
	short	*poly_Count;		/* pointer to polygon counts */
	int	poly_Count_size;	/* max num polygon counts */
#endif

	} PEXTK_BUFFER;

	/*
	 * PEXtk Rendering Targets
	 */
typedef struct PEXTK_MRT_OBJECT {
	int		depth;		/* depths 24, 12, or 8   */
	int		type;		/* type of window:	 */
					/*   PEXAnyDrawable      */
					/*   PEXWindowDrawable   */
					/*   PEXPixmapDrawable   */
					/*   PEXBufferDrawable   */
	int		db;		/* double buffering method:	 */
					/*   0 - None			 */
					/*   1 - MBX Double Buffering	 */
					/*   2 - E&S Double Buffering	 */
					/*   3 - Pixmap Double Buffering */
	} PEXTK_MRT_OBJECT;

	/*
	 * PEXtk Display List buffer 
	 */
typedef struct  PEXTK_DLS_OBJECT {
	DPY_LIST_NODE	*cur_dpy_node_ptr;/* current dls node */
	DLS_ALLOC_NODE	*node_ptr;	/* node ptr block (allocated nodes) */
	int		node_block_cnt;	/* current number of ptr blocks */
	DLS_ALLOC_NODE	*cur_node_ptr;	/* currently open node block ptr */

	DLS_NAME	*name_block;	/* base name block */
	int		name_block_cnt;	/* current number of name blocks */
	DLS_NAME	*cur_name_ptr;	/* current name block */

	PEXtk_VIEW      *view_block;	/* view table, alloc'd/realloc'd */
	int             view_block_size;/* current size of block */
	int             view_block_cnt; /* current items in block */
	} PEXTK_DLS_OBJECT;

typedef struct CTABLE_OBJECT {
	CTABLE_LIST	*clr_block;	/* first block in list */
	CTABLE_LIST	*cur_block;	/* current block in list */
	int		block_cnt;	/* number of blocks allocated */
	} CTABLE_OBJECT;

	/*
	 * PEXtk supported data (PEX Implementation Dependant Data
	 *	 		     and Enumerated Type Data.
	 */
typedef struct  PEXTK_SUPPORT {

	int	hlhsr[5];		/* Available Hlhsr modes	*/
	int	line_type[10];		/* Available Line types 	*/
	int	edge_type[10];		/* Available Edge types 	*/
	int	light_type[8];		/* Available Light types	*/
	int	curve_approx[12];	/* Available curve approx types */
	int	surf_approx[12];	/* Available surf approx types  */
	int	allow_zbuffer;		/* True if zbuffer is supported */
	int	allow_antialias;	/* True if antialias is supported    */
	int	allow_dither;		/* True if dither is supported  */
	int	allow_transparency;	/* True if transparency is supported */
	int	dbl_buf;		/* Double Buffer method, one of:     */
					/* 	PEXtk_DBUF_PIXMAP 	*/
					/* 	PEXtk_DBUF_MBUF 	*/
					/* 	PEXtk_DBUF_ES_ESCAPE 	*/
	int	max_lgts;		/* max num of lights supported  */
	int	max_views;		/* max num of views supported 	*/
	int	max_mrkr_sz;		/* max size of marker supported */
	int	max_line_sz;		/* max width of lines supported */
	int	max_names;		/* max number of namesets supported */
	int	gdp_sphere;		/* True if GDP Sphere is supported  */

	} PEXTK_SUPPORT;

typedef struct PEXTK_OBJECT {
	PEXTK_STATE	*state;
	PEXTK_BUILD	*build;
	PEXTK_SUPPORT	*support;
	PEXTK_BUFFER	*buffer;
	PEXTK_DLS_OBJECT *dpy_list;
	CTABLE_OBJECT	*clr_tbl;
	PEXTK_X		*x_data;

	int	(*clear_buffer)();	/* user settable clear buffer */
	int	(*create_buffer)();	/* user settable create buffer */
	int	(*delete_buffer)();	/* user settable delete buffer */
	int	(*swap_buffer)();	/* user settable swap buffer */
	int	(*clear_zbuffer)();	/* user settable clear zbuffer */
	int	(*antialias)();		/* user settable antialias routine */
	int	(*dither)();		/* user settable dither routine */
	int	(*transparency)();	/* user settable transparency routine*/
	int	(*push_clear_buffer)(); /* saved clear buffer func ptr */
	int	(*mpex_reset_state)();
	int	(*surface_color)();
	int	(*surface_color_ind)();
	int	(*pixel_color_ind)();	/* converts index to pixel value */
	int	(*pixel_color_rgb)();   /* converts rgb (3 floats) to pixel */
	int	(*concat)();
	int	(*preconcat)();
	} PEXTK_OBJECT;

#include "pextk_protos.h"

/* PEXtk macro definitions */

#define fpixelvalue(r,g,b) (unsigned long) (((unsigned long)((r) * 255)<<16) \
  | ((unsigned long)((g) * 255)<< 8) | (unsigned long)((b) * 255))
#define ipixelvalue(r,g,b) (unsigned long) (((unsigned long)(r )<<16) \
  | ((unsigned long)(g)<< 8) | (unsigned long)((b)))

/* This macro is to hide the conversion of a PEXtk defined 'preconcat' to
 * a preconcat used by PEX. Since PEXtk considers the points *column* array
 * to be postmultiplied by the transformations, and PEX considers the 
 * points *row* array to be premultiplied by the transformations, the use
 * of postconcatenation and preconcatenation must be reversed.
 */
#define POSTFLAG_IS_TRUE(flag)	(flag == 0 ? 1 : 0)

#define IS_LABEL_PENDING( pst )  \
			if( pst->label_flag == 1 ){ \
			    DB_update_label_offset(); \
			    pst->label_flag = 0; \
			}

#define IMMEDIATE_MODE(flag)	(((flag) & 0x01) == 0)
#define STRUCTURE_MODE(flag)	(((flag) & 0x01))

#define	PEXtk_DEFAULT_STRUCT_SIZE	100

/* 
 * Default number of vertices in a polygon buffer
 */
#define	PEXtk_POLY_SIZE			512

/* Possible PEXtk states: sys_mode. low bit clear implies immediate mode */

#define PEXtk_RENDERING_MODE	0
#define PEXtk_STRUCTURE_MODE	1
#define PEXtk_PICKING_MODE	2
#define PEXtk_DLS_MODE		3

/* Possible PEXtk edit modes: edit_mode */

#define PEXtk_EDITING_MODE	4
#define PEXtk_INSERT_MODE	5
#define PEXtk_REPLACE_MODE	6

/* Possible PEXtk double buffering methods */

#define PEXtk_DBUF_PIXMAP	0
#define PEXtk_DBUF_MBUF		1
#define PEXtk_DBUF_ES_ESCAPE	2

#define PEXtk_BLD_NONE		0
#define PEXtk_BLD_POLY		1
#define PEXtk_BLD_LINE		2
#define PEXtk_BLD_MESH		3
#define PEXtk_BLD_POINT		4

#define PEXtk_BLD_SET_VERT	0x0001
#define PEXtk_BLD_SET_COLOR	0x0002
#define PEXtk_BLD_SET_NORM	0x0004
#define PEXtk_BLD_SET_TEXTURE	0x0008
#define PEXtk_BLD_SET_TEXTURE3	0x0010
#define PEXtk_BLD_PNT_2D	0x0020
#define PEXtk_BLD_PNT_3D	0x0040
#define PEXtk_BLD_PNT_NV	0x0080
#define PEXtk_BLD_PNT_UV	0x0100
#define PEXtk_BLD_PNT_NV_UV	0x0200

#define	SURFACE_AMBIENT	0.5
#define	SURFACE_DIFFUSE	0.7
#define	SURFACE_SPEC   	0.8
#define	SURFACE_EXP    	.0
#define ATTENUATION1	.70
#define ATTENUATION2	.00

#define PEXtk_DIRTY_GEOM 	0x0001
#define PEXtk_DIRTY_VPT 	0x0002
#define PEXtk_DIRTY_VIEW 	0x0004
#define PEXtk_DIRTY_MTX 	0x0008
#define PEXtk_DIRTY_COLOR 	0x0010
#define PEXtk_DIRTY_REFL_AMB	0x0020
#define PEXtk_DIRTY_REFL_DIF	0x0040
#define PEXtk_DIRTY_REFL_SPEC	0x0080
#define PEXtk_DIRTY_REFL_EXP	0x0100
#define PEXtk_DIRTY_SHADE	0x0200
#define PEXtk_DIRTY_MRKR_TYPE	0x0400
#define PEXtk_DIRTY_PNT 	0x0800

#define PEXtk_DLS_EMPTY_NAME	-1
#define PEXtk_STR_EMPTY_NAME	-1
#define PEXtk_STR_EMPTY_LABEL	-1
