/*
 *  $Header: /auto/nfs/sequoia/u/an/work/rws/pextk/src/pextk_bld.c,v 1.2 1994/01/04 18:45:45 rws Exp $
 ******************************************************************************
 ******************************************************************************
 *****                          pextk_bld.c                               *****
 *****									  *****
 *****									  *****
 *****									  *****
 *                                                                            *
 *	Copyright 1992, 1993. All Rights Reserved                             *
 *			SHOgraphics					      *
 *			321 Soquel Way           			      *
 *			Sunnyvale, CA     94086   			      *
 *                                                                            *
 *****									  *****

    Permission to use, copy, modify, and distribute this software and
 its documentation for any purpose and without fee is hereby granted,
 provided that the above copyright notice appear in all copies and that
 both that copyright notice and this permission notice appear in
 supporting documentation, and that the name of SHOgraphics not be
 used in advertising or publicity pertaining to distribution of the 
 software without specific, written prior permission.

 SHOGRAPHICS DISCLAIMS ANY AND ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 INCLUDING ALL EXPRESSED AND IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS. IN NO EVENT SHALL SHOGRAPHICS BE LIABLE FOR ANY DAMAGES OF ANY
 KIND, INCLUDING BUT NOT LIMITED TO SPECIAL, INDIRECT OR CONSEQUENTIAL 
 DAMAGES RESULTING FROM LOSS OF USE, DATA OR PROSPECTIVE PROFITS, WHETHER 
 IN AN ACTION OF CONTRACT, NEGLIENCE OR OTHER TORTIOUS ACTION, ARISING OUT 
 OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

 ************************************************************************ */
/*
 *
 *  Module Purpose:
 *	This is the module which contains the PEXtk line, polygon and
 *	triangle mesh building routines.
 *
 *  Module Notes:
 *	The Ppoly_normal for triangle mesh is not yet supported.
 *
 *  Module Contents:
 *
 *    int   PEXtk_init_bld    Initializes the required pointers
 *    void  Pbegin_line       Begins polyline building mode.
 *    void  Pbegin_poly       Begins polygon building mode.
 *    void  Pbegin_tmesh      Begins triangle mesh building mode.
 *    int   Pend_line         Ends polyline building mode.
 *    int   Pend_poly         Ends polygon building mode.
 *    int   Pend_tmesh        Ends triangle mesh building mode.
 *    void  Pdraw_2d          Draws a 2D line to a point x,y.
 *    void  Pdraw_3d          Draws a 3D line to a point x,y,z.
 *    void  Pmove_2d          Moves position to 2D point x,y.
 *    void  Pmove_3d          Moves position to a 3D point x,y,z.
 *    void  Ppoint_2d         Draws a 2D point.
 *    void  Ppoint_3d         Draws a 3D point.
 *    int   Ppoly_close       Closes Ppoly_point... polygon building mode.
 *    void  Ppoly_normal      Sets a polygon normal.
 *    void  Ppoly_point_2d    Puts a 2D polygon vertex in geometry building b
 *    void  Ppoly_point_3d    Puts a 3D polygon vertex in geometry building b
 *    void  Ppoly_point_nv    Puts a 3D polygon vertex and vertex normal in g
 *    void  Ppoly_point_uv    Puts a 3D polygon vertex and texture vertex in 
 *    void  Ppoly_point_nv_uv Puts a 3D polygon vertex, vertex normal, and te
 *    void  Pswap_tmesh       Swaps the last two 3D verticies in geom buffer.
 *    void  Pvertex           Puts a 3D vertex in geometry building buffer.
 *    void  Pvertex_color     Puts a vertex color in geometry building buf
 *    void  Pvertex_color_hsv Puts a vertex color hsv in buffer
 *    void  Pvertex_color_ind Puts a vertex color ind in buffer
 *    void  Pvertex_color_hsv Puts a vertex color rgb in buffer
 *    void  Pvertex_color_array     Puts a vertex color array in buffer
 *    void  Pvertex_color_array_hsv Puts a vertex color hsv array in buffer
 *    void  Pvertex_color_array_ind Puts a vertex color index array in buffer
 *    void  Pvertex_color_array_rgb Puts a vertex color rgb array in buffer
 *    void  Pvertex_normal    Puts a 3D vertex normal in geometry buffer
 *    void  Pvertex_normals   Puts 'n' 3D vertex normals in geometry buffer
 *    void  Pvertex_texture   Puts a 2D texture vertex in geometry buffer
 *    void  Pvertex_texture   Puts 'n' 2D texture vertex in geometry buffer
 *
 *  Module Description:
 *	This module handles the primitive building capabilities of PEXtk.
 *
 */

#include <stdio.h>
#include <math.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <X11/PEX5/PEX.h>
#include <X11/PEX5/PEXproto.h>
#include <X11/PEX5/PEXlib.h>

#ifdef MPEX
#include <X11/MPEX.h>
#include <X11/MPEXproto.h>
#include <X11/MPEXlib.h>
#endif

#include <PEXtk.h>

#include "pextk.h"
#include "pextk_lcl.h"

/*
 *
 *                       +--------------------------+
 * ----------------------+      PEXtk_init_bld      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Initializes primitive building pointers.
 *
 *  USAGE:
 *	PEXtk_init_bld( p_obj )
 *	PEXtk_OBJECT    *p_obj	(in)	Pointer to PEXtk object.
 *
 *  DESCRIPTION:
 *	This routine initializes pointers for p_build and p_state.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_init_bld(p_obj)
PEXTK_OBJECT		*p_obj;
{
    /*
     * Save the various object ptrs
     */
    p__object	= p_obj;
    p_state	= p_obj->state;
    p_build	= p_obj->build;
    p_buffer	= p_obj->buffer;
    p_support	= p_obj->support;
    p_dl	= p_obj->dpy_list;
    p_clr	= p_obj->clr_tbl;
    p_x_data	= p_obj->x_data;

    dpy = p_x_data->dpy;

    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+       Pbegin_line        +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Begins polyline building mode.
 *
 *  USAGE:
 *	Pbegin_line( )
 *
 *  DESCRIPTION:
 *	The begin_line mode is initiated by initializing all counts to zero
 *	and by setting the build mode (p_build->bld_mode) to PEXtk_BLD_LINE.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pbegin_line( )
{
	p_build->bld_mode    = PEXtk_BLD_LINE;
	p_build->vert_cnt    = 0;
	p_build->color_cnt   = 0;
	p_build->normal_cnt  = 0;
	p_build->texture_cnt = 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+       Pbegin_poly        +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Begins polygon building mode.
 *
 *  USAGE:
 *	Pbegin_poly( style )
 *      int	    style	(in)	fill style of poly:
 *					    PEXtk_FILL_EMPTY
 *					    PEXtk_FILL_SOLID
 *
 *  DESCRIPTION:
 *	The begin_poly mode is initiated by initializing all counts to zero
 *	and by setting the build mode (p_build->bld_mode) to PEXtk_BLD_POLY.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pbegin_poly( style )
int	style;
{
	p_build->bld_mode    = PEXtk_BLD_POLY;
	p_build->bld_style   = style;
	p_build->vert_cnt    = 0;
	p_build->color_cnt   = 0;
	p_build->normal_cnt  = 0;
	p_build->texture_cnt = 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+       Pbegin_tmesh       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Begins triangle mesh building mode.
 *
 *  USAGE:
 *	Pbegin_tmesh( style )
 *      int	    style	(in)	fill style of poly:
 *					    PEXtk_FILL_EMPTY
 *					    PEXtk_FILL_SOLID
 *
 *  DESCRIPTION:
 *	The begin_tmesh mode is initiated by initializing all counts to zero
 *	and by setting the build mode (p_build->bld_mode) to PEXtk_BLD_MESH.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pbegin_tmesh( style )
int	style;
{
	p_build->bld_mode    = PEXtk_BLD_MESH;
	p_build->bld_style   = style;
	p_build->vert_cnt    = 0;
	p_build->color_cnt   = 0;
	p_build->normal_cnt  = 0;
	p_build->texture_cnt = 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+         Pdraw_2d         +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Draws a 2D line to a point x,y.
 *
 *  USAGE:
 *	Pdraw_2d( x, y )
 *	Fcoord	    x		(in)	The x position to draw to
 *	Fcoord	    y		(in)	The y position to draw to
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pdraw_2d( x, y )
Fcoord	x;
Fcoord	y;
{
    float points[2][3];

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);

	points[0][0] = p_build->cursor_pos.x;
	points[0][1] = p_build->cursor_pos.y;
	points[0][2] = p_build->cursor_pos.z;

	points[1][0] = x;
	points[1][1] = y;
	points[1][2] = 0.0;

	PEXPolyline( dpy, p_state->cur_id, p_state->req_type, 
		2, (PEXCoord *) points);

	p_build->cursor_pos.x = x;
	p_build->cursor_pos.y = y;
	p_build->cursor_pos.z = 0.0;
}
 
/*
 *
 *                       +--------------------------+
 * ----------------------+         Pdraw_3d         +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Draws a 3D line to a point x,y,z.
 *
 *  USAGE:
 *	Pdraw_3d( x, y, z )
 *	Fcoord	    x		(in)	The x position to draw to
 *	Fcoord	    y		(in)	The y position to draw to
 *	Fcoord	    z		(in)	The z position to draw to
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pdraw_3d( x, y, z )
Fcoord	x;
Fcoord	y;
Fcoord	z;
{
    float points[2][3];

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);

	points[0][0] = p_build->cursor_pos.x;
	points[0][1] = p_build->cursor_pos.y;
	points[0][2] = p_build->cursor_pos.z;

	points[1][0] = x;
	points[1][1] = y;
	points[1][2] = z;

	PEXPolyline( dpy, p_state->cur_id, p_state->req_type, 
		2, (PEXCoord *) points);

	p_build->cursor_pos.x = x;
	p_build->cursor_pos.y = y;
	p_build->cursor_pos.z = z;
}
 
/*
 *
 *                       +--------------------------+
 * ----------------------+        Pend_line         +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Ends polyline building mode.
 *
 *  USAGE:
 *	Pend_line( )
 *
 *  RETURNS:
 *	0 if successful,  non-zero if an error occurs.
 *			  1 - memory alloc error
 *			  2 - in poly or tmesh mode
 *			  3 - incorrect counts
 */
 
/*ARGSUSED*/
int 
Pend_line( )
{
    int i;
    char *ptr;

    if( p_build->bld_mode != PEXtk_BLD_LINE ){
	if( p_state->debug_mode )
          fprintf(stderr,
	     "ERROR... Pend_line called when in %s mode\n",
	    (p_build->bld_mode == PEXtk_BLD_POLY) ? "Pbegin_poly" :
	    (p_build->bld_mode == PEXtk_BLD_MESH) ? "Pbegin_tmesh" :
	    "no Pbegin...");
	DB_init_bld_counts();
	return 2;
    }
	
    if(DB_check_build_mask()){
	DB_init_bld_counts();
	return 3;
    }

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
    if( p_build->color_cnt > 0 )
    {
	    i = p_build->vert_cnt * 2 *
		  sizeof(BLD_VERT)*sizeof(BLD_COLOR);
	    if( DB_check_build_buffer_size(i))
		return 1;	/* memory alloc error */
	    p_build->plines.count = p_build->vert_cnt;
	    p_build->plines.vertices.no_data =(PEXCoord *)p_build->buffer;
	    ptr = (char *) p_build->buffer;

	    for(i=0;i<p_build->vert_cnt;i++){
	      memcpy(ptr, (char*) &p_build->bld_vert[i],sizeof(BLD_VERT));
	      ptr += sizeof(BLD_VERT);
	      memcpy(ptr, (char*) &p_build->bld_color[i],sizeof(BLD_COLOR));
	      ptr += sizeof(BLD_COLOR);
	    }
	PEXPolylineSetWithData( dpy, p_state->cur_id, p_state->req_type, 
		PEXGAColor, p_state->color_type, 1, &p_build->plines);

    } else {
	PEXPolyline( dpy, p_state->cur_id, p_state->req_type, 
		p_build->vert_cnt, (PEXCoord *) p_build->bld_vert);
    }
	DB_init_bld_counts();

    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+        Pend_poly         +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Ends polygon building mode.
 *
 *  USAGE:
 *	Pend_poly( )
 *
 *  RETURNS:
 *	0 if successful,  non-zero if an error occurs.
 *			  1 - memory alloc error
 *			  2 - in line or tmesh mode
 *			  3 - incorrect counts
 */
 
/*ARGSUSED*/
int
Pend_poly( )
{
    int err;

    if( p_build->bld_mode != PEXtk_BLD_POLY ){
	if( p_state->debug_mode )
          fprintf(stderr,
	     "ERROR... Pend_poly called when in %s mode\n",
	    (p_build->bld_mode == PEXtk_BLD_LINE) ? "Pbegin_line" :
	    (p_build->bld_mode == PEXtk_BLD_MESH) ? "Pbegin_tmesh" :
	    "no Pbegin...");
	DB_init_bld_counts();
	return 2;
    }
	
    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);

	PEXtk_check_build_attrib();

	err = PEXtk_build_primitive( PEXtk_BLD_POLY );

	DB_init_bld_counts();

    return err;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+        Pend_tmesh        +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Ends triangle mesh building mode.
 *
 *  USAGE:
 *	Pend_tmesh( )
 *
 *  RETURNS:
 *	0 if successful,  non-zero if an error occurs.
 *			  1 - memory alloc error
 *			  2 - in line or poly mode
 *			  3 - incorrect counts
 */
 
/*ARGSUSED*/
int
Pend_tmesh( )
{
    int err;

    /* Do what you need to here */
    if( p_build->bld_mode != PEXtk_BLD_MESH ){
	if( p_state->debug_mode )
          fprintf(stderr,
	     "ERROR... Pend_tmesh called when in %s mode\n",
	    (p_build->bld_mode == PEXtk_BLD_POLY) ? "Pbegin_poly" :
	    (p_build->bld_mode == PEXtk_BLD_LINE) ? "Pbegin_line" :
	    "no Pbegin...");
	DB_init_bld_counts();
	return 2;
    }
	
    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);

	PEXtk_check_build_attrib();

	err = PEXtk_build_primitive( PEXtk_BLD_MESH );
	DB_init_bld_counts();

    return err;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+         Pmove_2d         +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Moves position to 2D point x,y.
 *
 *  USAGE:
 *	Pmove_2d( x, y )
 *	Fcoord	    x		(in)	The x position to move to
 *	Fcoord	    y		(in)	The y position to move to
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pmove_2d( x, y )
Fcoord	x;
Fcoord	y;
{
	p_build->cursor_pos.x = x;
	p_build->cursor_pos.y = y;
	p_build->cursor_pos.z = 0.0;
	return;
}
 
/*
 *
 *                       +--------------------------+
 * ----------------------+         Pmove_3d         +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Moves position to a 3D point x,y,z.
 *
 *  USAGE:
 *	Pmove_3d( x, y, z )
 *	Fcoord	    x		(in)	The x position to move to
 *	Fcoord	    y		(in)	The y position to move to
 *	Fcoord	    z		(in)	The z position to move to
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pmove_3d( x, y, z )
Fcoord	x;
Fcoord	y;
Fcoord	z;
{
	p_build->cursor_pos.x = x;
	p_build->cursor_pos.y = y;
	p_build->cursor_pos.z = z;
	return;
}
 
/*
 *
 *                       +--------------------------+
 * ----------------------+        Ppoint_2d         +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Draws a 2D point.
 *
 *  USAGE:
 *	Ppoint_2d( x, y )
 *	Fcoord	    x		(in)	The x position of the point
 *	Fcoord	    y		(in)	The y position of the point
 *
 *  NOTE:
 *	The point routines use markers to draw the point, so there is a
 *	conflict with the point (poly) color, and the marker color. A
 *	mechanism is needed to efficiently handle the color state.
 *	Because of this fact, the Ppoint routines will use the current
 *	marker color.
 *
 *	The same is true for the marker type state, since marker type dot
 *	is needed, it has to be the current type, but the marker type
 *	state has to be restored for any subsequent marker calls. This
 *	has different effects for immediate and structure mode. PEXtk
 *	will automatically set and reset the marker type.
 *
 *	To avoid these conflicts, use the marker routine with the type
 *	set to PEXtk_MRKR_POINT, and set the marker color accordingly.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ppoint_2d( x, y )
Fcoord	x;
Fcoord	y;
{
    float points[3];

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);

	points[0] = x;
	points[1] = y;
	points[2] = 0.0;

	if( p_state->attr->mrkr_type !=  PEXtk_MRKR_POINT ){
		PEXSetMarkerType( dpy, p_state->cur_id, p_state->req_type, 
			PEXMarkerDot);
		PEXMarkers( dpy, p_state->cur_id, p_state->req_type, 
			1, (PEXCoord *) points);
		PEXSetMarkerType( dpy, p_state->cur_id, p_state->req_type, 
			p_state->pcAttrib.marker_type);
		/* p_state->dirty_mask |= PEXtk_DIRTY_MRKR_TYPE; */
		/* p_state->dirty_mask |= PEXtk_DIRTY_COLOR; */
	} else {
		PEXMarkers( dpy, p_state->cur_id, p_state->req_type, 
			1, (PEXCoord *) points);
		/* p_state->dirty_mask |= PEXtk_DIRTY_COLOR; */
	}
	p_build->cursor_pos.x = x;
	p_build->cursor_pos.y = y;
	p_build->cursor_pos.z = 0.0;

	return;
}
 
/*
 *
 *                       +--------------------------+
 * ----------------------+        Ppoint_3d         +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Draws a 3D point.
 *
 *  USAGE:
 *	Ppoint_3d( x, y, z )
 *	Fcoord	    x		(in)	The x position of the point
 *	Fcoord	    y		(in)	The y position of the point
 *	Fcoord	    z		(in)	The z position of the point
 *
 *  NOTE:
 *	The point routines use markers to draw the point, so there is a
 *	conflict with the point (poly) color, and the marker color. A
 *	mechanism is needed to efficiently handle the color state.
 *	Because of this fact, the Ppoint routines will use the current
 *	marker color.
 *
 *	The same is true for the marker type state, since marker type dot
 *	is needed, it has to be the current type, but the marker type
 *	state has to be restored for any subsequent marker calls. This
 *	has different effects for immediate and structure mode. PEXtk
 *	will automatically set and reset the marker type.
 *
 *	To avoid these conflicts, use the marker routine with the type
 *	set to PEXtk_MRKR_POINT, and set the marker color accordingly.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ppoint_3d( x, y, z )
Fcoord	x;
Fcoord	y;
Fcoord	z;
{
    float points[3];

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);

	points[0] = x;
	points[1] = y;
	points[2] = z;

	if( p_state->attr->mrkr_type !=  PEXtk_MRKR_POINT ){
	    /* if( p_state->dirty_mask & PEXtk_DIRTY_MRKR_TYPE ){ */
		PEXSetMarkerType( dpy, p_state->cur_id, p_state->req_type, 
			PEXMarkerDot);
		PEXSetMarkerColor( dpy, p_state->cur_id, p_state->req_type, 
			p_state->color_type, &p_state->attr->colorPoly);
		PEXMarkers( dpy, p_state->cur_id, p_state->req_type, 
			1, (PEXCoord *) points);
		PEXSetMarkerType(dpy, p_state->cur_id, p_state->req_type, 
			p_state->pcAttrib.marker_type);
		PEXSetMarkerColor( dpy, p_state->cur_id, p_state->req_type, 
			p_state->color_type, &p_state->attr->colorMrkr);
		/* p_state->dirty_mask |= PEXtk_DIRTY_MRKR_TYPE; */
		/* p_state->dirty_mask |= PEXtk_DIRTY_COLOR; */
	    /* } */
	} else {
		PEXSetMarkerColor( dpy, p_state->cur_id, p_state->req_type, 
			p_state->color_type, &p_state->attr->colorPoly);
		PEXMarkers( dpy, p_state->cur_id, p_state->req_type, 
			1, (PEXCoord *) points);
		PEXSetMarkerColor( dpy, p_state->cur_id, p_state->req_type, 
			p_state->color_type, &p_state->attr->colorMrkr);
		/* p_state->dirty_mask |= PEXtk_DIRTY_COLOR; */
	}
	p_build->cursor_pos.x = x;
	p_build->cursor_pos.y = y;
	p_build->cursor_pos.z = z;

	return;
}
 
/*
 *
 *                       +--------------------------+
 * ----------------------+       Ppoly_close        +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Closes Ppoly_point... polygon building mode.
 *
 *  USAGE:
 *	Ppoly_close( )
 *
 *  DESCRIPTION:
 *	When this routine is called,the current state is checked to
 *	make sure no other building routines have been invoked.
 *	The vertex counts for each Ppoly_point type is checked
 *	for consistency.
 *
 *  RETURNS:
 *	0 if successful,  non-zero if an error occurs.
 *			  1 - memory alloc error
 *			  2 - in poly, line or tmesh mode
 *			  3 - incorrect counts
 */
 
/*ARGSUSED*/
int
Ppoly_close( )
{
    int i;
    long fa, va;
    char *ptr, *pv, *p_bld, *p_norm;

    if((p_build->bld_mode == PEXtk_BLD_POLY ) ||
      ( p_build->bld_mode == PEXtk_BLD_LINE ) ||
      ( p_build->bld_mode == PEXtk_BLD_MESH )){
	if( p_state->debug_mode )
          fprintf(stderr,
	     "ERROR... Ppoly_close called when in %s mode\n",
	    (p_build->bld_mode == PEXtk_BLD_POLY) ? "Pbegin_poly" :
	    (p_build->bld_mode == PEXtk_BLD_LINE) ? "Pbegin_line" :
	    (p_build->bld_mode == PEXtk_BLD_MESH) ? "Pbegin_tmesh" :
	    "no Pbegin...");
	DB_init_bld_counts();
	return 2;
    }
	
    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
    i = (p_build->color_cnt) ? PEXSurfaceInterpColor : 0;
    PEXtk_check_poly_attrib(  PEXInteriorStyleSolid, i);

    switch( p_build->bld_mask ){

      case PEXtk_BLD_PNT_2D:
      case PEXtk_BLD_PNT_3D:
      case PEXtk_BLD_PNT_UV:

	    PEXFillArea( dpy, p_state->cur_id, p_state->req_type, 
			p_state->shape, False, 
			p_build->vert_cnt, (PEXCoord *) p_build->bld_vert);
	  break;
      case PEXtk_BLD_PNT_2D|PEXtk_BLD_SET_COLOR:
      case PEXtk_BLD_PNT_3D|PEXtk_BLD_SET_COLOR:
      case PEXtk_BLD_PNT_UV|PEXtk_BLD_SET_COLOR:
	    if(DB_check_build_mask()){
		DB_init_bld_counts();
		return 3;
	    }

	    i = p_build->vert_cnt * 
		  (sizeof(BLD_VERT)+sizeof(BLD_COLOR));
	    if( DB_check_build_buffer_size(i))
		return 1;	/* memory alloc error */

	    fa = 0;
	    va = PEXGAColor;
	    p_build->ppoly.count = p_build->vert_cnt;  /* n verts */
	    p_build->ppoly.vertices.no_data = (PEXCoord *)p_build->buffer;
	    ptr = (char *) p_build->buffer;
	    pv = (char *) p_build->bld_vert;
	    p_bld = (char *) p_build->bld_color;

	    for(i=0;i<p_build->vert_cnt;i++){
	      memcpy( ptr, (char *) pv, sizeof(BLD_VERT));
	      ptr += sizeof(BLD_VERT); pv += sizeof(BLD_VERT);
	      memcpy( ptr, (char *) p_bld, sizeof(BLD_COLOR));
	      ptr += sizeof(BLD_COLOR); p_bld += sizeof(BLD_COLOR);
	    }
	    if( p_build->polynorm_cnt ){
		fa = PEXGANormal;
	    } 
	    PEXFillAreaSetWithData( dpy, p_state->cur_id, p_state->req_type, 
		    p_state->shape, False, PEXContourDisjoint, fa, va, 
		    p_state->color_type, 1,
		    (PEXFacetData *)&p_build->bld_polynorm[0], &p_build->ppoly);
	  break;
      case PEXtk_BLD_PNT_NV:
      case PEXtk_BLD_PNT_NV | PEXtk_BLD_SET_COLOR:
      case PEXtk_BLD_PNT_NV_UV:
      case PEXtk_BLD_PNT_NV_UV | PEXtk_BLD_SET_COLOR:
	    if(DB_check_build_mask()){
		DB_init_bld_counts();
		return 3;
	    }

	    fa = 0;
	    va = PEXGANormal;
	    va |= (p_build->bld_mask&PEXtk_BLD_SET_COLOR) ? PEXGAColor : 0;

	    i = (p_build->bld_mask&PEXtk_BLD_SET_COLOR) ? sizeof(BLD_COLOR) :0;
	    i = p_build->vert_cnt * 
		  (sizeof(BLD_NORMAL)+sizeof(BLD_VERT) + i);
	    if( DB_check_build_buffer_size(i))
		return 1;	/* memory alloc error */

	    p_build->ppoly.count = p_build->vert_cnt;  /* n verts */
	    p_build->ppoly.vertices.no_data = (PEXCoord *)p_build->buffer;
	    ptr = (char *) p_build->buffer;
	    pv = (char *) p_build->bld_vert;
	    p_bld = (char *) p_build->bld_color;
	    p_norm = (char *) p_build->bld_normal;

	    for(i=0;i<p_build->vert_cnt;i++){
	      memcpy( ptr, (char *) pv, sizeof(BLD_VERT));
	      ptr += sizeof(BLD_VERT); pv += sizeof(BLD_VERT);
	      if( p_build->bld_mask & PEXtk_BLD_SET_COLOR ){
	          memcpy( ptr,(char *) p_bld, sizeof(BLD_COLOR));
	          ptr += sizeof(BLD_COLOR); p_bld += sizeof(BLD_COLOR);
	      }
	      memcpy( ptr, (char *) p_norm, sizeof(BLD_NORMAL));
	      ptr += sizeof(BLD_NORMAL); p_norm += sizeof(BLD_NORMAL);
	    }
	    if( p_build->polynorm_cnt ){
/*
		p_build->ppoly.facetData = (char *)&p_build->bld_polynorm[0];
*/
		fa = PEXGANormal;
	    }
	    PEXFillAreaSetWithData( dpy, p_state->cur_id, p_state->req_type, 
		    p_state->shape, False, PEXContourDisjoint, fa, va,
		    p_state->color_type, 1,
		    (PEXFacetData *)&p_build->bld_polynorm[0], &p_build->ppoly);
	  break;
      default:
	  if( p_state->debug_mode )
	    fprintf(stderr,"ERROR... geometry building not in valid state\n");
    }

	/* reset state */
	DB_init_bld_counts();

    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+        Ppoly_normal      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets a polygon normal.
 *
 *  USAGE:
 *	Ppoly_normal( u, v, w )
 *	Fcoord	    u		(in)	Polygon normal u
 *	Fcoord	    v		(in)	Polygon normal v
 *	Fcoord	    w		(in)	Polygon normal w
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ppoly_normal( u, v, w )
Fcoord	u;
Fcoord	v;
Fcoord	w;
{
    int i;

	DB_check_bld_polynorm();
	i = p_build->polynorm_cnt;
	p_build->bld_polynorm[i].norm.x   = u;
	p_build->bld_polynorm[i].norm.y   = v;
	p_build->bld_polynorm[i].norm.z   = w;

	p_build->polynorm_cnt++;
}
 
/*
 *
 *                       +--------------------------+
 * ----------------------+      Ppoly_point_2d      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Puts a 2D polygon vertex in geometry building buffer.
 *
 *  USAGE:
 *	Ppoly_point_2d( x, y )
 *	Fcoord	    x		(in)	x position
 *	Fcoord	    y		(in)	y position
 *
 *  NOTE:
 *	Only one of the Ppoly_point... routines may be used during
 *	any Ppoly_point session.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ppoly_point_2d( x, y )
Fcoord	x;
Fcoord	y;
{
    int i;

	if( p_build->bld_mask & ~(PEXtk_BLD_PNT_2D|PEXtk_BLD_SET_COLOR) ){
	  if( p_state->debug_mode )
	    fprintf(stderr,
                 "ERROR... Ppoly_point_2d called when building %s\n",
		DB_get_bld_type_string());
	    return;
	}

	p_build->bld_mask |= PEXtk_BLD_PNT_2D;
	DB_check_bld_poly_point();
	i = p_build->vert_cnt;
	p_build->bld_vert[i].pts.x = x;
	p_build->bld_vert[i].pts.y = y;
	p_build->bld_vert[i].pts.z = 0.0;

	p_build->pt2d_cnt++;
	p_build->vert_cnt++;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      Ppoly_point_3d      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Puts a 3D polygon vertex in geometry building buffer.
 *
 *  USAGE:
 *	Ppoly_point_3d( x, y, z )
 *	Fcoord	    x		(in)	x position
 *	Fcoord	    y		(in)	y position
 *	Fcoord	    z		(in)	z position
 *
 *  NOTE:
 *	Only one of the Ppoly_point... routines may be used during
 *	any Ppoly_point session.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ppoly_point_3d( x, y, z )
Fcoord	x;
Fcoord	y;
Fcoord	z;
{
    int i;

	if( p_build->bld_mask & ~(PEXtk_BLD_PNT_3D|PEXtk_BLD_SET_COLOR) ){
	  if( p_state->debug_mode )
	    fprintf(stderr,
                 "ERROR... Ppoly_point_3d called when building %s\n",
		DB_get_bld_type_string());
	    return;
	}

	p_build->bld_mask |= PEXtk_BLD_PNT_3D;
	DB_check_bld_poly_point();
	i = p_build->vert_cnt;
	p_build->bld_vert[i].pts.x = x;
	p_build->bld_vert[i].pts.y = y;
	p_build->bld_vert[i].pts.z = z;

	p_build->pt3d_cnt++;
	p_build->vert_cnt++;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      Ppoly_point_nv      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Puts a 3D polygon vertex and vertex normal in geometry building buffer.
 *
 *  USAGE:
 *	Ppoly_point_nv( x, y, z, nx, ny, nv )
 *	Fcoord	    x		(in)	x position
 *	Fcoord	    y		(in)	y position
 *	Fcoord	    z		(in)	z position
 *	Fcoord	    nx		(in)	Normal x value
 *	Fcoord	    ny		(in)	Normal y value
 *	Fcoord	    nz		(in)	Normal z value
 *
 *  NOTE:
 *	Only one of the Ppoly_point... routines may be used during
 *	any Ppoly_point session.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ppoly_point_nv( x, y, z, nx, ny, nz )
Fcoord	x;
Fcoord	y;
Fcoord	z;
Fcoord	nx;
Fcoord	ny;
Fcoord	nz;
{
    int i;

	if( p_build->bld_mask & ~(PEXtk_BLD_PNT_NV|PEXtk_BLD_SET_COLOR) ){
	  if( p_state->debug_mode )
	    fprintf(stderr,
                 "ERROR... Ppoly_point_nv called when building %s\n",
		DB_get_bld_type_string());
	    return;
	}

	p_build->bld_mask |= PEXtk_BLD_PNT_NV;
	DB_check_bld_poly_point();
	i = p_build->vert_cnt;
	p_build->bld_vert[i].pts.x = x;
	p_build->bld_vert[i].pts.y = y;
	p_build->bld_vert[i].pts.z = z;

	p_build->bld_normal[i].norm.x = nx;
	p_build->bld_normal[i].norm.y = ny;
	p_build->bld_normal[i].norm.z = nz;

	p_build->nv_cnt++;
	p_build->vert_cnt++;
	p_build->normal_cnt++;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      Ppoly_point_uv      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Puts a 3D polygon vertex and texture vertex in geometry building buffer.
 *
 *  USAGE:
 *	Ppoly_point_uv( x, y, z, ux, uy )
 *	Fcoord	    x		(in)	x position
 *	Fcoord	    y		(in)	y position
 *	Fcoord	    z		(in)	z position
 *	Fcoord	    ux		(in)	Texture vertex x value
 *	Fcoord	    uy		(in)	Texture vertex y value
 *
 *  NOTE:
 *	Only one of the Ppoly_point... routines may be used during
 *	any Ppoly_point session.
 *	Support for texture mapping is platform specific.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ppoly_point_uv( x, y, z, ux, uy, uz )
Fcoord	x;
Fcoord	y;
Fcoord	z;
Fcoord	ux;
Fcoord	uy;
{
    int i;

	if( p_build->bld_mask & ~(PEXtk_BLD_PNT_UV|PEXtk_BLD_SET_COLOR) ){
	  if( p_state->debug_mode )
	    fprintf(stderr,
                 "ERROR... Ppoly_point_uv called when building %s\n",
		DB_get_bld_type_string());
	    return;
	}

	p_build->bld_mask |= PEXtk_BLD_PNT_UV;
	DB_check_bld_poly_point();
	i = p_build->vert_cnt;
	p_build->bld_vert[i].pts.x = x;
	p_build->bld_vert[i].pts.y = y;
	p_build->bld_vert[i].pts.z = z;

	p_build->bld_texture[i].texture.x = ux;
	p_build->bld_texture[i].texture.y = uy;
	p_build->bld_texture[i].texture.z = 0.0;

	p_build->uv_cnt++;
	p_build->vert_cnt++;
	p_build->texture_cnt++;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+    Ppoly_point_nv_uv     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Puts a 3D polygon vertex, vertex normal, and texture vertex
 *	in geometry building buffer.
 *
 *  USAGE:
 *	Ppoly_point_nv_uv( x, y, z, nx, ny, nv, ux, uy )
 *	Fcoord	    x		(in)	x position
 *	Fcoord	    y		(in)	y position
 *	Fcoord	    z		(in)	z position
 *	Fcoord	    nx		(in)	Normal x value
 *	Fcoord	    ny		(in)	Normal y value
 *	Fcoord	    nz		(in)	Normal z value
 *	Fcoord	    ux		(in)	Texture vertex x value
 *	Fcoord	    uy		(in)	Texture vertex y value
 *
 *  NOTE:
 *	Only one of the Ppoly_point... routines may be used during
 *	any Ppoly_point session.
 *	Support for texture mapping is platform specific.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ppoly_point_nv_uv( x, y, z, nx, ny, nz, ux, uy )
Fcoord	x;
Fcoord	y;
Fcoord	z;
Fcoord	nx;
Fcoord	ny;
Fcoord	nz;
Fcoord	ux;
Fcoord	uy;
{
    int i;

	if( p_build->bld_mask & ~(PEXtk_BLD_PNT_NV_UV|PEXtk_BLD_SET_COLOR) )
	{
	  if( p_state->debug_mode )
	    fprintf(stderr,
                 "ERROR... Ppoly_point_nv_uv called when building %s\n",
		DB_get_bld_type_string());
	    return;
	}

	p_build->bld_mask |= PEXtk_BLD_PNT_NV_UV;
	DB_check_bld_poly_point();
	i = p_build->vert_cnt;
	p_build->bld_vert[i].pts.x = x;
	p_build->bld_vert[i].pts.y = y;
	p_build->bld_vert[i].pts.z = z;

	p_build->bld_normal[i].norm.x = nx;
	p_build->bld_normal[i].norm.y = ny;
	p_build->bld_normal[i].norm.z = nz;

	p_build->bld_texture[i].texture.x = ux;
	p_build->bld_texture[i].texture.y = uy;
	p_build->bld_texture[i].texture.z = 0.0;

	p_build->nvuv_cnt++;
	p_build->vert_cnt++;
	p_build->normal_cnt++;
	p_build->texture_cnt++;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+        Pswap_tmesh       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Swaps the last two 3D verticies in geometry building buffer.
 *
 *  USAGE:
 *	Pswap_tmesh( )
 *
 *  DESCRIPTION:
 *	To implement swap_tmesh, the buffer must be flushed and restarted
 *	after each swap_tmesh, and the last vertex is put in the second
 *	slot, and the last-1 vertex is put in the first slot. The ordering
 *	then continues normally.
 *
 *  NOTE:
 *	There must be at least three vertices in the buffer for this routine
 *	to take effect.
 *	
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pswap_tmesh( )
{
	if( p_build->bld_mode != PEXtk_BLD_MESH )
	    return;		/* not in correct mode */
	if( p_build->vert_cnt < 2 )
	    return;		/* not enough vertices */
	PEXtk_swap_tmesh();

}

/*
 *
 *                       +--------------------------+
 * ----------------------+          Pvertex         +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Puts a 3D vertex in geometry building buffer.
 *
 *  USAGE:
 *	Pvertex( point )
 *	Fcoord	    point[3]	(in)	3D vertex
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pvertex( point )
Fcoord	point[3];
{
    int i;

	DB_check_bld_vert(1);
	i = p_build->vert_cnt;
	p_build->bld_mask |= PEXtk_BLD_SET_VERT;
	p_build->bld_vert[i].pts.x = point[0];
	p_build->bld_vert[i].pts.y = point[1];
	p_build->bld_vert[i].pts.z = point[2];

	p_build->vert_cnt++;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      Pvertex_color       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Puts a vertex color in buffer.
 *
 *  USAGE:
 *	Pvertex_color( point )
 *	PCOLOR	    *col	(in)	PCOLOR structure with vertex color
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pvertex_color( col )
PCOLOR	*col;
{
    int i;

	DB_check_bld_color(1);
	p_build->bld_mask |= PEXtk_BLD_SET_COLOR;
	i = p_build->color_cnt;
	if( col->type == PEXtk_COLOR_HSV && 
	    p_state->color_type == PEXColorTypeRGB)
	{
	    PEXtk_convert_hsv_to_rgb(  col->color.hsv.hue,
			col->color.hsv.saturation,
			col->color.hsv.value,
			&p_build->bld_color[i].color.rgb.red,
			&p_build->bld_color[i].color.rgb.green,
			&p_build->bld_color[i].color.rgb.blue);
	} else {
	    p_build->bld_color[i].color.rgb.red  = 
			col->color.rgb.red;
	    p_build->bld_color[i].color.rgb.green= 
			col->color.rgb.green;
	    p_build->bld_color[i].color.rgb.blue = 
			col->color.rgb.blue;
	}

	p_build->color_cnt++;
    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+   Pvertex_color_array    +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Puts a vertex color array in buffer.
 *
 *  USAGE:
 *	Pvertex_color_array( n, col )
 *	int	    n		(in)	Number of items in rgb list
 *	PCOLOR	    col[]	(in)	PCOLOR structure with vertex color
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pvertex_color_array( n, col )
int	n;
PCOLOR	col[];
{
    int i,j;

	/* Check against 'n' size */
	DB_check_bld_color(n);
	p_build->bld_mask |= PEXtk_BLD_SET_COLOR;
	i = p_build->color_cnt;
	for( j=0 ; j<n; i++, j++){
	    if( col->type == PEXtk_COLOR_HSV && 
	        p_state->color_type == PEXColorTypeRGB)
	    {
		PEXtk_convert_hsv_to_rgb(  col[j].color.hsv.hue,
			col[j].color.hsv.saturation,
			col[j].color.hsv.value,
			&p_build->bld_color[i].color.rgb.red,
			&p_build->bld_color[i].color.rgb.green,
			&p_build->bld_color[i].color.rgb.blue);
	    } else {
	        p_build->bld_color[i].color.rgb.red   =
			col[j].color.rgb.red;
	        p_build->bld_color[i].color.rgb.green =
			col[j].color.rgb.green;
	        p_build->bld_color[i].color.rgb.blue  =
			col[j].color.rgb.blue;
	    }
	}

	p_build->color_cnt += n;
    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+ Pvertex_color_array_hsv  +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Puts a vertex color hsv array in buffer.
 *
 *  USAGE:
 *	Pvertex_color_array_hsv( n, hsv )
 *	int	    n		(in)	Number of items in hsv list
 *	hsvFloat    *hsv	(in)	Array of hsv triplets
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pvertex_color_array_hsv( n, hsv )
int	 n;
hsvFloat *hsv;
{
    int i,j;

	/* Check against 'n' size */
    DB_check_bld_color(n);
    p_build->bld_mask |= PEXtk_BLD_SET_COLOR;
    i = p_build->color_cnt;
    for( j=0 ; j<n; i++,j++){
	p_build->bld_color[i].color.hsv.hue	  = hsv[j].hue;
	p_build->bld_color[i].color.hsv.saturation = 
							hsv[j].saturation;
	p_build->bld_color[i].color.hsv.value = hsv[j].value;
    }

    p_build->color_cnt += n;
    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+ Pvertex_color_array_ind  +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Puts a vertex color index array in buffer.
 *
 *  USAGE:
 *	Pvertex_color_array_hsv( n, index )
 *	int	    n		(in)	Number of items in index list
 *	int	    index[]	(in)	Array of indexed color values
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pvertex_color_array_ind( n, index )
int	n;
int	index[];
{
    int i,j;

	/* Check against 'n' size */
	DB_check_bld_color(n);
	p_build->bld_mask |= PEXtk_BLD_SET_COLOR;
	i = p_build->color_cnt;
	for( j=0 ; j<n; i++, j++){
	    p_build->bld_color[i].color.indexed.index   = index[j];
	}

	p_build->color_cnt += n;
    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+ Pvertex_color_array_rgb  +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Puts a vertex color rgb array in buffer.
 *
 *  USAGE:
 *	Pvertex_color_array_rgb( n, rgb )
 *	int	    n		(in)	Number of items in rgb list
 *	rgbFloat    *rgb	(in)	Array of rgb triplets
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pvertex_color_array_rgb( n, rgb )
int	 n;
rgbFloat *rgb;
{
    int i,j;

	/* Check against 'n' size */
	DB_check_bld_color(n);
	p_build->bld_mask |= PEXtk_BLD_SET_COLOR;
	i = p_build->color_cnt;
	for( j=0 ; j<n; i++, j++){
	    p_build->bld_color[i].color.rgb.red   = rgb[j].red;
	    p_build->bld_color[i].color.rgb.green = rgb[j].green;
	    p_build->bld_color[i].color.rgb.blue  = rgb[j].blue;
	}

	p_build->color_cnt += n;
    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+    Pvertex_color_hsv     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Puts a vertex color hsv in buffer.
 *
 *  USAGE:
 *	Pvertex_color_hsv( hue, sat, value )
 *	Fdata	    hsv		(in)	Hue
 *	Fdata	    sat		(in)	Saturation
 *	Fdata	    value	(in)	Value
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pvertex_color_hsv( hue, sat, value )
Fdata	hue;
Fdata	sat;
Fdata	value;
{
    int i;

	/* Check against '1' size */
	DB_check_bld_color(1);
	p_build->bld_mask |= PEXtk_BLD_SET_COLOR;
	i = p_build->color_cnt;
	p_build->bld_color[i].color.hsv.hue	        = hue;
	p_build->bld_color[i].color.hsv.saturation  = sat;
	p_build->bld_color[i].color.hsv.value	= value;

	p_build->color_cnt++;
    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+    Pvertex_color_ind     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Puts a vertex color index in buffer.
 *
 *  USAGE:
 *	Pvertex_color_hsv( index )
 *	int	    index	(in)	Indexed color value
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pvertex_color_ind( index )
int	index;
{
    int i;

	/* Check against '1' size */
	DB_check_bld_color(1);
	p_build->bld_mask |= PEXtk_BLD_SET_COLOR;
	i = p_build->color_cnt;
	p_build->bld_color[i].color.indexed.index   = index;

	p_build->color_cnt++;
    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+    Pvertex_color_rgb     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Puts a vertex color rgb in buffer.
 *
 *  USAGE:
 *	Pvertex_color_rgb( red, green, blue )
 *	Fdata	    red		(in)	Red value
 *	Fdata	    green	(in)	Green value
 *	Fdata	    blue	(in)	Blue value
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pvertex_color_rgb( red, green, blue )
Fdata	red;
Fdata	green;
Fdata	blue;
{
    int i;

	/* Check against '1' size */
	DB_check_bld_color(1);
	p_build->bld_mask |= PEXtk_BLD_SET_COLOR;
	i = p_build->color_cnt;
	p_build->bld_color[i].color.rgb.red   = red;
	p_build->bld_color[i].color.rgb.green = green;
	p_build->bld_color[i].color.rgb.blue  = blue;

	p_build->color_cnt++;
    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      Pvertex_normal      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Puts a 3D vertex normal in geometry building buffer.
 *
 *  USAGE:
 *	Pvertex_normal( nx, ny, nz )
 *	Fcoord	    nx		(in)	vertex normal x
 *	Fcoord	    ny		(in)	vertex normal y
 *	Fcoord	    nz		(in)	vertex normal z
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pvertex_normal( nx, ny, nz )
Fcoord	nx;
Fcoord	ny;
Fcoord	nz;
{
    int i;

	DB_check_bld_normal(1);
	i = p_build->normal_cnt;
	p_build->bld_mask |= PEXtk_BLD_SET_NORM;

	p_build->bld_normal[i].norm.x = nx;
	p_build->bld_normal[i].norm.y = ny;
	p_build->bld_normal[i].norm.z = nz;

	p_build->normal_cnt++;
    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      Pvertex_normals     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Puts 'n' 3D vertex normal in geometry building buffer.
 *
 *  USAGE:
 *	Pvertex_normals( n, norm )
 *	int	    n		(in)	number of vertex normals
 *	Fcoord3D    *norm	(in)	vertex normals
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pvertex_normals( n, norm )
int		n;
Fcoord3D	*norm;
{
    int i, j;

	DB_check_bld_normal(n);
	i = p_build->normal_cnt;
	p_build->bld_mask |= PEXtk_BLD_SET_NORM;

	for( j=0 ; j<n; i++, j++){
	    p_build->bld_normal[i].norm.x = norm[j].x;
	    p_build->bld_normal[i].norm.y = norm[j].y;
	    p_build->bld_normal[i].norm.z = norm[j].z;
	}

	p_build->normal_cnt += n;
    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+     Pvertex_texture      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Puts a 2D texture vertex in geometry building buffer.
 *
 *  USAGE:
 *	Pvertex_texture( u, v )
 *	Fcoord	    u		(in)	texture vertex u
 *	Fcoord	    v		(in)	texture vertex v
 *
 *  NOTE:
 *	Support for texture mapping is platform specific.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pvertex_texture( u, v )
Fcoord	u;
Fcoord	v;
{

#ifdef SUPPORT_TEXTURE
    int i;

	DB_check_bld_texture(1);
	i = p_build->texture_cnt;
	p_build->bld_mask |= PEXtk_BLD_SET_TEXTURE;
	p_build->bld_texture[i].texture.x = u;
	p_build->bld_texture[i].texture.y = v;

	p_build->texture_cnt++;
    return;
#else
	return;
#endif
}

/*
 *
 *                       +--------------------------+
 * ----------------------+     Pvertex_textures     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Puts n 2D texture vertices in geometry building buffer.
 *
 *  USAGE:
 *	Pvertex_textures( n, tvert )
 *	int	    n		(in)	number of texture vertices
 *	Tvar	    *tvert	(in)	texture vertices 
 *
 *  NOTE:
 *	Support for texture mapping is platform specific.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pvertex_textures( n, tvert )
int	n;
Tvar	*tvert;
{

#ifdef SUPPORT_TEXTURE
    int i,j;

	DB_check_bld_texture(n);
	i = p_build->texture_cnt;
	p_build->bld_mask |= PEXtk_BLD_SET_TEXTURE;
	for( j=0 ; j<n; i++, j++){
	    p_build->bld_texture[i].texture.x = tvert[j].u;
	    p_build->bld_texture[i].texture.y = tvert[j].v;
	}

	p_build->texture_cnt += n;
    return;
#else
	return;
#endif
}

/*
 *
 *                       +--------------------------+
 * ----------------------+     Pvertex_texture3     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Puts a 3D texture vertex in geometry building buffer.
 *
 *  USAGE:
 *	Pvertex_texture3( tvert )
 *	Fcoord	    tvert[3]	(in)	texture vertex 
 *
 *  NOTE:
 *	Support for texture mapping is platform specific.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pvertex_texture3( tvert )
Fcoord	tvert[3];
{

#ifdef SUPPORT_TEXTURE
    int i;

	DB_check_bld_texture();
	i = p_build->texture_cnt;
	p_build->bld_mask |= PEXtk_BLD_SET_TEXTURE3;
	p_build->bld_texture[i].texture.x = tvert[0];
	p_build->bld_texture[i].texture.y = tvert[1];
	p_build->bld_texture[i].texture.z = tvert[2];

	p_build->texture_cnt++;
    return;
#else
	return;
#endif
}

/*
 *
 *                       +--------------------------+
 * ----------------------+    PEXtk_bld_primitive   +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Builds a polygon or triangle mesh.
 *
 *  USAGE:
 *	PEXtk_bld_primitive( type )
 *	int	    type	(in)	type of primitive:
 *					   PEXtk_BLD_POLY = polygon
 *					   PEXtk_BLD_MESH = triangle mesh
 *
 *  RETURNS:
 *	0 if successful,  non-zero if an error occurs.
 *			  1 - memory alloc error
 *			  2 - in line or tmesh mode
 *			  3 - incorrect counts
 */
 
/*ARGSUSED*/
int
PEXtk_build_primitive( type )
int	type;
{
    int i;
    unsigned long fa;
    char *ptr;
    PEXArrayOfFacetData fct;
    PEXArrayOfVertex vtx;

    switch( p_build->bld_mask ){

      case PEXtk_BLD_SET_VERT:			/* points only */

	    i = p_build->vert_cnt * sizeof(BLD_VERT);
	    if( DB_check_build_buffer_size(i))
		return 1;	/* memory alloc error */

	    if( type == PEXtk_BLD_POLY ){
	      if( p_build->polynorm_cnt ){
	        p_build->ppoly.count = p_build->vert_cnt;  /* n verts */
	        p_build->ppoly.vertices.no_data = (PEXCoord *)p_build->bld_vert;
	        PEXFillAreaSetWithData( dpy, p_state->cur_id, p_state->req_type, 
		    p_state->shape, False, PEXContourDisjoint, PEXGANormal, 0,
		    p_state->color_type, 1,
		    (PEXFacetData *)&p_build->bld_polynorm[0], &p_build->ppoly);
	      } else 
	        PEXFillArea( dpy, p_state->cur_id, p_state->req_type, 
			p_state->shape, False, 
			p_build->vert_cnt, (PEXCoord *) p_build->bld_vert);
	    } else {
	        vtx.no_data = (PEXCoord *)p_build->bld_vert;
	      PEXTriangleStrip( dpy, p_state->cur_id, p_state->req_type, 0, 
		0, p_state->color_type, fct, p_build->vert_cnt, vtx);
	    }
	  break;

			/* points and colors */
      case PEXtk_BLD_SET_VERT|PEXtk_BLD_SET_COLOR:

	    if(DB_check_build_mask()){
		DB_init_bld_counts();
		return 3;
	    }

	    i = p_build->vert_cnt * 
		  (sizeof(BLD_VERT)+sizeof(BLD_COLOR));
	    if( DB_check_build_buffer_size(i))
		return 1;	/* memory alloc error */

	    ptr = p_build->buffer;

	    for(i=0;i<p_build->vert_cnt;i++){
	      memcpy(ptr,(char*)&p_build->bld_vert[i],sizeof(BLD_VERT));
	      ptr += sizeof(BLD_VERT);
	      memcpy(ptr,(char*)&p_build->bld_color[i],sizeof(BLD_COLOR));
	      ptr += sizeof(BLD_COLOR);
	    }
	    if( type == PEXtk_BLD_POLY ){
	        p_build->ppoly.count = p_build->vert_cnt;  /* n verts */
	        p_build->ppoly.vertices.no_data = (PEXCoord *) p_build->buffer;
	      if( p_build->polynorm_cnt ){
/*
	        p_build->ppoly.facetData = (char *) &p_build->bld_polynorm[0];
*/
		fa = PEXGANormal;
	      } else 
		fa = 0;
	      PEXFillAreaSetWithData( dpy, p_state->cur_id, p_state->req_type, 
		    p_state->shape, False, PEXContourDisjoint, fa, PEXGAColor,
		    p_state->color_type, 1,
		    (PEXFacetData *)&p_build->bld_polynorm[0], &p_build->ppoly);
	    } else {
	        vtx.no_data = (PEXCoord *) p_build->buffer;
	      PEXTriangleStrip( dpy, p_state->cur_id, p_state->req_type, 0, 
		PEXGAColor, p_state->color_type, fct, p_build->vert_cnt, vtx);
	    }
	  break;

			/* points, colors and normals */
      case PEXtk_BLD_SET_VERT|PEXtk_BLD_SET_COLOR|PEXtk_BLD_SET_NORM:

	    if(DB_check_build_mask()){
		DB_init_bld_counts();
		return 3;
	    }

	    i=p_build->vert_cnt * 
		  (sizeof(BLD_NORMAL)+sizeof(BLD_VERT)+
		  sizeof(BLD_COLOR));
	    if( DB_check_build_buffer_size(i))
		return 1;	/* memory alloc error */

	    ptr = p_build->buffer;

	    for(i=0;i<p_build->vert_cnt;i++){
	      memcpy(ptr,(char*)&p_build->bld_vert[i],sizeof(BLD_VERT));
	      ptr += sizeof(BLD_VERT);
	      memcpy(ptr,(char*)&p_build->bld_color[i],sizeof(BLD_COLOR));
	      ptr += sizeof(BLD_COLOR);
	      memcpy(ptr,(char*)&p_build->bld_normal[i],sizeof(BLD_NORMAL));
	      ptr += sizeof(BLD_NORMAL);
	    }

	    if( type == PEXtk_BLD_POLY ){
	        p_build->ppoly.count = p_build->vert_cnt;  /* n verts */
	        p_build->ppoly.vertices.no_data = (PEXCoord *) p_build->buffer;
	      if( p_build->polynorm_cnt ){
/*
	        p_build->ppoly.facetData = (char *) &p_build->bld_polynorm[0];
*/
		fa = PEXGANormal;
	      } else 
		fa = 0;
	      PEXFillAreaSetWithData( dpy, p_state->cur_id, p_state->req_type, 
		    p_state->shape, False, PEXContourDisjoint, fa,
		    PEXGANormal|PEXGAColor, p_state->color_type, 1, 
		    (PEXFacetData *)&p_build->bld_polynorm[0], &p_build->ppoly);
	    } else {
	        vtx.no_data = (PEXCoord *) p_build->buffer;
	      PEXTriangleStrip( dpy, p_state->cur_id, p_state->req_type, 0, 
		PEXGANormal|PEXGAColor, p_state->color_type, fct,
		p_build->vert_cnt, vtx);
	    }
	  break;

			/* points and normals */
      case PEXtk_BLD_SET_VERT|PEXtk_BLD_SET_NORM:

	    if(DB_check_build_mask()){
		DB_init_bld_counts();
		return 3;
	    }

	    i=p_build->vert_cnt * 
		  (sizeof(BLD_NORMAL)+sizeof(BLD_VERT));
	    if( DB_check_build_buffer_size(i))
		return 1;	/* memory alloc error */

	    ptr = p_build->buffer;

	    for(i=0;i<p_build->vert_cnt;i++){
	      memcpy(ptr,(char*)&p_build->bld_vert[i],sizeof(BLD_VERT));
	      ptr += sizeof(BLD_VERT);
	      memcpy(ptr,(char*)&p_build->bld_normal[i],sizeof(BLD_NORMAL));
	      ptr += sizeof(BLD_NORMAL);
	    }

	    if( type == PEXtk_BLD_POLY ){
	        p_build->ppoly.count = p_build->vert_cnt;  /* n verts */
	        p_build->ppoly.vertices.no_data = (PEXCoord *) p_build->buffer;
	      if( p_build->polynorm_cnt ){
/*
	        p_build->ppoly.facetData = (char *) &p_build->bld_polynorm[0];
*/
		fa = PEXGANormal;
	      } else 
		fa = 0;
	      PEXFillAreaSetWithData( dpy, p_state->cur_id, p_state->req_type, 
		    p_state->shape, False, PEXContourDisjoint, fa, PEXGANormal, 
		    p_state->color_type, 1,
		    (PEXFacetData *)&p_build->bld_polynorm[0], &p_build->ppoly);
	    } else {
	        vtx.no_data = (PEXCoord *) p_build->buffer;
	      PEXTriangleStrip( dpy, p_state->cur_id, p_state->req_type, 
		0, PEXGANormal, p_state->color_type, fct,
		p_build->vert_cnt, vtx);
	    }
	  break;

	/*
	 * Note that the following are not currently supported.
	 */
			/* points, colors, normals and textures */
      case PEXtk_BLD_SET_VERT|PEXtk_BLD_SET_COLOR|PEXtk_BLD_SET_NORM|
		PEXtk_BLD_SET_TEXTURE:
      case PEXtk_BLD_SET_VERT|PEXtk_BLD_SET_COLOR|PEXtk_BLD_SET_NORM|
		PEXtk_BLD_SET_TEXTURE3:

	    if(DB_check_build_mask()){
		DB_init_bld_counts();
		return 3;
	    }

	    i=p_build->vert_cnt * 
		(sizeof(BLD_NORMAL)+sizeof(BLD_VERT)+
		sizeof(BLD_COLOR)+sizeof(BLD_TEXTURE)-
		(p_build->bld_mask&PEXtk_BLD_SET_TEXTURE3 ? 0 : sizeof(float)));

	    if( DB_check_build_buffer_size(i))
		return 1;	/* memory alloc error */

	    ptr = p_build->buffer;

	    for(i=0;i<p_build->vert_cnt;i++){
	      memcpy(ptr,(char*)&p_build->bld_vert[i],sizeof(BLD_VERT));
	      ptr += sizeof(BLD_VERT);
	      memcpy(ptr,(char*)&p_build->bld_color[i],sizeof(BLD_COLOR));
	      ptr += sizeof(BLD_COLOR);
	      memcpy(ptr, (char *) &p_build->bld_normal[i],sizeof(BLD_NORMAL));
	      ptr += sizeof(BLD_NORMAL);
	      /*
	       * Some mechanism is needed to set the PEX server to expect
	       * 2D or 3D texture vertices
	       */
	      if( p_build->bld_mask & PEXtk_BLD_SET_TEXTURE3 ){
		memcpy( ptr, (char *) &p_build->bld_texture[i],
			sizeof(BLD_TEXTURE));
		ptr += sizeof(BLD_TEXTURE);
	      } else {
		memcpy(ptr, (char *) &p_build->bld_texture[i],
			sizeof(BLD_TEXTURE)-sizeof(float));
		ptr = ptr + sizeof(BLD_TEXTURE) - sizeof(float);
	      }
	    }

	    if( type == PEXtk_BLD_POLY ){
	        p_build->ppoly.count = p_build->vert_cnt;  /* n verts */
	        p_build->ppoly.vertices.no_data = (PEXCoord *) p_build->buffer;
	      if( p_build->polynorm_cnt ){
/*
	        p_build->ppoly.facetData = (char *) &p_build->bld_polynorm[0];
*/
		fa = PEXGANormal;
	      } else 
		fa = 0;
	      PEXFillAreaSetWithData( dpy, p_state->cur_id, p_state->req_type, 
		    p_state->shape, False, PEXContourDisjoint, fa,
		    PEXGANormal|PEXGAColor, p_state->color_type, 1, 
		    (PEXFacetData *)&p_build->bld_polynorm[0], &p_build->ppoly);
	    } else {
	        vtx.no_data = (PEXCoord *) p_build->buffer;
	      PEXTriangleStrip( dpy, p_state->cur_id, p_state->req_type, 0, 
		PEXGANormal|PEXGAColor, p_state->color_type, fct,
		p_build->vert_cnt, vtx);
	    }
	  break;
      default:
	  if( p_state->debug_mode )
	    fprintf(stderr,"ERROR... geometry building not in valid state\n");
    }
    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      PEXtk_swap_tmesh    +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Performs a flush on the tmesh buffer, and resets first two vertices.
 *
 *  PURPOSE:
 *	The buffer is flushed and restarted, and the last vertex is put in 
 *	the second slot, and the last-1 vertex is put in the first slot. The 
 *	ordering then continues normally.
 *
 *  USAGE:
 *	PEXtk_swap_tmesh( void )
 *	
 *  RETURNS:
 *	0 if successful, non-zero if error
 */
 
/*ARGSUSED*/
int
PEXtk_swap_tmesh()
{
    int	   i;
    int    old_mask;		/* save the bld_mask, it will be reset */

	if( p_build->bld_mode != PEXtk_BLD_MESH )
	    return 1;		/* not in correct mode */
	if( p_build->vert_cnt < 2 )
	    return 1;		/* not enough vertices */
	old_mask = p_build->bld_mask;
	i = p_build->vert_cnt-1;  /* get current number of "slots" used */

	Pend_tmesh();

	/* Reset the state to the way it was */
	p_build->bld_mode    = PEXtk_BLD_MESH;
	p_build->bld_mask    = old_mask;
	p_build->vert_cnt    = 2;
	memcpy( (char*)&p_build->bld_vert[0], (char*)&p_build->bld_vert[i], 
			sizeof(PEXCoord));
	if( i != 2 ) /* if three verts were in buffer, no need to move midpt */
	    memcpy((char*)&p_build->bld_vert[1],(char*)&p_build->bld_vert[i-1], 
			sizeof(PEXCoord));

	if( old_mask & PEXtk_BLD_SET_COLOR ){
	    p_build->color_cnt   = 2;
	    memcpy((char*)&p_build->bld_color[0],(char*)&p_build->bld_color[i], 
			2*sizeof(PCOLOR));
	    if( i != 2 )
		memcpy( (char *) &p_build->bld_color[1],
			(char*) &p_build->bld_color[i-1], 
			sizeof(PCOLOR));
	}
	if( old_mask & PEXtk_BLD_SET_NORM ){
	    p_build->normal_cnt  = 2;
	    memcpy( (char *) &p_build->bld_normal[0],
			(char*) &p_build->bld_normal[i], 
			sizeof(PEXCoord));
	    if( i != 2 )
		memcpy( (char *) &p_build->bld_normal[1],
			(char*) &p_build->bld_normal[i-1], 
			sizeof(PEXCoord));
	}
	if( old_mask & PEXtk_BLD_SET_TEXTURE ){
	    p_build->texture_cnt = 2;
	    memcpy( (char *) &p_build->bld_texture[0],
			(char*) &p_build->bld_texture[i], 
			sizeof(PEXCoord));
	    if( i != 2 )
		memcpy( (char *)&p_build->bld_texture[1],
			(char*) &p_build->bld_texture[i-1], 
			sizeof(PEXCoord));
	}
	return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+ PEXtk_check_build_attrib +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Checks the current state for the Pbegin_.. routines for
 *	the interior style and surface interp method.
 *
 *  PURPOSE:
 *	The current fill style and surface interpolation methods are
 *	checked to see if they agree with the required types.
 *
 *  USAGE:
 *	PEXtk_check_build_attrib( )
 *	
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_check_build_attrib()
{
    if( p_state->sys_mode == PEXtk_STRUCTURE_MODE ){
        if(( p_build->bld_style		== PEXtk_FILL_SOLID ) &&
           ( p_state->attr->i_style_str != PEXInteriorStyleSolid )){
	    PEXSetInteriorStyle( dpy, p_state->cur_id, p_state->req_type,
		PEXInteriorStyleSolid);
	    p_state->attr->i_style_str = PEXInteriorStyleSolid;
	}
        if(( p_build->bld_style		== PEXtk_FILL_HOLLOW ) &&
           ( p_state->attr->i_style_str != PEXInteriorStyleHollow )){
	    PEXSetInteriorStyle( dpy, p_state->cur_id, p_state->req_type,
		PEXInteriorStyleHollow);
	    p_state->attr->i_style_str = PEXInteriorStyleHollow;
	}
	if( p_build->bld_mask & PEXtk_BLD_SET_COLOR )
	{
	    if( p_state->attr->interp_meth_str != PEXSurfaceInterpColor )
	    {
		p_state->attr->interp_meth_str = PEXSurfaceInterpColor;
		PEXSetSurfaceInterpMethod( dpy, p_state->cur_id, 
		    p_state->req_type, PEXSurfaceInterpColor);
	    }
	} /* turn off when done ? */
    } else { /* immediate mode */
	if(( p_build->bld_style	== PEXtk_FILL_SOLID ) &&
              ( p_state->attr->i_style  != PEXInteriorStyleSolid ))
	{
	    PEXSetInteriorStyle( dpy, p_state->cur_id, p_state->req_type,
		PEXInteriorStyleSolid);
	    p_state->attr->i_style = PEXInteriorStyleSolid;
	} else if(( p_build->bld_style	== PEXtk_FILL_HOLLOW ) &&
              ( p_state->attr->i_style  != PEXInteriorStyleHollow ))
	{
	    PEXSetInteriorStyle( dpy, p_state->cur_id, p_state->req_type,
		PEXInteriorStyleHollow);
	    p_state->attr->i_style = PEXInteriorStyleHollow;
	}

	if( p_build->bld_mask & PEXtk_BLD_SET_COLOR )
	{
	    if( p_state->attr->interp_meth != PEXSurfaceInterpColor )
	    {
		p_state->attr->interp_meth = PEXSurfaceInterpColor;
		PEXSetSurfaceInterpMethod( dpy, p_state->cur_id, 
		    p_state->req_type, PEXSurfaceInterpColor);
	    }
	} /* turn off when done ? */
    }
    return 0;
}

print_buffer(num)
int num;
{
  int i;
  PEXCoord *ptmp;

  ptmp = &p_build->bld_vert[0].pts;
  for(i=0;i<num; i++){
    printf(" %d, %f %f %f\n",i,ptmp->x,ptmp->y,ptmp->z);ptmp++;
  }
    printf(" \n");
}
