/*
 *  $Header: /auto/nfs/sequoia/u/an/work/rws/pextk/src/pextk_cmap.c,v 1.2 1994/01/04 01:21:47 rws Exp $
 ******************************************************************************
 ******************************************************************************
 *****                            pextk_cmap.c                            *****
 *****									  *****
 *****									  *****
 *****									  *****
 *****									  *****
 *****									  *****
 *                                                                            *
 *	Copyright 1992, 1993. All Rights Reserved                             *
 *			SHOgraphics					      *
 *			321 Soquel Way           			      *
 *			Sunnyvale, CA     94086   			      *
 *                                                                            *
 *                                                                            *
 *****									  *****

    Permission to use, copy, modify, and distribute this software and
 its documentation for any purpose and without fee is hereby granted,
 provided that the above copyright notice appear in all copies and that
 both that copyright notice and this permission notice appear in
 supporting documentation, and that the name of SHOgraphics not be
 used in advertising or publicity pertaining to distribution of the 
 software without specific, written prior permission.

 SHOGRAPHICS DISCLAIMS ANY AND ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 INCLUDING ALL EXPRESSED AND IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS. IN NO EVENT SHALL SHOGRAPHICS BE LIABLE FOR ANY DAMAGES OF ANY
 KIND, INCLUDING BUT NOT LIMITED TO SPECIAL, INDIRECT OR CONSEQUENTIAL 
 DAMAGES RESULTING FROM LOSS OF USE, DATA OR PROSPECTIVE PROFITS, WHETHER 
 IN AN ACTION OF CONTRACT, NEGLIENCE OR OTHER TORTIOUS ACTION, ARISING OUT 
 OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

 ************************************************************************ */
/*
 *
 *  Module Purpose:
 *	This is the module which contains the PEXtk API routines.
 *
 *  Module Notes:
 *
 *  Module Contents:
 *
 *    int   PEXtk_init_cmap	 Initializes PEXtk
 *    int   Pset_color_equation  Sets the color equation coefficient values.
 *    int   Pset_colormap        This routine will set the current colormap
 *    long  Ppixel_value_rgb	 Returns pixel value for an rgb triplet
 *    int   PEXtk_convert_hsv_to_rgb    Converts hsv to rgb values
 *
 *  Other externally available utilities:
 *    long  Ptrue_color		 Calcs pixel for true color visuals
 *    long  PEXtk_true_index     Gets pixel value for color index
 *    long  Pindex_color              Calcs pixel for input index value
 *    long  Ppseudo_color	      Calcs pixel for pseudo color visuals
 *    int   PEXtk_rgb_index  	      Gets rgb for indexed color value from DB
 *
 *  Undocumented utilities:
 *    int   PEXtk_SetColorTable  Sets the color table entries for index color
 *        8-bit colormap support:
 *    Colormap PEXtk_CreatePsuedoColormap Creates the pseudo colormap
 *    int   PEXtk_PCSetColormap       Sets the pc and lut for a pseudo colormap
 *    int   PEXtk_Setup_Pixel_RGB_Data  Setups the XColor data for rgb colormap
 *    int   PEXtk_Setup_Pixel_BGR_Data  Setups the XColor data for bgr colormap
 *    long  PEXtk_pseudo_index        Gets pixel value for color index
 *    int   PEXtk_rgb_colormap_index  Gets rgb for internal index value from
 *				      8 bit colormap lookup table.
 *    
 *

 */

#include <stdio.h>
#include <math.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>

#include <X11/PEX5/PEX.h>
#include <X11/PEX5/PEXproto.h>
#include <X11/PEX5/PEXlib.h>

#include <PEXtk.h>

#include "pextk.h"
#include "pextk_lcl.h"

#define	DEFAULT_GAMMA			2.3
#define ENUF_CELLS 			5

/*
 *	PREALLOC is a min base from the previous colormap
 */
#define PREALLOC 			8

/*
 *	BASEALLOC is a good base from the previous colormap
 */
#define BASEALLOC			16	

	int 				PEXtk_FreeColormap();
	static PEXColorApproxEntry	col_approx;
	static Colormap			cmap = 0;
	static int			newmap=0; /* True if new map created */

	/*
	 * The colormap rgb values for the color ramp are save in
	 * lkup_rgb[i][j]. The pixel values are saved in lkup_pix[i].
	 */
        unsigned long lkup_rgb[256][3];
        unsigned long lkup_pix[256];
/*
 *
 *                       +--------------------------+
 * ----------------------+      PEXtk_init_cmap     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Initializes colormap routines.
 *
 *  USAGE:
 *	PEXtk_init_cmap( p_obj )
 *	PEXtk_OBJECT    *p_obj	(in)	Pointer to PEXtk object.
 *
 *  DESCRIPTION:
 *	This routine saves the required x_data.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_init_cmap(p_obj)
PEXTK_OBJECT		*p_obj;
{
    /*
     * Save the various object ptrs
     */
    p__object	= p_obj;
    p_state	= p_obj->state;
    p_build	= p_obj->build;
    p_buffer	= p_obj->buffer;
    p_support	= p_obj->support;
    p_dl	= p_obj->dpy_list;
    p_clr	= p_obj->clr_tbl;
    p_x_data	= p_obj->x_data;

    dpy = p_x_data->dpy;

    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+    Pset_color_equation   +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the color equation coefficient values.
 *
 *  DESCRIPTION:
 *	If a colormap is passed in, then it will be used and the Pipeline
 *	Context and the Color Approx LUT will be created.
 *	If the pointer to the colormap is NULL or the value of the colormap
 *	is zero, then a colormap will be created. If the value of the
 *	colormap is zero, the value of the created colormap will be put
 *	into *usemap, otherwise it is not returned.
 *
 *	If the pointer to base_pixel is not NULL, then if the value of 
 *	base_pixel is not -1, it is used as the base pixel value when 
 *	creating a new colormap. If it is -1, a 'good' value is chosen,
 *	and the one chosen is returned in *base_pixel.
 *
 *  USAGE:
 *	Pset_color_equation( usemap, base_pixel, num_red, num_green, num_blue,
 *		mult_red, mult_green, mult_blue )
 *	int	    *usemap	(in)	Pointer to colormap
 *					if not NULL and value is not zero, 
 *					use this as colormap, otherwise
 *					one will be created.
 *	int	    *base_pixel	(in)	Pointer to Base pixel for colormap
 *					(if NULL, ignore).
 *	int	    num_red	(in)	Number of items in ramp for red
 *	int	    num_green	(in)	Number of items in ramp for green
 *	int	    num_blue	(in)	Number of items in ramp for blue
 *	int	    mult_red	(in)	Multiplier to use for red
 *	int	    mult_green	(in)	Multiplier to use for green
 *	int	    mult_blue	(in)	Multiplier to use for blue
 *	Fdata	    gamma	(in)	gamma correction value 
 *
 *  NOTE:
 *	This routine is for 8-bit systems only.
 *
 *  RETURNS:
 *	0 if successful, non-zero if an error occurred.
 */
 
/*ARGSUSED*/
int
Pset_color_equation( usemap, base_pixel, num_red, num_green, num_blue,
	mult_red, mult_green, mult_blue, gamma_value )
Colormap *usemap;
int	*base_pixel;
int	num_red;
int	num_green;
int	num_blue;
int	mult_red;
int	mult_green;
int	mult_blue;
Fdata	gamma_value;
{
    PEXPipelineContext	pc;
    PEXLookupTable	lut;
    PEXColorApproxEntry	approx;
    Colormap PEXtk_CreatePseudoColormap();

	if( !p_x_data ) return 1;
	if( p_x_data->win_depth != 8 ) /* only supports 8-bit color */
		return 1;
	approx.type = PEXColorSpace;
	approx.model = PEXColorApproxRGB;
	approx.dither = ( p_support->allow_dither ) ? p_state->dither : 0;
	approx.max1 = num_red-1;
	approx.max2 = num_green-1;
	approx.max3 = num_blue-1;
	approx.mult1 = mult_red;
	approx.mult2 = mult_green;
	approx.mult3 = mult_blue;
	approx.weight1 = 1.0;
	approx.weight2 = 1.0;
	approx.weight3 = 1.0;
	if( base_pixel != NULL ){
	    if( *base_pixel != -1 )
		approx.base_pixel = *base_pixel;
	    else
		approx.base_pixel = BASEALLOC;
	} else
	    approx.base_pixel = BASEALLOC;

	pc = p_state->pc_ID; lut = p_state->lut_approx;

 	if( p_state->win[p_state->cur_w].Rdr_ID == 0 ){
	  if( p_state->num_w_id == 0 ){
	    PEXtk_create_all_id( 0, RootWindow(dpy,DefaultScreen(dpy)));
	    p_state->win[p_state->cur_w].w_id =
	    			RootWindow(dpy,DefaultScreen(dpy));
	  } else 	/* some type of error, try to recover */
	    PEXtk_copy_all_id( 0, p_state->cur_w);
	}

	if( usemap != NULL ){
	    cmap = *usemap;
	    PEXtk_PCSetColormap( p_state->win[p_state->cur_w].Rdr_ID, 
		&pc, &lut, &approx, gamma_value);
	} else
            PEXtk_CreatePseudoColormap( p_state->win[p_state->cur_w].Rdr_ID, 
		&pc, &lut, &approx, gamma_value);

	p_state->rdrAttrib.color_approx_table = lut;
	p_state->rdrAttrib.pipeline_context  = pc;

	p_state->pc_ID      = pc;
	p_state->lut_approx = lut;
	if( base_pixel != NULL ){
#ifdef DEBUG
	    if( *base_pixel != -1 && approx.base_pixel != *base_pixel )
	    fprintf( stderr, "Unable to use base_pixel of %d, changed to %d\n",
		*base_pixel,approx.base_pixel);
#endif
	    *base_pixel = approx.base_pixel;
	}

	p_x_data->cmap = cmap;
	if( usemap != NULL )
	    if( *usemap == 0 )
		*usemap = cmap;

	return (0);
}

/*
 *
 *                      +----------------------------+
 * ---------------------+ PEXtk_CreatePsuedoColormap +------------------------
 *                      +----------------------------+
 *
 *  PURPOSE:
 *	This routine will create a gamma corrected pseudo colormap.  
 *
 *  USAGE:
 *	PEXtk_CreatePsuedoColormap( rdr, pc, lut, approx, gamma_value )
 *
 *	PEXRenderer rdr		(in)	Renderer id
 *	PEXPipelineContext *pc	(in)	Pointer to pc associated with renderer
 *					(if 0, one will be created ).
 *	PEXLookupTable *lut	(in)	Pointer to Color Approx LUT 
 *	PEXColorApproxEntry 
 *			*approx (in)	Pointer to Color Approx Values 
 *	Fdata	    gamma_value	(in)	Gamma value, if 0 DEFAULT_GAMMA is used
 *
 *
 *  DESCRIPTION:
 *	This routine will generate a pseudo colormap by attempting to
 *	allocate color cells from the current (Default) colormap. If 
 *	there is not enough room, then a new colormap is allocated.
 *	Note that separate rgb gamma tables are kept if a client needs to
 *	use different values for each red, green, blue table.
 *
 *  RETURNS:
 *	Colormap id if successful, NULL if error
 */
 
/*ARGSUSED*/
Colormap 
PEXtk_CreatePseudoColormap(rdr,pc,lut,approx,gamma_value)
PEXRenderer		rdr;
PEXPipelineContext	*pc;
PEXLookupTable		*lut;
PEXColorApproxEntry	*approx;
Fdata			gamma_value;
{

	Window		root;		/* root window of display */
	Window		window;		/* current drawable */
	int		i, status;
	XColor		pix_color[256];
	unsigned long	pixels[256];
	unsigned long	planeMasks[24];
	unsigned short	rgb_table[256];
	double		igamma, value, num, inc;
	int		two_to_16th = 1 << 16;
	int		colorCount;	/* number of colors in color ramp */
	int		maxInten;
	int		breakPixel;	/* for finding contiguous cells */
	PEXPCAttributes pcAttribs;
	unsigned long	masks[3];	/* PC attribs mask */
	int		baseAlloc;
	int		baseCount=0;
	Colormap	default_cmap;
	PEXRendererAttributes	rdrAttribs;

#ifdef NO_REMAP
	if( cmap != 0 )
	    return (Colormap) (cmap);
#endif
/*
 * Initialize data
 */

	window = p_state->win[p_state->cur_w].w_id;

	root = RootWindow(dpy,DefaultScreen(dpy));

/*
 * Get the maximum intensity value
 */

	*lut = PEXCreateLookupTable(dpy, window, PEXLUTColorApprox);
	maxInten = CellsOfScreen(DefaultScreenOfDisplay(dpy)) - 1;

/*
 * Get the default colormap, it will be needed to get some base colors
 */
	default_cmap = DefaultColormapOfScreen( ScreenOfDisplay (dpy, 
		DefaultScreen (dpy)) );

/*
 * Allocate the colormap
 */

	colorCount = (approx->max1+1) * (approx->max2+1) * (approx->max3+1);
	if( (int) approx->base_pixel < (int) 0 )
	    baseAlloc = BASEALLOC;
	else
	    baseAlloc = approx->base_pixel;

	/*
	 *  If we're using the default colormap, then allocate a little more 
 	 *  than enough cells in order to insure a set of contiguous cells 
 	 *  (not just color planes).
 	 */
	 cmap = default_cmap;
	 status = XAllocColorCells(dpy,cmap,1,planeMasks,0,pixels,
		colorCount+ENUF_CELLS);
	 if (!status) {

	/* 
 	 *  An error occurred, we can't allocate the cells. Try to
 	 *  create a new colormap, and use the core cells from it.
 	 */

	    cmap = XCreateColormap(dpy, root, p_x_data->win_visual, AllocNone);

	    if( cmap == 0 ){
		if( p_state->debug_mode )
		    fprintf(stderr,"Error... creating base colormap\n");
		return (Colormap) 0;
	    }

	/* 
 	 *  Calculate how many left over colormap locations there are.
	 *  Try to get as many of the previous colormap entries into
	 *  the new colormap.
	 */
	    baseCount = 255 - colorCount - ENUF_CELLS;

	    if( baseCount < PREALLOC )
		baseCount = baseCount+ENUF_CELLS;
	    else if( baseCount > baseAlloc )
		baseCount = baseAlloc;

	/*
	 * allocate and set aside the first baseCount cells 
	 * for existing default colors
	 */

	    status = XAllocColorCells(dpy, cmap, 1, planeMasks, 0, pixels,
			baseCount);

	    if( status == 0 ){
		if( p_state->debug_mode )
		    fprintf(stderr,"Error... allocating base colormap\n");
		return (Colormap) NULL;
	    }

	    status = XAllocColorCells(dpy, cmap, 1, planeMasks, 0, 
			&pixels[baseCount], colorCount);

	    if( status == 0 ){
		if( p_state->debug_mode )
		    printf("Error... Allocating upper colormap\n");
		return (Colormap) NULL;
	    }

	    XSetWindowColormap(dpy, window, cmap);
	    newmap = 1;	/* set flag indicating a new colormap was allocated */

	} else {
	    newmap = 0;
	    baseCount = 0;
	}

	/*
	 * The color cells have been allocated, now look for the set of 
	 * contiguous cells because X may return back some cells which are 
	 * left over from other planes. The planes are continuous however.
	 */

	for (i=0,breakPixel=0;((i<colorCount-1) && (!breakPixel));i++) {
		if (pixels[i] != pixels[i+1]-1)
		   breakPixel = i+1;
	}
		
	/* for (i = baseCount; i < colorCount; i++)  */
	for (i = baseCount; i < colorCount+baseCount; i++) {
		pixels[i] = pixels[i+breakPixel];
		pix_color[i].pixel = pixels[i];
	}
	    /* approx->base_pixel = pixels[0]; */
	    approx->base_pixel = pixels[baseCount];

	/* compute a gamma-corrected color list */

	if (gamma_value == 0.0)
	    gamma_value = DEFAULT_GAMMA;
	igamma = 1.0/gamma_value;
	inc = (double) two_to_16th / (double) maxInten;
	num = inc / 2.0;
	for (i = 0; i <= maxInten; i++) {
		value = num / (double) (two_to_16th - 1);
		rgb_table[i] =
			(unsigned short) (pow(value, igamma) * two_to_16th);
		num += inc;
	}

	if( approx->mult1 > approx->mult3)
	    PEXtk_Setup_Pixel_RGB_Data( maxInten, baseCount, approx, 
		rgb_table, pix_color );
	else
	    PEXtk_Setup_Pixel_BGR_Data( maxInten, baseCount, approx, 
		rgb_table, pix_color );

/* 
 * Store the colors
 */

      {
#ifdef PRINT_CMAP
	FILE *fpfoo, *fopen();
	if(( fpfoo=fopen("cmap.out","w")) == (FILE *)NULL) 
	    printf("Can;t open cmap file\n");
	else fprintf(fpfoo," colormap has %d entries, base %d\n",
		colorCount, baseCount);
#endif

	/*
	 * If a new colormap was created, get the first "pre-allocated"
	 * colors, and save them
	 */
	if( newmap != 0 )
	{
           for (i = 0; i < baseCount; i++)
           {
                pix_color[i].pixel = i;
		XQueryColor( dpy, default_cmap, &pix_color[i]);
#ifdef PRINT_CMAP
	     if( fpfoo)
	      fprintf(fpfoo," Query %d, pix %d  rgb %d %d %d\n",i,
		pix_color[i].pixel,pix_color[i].red,pix_color[i].green,
		pix_color[i].blue);
#endif
           }
	}

	if( newmap )
	    XStoreColors(dpy, cmap, pix_color, (colorCount+baseCount));
	else
	    XStoreColors(dpy, cmap, &pix_color[baseCount], colorCount);

	/*
	 * What needs to be done is to get the colors that are stored,
	 * and put the 'real' colors into the rgb lookup table, which
	 * will correspond to what actually is available. TBD
	 */
	/* XQueryColors( dpy, cmap, pix_color, (colorCount+baseCount)); */

	/*
	 * Get the pixel values for the created colormap and save them
	 */
	for(i=0;i<colorCount;i++){
	    lkup_pix[i] = pix_color[i+baseCount].pixel;
	}
#ifdef PRINT_CMAP
	for(i=0;i<colorCount+baseCount;i++){
	   if( fpfoo)
	    fprintf(fpfoo," pix %d i %d rgb %d %d %d : def %d, %d %d %d\n",
		lkup_pix[i], i,lkup_rgb[i][0],lkup_rgb[i][1],lkup_rgb[i][2],
		pix_color[i].pixel,pix_color[i].red,pix_color[i].green,
		pix_color[i].blue);
	}
	fclose(fpfoo);
#endif
      }

	XFlush(dpy);

/* 
 * Set PEX to use this colormap
 */

	/* 
	 * Update the PipelineContext (pc) to new color approximation table.
	 */

	pcAttribs.color_approx_index = 1;
	masks[0] = masks[1] = masks[2] = 0;
	PEXSetPCAttributeMask(masks,PEXPCColorApproxIndex);
	if (*pc)
	    PEXChangePipelineContext(dpy,*pc,masks,&pcAttribs);
	else {
	    *pc = PEXCreatePipelineContext(dpy, masks, &pcAttribs);
	    rdrAttribs.pipeline_context = *pc;
	    PEXChangeRenderer(dpy, rdr, PEXRAPipelineContext, &rdrAttribs);
	}
	rdrAttribs.color_approx_table = *lut;
	PEXChangeRenderer(dpy, rdr, PEXRAColorApproxTable, &rdrAttribs);
	PEXSetTableEntries( dpy, *lut, 1, 1, PEXLUTColorApprox, approx);

	/*
	 * save the approx structure data
	 */
	memcpy((char*)&col_approx, (char*)approx,sizeof(PEXColorApproxEntry));

	return (Colormap) cmap;

}

/*
 *
 *                      +----------------------------+
 * ---------------------+    PEXtk_PCSetColormap     +------------------------
 *                      +----------------------------+
 *
 *  PURPOSE:
 *	This routine will setup the pc and lut for a pseudo colormap.  
 *
 *  USAGE:
 *	PEXtk_PCSetColormap( rdr, pc, lut, approx, gamma_value )
 *
 *	PEXRenderer rdr		(in)	Renderer id
 *	PEXPipelineContext *pc	(in)	Pointer to pc associated with renderer
 *					(if 0, one will be created ).
 *	PEXLookupTable *lut	(in)	Pointer to Color Approx LUT 
 *	PEXColorApproxEntry 
 *			*approx (in)	Pointer to Color Approx Values 
 *	Fdata	    gamma_value	(in)	Gamma value.
 *
 *
 *  DESCRIPTION:
 *	This routine will read a pseudo colormap by attempting to
 *	XQueryColors from the specified (p_state->cmap) colormap. 
 *	It will save the pixel values associated with the rgb table.
 *
 *  RETURNS:
 *	0 id if successful, 1 if error
 */
 
/*ARGSUSED*/
int 
PEXtk_PCSetColormap(rdr,pc,lut,approx,gamma_value)
PEXRenderer		rdr;
PEXPipelineContext	*pc;
PEXLookupTable		*lut;
PEXColorApproxEntry	*approx;
Fdata			gamma_value;
{

	Window		window;		/* current drawable */
	unsigned short	rgb_table[256];
	int		i;
	int		two_to_16th = 1 << 16;
	double		igamma, value, num, inc;
	int		maxInten;
	XColor		pix_color[256];
	PEXPCAttributes pcAttribs;
	unsigned long	masks[3];	/* PC attribs mask */
	int		colorCount;
	int		base_pixel;
	PEXRendererAttributes	rdrAttribs;


	window = p_state->win[p_state->cur_w].w_id;

	if( *lut == (PEXLookupTable) NULL )
	    *lut = PEXCreateLookupTable(dpy, window, PEXLUTColorApprox);

	colorCount = (approx->max1+1) * (approx->max2+1) * (approx->max3+1);
	base_pixel = approx->base_pixel;
	maxInten = CellsOfScreen(DefaultScreenOfDisplay(dpy)) - 1;

	if (gamma_value == 0.0)
	    gamma_value = DEFAULT_GAMMA;
	igamma = 1.0/gamma_value;
	inc = (double) two_to_16th / (double) maxInten;
	num = inc / 2.0;
	for (i = 0; i <= maxInten; i++) {
		value = num / (double) (two_to_16th - 1);
		rgb_table[i] =
			(unsigned short) (pow(value, igamma) * two_to_16th);
		num += inc;
	}

	if( approx->mult1 > approx->mult3)
	    PEXtk_Setup_Pixel_RGB_Data( maxInten, 0, approx, 
		rgb_table, pix_color );
	else
	    PEXtk_Setup_Pixel_BGR_Data( maxInten, 0, approx, 
		rgb_table, pix_color );

	/*
	 * What needs to be done is to get the colors that are stored,
	 * and put the 'real' colors into the rgb lookup table, which
	 * will correspond to what actually is available. TBD
	 */

	for(i=0;i<colorCount;i++){
	    pix_color[i].pixel = i + base_pixel;
	    /* XQueryColor( dpy, cmap, &pix_color[i]); */
	}
	XQueryColors( dpy, cmap, pix_color, colorCount);

	for(i=0;i<colorCount;i++){	/* save the pixel values */
	    lkup_pix[i] = pix_color[i].pixel;
	}

#ifdef PRINT_CMAP
	for(i=0;i<colorCount;i++){
	    printf(" Pix %d i %d rgb %d %d %d : def %d, %d %d %d\n",
		lkup_pix[i], i,lkup_rgb[i][0],lkup_rgb[i][1],lkup_rgb[i][2],
		pix_color[i].pixel,pix_color[i].red,pix_color[i].green,
		pix_color[i].blue);
	}
#endif
	/* 
	 * Update the PipelineContext (pc) to new color approximation table.
	 */

	pcAttribs.color_approx_index = 1;
	masks[0] = masks[1] = masks[2] = 0;
	PEXSetPCAttributeMask(masks,PEXPCColorApproxIndex);
	if (*pc)
	    PEXChangePipelineContext( dpy,*pc,masks,&pcAttribs);
	else {
	    *pc = PEXCreatePipelineContext( dpy, masks, &pcAttribs);
	    rdrAttribs.pipeline_context = *pc;
	    PEXChangeRenderer( dpy, rdr, PEXRAPipelineContext,&rdrAttribs);
	}
	rdrAttribs.color_approx_table = *lut;
        p_state->rdrAttrib.color_approx_table = *lut;
	PEXChangeRenderer( dpy, rdr, PEXRAColorApproxTable, &rdrAttribs);
	PEXSetTableEntries( dpy, *lut, 1, 1, PEXLUTColorApprox, approx);

	/*
	 * save the approx structure data
	 */
	memcpy((char*)&col_approx, (char*)approx,sizeof(PEXColorApproxEntry));

	return 0;
}

/*
 *
 *                      +----------------------------+
 * ---------------------+ PEXtk_Setup_Pixel_RGB_Data +------------------------
 *                      +----------------------------+
 *
 *  PURPOSE:
 *	This routine will setup the XColor data for a RGB ramp.
 *
 *  USAGE:
 *	PEXtk_Setup_Pixel_RGB_Data( maxInten, base, approx, Gamma, pix_color )
 *
 *	int 	    maxInten	(in)	max intensity for a pixel
 *	int	    base	(in)	base index to start at
 *	PEXColorApproxEntry 
 *			*approx (in)	Pointer to Color Approx Values 
 *	unsigned short rgb_tbl 	(in)	gamma corrected rgb table
 *	XColor	    *pix_color	(in)	Color values
 *
 *
 *  DESCRIPTION:
 *	This routine will calculate the XColor rgb color values for the
 *	XQueryColor routine.
 *
 *  NOTE:
 *	This module does not support HSV color.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_Setup_Pixel_RGB_Data( maxInten, baseCount, approx, rgb_table,pix_color)
int			maxInten;
int			baseCount;
PEXColorApproxEntry	*approx;
unsigned short		*rgb_table;
XColor			*pix_color;
{
	int		r, g, b;
	int		i;
	double		rinc, ginc, binc;

	rinc = maxInten / (double)( approx->max1 + 1);
	ginc = maxInten / (double)( approx->max2 + 1);
	binc = maxInten / (double)( approx->max3 + 1);
	i = 0;
	for (r = 0; r <= (int) approx->max1; r++)
	    for (g = 0; g <= (int) approx->max2; g++)
		for (b = 0; b <= (int) approx->max3; b++) {
		    pix_color[i+baseCount].red   = 
			rgb_table[(int)((double)r * rinc)];
		    pix_color[i+baseCount].green = 
			rgb_table[(int)((double)g * ginc)];
		    pix_color[i+baseCount].blue  = 
			rgb_table[(int)((double)b * binc)];
		    pix_color[i+baseCount].flags = DoRed | DoGreen | DoBlue;
		    lkup_rgb[i][0]= r*maxInten/approx->max1;
		    lkup_rgb[i][1]= g*maxInten/approx->max2;
		    lkup_rgb[i][2]= b*maxInten/approx->max3;
		    i++;
		}
	return 0;
}

/*
 *
 *                      +----------------------------+
 * ---------------------+ PEXtk_Setup_Pixel_BGR_Data +------------------------
 *                      +----------------------------+
 *
 *  PURPOSE:
 *	This routine will setup the XColor data for a BGR ramp.
 *
 *  USAGE:
 *	PEXtk_Setup_Pixel_BGR_Data( maxInten, base, approx, Gamma, pix_color )
 *
 *	int 	    maxInten	(in)	max intensity for a pixel
 *	int	    base	(in)	base index to start at
 *	PEXColorApproxEntry 
 *			*approx (in)	Pointer to Color Approx Values 
 *	unsigned short rgb_tbl 	(in)	gamma corrected rgb table
 *	XColor	    *pix_color	(in)	Color values
 *
 *
 *  DESCRIPTION:
 *	This routine will calculate the XColor rgb color values for the
 *	XQueryColor routine.
 *
 *  NOTE:
 *	This module does not support HSV color.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_Setup_Pixel_BGR_Data( maxInten, baseCount, approx, rgb_table,pix_color)
int			maxInten;
int			baseCount;
PEXColorApproxEntry	*approx;
unsigned short		*rgb_table;
XColor			*pix_color;
{
	int		r, g, b;
	int		i;
	double		rinc, ginc, binc;

	rinc = maxInten / (double)( approx->max1 + 1);
	ginc = maxInten / (double)( approx->max2 + 1);
	binc = maxInten / (double)( approx->max3 + 1);
	i = 0;
	for (b = 0; b <= (int) approx->max3; b++) 
	    for (g = 0; g <= (int) approx->max2; g++)
		for (r = 0; r <= (int) approx->max1; r++){
		    pix_color[i+baseCount].red   = 
			rgb_table[(int)((double)r * rinc)];
		    pix_color[i+baseCount].green = 
			rgb_table[(int)((double)g * ginc)];
		    pix_color[i+baseCount].blue  = 
			rgb_table[(int)((double)b * binc)];
		    pix_color[i+baseCount].flags = DoRed | DoGreen | DoBlue;
		    lkup_rgb[i][0]= r*maxInten/approx->max1;
		    lkup_rgb[i][1]= g*maxInten/approx->max2;
		    lkup_rgb[i][2]= b*maxInten/approx->max3;
		    i++;
		}
	return 0;
}

/*
 *
 *                      +----------------------------+
 * ---------------------+     PEXtk_SetColorTable    +------------------------
 *                      +----------------------------+
 *
 *  PURPOSE:
 *	This routine will setup the lut data for a color index LUT.  
 *
 *  USAGE:
 *	PEXtk_SetColorTable( rdr, index, col )
 *
 *	PEXRenderer rdr		(in)	Renderer id
 *	int	    index	(in)	Index into table to put color at
 *	PCOLOR	    *col	(in)	Color values
 *
 *
 *  DESCRIPTION:
 *	This routine will set the color lookup table values for the
 *	p_state->lut_index LUT. If the lut has not been created yet, it
 *	will be created.
 *
 *  NOTE:
 *	This module does not support HSV color.
 *
 *  RETURNS:
 *	0 id if successful, 1 if error
 */
 
/*ARGSUSED*/
int 
PEXtk_SetColorTable(rdr, index, col)
PEXRenderer	rdr;
int		index;
PCOLOR		*col;
{

	Window			window;		/* current drawable */
	PEXColorEntry 		clr_data;
	PEXRendererAttributes	rdrAttribs;

	window = p_state->win[p_state->cur_w].w_id;

	if( p_state->lut_index == (PEXLookupTable) 0 ){
	    p_state->lut_index = PEXCreateLookupTable(dpy, window, PEXLUTColor);
            p_state->rdrAttrib.color_table = p_state->lut_index;
	    rdrAttribs.color_table = p_state->lut_index;
	    PEXChangeRenderer(dpy, rdr, PEXRAColorTable, &rdrAttribs);
	}
	clr_data.type = PEXColorTypeRGB;
	if( col->type == PEXtk_COLOR_HSV ){
	    PEXtk_convert_hsv_to_rgb(  col->color.hsv.hue,
			col->color.hsv.saturation,
			col->color.hsv.value,
			&clr_data.value.rgb.red,
			&clr_data.value.rgb.green,
			&clr_data.value.rgb.blue);
	} else {
	    clr_data.value.rgb.red   = col->color.rgb.red;
	    clr_data.value.rgb.green = col->color.rgb.green;
	    clr_data.value.rgb.blue  = col->color.rgb.blue;
	}
	PEXSetTableEntries( dpy, p_state->lut_index, index, 1,
		PEXLUTColor, &clr_data);

	return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      Pset_colormap       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	This routine will set the current colormap.
 *
 *  USAGE:
 *	Pset_colormap( id )
 *	Colormap    id		(in)	colormap id.
 *
 *  DESCRIPTION:
 *	This module sets the current colormap. If the colormap is set
 *	before Pset_drawable is called, then Pset_color_equation will
 *	not be called. This routine would be used for 8-bit systems
 *	only.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
Pset_colormap( id )
Colormap	id;
{
    p_x_data->cmap = id;
    return 0;
}

/*
 *
 *                      +----------------------------+
 * ---------------------+      Ppixel_value_rgb      +------------------------
 *                      +----------------------------+
 *
 *  PURPOSE:
 *	Calculates the pixel value for a given rgb triplet.
 *
 *  USAGE:
 *	Ppixel_value_rgb( red, green, blue )
 *	Fdata 	    red		(in)	Red value
 *	Fdata 	    green	(in)	Green value
 *	Fdata 	    blue	(in)	Blue value
 *
 *
 *  DESCRIPTION:
 *	This routine will calculate the pixel value for the input
 *	red, green, blue values.
 *
 *  NOTE:
 *	This module does not support HSV color.
 *
 *  RETURNS:
 *	color pixel value 
 */

/*ARGSUSED*/
long
Ppixel_value_rgb( r, g, b)
Fdata	r;
Fdata	g;
Fdata	b;
{
    unsigned long pix;

    /*
     * For now, only 8-bit pseudo color is supported.
     */
    if( p_x_data->win_depth == 8 ){
	pix = Ppseudo_color( r, g, b);
	return pix;
    }
    switch( p_x_data->c_class ){
	case TrueColor:
	case DirectColor:
		pix = Ptrue_color( r, g, b);
		break;
	default:
		pix = Ppseudo_color( r, g, b);
		break;
    }
    return pix;
}

/*
 *
 *                      +----------------------------+
 * ---------------------+       Ppseudo_color        +------------------------
 *                      +----------------------------+
 *
 *  PURPOSE:
 *	Calculates the pixel value for rgb triplet.
 *
 *  USAGE:
 *	Ppseudo_color( red, green, blue )
 *	Fdata 	    red		(in)	Red value
 *	Fdata 	    green	(in)	Green value
 *	Fdata 	    blue	(in)	Blue value
 *
 *
 *  DESCRIPTION:
 *	This routine will calculate the pixel value for the input
 *	red, green, blue values.
 *
 *  NOTE:
 *	This module does not support HSV color.
 *
 *  RETURNS:
 *	color pixel value 
 */
 
/*ARGSUSED*/
long
Ppseudo_color( r, g, b)
Fdata	r;
Fdata	g;
Fdata	b;
{

    int index;
    long pix;

    if( p_x_data->win_depth == 8 ){
	index = (int) 
	  ((int)((int)((r)*col_approx.max1)  * col_approx.mult1) + 
	  ((int)((int)((g)*col_approx.max2)) * col_approx.mult2) +
	  ((int)((int)((b)*col_approx.max3)) * col_approx.mult3));
	pix =  (long) lkup_pix[index]; /* base pixel is zero */
    } else {
	pix = (unsigned long)
	      (((unsigned long)((r) * 255) << p_x_data->red_shift) |
	      (( unsigned long)((g) * 255) << p_x_data->green_shift) | 
	      (  unsigned long)((b) * 255) << p_x_data->blue_shift);
    }

#ifdef PSEUDO_8_TO_24_COLOR
	/* good values for an 8 bit colormap to simulate 24 bit truecolor */
	index =  ((int) ((int)((int)((r)*5) * 8) + 
	    ((int)((int)((g)*7))) + (int)(((int)((b)*4)) * 48));

	index = (int)  ((int)((int)((r)*5) * 36) + 
	  ((int)((int)((b)*5))) + (int)(((int)((g)*5)) * 6));
#endif

	return pix;
}

/*
 *
 *                      +----------------------------+
 * ---------------------+        Ptrue_color         +------------------------
 *                      +----------------------------+
 *
 *  PURPOSE:
 *	Calculates the pixel value for true color visuals.
 *
 *  USAGE:
 *	Ptrue_color( red, green, blue )
 *	Fdata 	    red		(in)	Red value
 *	Fdata 	    green	(in)	Green value
 *	Fdata 	    blue	(in)	Blue value
 *
 *
 *  DESCRIPTION:
 *	This routine will calculate the pixel value for the input
 *	red, green, blue values.
 *
 *  NOTE:
 *	This module does not support HSV color.
 *
 *  RETURNS:
 *	color pixel value 
 */
 
/*ARGSUSED*/
long
Ptrue_color( r, g, b)
	Fdata	r;
	Fdata	g;
	Fdata	b;
{
    long pix;

	pix = (unsigned long)
	      (((unsigned long)((r) * 255) << p_x_data->red_shift) |
	      (( unsigned long)((g) * 255) << p_x_data->green_shift) | 
	      (  unsigned long)((b) * 255) << p_x_data->blue_shift);

	return pix;
}

/*
 *
 *                      +----------------------------+
 * ---------------------+        Pindex_color        +------------------------
 *                      +----------------------------+
 *
 *  PURPOSE:
 *	Calculates the pixel value for an input index value.
 *
 *  USAGE:
 *	Pindex_color( index )
 *	int 	    index	(in)	index location
 *
 *
 *  DESCRIPTION:
 *	This routine will calculate the pixel value for the input
 *	index value.
 *
 *  NOTE:
 *	This is the default module for PEXtk_FUNC_PIXEL_INDEX
 *	This module only supports 8 bit color
 *
 *  RETURNS:
 *	pixel value 
 */
 
/*ARGSUSED*/
long
Pindex_color( index )	/* get pixel at index */
int	index;
{
    if( p_x_data->win_depth == 8 ){
	return (long) lkup_pix[index];
    } else
	return 0;
}

/*
 *
 *                      +----------------------------+
 * ---------------------+      PEXtk_rgb_index       +------------------------
 *                      +----------------------------+
 *
 *  PURPOSE:
 *	Calculates the rgb value for an input index color value.
 *
 *  USAGE:
 *	PEXtk_rgb_index( col )
 *	PCOLOR 	    *col	(in)	PCOLOR structure with index in 
 *					col->color.index.value.
 *					The rgb values are placed in 
 *					col->color.rgb.red,col->color.rgb.green, 
 *					col->color.rgb.blue.
 *
 *  DESCRIPTION:
 *	This routine will return the rgb values for the index color
 *	value stored in the color table set with Pset_color_table. If
 *	no color table has been setup, 0 is returned.
 *
 *  NOTE:
 *	This module only supports index color mode
 *
 *  RETURNS:
 *	RGB values in color_table
 *	0 if successful, 1 if error
 */
 
/*ARGSUSED*/
int
PEXtk_rgb_index( col )		/* get rgb at index color location */
PCOLOR	*col;
{

    if(DB_get_color_values( col->color.index.value, col)){
	return 1;
    } else
	return 0;
}

/*
 *
 *                      +----------------------------+
 * ---------------------+    PEXtk_pseudo_index      +------------------------
 *                      +----------------------------+
 *
 *  PURPOSE:
 *	Calculates the rgb value for an input index value.
 *
 *  USAGE:
 *	PEXtk_pseudo_index( index )
 *	int 	    index	(in)	index location
 *
 *
 *  DESCRIPTION:
 *	This routine will return the pixel value for the rgb data
 *	stored in the color table set with Pset_color_table. If
 *	no color table has been setup, 0 is returned.
 *
 *  NOTE:
 *	This module only supports 8 bit color
 *
 *  RETURNS:
 *	best possible pixel value
 */
 
/*ARGSUSED*/
long
PEXtk_pseudo_index( index )	/* get pixel at index color location */
int	index;
{
    PCOLOR col;
    long pix;

    if( p_x_data->win_depth == 8 ){
	if(DB_get_color_values( index, &col)){
	    return (long) lkup_pix[index];
	}
	pix = Ppseudo_color( col.color.rgb.red,
		col.color.rgb.green, col.color.rgb.blue);
	return pix;
    } else
	return 0;
}

/*
 *
 *                      +----------------------------+
 * ---------------------+      PEXtk_true_index      +------------------------
 *                      +----------------------------+
 *
 *  PURPOSE:
 *	Calculates the rgb value for an input index value for true color.
 *
 *  USAGE:
 *	PEXtk_true_index( index )
 *	int 	    index	(in)	index location
 *
 *
 *  DESCRIPTION:
 *	This routine will return the pixel value for the rgb data
 *	stored in the color table set with Pset_color_table. If
 *	no color table has been setup, 0 is returned.
 *
 *  NOTE:
 *	This module could return 0 for a valid index location
 *
 *  RETURNS:
 *	0 if no color table setup, otherwise pixel value
 */
 
/*ARGSUSED*/
long
PEXtk_true_index( index )	/* get pixel at index color location */
int	index;
{
    PCOLOR col;
    long pix;

	if(DB_get_color_values( index, &col)){
	    return 0;
	}
	pix = Ptrue_color( col.color.rgb.red,
		col.color.rgb.green, col.color.rgb.blue);
	return pix;
}

/*
 *
 *                      +----------------------------+
 * ---------------------+  PEXtk_rgb_colormap_index  +------------------------
 *                      +----------------------------+
 *
 *  PURPOSE:
 *	Calculates the rgb value for an input index value into the
 *	colormap table.
 *
 *  USAGE:
 *	PEXtk_rgb_colormap_index( index, red, green, blue )
 *	int 	    index	(in)	index location
 *	Fdata 	    *red	(out)	Red value
 *	Fdata 	    *green	(out)	Green value
 *	Fdata 	    *blue	(out)	Blue value
 *
 *
 *  DESCRIPTION:
 *	This routine will calculate the rgb value for the input
 *	index value. It is computed from the X Colormap table.
 *
 *  NOTE:
 *	This module only supports 8 bit color
 *
 *  RETURNS:
 *	0 if successful, non-zero if error
 */
 
/*ARGSUSED*/
int
PEXtk_rgb_colormap_index( index, r, g, b)	/* get rgb at index */
int	index;
Fdata	*r;
Fdata	*g;
Fdata	*b;
{
    if( p_x_data->win_depth == 8 ){
	*r = (float) ((float) lkup_rgb[index][0])/255.0;
	*g = (float) ((float) lkup_rgb[index][1])/255.0;
	*b = (float) ((float) lkup_rgb[index][2])/255.0;
	return 0;
    } else
	return 1;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+ PEXtk_convert_hsv_to_rgb +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Converts an HSV triplet to RGB triplet.
 *
 *  USAGE:
 *      PEXtk_convert_hsv_to_rgb( h, s, v, r, g, b )
 *      Fdata	    h		(in)	Hue 
 *      Fdata	    s		(in)	Saturation 
 *      Fdata	    v		(in)	Value  (Intensity)
 *      Fdata	    *r		(in)	Pointer to red value
 *      Fdata	    *g		(in)	Pointer to green value
 *      Fdata	    *b		(in)	Pointer to blue value
 *
 *  DESCRIPTION:
 *	Routine to convert HSV to RGB
 *	Ref: Computer Graphics, Principle and Practice
 *			Foley, van Dam, Feiner, Hughes
 *	Page 593
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_convert_hsv_to_rgb( h, s, v, r, g, b)
Fdata	h;
Fdata	s;
Fdata	v;
Fdata	*r;
Fdata	*g;
Fdata	*b;
{
    int i;
    float f, p, q, t;

    if( s == 0.0 ){	/* if h is defined, this would be in error */
	*r = *g = *b = v;
    } else {
	if( h == 360.0 ) h = 0.0;
	h = h / 60.0;
	i = floor((double) h);
	f = h - i;
	p = v * ( 1.0 - s);
	q = v * ( 1.0 - ( s * f ));
	t = v * ( 1.0 - ( s * ( 1.0 - f ) ) );
	switch (i) {
	    case 0:
		*r = v; *g = t; *b = p;
		break;
	    case 1:
		*r = q; *g = v; *b = p;
		break;
	    case 2:
		*r = p; *g = v; *b = t;
		break;
	    case 3:
		*r = p; *g = q; *b = v;
		break;
	    case 4:
		*r = t; *g = p; *b = v;
		break;
	    case 5:
		*r = v; *g = p; *b = q;
		break;
	} /* end switch */
    } /* end if */

    return 0;
}

int
PEXtk_FreeColormap()
{
    if( newmap )
	XFreeColormap( dpy, cmap);
    newmap = 0;
    cmap = 0;
    return 0;
}
