/*
 *  $Header: /auto/nfs/sequoia/u/an/work/rws/pextk/src/pextk_pck.c,v 1.3 1994/01/04 00:05:58 rws Exp $
 ******************************************************************************
 ******************************************************************************
 *****                          pextk_pck.c                               *****
 *****									  *****
 *****									  *****
 *****									  *****
 *                                                                            *
 *	Copyright 1992, 1993. All Rights Reserved                             *
 *			SHOgraphics					      *
 *			321 Soquel Way           			      *
 *			Sunnyvale, CA     94086   			      *
 *                                                                            *
 *****									  *****

    Permission to use, copy, modify, and distribute this software and
 its documentation for any purpose and without fee is hereby granted,
 provided that the above copyright notice appear in all copies and that
 both that copyright notice and this permission notice appear in
 supporting documentation, and that the name of SHOgraphics not be
 used in advertising or publicity pertaining to distribution of the 
 software without specific, written prior permission.

 SHOGRAPHICS DISCLAIMS ANY AND ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 INCLUDING ALL EXPRESSED AND IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS. IN NO EVENT SHALL SHOGRAPHICS BE LIABLE FOR ANY DAMAGES OF ANY
 KIND, INCLUDING BUT NOT LIMITED TO SPECIAL, INDIRECT OR CONSEQUENTIAL 
 DAMAGES RESULTING FROM LOSS OF USE, DATA OR PROSPECTIVE PROFITS, WHETHER 
 IN AN ACTION OF CONTRACT, NEGLIENCE OR OTHER TORTIOUS ACTION, ARISING OUT 
 OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

 ************************************************************************ */
/*
 *
 *  Module Purpose:
 *	This is the module which contains the PEXtk Picking routines
 *
 *  Module Notes:
 *
 *  Module Contents:
 *
 *    void  PEXtk_init_pck    PEXtk internal routine. Initializes pick module.
 *    void  Pbegin_pick       Begins pick mode
 *    int   Pend_pick         Ends pick mode
 *    void  Pbegin_pick_mpex  Begins pick mode for DEC MPEX
 *    int   Pend_pick_mpex    Ends pick mode for DEC MPEX
 *
 *
 *  Module Description:
 *
 */

#include <stdio.h>
#include <math.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <X11/PEX5/PEX.h>
#include <X11/PEX5/PEXproto.h>
#include <X11/PEX5/PEXlib.h>

#ifdef MPEX
#include <X11/MPEX.h>
#include <X11/MPEXproto.h>
#include <X11/MPEXlib.h>
#endif

#include <PEXtk.h>

#include "pextk.h"
#include "pextk_lcl.h"

#ifdef MPEX
	mpxlRendererAttributes 	pickAttrib;
#endif
	PEXPickRecord		pick_record;
	int			max_buf_size;
	int			max_hit_size;

/*
 *
 *                       +--------------------------+
 * ----------------------+      PEXtk_init_pck      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Initializes picking pointers.
 *
 *  USAGE:
 *	PEXtk_init_pck( p_obj )
 *	PEXtk_OBJECT    *p_obj	(in)	Pointer to PEXtk object.
 *
 *  DESCRIPTION:
 *	This routine initializes pointers for p_build and p_state.
 *
 *  RETURNS:
 *	0 if successful, 1 if error
 */
 
/*ARGSUSED*/
int
PEXtk_init_pck(p_obj)
PEXTK_OBJECT		*p_obj;
{
    int i,j;
    /*
     * Save the various object ptrs
     */
    p__object	= p_obj;
    p_state	= p_obj->state;
    p_build	= p_obj->build;
    p_buffer	= p_obj->buffer;
    p_support	= p_obj->support;
    p_dl	= p_obj->dpy_list;
    p_clr	= p_obj->clr_tbl;
    p_x_data	= p_obj->x_data;

    dpy = p_x_data->dpy;

    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+        Pbegin_pick       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Begins pick mode.
 *
 *  USAGE:
 *	Pbegin_pick( x, y, xsize, ysize, max_hits )
 *	int	    x		(in)	The center x position of the pick box
 *	int	    y		(in)	The center y position of the pick box
 *	int	    xsize	(in)	The x size (half width) of the pick box
 *	int	    ysize	(in)	The y size (half width) of the pick box
 *	int	    max_hits	(in)	The maximum number of hits to record
 *
 *  RETURNS:
 *	Nothing
 *
 *  DESCRIPTION:
 *	All primitives that follow a Pbegin_pick until a Pend_pick 
 *	is specified will not be drawn but will be 'hit' tested. Any 
 *	primitive that falls in the pick 'window' will create a 'hit'.
 *
 */
 
/*ARGSUSED*/
void
Pbegin_pick( x, y, xsize, ysize, max_hits )
int	x;
int	y;
int	xsize;
int	ysize;
int	max_hits;
{
    float size;

    PEXEndRendering( dpy, p_state->cur_rdr, True);
    PEXChangeNameSet( dpy, p_state->nameSet,PEXNSReplace, 1, p_state->names);

    size = xsize;
    if( ysize < xsize ) size = ysize; /* use smaller of size */

	pick_record.box.position.x = x;
	pick_record.box.position.y = y;
	pick_record.box.distance   = size/2.0;

	max_hit_size = max_hits;

	PEXtk_check_pc();
        PEXBeginPickAll( dpy, p_state->cur_draw, p_state->cur_rdr, 
		1, PEXPickAllVisible, False, max_hits, PEXPickDeviceDCHitBox,
		&pick_record);

	p_state->sys_mode = PEXtk_PICKING_MODE;
	PEXAddToNameSet( dpy, p_state->cur_rdr, PEXOCRender,
		1, p_state->names);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+          Pend_pick       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Ends pick mode.
 *
 *  USAGE:
 *	Pend_pick( buffer, buf_size, anyMore )
 *	int	    *buffer	(in)	The pick buffer
 *	int	    buf_size	(in)	The array size of the pick buffer.
 *	int	    *anyMore	(in)	True = more hits detected, False = none
 *
 *  RETURNS:
 *	Number of items (hits) in pick buffer
 *
 *  DESCRIPTION:
 *	This routine will return multiple 'sets' of hits, where each set contains
 *	an integer representing the number of hits in the set, followed by
 *	integers describing the pick_id hit list.
 *
 *	More specifically, the pick buffer contains the following:
 *	   for (i=1,number of hit sets in pick buffer)
 *		n	First hit, number of pick ids to follow in this hit set
 *	   	for (i=1,n)
 *			id_1	First pick id
 *			...
 *			id_n	Last pick id for current hit set
 *
 *  EXAMPLE:
 *	For the following (pseuso-code) example:
 *		Pcreate_struct(1)
 *		    Ppick_id(10)
 *			Ppoly(a)
 *		Pclose_struct()
 *		Pcreate_struct(2)
 *		    Ppick_id(20)
 *			Ppoly(b)
 *		Pclose_struct()
 *		# do picking now 
 *		while(1)
 *			Pbegin_pick( pick_box large enough to hit poly a and b)
 *			    Ppick_id(1)
 *			    Pexec_struct(1)
 *			    Ppick_id(2)
 *			    Pexec_struct(2)
 *			n = Pend_pick( hit_list, HIT_LIST_SIZE, &more)
 *			for( n )
 *			  m = hit_list[i]
 *			  i++
 *			  for( m )
 *			    printf hit_list[i], i++
 *
 *	The following would be the hit list (end_pick would return 2):
 *		2
 *		1 10
 *		2
 *		2 20
 *		
 */
 
/*ARGSUSED*/
int
Pend_pick( buffer, buf_size, anyMore )
int	*buffer;
int	buf_size;
int	*anyMore;
{

    int i, j, cnt, stat, moreRet;
    unsigned long count;
    PEXPickPath *pPath;
    PEXPickElementRef *pRef;

    p_state->sys_mode = PEXtk_RENDERING_MODE;

    if((pPath = PEXEndPickAll(dpy, p_state->cur_rdr, &stat,
		&moreRet, &count)) == (PEXPickPath *) NULL){
	PEXtk_reset_pex_state(1, 0);
	return 0;		/* error */
    }

    max_buf_size = buf_size;
    *anyMore = 0;
    cnt = 0;
    for(i=0;i<count; i++){
	pRef = pPath[i].elements;
	if( pRef != NULL ){
		buffer[cnt] = pPath[i].count; cnt++;
	    for(j=0;j<pPath[i].count; j++){
		buffer[cnt] = pRef[j].pick_id;
		if( cnt < max_buf_size )
		    cnt++;
		else {
		    *anyMore = 1;
		    /* save the rest in the buffer, save current state */
		    /* TBD */
		    PEXtk_reset_pex_state(1, 0);
		    PEXFreePickPaths( count, pPath);
		    /* return cnt; */
		    return i;
		}
	    }
	}
    }

    if( moreRet == PEXMoreHits || moreRet == PEXMayBeMoreHits){
	*anyMore = 1;
	if( cnt < max_buf_size ){
		/* get rest of data and put in 'buffer' */
	} else {
		/* get data and put only in local buffer */
	}
    }
    PEXFreePickPaths( count, pPath);
    PEXtk_reset_pex_state(1, 0);
    return count;
}

#ifdef MPEX
/*
 *
 *                       +--------------------------+
 * ----------------------+     Pbegin_pick_mpex     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Begins pick mode.
 *
 *  USAGE:
 *	Pbegin_pick_mpex( x, y, xsize, ysize, buffer, buf_size )
 *	int	    x		(in)	The center x position of the pick box
 *	int	    y		(in)	The center y position of the pick box
 *	int	    xsize	(in)	The x size (half width) of the pick box
 *	int	    ysize	(in)	The y size (half width) of the pick box
 *	int	    max_hits	(in)	The maximum number of hits to record
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pbegin_pick_mpex( x, y, xsize, ysize, max_hits )
int	x;
int	y;
int	xsize;
int	ysize;
int	max_hits;
{
    float size;

    PEXEndRendering( dpy, p_state->cur_rdr, True);
    PEXChangeNameSet( dpy, p_state->nameSet,PEXNSReplace, 1, p_state->names);
    size = xsize;
    if( ysize < xsize ) size = ysize;

	max_hit_size = max_hits;

	PEXtk_check_pc();
	pickAttrib.pickPosition.x = x;
	pickAttrib.pickPosition.y = y;
	pickAttrib.pickDistance   = size/2.0;
	pickAttrib.echoSwitch     = PEXOff;
	pickAttrib.pickInclusion  = p_state->nameSet;
	pickAttrib.pickMaxHits    = 100;
	pickAttrib.pickMaxHits    = max_hits;
	pickAttrib.pickStartPath.pathLength   = 1;
	/* elRef.structure = 0; elRef.offset = 0; */
	/* pickAttrib.pickStartPath.elementRefs   = &elRef; */

        MPEXChangeRenderer( dpy, p_state->cur_rdr,
	    MPEXNRAPickPosition |
	    MPEXNRAPickInclusion|MPEXNRAPickDistance|MPEXNRAEchoSwitch|
	    MPEXNRAPickMaxHits, &pickAttrib);
	    /* MPEXNRAPickStartPath| */

	max_buf_size = buf_size;
        MPEXBeginPickAll( dpy, p_state->cur_rdr, 
		p_state->cur_draw, 1, False);

	p_state->sys_mode = PEXtk_PICKING_MODE;
	PEXtk_reset_attributes();
	PEXAddToNameSet( dpy, p_state->cur_rdr, p_state->req_type, 
		1, p_state->names);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      Pend_pick_mpex      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Ends pick mode.
 *
 *  USAGE:
 *	Pend_pick_mpex( buffer, buf_size, anyMore )
 *	int	    *buffer	(in)	The pick buffer
 *	int	    buf_size	(in)	The array size of the pick buffer.
 *	int	    *anyMore	(in)	True = more hits detected, False = none
 *
 *  RETURNS:
 *	Number of items (hits) in pick buffer
 */
 
/*ARGSUSED*/
int
Pend_pick_mpex( buffer, buf_size, anyMore )
int	*buffer;
int      buf_size;
int	*anyMore;
{

    int i, j, cnt, count, moreRet;
    pxlPickPath *pPath, *last_pick;
    mpxlPickAllPath *paPath;

    if(MPEXEndPickAll(dpy, p_state->cur_rdr,
		&moreRet, &paPath, &count))
	return 0;		/* error */

    p_state->sys_mode = PEXtk_RENDERING_MODE;

    max_buf_size = buf_size;
    cnt = 0;
    for(i=0;i<count; i++){
	pPath = paPath[i].elementRefs;
	if( pPath != NULL ){
	    for(j=0;j<paPath[i].pathLength; j++){
		buffer[cnt] = pPath[j].offset;
		if( cnt < max_buf_size )
		    cnt++;
		else {
		    *anyMore = 1;
		    /* save the rest in the buffer, save current state */
		    /* TBD */
		    /* PEXtk_Begin_Rendering(); */
		    PEXtk_reset_pex_state(1, 0);
		    return i;
		}
	    }
	}
    }

    if( moreRet ){
	if( cnt < max_buf_size ){
		/* get rest of data and put in 'buffer' */
	} else {
		/* get data and put only in local buffer */
	}
    }
    PEXtk_reset_pex_state(1, 0);
    return count;
}
#endif
