/*
 *  $Header: /auto/nfs/sequoia/u/home/rws/pextk/src/pextk_vue.c,v 1.1 1993/10/12 00:07:09 rws Exp $
 ******************************************************************************
 ******************************************************************************
 *****                          pextk_vue.c                               *****
 *****									  *****
 *****									  *****
 *****									  *****
 *                                                                            *
 *	Copyright 1992, 1993. All Rights Reserved                             *
 *			SHOgraphics					      *
 *			321 Soquel Way           			      *
 *			Sunnyvale, CA     94086   			      *
 *                                                                            *
 *****									  *****

    Permission to use, copy, modify, and distribute this software and
 its documentation for any purpose and without fee is hereby granted,
 provided that the above copyright notice appear in all copies and that
 both that copyright notice and this permission notice appear in
 supporting documentation, and that the name of SHOgraphics not be
 used in advertising or publicity pertaining to distribution of the 
 software without specific, written prior permission.

 SHOGRAPHICS DISCLAIMS ANY AND ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 INCLUDING ALL EXPRESSED AND IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS. IN NO EVENT SHALL SHOGRAPHICS BE LIABLE FOR ANY DAMAGES OF ANY
 KIND, INCLUDING BUT NOT LIMITED TO SPECIAL, INDIRECT OR CONSEQUENTIAL 
 DAMAGES RESULTING FROM LOSS OF USE, DATA OR PROSPECTIVE PROFITS, WHETHER 
 IN AN ACTION OF CONTRACT, NEGLIENCE OR OTHER TORTIOUS ACTION, ARISING OUT 
 OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

 ************************************************************************ */
/*
 *
 *  Module Purpose:
 *	This is the module which contains the PEXtk View routines
 *
 *  Module Notes:
 *
 *  Module Contents:
 *
 *    void  Pclip_rectangle   Sets viewport clip rectangle to the rectangle s
 *    void  Pget_clip_rectangle   Returns the current viewport clip rectangle.
 *    int   Pget_view         Returns current view parameters
 *    int   Pget_viewport     Returns current window viewport parameters
 *    void  Plookat_view      Sets the lookat view.
 *    void  Portho_2d_view    Sets the 2d orthographic view position to l,r,b,t
 *    void  Portho_view       Sets the orthographic view position to the frus
 *    void  Ppersp_view       Sets the perspective view to the frustum specif
 *    void  Ppolar_view       Sets the polar view.
 *    void  Ppop_view         Pops the view stack stack. Only Immediate mode 
 *    void  Ppush_view        Pushes the view stack stack. Only Immediate mode
 *    void  Ppop_viewport     Pops the viewport stack.
 *    void  Ppush_viewport    Pushes the viewport stack.
 *    int   Pset_view_index   Sets the view index id.
 *    void  Pviewport         Sets window viewport to the rectangle specified
 *    void  Pwindow_view      Sets the perspective view to the frustum
 *
 *
 *  Module Description:
 *
 *	The viewport stack, view stack, and clip rectangle stack are all
 *	maintained by the routines in this module.
 *
 *	The viewport and clip rectangle stacks use the same stack pointer, 
 *	p_state->vpt_loc.
 *
 *	The view stack, which is accessed using matrix mode PEXtk_MATRIX_PROJ
 *	to modify the view_table.mapping matrix, or the matrix mode
 *	PEXtk_MATRIX_VIEWING to update the view_table.orientation matrix
 *	use the same stack pointer, p_state->view_loc.
 *
 *	When a projection routine is called (Portho_2d_view, Portho_view,
 *	Ppersp_view, or Pwindow_view), an 'id' is passed also. If this 'id'
 *	is zero, then the current stack location, p_state->view_loc, is
 *	used as the index into the view table at which to place the resulting
 *	matrix. This allows for users to simply ignore the fact that the
 *	mapping matrix is being placed in a table, and to have PEXtk
 *	continuously replacing the last projection matrix. 
 *
 */

#include <stdio.h>
#include <math.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <X11/PEX5/PEX.h>
#include <X11/PEX5/PEXproto.h>
#include <X11/PEX5/PEXlib.h>

#ifndef SYSV
#include <memory.h>
#endif

#include <PEXtk.h>

#include "pextk.h"
#include "pextk_lcl.h"
#include "pextk_ops.h"

	static PEXViewEntry	view_table[PEXtk_VIEW_SIZE];
	static PEXtk_VIEWPORT	viewport[PEXtk_VPT_SIZE];
	static PEXDeviceRect	clip[PEXtk_VPT_SIZE];
	static PEXtk_VIEW	view_stk[PEXtk_VIEW_SIZE];
	static Matrix ident_mat  =     {{1.0, 0.0, 0.0, 0.0},
                                        {0.0, 1.0, 0.0, 0.0},
                                        {0.0, 0.0, 1.0, 0.0},
                                        {0.0, 0.0, 0.0, 1.0}};

/*
 *
 *                       +--------------------------+
 * ----------------------+      PEXtk_init_vue      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Initializes picking pointers.
 *
 *  USAGE:
 *	PEXtk_init_vue( p_obj )
 *	PEXtk_OBJECT    *p_obj	(in)	Pointer to PEXtk object.
 *
 *  DESCRIPTION:
 *	This routine initializes pointers for p_build and p_state.
 *
 *  RETURNS:
 *	0 if successful, 1 if error
 */
 
/*ARGSUSED*/
int
PEXtk_init_vue(p_obj)
PEXTK_OBJECT		*p_obj;
{
    int i,j;
    /*
     * Save the various object pointers
     */
    p__object	= p_obj;
    p_state	= p_obj->state;
    p_build	= p_obj->build;
    p_buffer	= p_obj->buffer;
    p_support	= p_obj->support;
    p_dl	= p_obj->dpy_list;
    p_clr	= p_obj->clr_tbl;
    p_x_data	= p_obj->x_data;

    dpy = p_x_data->dpy;

    p_state->view = &view_stk[0];	/* current view stack pointer */
    p_state->vpt  = &viewport[0];	/* current viewport stack pointer */
    p_state->clip = &clip[0];		/* current clip rectangle stack pointer */
    p_state->vue = &view_table[0];

    p_state->view_stk = &view_stk[0];	/* base view stack pointer */
    p_state->vue_tbl  = &view_table[0];	/* base view table pointer */

    /* Initialize the viewing matrices to identity */
    for( i=0; i < PEXtk_VIEW_SIZE; i++){
	PEXtk_matrix_copy( view_table[i].orientation, ident_mat);
	PEXtk_matrix_copy( view_table[i].mapping, ident_mat);
    }

    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+     Pclip_rectangle      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets viewport clip rectangle to the rectangle specified by 
 *	(left,bottom) and (right,top).
 *
 *  USAGE:
 *	Pclip_rectangle( left, right, bottom, top )
 *	Screencoord    left	(in)	lower left x position to clip
 *	Screencoord    right	(in)	upper right x position to clip
 *	Screencoord    bottom	(in)	lower left y position to clip
 *	Screencoord    top	(in)	upper right y position to clip
 *
 *  DESCRIPTION:
 *	This routine sets the clip rectangle, which should be a subregion of
 *	the current viewport.
 *	Since the NPC Subvolume for the full viewport is the subvolume
 *	defined by the full [0.0-1.0] space, the NPC Subvolume is 
 *	calculated to be that portion of the [0.0-1.0] space which will
 *	provide the desired clip rectangle.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pclip_rectangle( left, right, bottom, top )
Screencoord left;
Screencoord right;
Screencoord bottom;
Screencoord top;
{
    int loc, r_l;

    if( p_state->sys_mode == PEXtk_DLS_MODE ){
	DLS_view_box( PEXtk_OP_CLIP_RECT, left, right, bottom, top);
	return;
    }

    loc = p_state->vpt_loc;

    p_state->rdrAttrib.clip_list.count = 1;
    p_state->rdrAttrib.clip_list.rectangles = &clip[loc];

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_VPT);

    clip[loc].xmin	= left;		/* clip x_ll */
    clip[loc].xmax	= right;	/* clip x_ur */
    clip[loc].ymin	= bottom;	/* clip y_ll */
    clip[loc].ymax	= top;		/* clip y_ur */

    PEXtk_reset_pex_state( 1, PEXRAClipList);

    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+    Pget_clip_rectangle   +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Returns the current viewport clip rectangle.
 *
 *  USAGE:
 *	Pget_clip_rectangle( left, right, bottom, top )
 *	Screencoord    *left	(in)	lower left x position to clip
 *	Screencoord    *right	(in)	upper right x position to clip
 *	Screencoord    *bottom	(in)	lower left y position to clip
 *	Screencoord    *top	(in)	upper right y position to clip
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pget_clip_rectangle( left, right, bottom, top )
Screencoord *left;
Screencoord *right;
Screencoord *bottom;
Screencoord *top;
{
    int loc;

    loc = p_state->vpt_loc;

    *left	= clip[loc].xmin;  	/* clip x_ll */
    *right	= clip[loc].xmax; 	/* clip x_ur */
    *bottom	= clip[loc].ymin;	/* clip y_ll */
    *top	= clip[loc].ymax;   	/* clip y_ur */

    return;

}

/*
 *
 *                       +--------------------------+
 * ----------------------+        Pget_view         +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Returns current view paramters
 *
 *  USAGE:
 *	Pget_view( left, right, bottom, top, near, far )
 *	int    		*type	(in)	viewing type
 *	Fcoord		*left	(in)	lower left x position
 *	Fcoord		*right	(in)	upper right x position
 *	Fcoord		*bottom	(in)	lower left y position 
 *	Fcoord		*top	(in)	upper right y position
 *	Fcoord		*fov	(in)	field of view
 *	Fcoord		*aspect	(in)	aspect ratio
 *	Fcoord		*near	(in)	near clipping plane
 *	Fcoord		*far	(in)	far clipping plane
 *
 *  RETURNS:
 *	Current stack location
 */
 
/*ARGSUSED*/
int
Pget_view( type, left, right, bottom, top, fov, aspect, near, far )
int    *type;
Fcoord *left;
Fcoord *right;
Fcoord *bottom;
Fcoord *top;
Fcoord *fov;
Fcoord *aspect;
Fcoord *near;
Fcoord *far;
{
    int i;

    i = p_state->view_loc;

    switch( view_stk[i].type ){
	case PEXtk_VIEW_ORTHO:
	case PEXtk_VIEW_WINDOW:
		*type   = view_stk[i].type;
		*left   = view_stk[i].left;
		*right  = view_stk[i].right;
		*bottom = view_stk[i].bottom;
		*top    = view_stk[i].top;
		*near   = view_stk[i].near;
		*far    = view_stk[i].far;
	    break;
	case PEXtk_VIEW_ORTHO_2D:
		*type   = view_stk[i].type;
		*left   = view_stk[i].left;
		*right  = view_stk[i].right;
		*bottom = view_stk[i].bottom;
		*top    = view_stk[i].top;
	    break;
	case PEXtk_VIEW_PERSP:
		*type   = view_stk[i].type;
		*near   = view_stk[i].near;
		*far    = view_stk[i].far;
		*fov    = view_stk[i].fov;
		*aspect = view_stk[i].aspect;
	    break;
    }

    return i;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      Pget_viewport       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Returns current window viewport parameters to 
 *	the pointers specified by (left,bottom) and (right,top).
 *
 *  USAGE:
 *	Pget_viewport( left, right, bottom, top )
 *	Screencoord    *left	(in)	lower left x position of viewport
 *	Screencoord    *right	(in)	upper right x position of viewport
 *	Screencoord    *bottom	(in)	lower left y position of viewport
 *	Screencoord    *top	(in)	upper right y position of viewport
 *
 *  RETURNS:
 *	Current stack location
 */
 
/*ARGSUSED*/
int
Pget_viewport( left, right, bottom, top )
Screencoord *left;
Screencoord *right;
Screencoord *bottom;
Screencoord *top;
{
    int loc;

    loc = p_state->vpt_loc;

    *left   = viewport[loc].left;		/* viewport x_ll */
    *right  = viewport[loc].right;		/* viewport x_ur */
    *bottom = viewport[loc].bottom;		/* viewport y_ll */
    *top    = viewport[loc].top;		/* viewport y_ur */

    return loc;

}

/*
 *
 *                       +--------------------------+
 * ----------------------+       Plookat_view       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the lookat view.
 *
 *  USAGE:
 *	Plookat_view( vx, vy, vz, px, py, pz, twist )
 *	Fcoord	    vx		(in)	the x viewpoint coordinate
 *	Fcoord	    vx		(in)	the y viewpoint coordinate
 *	Fcoord	    vx		(in)	the z viewpoint coordinate
 *	Fcoord	    px		(in)	the x reference coordinate
 *	Fcoord	    px		(in)	the y reference coordinate
 *	Fcoord	    px		(in)	the z reference coordinate
 *	Angle	    twist	(in)	the twist angle
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Plookat_view( vx, vy, vz, px, py, pz, twist)
Fcoord	vx;
Fcoord	vy;
Fcoord	vz;
Fcoord	px;
Fcoord	py;
Fcoord	pz;
Angle	twist;
{
    float delta_x, delta_y, delta_z;
    float len, c, s, mag, dx_m, dz_m, Cx;
    PEXCoord Rx, Ry, Rz, norm;
    double rad;
    float *ptr;

    if( p_state->sys_mode == PEXtk_DLS_MODE ){
	DLS_view_lookat( vx, vy, vz, px, py, pz, twist);
	return;
    }

#ifdef VIEWDEBUG_L
printf(" lookat_view v %f %f %f p %f %f %f twist %f\n",
	vx,vy,vz,px,py,pz,twist);
#endif
	if( p_state->mat_mode == PEXtk_MATRIX_MODELING )
	    ptr = &p_state->mat_stk[p_state->mtx_loc][0][0];
	else
	    ptr = &p_state->vue->orientation[0][0];

	delta_x = px - vx;
	delta_y = py - vy;
	delta_z = pz - vz;

	rad = delta_x*delta_x + delta_y*delta_y + delta_z*delta_z;
	len = (float) sqrt( (double)rad);
	if( len == 0.0)
	    return;

	rad = delta_x*delta_x + delta_z*delta_z;
	mag = (float) sqrt( (double)rad);
	if( mag == 0.0)
	    return;

	norm.x = delta_x / len;
	norm.y = delta_y / len;
	norm.z = delta_z / len;

	rad = -twist * 3.141592653589793 / 180.0;
	c = cos(rad);
	s = sin(rad);

	Cx = mag / len;
	dx_m = delta_x / mag;
	dz_m = delta_z / mag;

	Rx.x = -dz_m * c + dx_m * s * norm.y;
	Rx.y = dx_m * s - dx_m * c * norm.y;
	Rx.z = -dx_m*Cx;

	Ry.x = -s * Cx;
	Ry.y = c * Cx;
	Ry.z = -norm.y;

	Rz.x = dx_m*c + dz_m*norm.y*s;
	Rz.y = dz_m*s - dz_m*norm.y*c;
	Rz.z = -Cx*dz_m;

	/* Transpose the above terms */

	*ptr		= Rx.x;
	*(ptr+4)	= Rx.y;
	*(ptr+8)	= Rx.z;
	*(ptr+1)	= Ry.x;
	*(ptr+5)	= Ry.y;
	*(ptr+9)	= Ry.z;
	*(ptr+2)	= Rz.x;
	*(ptr+6)	= Rz.y;
	*(ptr+10)	= Rz.z;

	*(ptr+3)	= -vx*Rx.x - vy*Ry.x - vz*Rz.x;

	*(ptr+7)	= -vx*Rx.y - vy*Ry.y - vz*Rz.y;
	*(ptr+11)	= -vx*Rx.z - vy*Ry.z - vz*Rz.z;
	*(ptr+12) = *(ptr+13) = *(ptr+14) = 0.0;

	*(ptr+15)	= 1.0;

	PEXSetGlobalTransform( dpy, p_state->cur_id, p_state->req_type, 
	    p_state->mat_stk[p_state->mtx_loc]);
	memcpy( p_state->pcAttrib.global_transform,
	    p_state->mat_stk[p_state->mtx_loc], sizeof(PEXMatrix));
	PEXSetPCAttributeMask( p_state->pc_mask, PEXPCGlobalTransform);

#ifdef VIEWDEBUG_L
    {  int i; float *f_ptr;
    printf("Lookat view matrix\n"); f_ptr = ptr;
    for(i=0;i<4;i++,f_ptr += 4)
    printf(" %f %f %f %f\n",
	*(f_ptr), *(f_ptr+1), *(f_ptr+2), *(f_ptr+3));
    }
#endif

    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+     Portho_2d_view       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the 2d orthographic view position to the frustum (left,bottom)
 *	to (right,top).
 *
 *  USAGE:
 *	Portho_2d_view( id, left, right, bottom, top )
 *	int	    id		(in)	view id number
 *	Fcoord	    left	(in)	lower left x position of frustum
 *	Fcoord	    right	(in)	upper right x position of frustum
 *	Fcoord	    bottiom	(in)	lower left y position of frustum
 *	Fcoord	    top		(in)	upper right y position of frustum
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Portho_2d_view( id, left, right, bottom, top )
int	id;
Fcoord	left;
Fcoord	right;
Fcoord	bottom;
Fcoord	top;
{
    int i;

    if( p_state->sys_mode == PEXtk_DLS_MODE ){
	DLS_view_proj( PEXtk_OP_VIEW_ORTHO_2D, id, left, right, 
		bottom, top, 0.0, 0.0 );
	return;
    }
    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_VIEW);
    XFlush( dpy);

#ifdef VIEWDEBUG_O
printf(" ortho_2d_view l %f r %f b %f t %f\n",
	left,right,bottom,top);
#endif
	if( id > 0 )
	    i = id;
	else
	    i = p_state->view_loc;

	view_table[i].mapping[0][1] =
	view_table[i].mapping[0][2] =
	view_table[i].mapping[1][0] =
	view_table[i].mapping[1][2] =
	view_table[i].mapping[2][0] =
	view_table[i].mapping[2][1] =
	view_table[i].mapping[3][0] =
	view_table[i].mapping[3][1] =
	view_table[i].mapping[3][2] =	0.0;

	view_table[i].mapping[0][0] =  1.0 / (right - left);
	view_table[i].mapping[1][1] =  1.0 / (top - bottom);
	view_table[i].mapping[2][2] =  .00001;
	/* 
	 * The mapping was fixed to .5, but this does not align
	 * a 'min=0,max=n' viewing in the center
	 */
	/* view_table[i].mapping[0][3] = 0.5; */
	/* view_table[i].mapping[1][3] = 0.5; */
	view_table[i].mapping[0][3] = 
	                           0.5 - (right + left) /
					     (2.0 * (right - left));
	view_table[i].mapping[1][3] = 
	                           0.5 - (top + bottom) /
					     (2.0 * (top - bottom));
	view_table[i].mapping[2][3] = 0.0;
	view_table[i].mapping[3][3] = 1.0;

    view_table[i].clip_flags = 0;
    view_table[i].clip_flags = (PEXClipXY | PEXClipBack | PEXClipFront);
    view_table[i].clip_limits.min.x = .0; 
    view_table[i].clip_limits.max.x = 1.;
    view_table[i].clip_limits.min.y = .0; 
    view_table[i].clip_limits.max.y = 1.;
    view_table[i].clip_limits.min.z = .0; 
    view_table[i].clip_limits.max.z = 1.0;

	view_stk[i].type  	= PEXtk_VIEW_ORTHO_2D;
	view_stk[i].left	= left;
	view_stk[i].right	= right;
	view_stk[i].bottom	= bottom;
	view_stk[i].top		= top;

	p_state->rdrAttrib.view_table = p_state->win[p_state->cur_w].View_ID;

	PEXSetTableEntries( dpy, p_state->win[p_state->cur_w].View_ID, 
		id, 1, PEXLUTView, &view_table[i]);

    PEXtk_reset_pex_state( 0, PEXRAViewTable);

#ifdef VIEWDEBUG_O
{  int l;
printf("View2d matrix\n");
for(l=0;l<4;l++)
printf(" %f %f %f %f\n",
	view_table[i].mapping[l][0], view_table[i].mapping[l][1],
	view_table[i].mapping[l][2], view_table[i].mapping[l][3]);
}
#endif

    return;

}

/*
 *
 *                       +--------------------------+
 * ----------------------+       Portho_view        +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the orthographic view position to the frustum (left,bottom)
 *	to (right,top), and the near and far clipping planes.
 *
 *  USAGE:
 *	Portho_view( id, left, right, bottom, top, near, far )
 *	int	    id		(in)	view stack id number
 *	Fcoord	    left	(in)	lower left x position of frustum
 *	Fcoord	    right	(in)	upper right x position of frustum
 *	Fcoord	    bottiom	(in)	lower left y position of frustum
 *	Fcoord	    top		(in)	upper right y position of frustum
 *	Fcoord	    near	(in)	near clipping plane
 *	Fcoord	    far		(in)	far clipping plane
 *
 *  DESCRIPTION:
 *	The near value must be >= the far value
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Portho_view( id, left, right, bottom, top, near, far )
int	id;
Fcoord	left;
Fcoord	right;
Fcoord	bottom;
Fcoord	top;
Fcoord	near;
Fcoord	far;
{
    int i;

    if( p_state->sys_mode == PEXtk_DLS_MODE ){
	DLS_view_proj( PEXtk_OP_VIEW_ORTHO, id, left, right, 
		bottom, top, near, far);
	return;
    }
    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_VIEW);
    XFlush( dpy);

#ifdef VIEWDEBUG_O
printf(" ortho_view l %f r %f b %f t %f near %f far %f\n",
	left,right,bottom,top,near,far);
#endif
	if( id > 0 )
	    i = id;
	else
	    i = p_state->view_loc;

	view_table[i].mapping[0][1] =
	view_table[i].mapping[0][2] =
	view_table[i].mapping[1][0] =
	view_table[i].mapping[1][2] =
	view_table[i].mapping[2][0] =
	view_table[i].mapping[2][1] =
	view_table[i].mapping[3][0] =
	view_table[i].mapping[3][1] =
	view_table[i].mapping[3][2] =	0.0;

	view_table[i].mapping[0][0] =  1.0 / (right - left);
	view_table[i].mapping[1][1] =  1.0 / (top - bottom);
	view_table[i].mapping[2][2] =  1.0 / (far - near);

	/* 
	 * The mapping was fixed to .5, but this does not align
	 * a 'min=0,max=n' viewing in the center
	 */
	/* view_table[i].mapping[0][3] = 0.5; */
	/* view_table[i].mapping[1][3] = 0.5; */
	view_table[i].mapping[0][3] = 
	                           0.5 - (right + left) /
					     (2.0 * (right - left));
	view_table[i].mapping[1][3] = 
	                           0.5 - (top + bottom) /
					     (2.0 * (top - bottom));
	view_table[i].mapping[2][3] =
               1.0 + (near) / (far - near);
               /* 0.5 + (far + near) / (2.0 * (far - near)); */
	view_table[i].mapping[3][3] = 1.0;

    view_table[i].clip_flags = 0;
    view_table[i].clip_flags = (PEXClipXY | PEXClipBack | PEXClipFront);
    view_table[i].clip_limits.min.x = .0; 
    view_table[i].clip_limits.max.x = 1.;
    view_table[i].clip_limits.min.y = .0; 
    view_table[i].clip_limits.max.y = 1.;
    view_table[i].clip_limits.min.z = .0; 
    view_table[i].clip_limits.max.z = 1.0;

	view_stk[i].type  	= PEXtk_VIEW_ORTHO;
	view_stk[i].near	= near;
	view_stk[i].far		= far;
	view_stk[i].left	= left;
	view_stk[i].right	= right;
	view_stk[i].bottom	= bottom;
	view_stk[i].top		= top;

	p_state->rdrAttrib.view_table = p_state->win[p_state->cur_w].View_ID;

	PEXSetTableEntries( dpy, p_state->win[p_state->cur_w].View_ID, 
		id, 1, PEXLUTView, &view_table[i]);

    PEXtk_reset_pex_state( 0, PEXRAViewTable);

#ifdef VIEWDEBUG_O
{  int l;
printf("Ortho view matrix\n");
for(l=0;l<4;l++)
printf(" %f %f %f %f\n",
	view_table[i].mapping[l][0], view_table[i].mapping[l][1],
	view_table[i].mapping[l][2], view_table[i].mapping[l][3]);
}
#endif

    return;

}

/*
 *
 *                       +--------------------------+
 * ----------------------+       Ppersp_view        +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the perspective view to the frustum specified by fov (the
 *	field of view), aspect (the aspect ratio for viewing), and the
 *	near and far clipping planes.
 *
 *  USAGE:
 *	Ppersp_view( id, fov, aspect, near, far )
 *	int	    id		(in)	view id number
 *	Fcoord	    fov		(in)	the field of view
 *	Fcoord	    aspect	(in)	viewing aspect ratio
 *	Fcoord	    near	(in)	near clipping plane
 *	Fcoord	    far		(in)	far clipping plane
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ppersp_view( id, fov, aspect, near, far)
int	id;
Fcoord	fov;
Fcoord	aspect;
Fcoord	near;
Fcoord	far;
{
    int i;

    if( p_state->sys_mode == PEXtk_DLS_MODE ){
	DLS_view_proj( PEXtk_OP_VIEW_PERSP, id, fov, aspect, near, far, 0.,0.);
	return;
    }
    XFlush( dpy);

	if( id > 0 )
	    i = id;
	else
	    i = p_state->view_loc;

#ifdef VIEWDEBUG_P
printf(" persp_view %f %f %f %f\n",
	fov,aspect,near,far);
#endif

	view_table[i].mapping[0][1] =
	view_table[i].mapping[0][3] =
	view_table[i].mapping[1][0] =
	view_table[i].mapping[1][3] =
	view_table[i].mapping[2][0] =
	view_table[i].mapping[2][1] =
	view_table[i].mapping[3][0] =
	view_table[i].mapping[3][1] =
	view_table[i].mapping[3][3] =	0.0;

#define USE_OLD_PERSP
#ifdef USE_OLD_PERSP
	view_table[i].mapping[0][0] = 0.50 /
		((float)tan((double)(fov * 3.141592653589793 /(2.0 * 180.0)))*
		(double)aspect);

	view_table[i].mapping[1][1] = 0.50 / (float)tan((double)fov *
		3.141592653589793 / (2.0 * 180.0));
	view_table[i].mapping[2][2] =
		(near) / (far - near);
		/* -(far + near) / (2 * (far - near)) +0.5; */

	view_table[i].mapping[3][2] = -1.0;
	view_table[i].mapping[2][3] =
		  (far * near) / (far - near);
        view_table[i].mapping[0][2] = -0.50;
        view_table[i].mapping[1][2] = -0.50;
#else
	view_table[i].mapping[0][0] = 0.50 / (near*(float)
		tan((double)(fov * 3.141592653589793 /(2.0 * 180.0)))*
		(double)aspect);

	view_table[i].mapping[1][1] = 0.50 / (near*(float)tan((double)fov *
		3.141592653589793 / (2.0 * 180.0)));
	view_table[i].mapping[2][2] =
		1.0 / (near - far);

        view_table[i].mapping[3][2] = -1.0/near;
	view_table[i].mapping[2][3] =
		  far / (far - near);
        view_table[i].mapping[0][2] = -0.5/near;
        view_table[i].mapping[1][2] = -0.5/near;
#endif

#ifdef VIEWDEBUG_P
{  int i;
printf("Persp view matrix\n");
for(i=0;i<4;i++)
printf(" %f %f %f %f\n",
	view_table[i].mapping[i][0], view_table[i].mapping[i][1],
	view_table[i].mapping[i][2], view_table[i].mapping[i][3]);
}
#endif

    view_table[i].clip_flags = (PEXClipXY | PEXClipBack | PEXClipFront);
    view_table[i].clip_limits.min.x = .0; 
    view_table[i].clip_limits.max.x = 1.;
    view_table[i].clip_limits.min.y = .0; 
    view_table[i].clip_limits.max.y = 1.;
    view_table[i].clip_limits.min.z = .0; 
    view_table[i].clip_limits.max.z = 1.0;

	view_stk[i].type	= PEXtk_VIEW_PERSP;
	view_stk[i].near	= near;
	view_stk[i].far		= far;
	view_stk[i].fov		= fov;
	view_stk[i].aspect	= aspect;

	p_state->rdrAttrib.view_table = p_state->win[p_state->cur_w].View_ID;

	PEXSetTableEntries( dpy, p_state->win[p_state->cur_w].View_ID, 
		id, 1, PEXLUTView, &view_table[i]);

    PEXtk_reset_pex_state( 0, PEXRAViewTable);

    return;

}

/*
 *
 *                       +--------------------------+
 * ----------------------+        Ppolar_view       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the polar view.
 *
 *  USAGE:
 *	Ppolar_view( dist, azim, inc, twist )
 *	Fcoord	    dist	(in)	the distance from the origin
 *	Angle	    azim	(in)	the angle in the x-y plane
 *	Angle	    inc		(in)	the angle in the y-z plane
 *	Angle	    twist	(in)	the twist angle
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ppolar_view( dist, azim, inc, twist)
Fcoord	dist;
Angle	azim;
Angle	inc;
Angle	twist;
{
    if( p_state->sys_mode == PEXtk_DLS_MODE ){
	DLS_view_polar( dist, azim, inc, twist );
	return;
    }

#ifdef VIEWDEBUG_PO
printf(" lookat_view v %f %f %f p %f %f %f twist %f\n",
	vx,vy,vz,px,py,pz,twist);
#endif

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);

	Pload_identity();
	Ptranslate_z( -dist, 1);
	Protate_z( -twist, 1);
	Protate_x( -inc, 1);
	Protate_y( -azim, 1);

    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+         Ppop_view        +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Pops the view stack stack.
 *
 *  USAGE:
 *	Ppop_view( )
 *
 *  DESCRITPTION:
 *	The current view state pointer is reset, and Pdepth is called
 *	with the current near and far values (this routine is used
 *	simply because it is a convenient mechanism to restore the view.
 *
 *  NOTE:
 *	The routine is for Immediate Mode only.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ppop_view( )
{
    if( p_state->sys_mode == PEXtk_RENDERING_MODE ||
        p_state->sys_mode == PEXtk_PICKING_MODE ){
	if( p_state->view_loc > 0 ){
	    p_state->view_loc--;
	    p_state->view = &view_stk[p_state->view_loc];
	    p_state->vue  = &view_table[p_state->view_loc];
		/* restore the previous view stack */
	    Pdepth( p_state->view->near, p_state->view->far );
	}
	return;
    } else
    if( p_state->sys_mode == PEXtk_DLS_MODE ){
	DLS_stack( PEXtk_OP_POP_VIEW);
	return;
    }

    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+         Ppush_view       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Pushes the view stack stack. Only Immediate mode is supported.
 *
 *  USAGE:
 *	Ppush_view( )
 *
 *  NOTE:
 *	This routine is for Immediate Mode only.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ppush_view( )
{
    if( p_state->sys_mode == PEXtk_RENDERING_MODE ||
        p_state->sys_mode == PEXtk_PICKING_MODE ){
	if( p_state->view_loc < PEXtk_VIEW_SIZE-1 ){
	    p_state->view_loc++;
	    p_state->view = &view_stk[p_state->view_loc];
	    p_state->vue  = &view_table[p_state->view_loc];
	    memcpy( (char*)&view_stk[p_state->view_loc], 
		(char*)&view_stk[p_state->view_loc-1], sizeof(PEXtk_VIEW));
	    memcpy( (char*)&view_table[p_state->view_loc],
		(char*)&view_table[p_state->view_loc-1], sizeof(PEXViewEntry));
	}
	return;
    } else
    if( p_state->sys_mode == PEXtk_DLS_MODE ){
	DLS_stack( PEXtk_OP_PUSH_VIEW);
	return;
    }

    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+       Ppop_viewport      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Pops the viewport stack.
 *
 *  USAGE:
 *	Ppop_viewport( )
 *
 *  NOTE:
 *	This routine is for Immediate Mode only.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ppop_viewport( )
{
    int loc;

    if( p_state->sys_mode == PEXtk_RENDERING_MODE ||
        p_state->sys_mode == PEXtk_PICKING_MODE ){
	if( p_state->vpt_loc > 0 ){
	    p_state->vpt_loc--;
	    p_state->vpt = &viewport[p_state->vpt_loc];
	    p_state->clip= &clip[p_state->vpt_loc];
	}
	loc = p_state->vpt_loc;
	Pviewport( viewport[loc].left, viewport[loc].right, 
		viewport[loc].bottom, viewport[loc].top );
	return;
    } else
    if( p_state->sys_mode == PEXtk_DLS_MODE ){
	DLS_stack( PEXtk_OP_POP_VPT);
	return;
    }

    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+       Ppush_viewport     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Pushes the viewport stack.
 *
 *  USAGE:
 *	Ppush_viewport( )
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ppush_viewport( )
{
    if( p_state->sys_mode == PEXtk_RENDERING_MODE ||
        p_state->sys_mode == PEXtk_PICKING_MODE ){
	if( p_state->vpt_loc < PEXtk_VPT_SIZE-1 ){
	    p_state->vpt_loc++;
	    p_state->vpt = &viewport[p_state->vpt_loc];
	    p_state->clip= &clip[p_state->vpt_loc];
	    memcpy( (char *) &viewport[p_state->vpt_loc],
		(char *) &viewport[p_state->vpt_loc-1],
		sizeof(PEXtk_VIEWPORT));
	    memcpy( (char *) &clip[p_state->vpt_loc],
		(char *) &clip[p_state->vpt_loc-1],
		sizeof(PEXtk_VIEWPORT));
	}
	return;
    } else
    if( p_state->sys_mode == PEXtk_DLS_MODE ){
	DLS_stack( PEXtk_OP_PUSH_VPT);
	return;
    }

    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+     Pset_view_index      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the view index id and the view stack location.
 *
 *  USAGE:
 *	Pset_view_index( id )
 *	int	    id		(in)	view id
 *
 *  RETURNS:
 *	0 if successful, non-zero if error
 */
 
/*ARGSUSED*/
int
Pset_view_index( id )
int	id;
{
    if( p_state->sys_mode == PEXtk_DLS_MODE ){
	DLS_attrib( PEXtk_OP_SET_VIEW, id );
	return 0;
    }
	p_state->cur_view = id;
	PEXSetViewIndex( dpy, p_state->cur_id, p_state->req_type, 
	    p_state->cur_view);
	p_state->pcAttrib.view_index = id;
	PEXSetPCAttributeMask( p_state->pc_mask, PEXPCViewIndex);

	if( ( id >= 0 ) && ( id < PEXtk_VIEW_SIZE ) ){
	    p_state->view_loc = id;
	    p_state->view = &view_stk[p_state->view_loc];
	    p_state->vue  = &view_table[p_state->view_loc];
	}
	return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+        Pviewport         +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets window viewport to the rectangle specified by (left,bottom)
 *	and (right,top). The clip rectangle is also set.
 *
 *  USAGE:
 *	Pviewport( left, right, bottom, top )
 *	Screencoord    left	(in)	lower left x position of viewport
 *	Screencoord    right	(in)	upper right x position of viewport
 *	Screencoord    bottom	(in)	lower left y position of viewport
 *	Screencoord    top	(in)	upper right y position of viewport
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pviewport( left, right, bottom, top )
Screencoord left;
Screencoord right;
Screencoord bottom;
Screencoord top;
{
    int loc;

    if( p_state->sys_mode == PEXtk_DLS_MODE ){
	DLS_view_box( PEXtk_OP_VPT, left, right, bottom, top );
	return;
    }

    p_state->rdrAttrib.viewport.min.x = left ; 
    p_state->rdrAttrib.viewport.min.y = bottom;
    p_state->rdrAttrib.viewport.min.z = 0.0;
    p_state->rdrAttrib.viewport.max.x = right ;
    p_state->rdrAttrib.viewport.max.y = top;
    p_state->rdrAttrib.viewport.max.z = 1.0;
    p_state->rdrAttrib.viewport.use_drawable = False;

    loc = p_state->vpt_loc;
    p_state->rdrAttrib.clip_list.count = 1;
    p_state->rdrAttrib.clip_list.rectangles = &clip[loc];

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_VPT);
    clip[loc].xmin = viewport[loc].left   = left;	/* viewport x_ll */
    clip[loc].xmax = viewport[loc].right  = right;	/* viewport x_ur */
    clip[loc].ymin = viewport[loc].bottom = bottom; 	/* viewport y_ll */
    clip[loc].ymax = viewport[loc].top    = top;	/* viewport y_ur */

    PEXtk_reset_pex_state(1,PEXRAViewport|PEXRAViewTable|PEXRAClipList);

    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+       Pwindow_view       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the perspective view to the frustum specified by (left,bottom)
 *	and (right,top), and the near and far clipping planes.
 *
 *  USAGE:
 *	Pwindow_view( id, left, right, bottom, top, near, far )
 *	int	    id		(in)	view id number
 *	Fcoord	    left	(in)	lower left x position of frustum
 *	Fcoord	    right	(in)	upper right x position of frustum
 *	Fcoord	    bottom	(in)	lower left y position of frustum
 *	Fcoord	    top		(in)	upper right y position of frustum
 *	Fcoord	    near	(in)	near clipping plane
 *	Fcoord	    far		(in)	far clipping plane
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pwindow_view( id, left, right, bottom, top, near, far)
int	id;
Fcoord	left;
Fcoord	right;
Fcoord	bottom;
Fcoord	top;
Fcoord	near;
Fcoord	far;
{
    int i;

    if( p_state->sys_mode == PEXtk_DLS_MODE ){
	DLS_view_proj( PEXtk_OP_VIEW_WINDOW, id, left, right, bottom, 
		top, near, far );
	return;
    }
    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_VIEW);
    XFlush( dpy);

	if( id > 0 )
	    i = id;
	else
	    i = p_state->view_loc;

#ifdef VIEWDEBUG_W
printf(" window_view l %f r %f b %f t %f, near %f, far %f\n",
	left,right,bottom,top,near,far);
#endif

	view_table[i].mapping[0][1] =
	view_table[i].mapping[0][3] =
	view_table[i].mapping[1][0] =
	view_table[i].mapping[1][3] =
	view_table[i].mapping[2][0] =
	view_table[i].mapping[2][1] =
	view_table[i].mapping[3][0] =
	view_table[i].mapping[3][1] =
	view_table[i].mapping[3][3] =	0.0;

	view_table[i].mapping[0][0] = near / ( right - left );
		
	view_table[i].mapping[1][1] = near / ( top - bottom );

	view_table[i].mapping[2][2] =
		(near) / (far - near);
		/* -(far + near) / (2 * (far - near)) +0.5; */

	view_table[i].mapping[3][2] = -1.0;
	view_table[i].mapping[2][3] =
		  (far * near) / (far - near);
	view_table[i].mapping[0][2] = -0.5 - (right + left) /
					     (2.0 * (right - left));
	view_table[i].mapping[1][2] = -0.5 - (top + bottom) /
					     (2.0 * (top - bottom));
#ifdef IF_ORTHO1
	view_table[i].mapping[0][3] = 0.5 - (right + left) /
					     (2.0 * (right - left));
	view_table[i].mapping[1][3] = 0.5 - (top + bottom) /
					     (2.0 * (top - bottom));
#endif
#ifdef IF_PERSP1
        view_table[i].mapping[0][2] = -0.50;
        view_table[i].mapping[1][2] = -0.50;
#endif

#ifdef VIEWDEBUG_W
{  int i;
printf("Window view matrix\n");
for(i=0;i<4;i++)
printf(" %f %f %f %f\n",
	view_table[i].mapping[i][0], view_table[i].mapping[i][1],
	view_table[i].mapping[i][2], view_table[i].mapping[i][3]);
}
#endif

    view_table[i].clip_flags = (PEXClipXY | PEXClipBack | PEXClipFront);
    view_table[i].clip_limits.min.x = .0; 
    view_table[i].clip_limits.max.x = 1.;
    view_table[i].clip_limits.min.y = .0; 
    view_table[i].clip_limits.max.y = 1.;
    view_table[i].clip_limits.min.z = .0; 
    view_table[i].clip_limits.max.z = 1.0;

	view_stk[i].type	= PEXtk_VIEW_WINDOW;
	view_stk[i].near	= near;
	view_stk[i].far		= far;
	view_stk[i].left	= left;
	view_stk[i].right	= right;
	view_stk[i].bottom	= bottom;
	view_stk[i].top		= top;

	p_state->rdrAttrib.view_table = 
		p_state->win[p_state->cur_w].View_ID;

	PEXSetTableEntries( dpy, p_state->win[p_state->cur_w].View_ID,
		id, 1, PEXLUTView, &view_table[i]);

    PEXtk_reset_pex_state( 0, PEXRAViewTable);

    return;

}

/*
 *
 *                       +--------------------------+
 * ----------------------+    PEXtk_set_viewport    +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Called by the set_drawable routine to set the default window 
 *	viewport to a rectangle the size of the drawable win_width,
 *	win_height, (0,win_width-1) and (0,win_height-1).
 *
 *  USAGE:
 *	PEXtk_set_viewport( )
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_set_viewport( )
{
    int loc, width, height;

    width = p_state->win[p_state->cur_w].width;
    height = p_state->win[p_state->cur_w].height;

    p_state->rdrAttrib.viewport.min.x = 0 ; 
    p_state->rdrAttrib.viewport.min.y = 0;
    p_state->rdrAttrib.viewport.min.z = 0.0;
    p_state->rdrAttrib.viewport.max.x = width-1 ;
    p_state->rdrAttrib.viewport.max.y = height-1;
    p_state->rdrAttrib.viewport.max.z = 1.0;
    p_state->rdrAttrib.viewport.use_drawable = True;

    loc = p_state->vpt_loc;
    viewport[loc].left   = 0;		/* viewport x_ll */
    viewport[loc].right  = width-1;	/* viewport x_ur */
    viewport[loc].bottom = 0;		/* viewport y_ll */
    viewport[loc].top    = height-1;	/* viewport y_ur */

    return 0;
}
