/*
 *  $Header: /auto/nfs/sequoia/u/home/rws/pextk/src/pextk_win.c,v 1.1 1993/10/12 00:07:09 rws Exp $
 ******************************************************************************
 ******************************************************************************
 *****                          pextk_win.c                               *****
 *****									  *****
 *****									  *****
 *****									  *****
 *                                                                            *
 *	Copyright 1992, 1993. All Rights Reserved                             *
 *			SHOgraphics					      *
 *			321 Soquel Way           			      *
 *			Sunnyvale, CA     94086   			      *
 *                                                                            *
 *****									  *****

    Permission to use, copy, modify, and distribute this software and
 its documentation for any purpose and without fee is hereby granted,
 provided that the above copyright notice appear in all copies and that
 both that copyright notice and this permission notice appear in
 supporting documentation, and that the name of SHOgraphics not be
 used in advertising or publicity pertaining to distribution of the 
 software without specific, written prior permission.

 SHOGRAPHICS DISCLAIMS ANY AND ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 INCLUDING ALL EXPRESSED AND IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS. IN NO EVENT SHALL SHOGRAPHICS BE LIABLE FOR ANY DAMAGES OF ANY
 KIND, INCLUDING BUT NOT LIMITED TO SPECIAL, INDIRECT OR CONSEQUENTIAL 
 DAMAGES RESULTING FROM LOSS OF USE, DATA OR PROSPECTIVE PROFITS, WHETHER 
 IN AN ACTION OF CONTRACT, NEGLIENCE OR OTHER TORTIOUS ACTION, ARISING OUT 
 OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

 ************************************************************************ */
/*
 *
 *  Module Purpose:
 *	This is the module which contains the PEXtk Window management routines
 *
 *  Module Notes:
 *
 *  Module Contents:
 *
 *    void  Pget_identity_matrix  Copies a the 4x4 identity matrix into mat.
 *    int   PEXtk_Update_Matrix   Sets the global transform matrix
 *
 *
 *  Module Notes:
 *
 */

#include <stdio.h>
#include <math.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>

#include <X11/PEX5/PEX.h>
#include <X11/PEX5/PEXproto.h>
#include <X11/PEX5/PEXlib.h>

#include <PEXtk.h>

    static GC 			gc;
    static Display		*dpy_id = NULL;
    static Window		window_id = 0;
    static int                  win_screen;
    static char                 *win_title = "PEXtk Program";
    static char                 *win_icon_name   = "PEXtk";
    static Pixmap               win_icon_pixmap;
    static XIconSize            **win_icon_list;
    static XFontStruct 	 	*xfont_info;
    static XGCValues 		values;
    static int			win_window_width;
    static int			win_window_height;
    static int                  win_icon_width = 32;
    static int                  win_icon_height = 32;
    static XClassHint           win_class_hints;
    static XSizeHints           win_size_hints;   /* preferred window sizes */
    static XWMHints             win_icon_hints;
    static Visual               *win_visual;
    static Colormap       	win_colormap;
    static XColor               win_fg;
    static XColor               win_bg;
    static XSetWindowAttributes win_atts;
    static unsigned long        win_mask;
    static unsigned int         win_class = InputOutput;
    static int                  win_xpos = 0;		/* window position */
    static int                  win_ypos = 0;
    static int                  win_depth;
    static unsigned int         win_border_width = 4;
    static unsigned long        pix_fg, pix_bg; 

/*
 *
 *                       +--------------------------+
 * ----------------------+      Pcreate_window      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Initializes window pointers.
 *
 *  USAGE:
 *	Pcreate_window( disp, width, height )
 *	Display    	*disp	(in)	Pointer to display structure.
 *	int    		width	(in)	Width of window.
 *	int    		height	(in)	Height of window.
 *
 *  DESCRIPTION:
 *	This routine will create a window.
 *
 *  RETURNS:
 *	window id if successfull, 0 if error occurred.
 */
 
/*ARGSUSED*/
Window
Pcreate_window( disp, win_width, win_height)
Display *disp;
int     win_width, win_height;
{
    Window window;
    XEvent 		 report;
    Colormap get_standard_colormap();

    dpy_id 	 = disp;		/* save data */
    win_window_width = win_width;
    win_window_height = win_height;

    win_screen   = DefaultScreen (disp);


#if 0
/* this is a problem on LEO running with default depth */
    win_colormap = DefaultColormapOfScreen(
			ScreenOfDisplay(disp, win_screen));
#endif
 
    /*
     * This is a hack to make the HP work properly.
     * The low end HP 700's can not double buffer
     * the 24 bit visual, even if it has it.
     */
    win_depth = PEXtk_Get_Best_Visual( disp, &win_visual);
    win_colormap = get_standard_colormap(disp, win_visual, win_depth, win_screen);

    if( win_depth == 0 ){
	win_visual   = DefaultVisual (disp, win_screen);
	win_depth    = DefaultDepth (disp, win_screen);
    }
    /*
     *  E&S Freedom graphics doesn't give proper colormap
     *  from XmuLookup routine, so always get the default.
     */
#ifndef ES
    if( win_colormap == 0 ) /* make sure colormap is not null */
#endif
    win_colormap = DefaultColormapOfScreen(
			ScreenOfDisplay(disp, win_screen));

   /*
    * Get background and foreground pixel values
    */
#ifdef WHITEBG
    pix_bg = WhitePixel(dpy_id, win_screen); /* White */
    pix_fg = BlackPixel(dpy_id, win_screen); /* Black */
#else
  /* Use Black Background */
    pix_fg = WhitePixel(dpy_id, win_screen); /* White */
    pix_bg = BlackPixel(dpy_id, win_screen); /* Black */
#endif

    win_mask     = CWBorderPixel | CWBackPixel | CWColormap | CWBackingStore;
    win_atts.backing_store    = WhenMapped;
    win_atts.colormap         = win_colormap;
    win_atts.background_pixel = pix_bg;
    win_atts.border_pixel     = pix_bg;

    window = XCreateWindow
                       (disp,
                        RootWindow (disp, win_screen),
                        win_xpos, win_ypos,
                        win_width, win_height,
                        win_border_width,
                        win_depth,
                        win_class,
                        win_visual,
               		win_mask,
			&win_atts); 

    window_id = window;

    win_size_hints.flags = USPosition | PPosition | PMinSize | PMaxSize;
    win_size_hints.x		 = 20;
    win_size_hints.y		 = 20;
    win_size_hints.max_width     = win_width;
    win_size_hints.max_height    = win_height;
    win_size_hints.min_width     = win_width;
    win_size_hints.min_height    = win_height;
	
    XStoreName      (disp, window, win_title);
    XSetClassHint   (disp, window, &win_class_hints);
    XSetNormalHints (disp, window, &win_size_hints);
	
   /*
    * Create resources
    */
    if ((xfont_info = XLoadQueryFont (disp, "courier14")) == NULL)
     if ((xfont_info = XLoadQueryFont (disp, "9x15")) == NULL)
      if ((xfont_info = XLoadQueryFont (disp, "fixed")) == NULL)
        {
        printf("Error on XLoadQueryFont\n");
        exit(2);
        }

    gc = XCreateGC (disp, window, 0L, &values);

    XSetFont (disp, gc, xfont_info->fid);
    XSetForeground (disp, gc, pix_fg);
    XSetBackground (disp, gc, pix_bg);
    XSetLineAttributes (disp, gc, 0, LineSolid, CapButt, JoinMiter);
#ifdef DO_GRAPHICS_EXPOSURES
    XSetGraphicsExposures(disp, gc, False);
#endif

    XSelectInput(disp, window, ExposureMask);
    XMapWindow(disp,window);
    XWindowEvent(disp, window, ExposureMask, &report);

    XSelectInput (disp, window, ExposureMask | FocusChangeMask |
		OwnerGrabButtonMask | PointerMotionMask
		| KeyPressMask | ButtonPressMask | ButtonReleaseMask);

    return window;
}

int
PEXtk_util_find_visual( dpy, win_visual)
Display *dpy;
Visual **win_visual;
{
    int	m_op, f_ev, f_err, i, j, ret_class, c_class;
    XVisualInfo vTemplate, *visualList;

    /* Check for the HPExtension to use HP type double-buffered
     * visuals
     */
    if(!XQueryExtension( dpy, "HPExtension",
                                &m_op, &f_ev, &f_err))
        return 0;
    /* Find the 12 bit visual
     */

    win_depth    = 0;
    /*
     *  See if any 12-bit visuals are available
     *  If there is a 12-bit, then look for a truecolor visual if
     *  it exists.
     */
        vTemplate.screen = DefaultScreen( dpy);
        vTemplate.depth  = 12;
        visualList = XGetVisualInfo( dpy, VisualScreenMask |
                        VisualDepthMask, &vTemplate, &i);
	c_class = 0;
        for(j=0;j<i;j++){
#if defined(__cplusplus) || defined(c_plusplus)
            ret_class = visualList[j].c_class;
#else
	    ret_class = visualList[j].class;
#endif
            if( ret_class > c_class ){
                *win_visual = visualList[j].visual;
                win_depth = 12;
		c_class = ret_class;
                /*
                 * Calculate the shift amount for each rgb mask
                 */
		if( c_class == TrueColor ){
		    /* free the visualinfo */
		    return 1;
		}
            }
        }
}

int
PGet_Queue()
{
    XEvent 		 report;

        XCheckMaskEvent( dpy_id, ExposureMask | FocusChangeMask |
		OwnerGrabButtonMask | PointerMotionMask
		| KeyPressMask | ButtonPressMask | ButtonReleaseMask,
		&report);
        switch(report.type){
            case KeyPress:
		return 7;
            case ButtonPress:
		return 103;
	    default:
		break;
	}
	return 0;
}

Pset_window_title(str)
char *str;
{
    win_title = str;
    if( dpy_id == NULL ) return;
    if( window_id == NULL ) return;
    XStoreName( dpy_id, window_id, win_title);
}

PSet_Window_Size( xsize, ysize, xpos, ypos)
int xsize, ysize, xpos, ypos;
{
    if( xsize != 0 )
	win_window_width  = xsize;
    if( ysize != 0 )
	win_window_height = ysize;
    win_xpos   = xpos;
    win_ypos   = ypos;
    win_size_hints.flags = USPosition | PPosition | PMinSize | PMaxSize;
    win_size_hints.x	         = win_xpos;
    win_size_hints.y	         = win_xpos;
    win_size_hints.max_width     = win_window_width;
    win_size_hints.max_height    = win_window_height;
    win_size_hints.min_width     = win_window_width;
    win_size_hints.min_height    = win_window_height;
	
    if( dpy_id == NULL )
    if( window_id == NULL ) return;

    XSetNormalHints( dpy_id, window_id, &win_size_hints);
}

Colormap
get_standard_colormap(disp, win_visual, win_depth, win_screen)
Display *disp;
Visual *win_visual;
int win_depth;
int win_screen;
{
    XStandardColormap   cmap_info,*std_cmaps;
    Atom                property;
    int                 i, num_cmaps;

    switch(win_visual->class) {
    case TrueColor:
    case StaticColor:
#ifdef SUN
        property = XA_RGB_BEST_MAP;
#else
        property = XA_RGB_DEFAULT_MAP;
#endif
        break;
    case PseudoColor:
    case DirectColor:
    default:
        property = XA_RGB_DEFAULT_MAP;
        break;
    case StaticGray:
    case GrayScale:
        property = XA_RGB_GRAY_MAP;
        break;
    }

      /* Ensure that the property is defined. */
      if ( XmuLookupStandardColormap( disp, win_screen,
        win_visual->visualid, win_depth, property,
        False, True ) ) {

        /* Get the standard colormap properties. */
        if ( XGetRGBColormaps( disp,
              RootWindow( disp, DefaultScreen( disp ) ),
              &std_cmaps, &num_cmaps, property ) ) {

            /* Find the properties for the specified visual. */
            for ( i = 0; i < num_cmaps; i++ ) {
                if ( win_visual->visualid == std_cmaps[i].visualid ) {
                    cmap_info = std_cmaps[i];
                    break;
                }
            }
            XFree( (char *)std_cmaps );
        }    
      }
    return (cmap_info.colormap);
}

