/*
 ******************************************************************************
 ******************************************************************************
 *****                          dbl_test.c                                *****
 *****									  *****
 *****		 Simple PEXtk double buffering test program		  *****
 *****									  *****
 *****									  *****
 *                                                                            *
 *	Copyright 1991, 1993. All Rights Reserved                             *
 *			SHOgraphics					      *
 *			321 Soquel Way           			      *
 *			Sunnyvale, CA      94086  			      *
 *                                                                            *
 *****									  *****

    Permission to use, copy, modify, and distribute this software and
 its documentation for any purpose and without fee is hereby granted,
 provided that the above copyright notice appear in all copies and that
 both that copyright notice and this permission notice appear in
 supporting documentation, and that the name of SHOgraphics not be
 used in advertising or publicity pertaining to distribution of the 
 software without specific, written prior permission.

 SHOGRAPHICS DISCLAIMS ANY AND ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 INCLUDING ALL EXPRESSED AND IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS. IN NO EVENT SHALL SHOGRAPHICS BE LIABLE FOR ANY DAMAGES OF ANY
 KIND, INCLUDING BUT NOT LIMITED TO SPECIAL, INDIRECT OR CONSEQUENTIAL 
 DAMAGES RESULTING FROM LOSS OF USE, DATA OR PROSPECTIVE PROFITS, WHETHER 
 IN AN ACTION OF CONTRACT, NEGLIENCE OR OTHER TORTIOUS ACTION, ARISING OUT 
 OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

 ************************************************************************** */

/*
	Program:   dbl_test.c

	Synopsis   This program is a simple test of PEXtk.

	Usage:	   stest [options] 

	    where:
		options are:
		    -s 			Uses standard method of dbl buffering
		    -d 			Uses PEXtk choice of dbl buffering
		    -i num		sets iteration count
		    -f 			uses Ppoly_fill instead of Ppoly
		    -w size		sets the size of the window

*/

#include <stdio.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <PEXtk.h>
#include <X11/extensions/multibuf.h>

float line[2][3] = {{ 0.0, 0.0, -5.0},{100.0, 100.0, -15.0}};
float poly[4][3] = {{ 5.0, 5.0, -15.0},{-5.0, 5.0, -15.0},
                    { -5.0, -5.0, -15.0},{ 5.0, -5.0, -15.0}};
float ply2[4][3] = {{ 6.0, 4.0, -16.0},{-4.0, 4.0, -16.0},
                    { -4.0, -6.0, -16.0},{ 6.0, -6.0, -16.0}};

    int			 win_width, win_height;
    Multibuffer 	 mbuf[2];
    int			 cur_buf;
    Display		 *dpy; 

main(argc,argv)
int	argc;
char	*argv[];
{
	Window window;
	int i,iter=10,dbl_buf, clear_flag=0,fill_flag=0;
	PCOLOR	col;
	int upd_act;
	int dbl_flag = 0;
	int dbl_clear();

	win_width=500; win_height=500;

	for(i=1;i<argc;i++){
	    if( strcmp(argv[i],"-i") == 0 ){
		i++;
		iter = atoi(argv[i]);
	    }
	    if( strcmp(argv[i],"-c") == 0 ){	/* if true, call PEXtk clear */
		clear_flag = 1;
	    }
	    if( strcmp(argv[i],"-f") == 0 ){
		fill_flag = 1;
	    }
	    if( strcmp(argv[i],"-s") == 0 ){	/* type of dbl buf to use */
		dbl_flag = 1;
	    }
	    if( strcmp(argv[i],"-w") == 0 ){
		i++;
		win_width = atoi(argv[i]);
		win_height = win_width;
	    }
	}

        if ((dpy = XOpenDisplay ("")) == NULL) {
            printf(" Error opening display\n");
	    return 1;
        }

	if(Pinit(dpy)){
	    printf("ERROR... initializing PEX server\n");
	    exit(1);
	}

	if( (window = Pcreate_window(dpy, win_width, win_height)) == NULL){
	    printf("ERROR... creating window\n");
	    exit(1);
	}
	col.color.rgb.red=1.0;col.color.rgb.green=col.color.rgb.blue=0.;


	if( clear_flag) upd_act = MultibufferUpdateActionUntouched;
	else upd_act = MultibufferUpdateActionBackground;
    
    if( dbl_flag == 0 )
	if(XmbufCreateBuffers( dpy, window, 2, 
	    upd_act, MultibufferUpdateHintFrequent, 
	    mbuf) != 2)
	{
	        printf(" Error... Unable to create multibuffers\n");
        	exit(1);
	}
	cur_buf = 1;	/* the display buffer */

	Pset_drawable(window);

	if( dbl_flag == 0 ){
	    Pset_update_buffer(mbuf[0]);
	    Pset_display_buffer(mbuf[1]);
	    Pset_function( PEXtk_FUNC_CLEAR_BUFFER, dbl_clear);
	} else Pbuffer_mode(PEXtk_DOUBLE_BUFFER);
	Pzbuffer(1);

	Pviewport( 0, win_width-1, 0, win_height-1);
	Ppersp_view( 0, 90.0, 1.0, 1.0, 1000.0);
	for(i=0;i<iter;i++)
	{
	    if( !clear_flag )
		Pclear_rgb( 0.0, 0.0, 0.0);
	    Pline_color(&col);
	    Ppoly_line(2, line);
	    Pcolor(&col);
	    if( fill_flag ){
		Ppoly_fill(4, poly);
		Pcolor_rgb( 0.0, 1.0, 0.0);
		Ppoly_fill(4, ply2);
	    } else {
		Ppoly(4, poly);
		Pcolor_rgb( 0.0, 1.0, 0.0);
		Ppoly(4, ply2);
	    }
	    Protate_z(5.0, 1);
	    if( dbl_flag == 0 ){
	        Pset_update_buffer( mbuf[1-cur_buf]);
	        Pset_display_buffer( mbuf[cur_buf]);
	        cur_buf = 1-cur_buf;
	    } else Pswap_buffers();
	    Pflush(1);
	}
	Pflush(1);
	printf("Press return\n");scanf("%c",&iter);
	Pexit();
	XDestroyWindow( dpy, window);
	XCloseDisplay(dpy);
	exit(0);
}

int
dbl_clear(pix)
long pix;
{
    XSetForeground( dpy, DefaultGC(dpy,DefaultScreen(dpy)), pix);
    XFillRectangle( dpy, mbuf[cur_buf], DefaultGC(dpy,DefaultScreen(dpy)),
	0, 0, win_width, win_height);
}
