/*
 ******************************************************************************
 ******************************************************************************
 *****                             PEXtk                                  *****
 *****									  *****
 *****		    An Immediate Mode Interface to PEX			  *****
 *****									  *****
 *****				Test Programs				  *****
 *****									  *****
 *                                                                            *
 *	Copyright 1991, 1993. All Rights Reserved                             *
 *			SHOgraphics					      *
 *			321 Soquel Way           			      *
 *			Sunnyvale, CA      94086  			      *
 *                                                                            *
 *****									  *****

    Permission to use, copy, modify, and distribute this software and
 its documentation for any purpose and without fee is hereby granted,
 provided that the above copyright notice appear in all copies and that
 both that copyright notice and this permission notice appear in
 supporting documentation, and that the name of SHOgraphics not be
 used in advertising or publicity pertaining to distribution of the 
 software without specific, written prior permission.

 SHOGRAPHICS DISCLAIMS ANY AND ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 INCLUDING ALL EXPRESSED AND IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS. IN NO EVENT SHALL SHOGRAPHICS BE LIABLE FOR ANY DAMAGES OF ANY
 KIND, INCLUDING BUT NOT LIMITED TO SPECIAL, INDIRECT OR CONSEQUENTIAL 
 DAMAGES RESULTING FROM LOSS OF USE, DATA OR PROSPECTIVE PROFITS, WHETHER 
 IN AN ACTION OF CONTRACT, NEGLIENCE OR OTHER TORTIOUS ACTION, ARISING OUT 
 OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

 ************************************************************************ 
 *		PEXtk Test Program for primitive routines
 */
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <stdio.h>
#include <PEXtk.h>

#include "read_demo_obj.c"

    Fcoord3D	 	ctr;
    Fcoord3D	 	*v_norm;
    int			iter, verbose_flag, ModelType, prim_type, l_type;
    int			pause_flag, str_flag, color_flag;
    float		near, far;
    float		fov, aspect;
    int			win_width, win_height;
    PCOLOR		colorObj, colorList[256];


main(argc, argv)
int argc; char **argv;
{
    Display 		 *dpy;
    Window 		 window;
    int 		 i;
    float		 rinc;

    ModelType = 0; pause_flag = verbose_flag = 0;
    iter = 100;
    win_width=500; win_height=500; prim_type = 0; l_type = 0; str_flag = 0;
    near = 1.0; far = 1000.0; aspect = 1.0; fov = 90.0; color_flag = 0;
    v_norm = NULL;

    parse_line(argc,argv);

    if ((dpy = XOpenDisplay ("")) == NULL) {
	printf(" Error opening display\n");
	return 1;
    }

    if(Pinit(dpy)){
	printf("ERROR... initializing PEX server\n");
	exit(1);
    }

    if( (window = Pcreate_window(dpy, win_width, win_height)) == NULL){
	printf("ERROR... creating window\n");
	exit(1);
    }

    Pset_drawable( window );
    Pzbuffer(1);
    Pbuffer_mode(  PEXtk_DOUBLE_BUFFER );
    Pbackface( PEXtk_CULL_NONE);

    colorObj.color.rgb.red   = 0.9; colorObj.color.rgb.green = 0.0;
    colorObj.color.rgb.blue  = 0.4;
    colorList[0].color.rgb.red   = 1.0; colorList[0].color.rgb.green = 0.0;
    colorList[0].color.rgb.blue  = 0.0;
    colorList[1].color.rgb.red   = 0.0; colorList[1].color.rgb.green = 1.0;
    colorList[1].color.rgb.blue  = 0.0;
    colorList[2].color.rgb.red   = 0.0; colorList[2].color.rgb.green = 0.0;
    colorList[2].color.rgb.blue  = 1.0;
    rinc = 1.0/(256-3);
	colorList[3].color.rgb.red   = 1.0;
	colorList[3].color.rgb.green = 0.0;
	colorList[3].color.rgb.blue  = 0.0;
    for(i=3;i<256;i++){
	colorList[i].color.rgb.red   = colorList[3].color.rgb.red - rinc*i;
	colorList[i].color.rgb.green = colorList[3].color.rgb.green + rinc*i;
	colorList[i].color.rgb.blue  = 0.0;
    }

	Set_view();
	if( ModelType ){
	    Pshade_mode( ModelType);
	    Setup_mode();
	}
	/* } else */
	    /* Pshade_mode( PEXtk_SHADE_WIRE); */

	if( prim_type == 2 ){
		Fcoord3D *tmp;
	    if(PEXtk_calc_vertex_normals( vct, poly_cnt, verts, facet_Count,
			vert_indices, &v_norm) == 0 )
		    printf("ERROR calculating vertex normals\n");
		for(tmp=v_norm,i=0;i<vct;i++){
		    /* printf(" %f %f %f\n",tmp->x,tmp->y,tmp->z);tmp++; */
		}
	}
	Pcolor( &colorObj);
	Pline_color( &colorObj);

    i = 0;
	Create_str();
	Ptranslate( -ctr.x, -ctr.y, 0.0, 1);
    while( i < iter ){
	Pclear_rgb( 0.0, 0.0, 0.0);
	Protate_z( 3.0, 1 );
	Exec_str( );
	Pswap_buffers(); 
	Pflush(1);
	i++;
    }
    if( pause_flag ){
	printf("enter a char ");
	fscanf( stdin, "%c",&i);
	}
      Delete_str();
    Pexit();
    XDestroyWindow( dpy, window);
    XCloseDisplay( dpy);
    if( v_norm != NULL ) free(v_norm);
    exit(0);
}

Setup_mode()
{
    PEXtk_LIGHT lgt;

    if(!( ModelType & PEXtk_SHADE_LIGHTS) ) return;
    Plight_color_rgb( 0.6, 0.6, 0.6);
    Plight_ambient( .6);
    switch(l_type ){
      default:
      case PEXtk_LIGHT_INFINITE:
	lgt.type = PEXtk_LIGHT_INFINITE;
	Plight_color_rgb( 1.0, 1.0, 1.0);
	lgt.dir.x = -.707; lgt.dir.y = 0.0; lgt.dir.z = -.707;
	break;
      case PEXtk_LIGHT_POINT:
	lgt.type = l_type;
	Plight_color_rgb( 0.6, 0.6, 0.6);
	/* lgt.dir.x = 0.0; lgt.dir.y = 0.0; lgt.dir.z = -1.0; */
	lgt.pos.x = 20.0; lgt.pos.y = 20.0; lgt.pos.z = 21.0;
	lgt.atten1 = .5;
	lgt.atten2 = 0.0;
	break;
      case PEXtk_LIGHT_SPOT:
	lgt.type = l_type;
	lgt.dir.x = 0.0; lgt.dir.y = 0.0; lgt.dir.z = -1.0;
	lgt.pos.x = 20.0; lgt.pos.y = 20.0; lgt.pos.z = 21.0;
	lgt.conc = 1.;
	lgt.angle = 60.0*3.141592/180.0;
	lgt.atten1 = .5;
	lgt.atten2 = 0.0;
	Plight_color_rgb( 0.6, 0.6, 0.6);
	break;
    }
    Plight( 1, &lgt);
    Pspecular_color_rgb( 1.0, 1.0, 1.0);            /* white specular */
    Preflection_property( PEXtk_REFL_SPECULAR, .5);
    Preflection_property( PEXtk_REFL_EXPONENT, 5.);
}
	
Set_view()
{
    Ppersp_view( 0, fov, aspect, near, far);
}

Create_str()
{
    int i, j, k, m, n;
    Fcoord3D	 vn[255], vt[255];
    Fcoord3D *vdata; short *edges; short *fC;

    if( str_flag )
	Pcreate_struct(1);
      switch(prim_type){
       case 0:
         for(i=0,j=0; i<poly_cnt; i++) {
           k = vert_indices[j];
           for( m=0;m<facet_Count[i]; m++){
             n = vert_indices[j++];
             vt[m].x = verts[n].x; vt[m].y = verts[n].y; vt[m].z = verts[n].z;
	   }
	  Ppoly( m, vt);
	 }
	break;
       case 1:
         for(i=0,j=0; i<poly_cnt; i++) {
           k = vert_indices[j];
           for( m=0;m<facet_Count[i]; m++){
             n = vert_indices[j++];
	     if( color_flag )
		Pvertex_color( &colorList[n%255]);
             vt[m].x = verts[n].x; vt[m].y = verts[n].y; vt[m].z = verts[n].z;
	   }
	   Ppoly_fill( (int) facet_Count[i], vt);
	 }
	break;
       case 2:
         for(i=0,j=0; i<poly_cnt; i++) {
	   if( color_flag )
            for( m=0;m<facet_Count[i]; m++){
              n = vert_indices[j++];
              vt[m].x = verts[n].x; vt[m].y = verts[n].y; vt[m].z = verts[n].z;
	   }
           for( m=0;m<facet_Count[i]; m++){
             n = vert_indices[j++];
	     if( color_flag )
		Pvertex_color( &colorList[n%255]);
             vt[m].x = verts[n].x; vt[m].y = verts[n].y; vt[m].z = verts[n].z;
             vn[m].x =v_norm[n].x; vn[m].y =v_norm[n].y; vn[m].z = v_norm[n].z;
	   }
	   Ppoly_shade( (int) facet_Count[i], vt, vn);
	 }
	break;
       case 3:
         for(i=0,j=0; i<poly_cnt; i++) {
/*
for(k=0;k<numEdges;k++)
printf(" %d",vert_indices[k]);
printf("\n");
*/
           k = vert_indices[j];
           for( m=0;m<facet_Count[i]; m++){
             n = vert_indices[j++];
             vt[m].x = verts[n].x; vt[m].y = verts[n].y; vt[m].z = verts[n].z;
	   }
	   Ppoly_fill_area(0,0,(int)facet_Count[i], PEXtk_COLOR_RGB, NULL, vt);
	 }
	break;
       case 4:
	if((vdata = (Fcoord3D *)malloc(sizeof(Fcoord3D)*vct))
		== NULL){
	    printf("ERROR... can't get any memory\n");
	    exit(1);
	}
	if((edges = (short *)malloc(sizeof(short)*numEdges))
		== NULL){
	    printf("ERROR... can't get any memory\n");
	    exit(1);
	}
	if((fC = (short *)malloc(sizeof(short)*poly_cnt))
		== NULL){
	    printf("ERROR... can't get any memory\n");
	    exit(1);
	}
	for(i=0;i<poly_cnt;i++)
	    fC[i] = facet_Count[i];
	for(i=0;i<numEdges;i++)
	    edges[i] = vert_indices[i];

	for(i=0;i<vct;i++){
	    vdata[i].x = verts[i].x;
	    vdata[i].y = verts[i].y;
	    vdata[i].z = verts[i].z;
	}
	  /* Ppoly_index( 0, 0, poly_cnt, vct, numEdges,PEXtk_COLOR_RGB, */
	    /* (short *) fC, NULL, (short *) edges, (float *)vdata); */

/*
for(i=0;i<numEdges;i++)
printf(" %d",edges[i]);
printf("\n");
for(i=0;i<vct;i++)
printf(" %f %f %f\n",vdata[i].x,vdata[i].y,vdata[i].z);
printf("poly_index, poly %d, vct %d, ne %d,Fc0 %d, fC1 %d\n",
		poly_cnt,vct,numEdges,facet_Count[0],facet_Count[1]);
*/
	  Ppoly_index( 0, 0, poly_cnt, vct, numEdges,PEXtk_COLOR_RGB,
	    facet_Count, NULL, vert_indices, verts);
	break;
       case 5:
         for(i=0,j=0; i<poly_cnt; i++) {
           k = vert_indices[j];
           for( m=0;m<facet_Count[i]; m++){
             n = vert_indices[j++];
		if( m == 0 )
		  Pmove_3d( verts[n].x, verts[n].y,verts[n].z);
		else
		  Pdraw_3d( verts[n].x, verts[n].y,verts[n].z);
	   }
	  Pdraw_3d( verts[k].x, verts[k].y,verts[k].z); /* join to 1st pt */
	 }
	break;
      } /* end switch */
   if( str_flag )
    Pclose_struct();
}

Delete_str()
{
  if( str_flag )
   if(Pdelete_struct(1))
	printf("ERROR... deleting structure 1\n");
}

Exec_str()
{
    if( str_flag )
	Pcall_struct( 1);
    else
	Create_str( );
}

/*
 *                      +---------------------------+
 * ---------------------+         parse_line        +------------------------
 *                      +---------------------------+
 *  FUNCTION:
 *      Parses the command line.
 *
 *  USAGE:
 *      parse_line(argc, argv,options)
 *	int	   argc 	   (in)  - number of arg's
 *	char   	   argv 	   (in)  - pointer to arg list
 *
 *  DESCRIPTION:
 *	This routine will parse the command line for the following options:
 *	   opt  argument	type	  comment
 *	   ---  ------------    ----      ----------------------------------
 *	    -f filename.obj	string	  .obj file to read
 *	    -i iter		int	  number of iterations
 *	    -c 			void	  sets color per vertex flag
 *	    -p 			void	  sets pause flag
 *	    -v 			void	  sets verbose flag
 *	    -l type		int	  light type
 *					  the shade mode will be set to
 *					  PEXtk_SHADE_LIGHTS
 *	    -m mode		int	  shade mode, in decimal
 *	    -t type		int	  primitive type
 *	    -s 			void	  sets structure mode
 *	    -w size		int	  window size
 *	    -z trans		float	  z translation value
 *	    -h near		float	  near (hither) plane
 *	    -y far		float	  far (yon) plane
 *	    -a aspect		float	  aspect ratio
 *                                
*/
int parse_line(argc,argv)
int argc;
char **argv;
{
     char *temp,ffile[32];
     extern char *optarg;
     extern int optind;
     int file_err;
     int c;

    if( argc < 2 ){
	Usage();
	exit(1);
    }

    file_err = 1;

    while((c=getopt(argc,argv,"a:t:i:h:y:l:m:f:w:cpsvz:")) != EOF)
    {
	switch(c){
	    case 'f':
		  read_demo_obj(optarg, &ctr, 1.0);
		  file_err = 0;
		break;
	    case 'i':
		  sscanf(optarg,"%d",&iter);
		break;
	    case 'a':
		  sscanf(optarg,"%f",&aspect);
		break;
	    case 'l':
		  sscanf(optarg,"%d",&l_type);
		  ModelType = ModelType | PEXtk_SHADE_LIGHTS;
		break;
	    case 'm':
		  sscanf(optarg,"%d",&ModelType);
		break;
	    case 'c':
		  color_flag = 1;
		break;
	    case 'p':
		  pause_flag = 1;
		break;
	    case 's':
		  str_flag = 1;
		break;
	    case 't':
		  sscanf(optarg,"%d",&prim_type);
		break;
	    case 'v':
		  verbose_flag = 1;
		break;
	    case 'z':
		  sscanf(optarg,"%f",&z_trans_factor);
		break;
	    case 'h':
		  sscanf(optarg,"%f",&near);
		break;
	    case 'y':
		  sscanf(optarg,"%f",&far);
		break;
	}
    }

    for(; optind < argc;optind++ ){
		  read_demo_obj(argv[optind], &ctr, 1.0);
		  file_err = 0;
    }
    if( file_err ){
	Usage();
	exit(1);
    }
}

Usage()
{
    printf("Error...  Usage is:\n");
    printf(" prm_test [options] file.obj\n");
    printf(" options: \n");
    printf("\t -m mode {0-9}\t(shade mode, sum of desired modes):\n");
	printf("\t\t PEXtk_SHADE_NONE \t\t%d\n",PEXtk_SHADE_NONE);
	printf("\t\t PEXtk_SHADE_WIRE \t\t%d\n",PEXtk_SHADE_WIRE);
	printf("\t\t PEXtk_SHADE_FLAT \t\t%d\n",PEXtk_SHADE_FLAT);
	printf("\t\t PEXtk_SHADE_COLORS \t\t%d\n",PEXtk_SHADE_COLORS);
	printf("\t\t PEXtk_SHADE_NORMALS \t\t%d\n",PEXtk_SHADE_NORMALS);
	printf("\t\t PEXtk_SHADE_SPECULAR \t\t%d\n",PEXtk_SHADE_SPECULAR);
	printf("\t\t PEXtk_SHADE_LIGHTS \t\t%d\n",PEXtk_SHADE_LIGHTS);
	printf("\t\t PEXtk_SHADE_HIDDENLINE \t%d\n",PEXtk_SHADE_HIDDENLINE);
	printf("\t\t PEXtk_SHADE_ANTIALIAS \t\t%d\n",PEXtk_SHADE_ANTIALIAS);
	printf("\t\t PEXtk_SHADE_DOTPRODUCT \t%d\n",PEXtk_SHADE_DOTPRODUCT);
    printf("\t -l type {1-3}\t(light type)\n");
	printf("\t\t PEXtk_LIGHT_INFINITE \t\t%d\n",PEXtk_LIGHT_INFINITE);
	printf("\t\t PEXtk_SHADE_POINT \t\t%d\n",PEXtk_LIGHT_POINT);
	printf("\t\t PEXtk_SHADE_SPOT \t\t%d\n",PEXtk_LIGHT_SPOT);
    printf("\t -i iter\t(number of iterations)\n");
    printf("\t -t type {0-5}\n");
	printf("\t\t Ppoly \t\t\t0\n");
	printf("\t\t Ppoly_fill \t\t1\n");
	printf("\t\t Ppoly_shade \t\t2\n");
	printf("\t\t Ppoly_fill_area \t3\n");
	printf("\t\t Ppoly_index \t\t4\n");
	printf("\t\t move/draw \t\t5\n");
    printf("\t -c\t\t(color per vertex flag)\n");
    printf("\t -p\t\t(pause flag)\n");
    printf("\t -w size\t(set window width and height)\n");
    printf("\t -h hither\t(set hither plane)\n");
    printf("\t -y yon\t\t(set yon plane)\n");
    printf("\t -v\t\t(verbose flag)\n");
    printf("\t -f file.obj\n");
}
