/*
 ******************************************************************************
 ******************************************************************************
 *****                        read_obj.c                                  *****
 *****									  *****
 *****		    File reading utility for .obj files			  *****
 *****									  *****
 *****									  *****
 *                                                                            *
 *	Copyright 1992, 1993. All Rights Reserved                             *
 *			SHOgraphics					      *
 *			321 Soquel Way           			      *
 *			Sunnyvale, CA     94086   			      *
 *                                                                            *
 *****									  *****

    Permission to use, copy, modify, and distribute this software and
 its documentation for any purpose and without fee is hereby granted,
 provided that the above copyright notice appear in all copies and that
 both that copyright notice and this permission notice appear in
 supporting documentation, and that the name of SHOgraphics not be
 used in advertising or publicity pertaining to distribution of the 
 software without specific, written prior permission.

 SHOGRAPHICS DISCLAIMS ANY AND ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 INCLUDING ALL EXPRESSED AND IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS. IN NO EVENT SHALL SHOGRAPHICS BE LIABLE FOR ANY DAMAGES OF ANY
 KIND, INCLUDING BUT NOT LIMITED TO SPECIAL, INDIRECT OR CONSEQUENTIAL 
 DAMAGES RESULTING FROM LOSS OF USE, DATA OR PROSPECTIVE PROFITS, WHETHER 
 IN AN ACTION OF CONTRACT, NEGLIENCE OR OTHER TORTIOUS ACTION, ARISING OUT 
 OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

 ************************************************************************ */

#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <math.h>

typedef struct rgbFloat {
	float	r,g,b;
	} rgbFloat;

typedef struct {
	float	x;
	float	y;
	float	z;
	} Fcoord3D;

typedef struct {
	float	x;
	float	y;
	float	z;
	float	w;
	} Fcoord4D;

#include "obj_def.h"

#define USE_DB
#ifdef USE_DB
#include "include/database.h"
#endif

	char text_string[256];
	static char *IO_get_next_float();
	static char *IO_parse_to_next_token();

#define DEF_PARM_SIZE	100
#define BUF_SIZE	1024

/*
 *  ROUTINE:	read_obj
 *  PURPOSE:	reads a Wavefront .obj file
 */

read_obj(argv, Obj_data)
char		*argv;
ObjData		*Obj_data;
{
    FILE *fp2, *fopen();
    char buffer[BUF_SIZE+1],objname2[128], *bufptr, *ptr;
    char *objname;
    int value;
    int i, j, k, lcount, itmp;
    float *f_ptr, f_value;
    double atof();
    int flag, el_type;

    int		flip_flag;
    Fcoord3D    *vertt, *vertn, *vertp;
    Fcoord4D    *verts;
    unsigned short  *vert_indices;
    unsigned short  *vt_indices;
    unsigned short  *vn_indices;
    unsigned short  *vc_indices;
    unsigned short  *facet_Count, *norm_Count;
    int             *poly_cnt, *vct, *vtct, *vnct, *vpct, *numEdges;
    float		z_trans;
    rgbFloat		*col;
    Fcoord3D 		*ctr;
    Fcoord3D 		*global_pos;
    Fcoord3D 		*global_min;
    Fcoord3D 		*global_max;
    ELEM_DATA		el_data;

	z_trans	   = Obj_data->z_trans;
	ctr	   = &Obj_data->ctr;
	global_pos = &Obj_data->global_pos;
	global_max = &Obj_data->global_max;
	global_min = &Obj_data->global_min;
	poly_cnt   = &Obj_data->poly_cnt;
	vct	   = &Obj_data->vct;
	vtct	   = &Obj_data->vtct;
	vnct	   = &Obj_data->vnct;
	vpct	   = &Obj_data->vpct;
	numEdges   = &Obj_data->numEdges;
	flip_flag  = Obj_data->flip_flag;
	verts	   = Obj_data->verts;
	vertt	   = Obj_data->vertt;
	vertn	   = Obj_data->vertn;
	vert_indices= Obj_data->vert_indices;
	vt_indices = Obj_data->vt_indices;
	vn_indices = Obj_data->vn_indices;
	vc_indices = Obj_data->vc_indices;
	facet_Count= Obj_data->facet_Count;
	norm_Count = Obj_data->norm_Count;

	Obj_data->mask = 0;


    objname = argv;

	strcpy(objname2,objname);
						/* read in file.obj */
    if((fp2 = fopen(objname, "r")) == NULL)  {
	strcat(objname2,".obj");
	
    	if((fp2 = fopen(objname2, "r")) == NULL)  {
	  printf("Error: can't open object %s\n", objname);
	  exit(-1);
	}
    }

    lcount = 0;
    while(1) {
      if(fgets(buffer, BUF_SIZE, fp2) == NULL)
	    break;

      lcount++;
      switch(*buffer)  {
	default:
	    break;
	case 'v':
	  if( *(buffer+1) == ' ' ){
		itmp = *vct;
		bufptr = buffer+2;
	        bufptr = IO_get_next_float( fp2, bufptr, buffer, &f_value);
		verts[itmp].x = f_value;
	        bufptr = IO_get_next_float( fp2, bufptr, buffer, &f_value);
		verts[itmp].y = f_value;
	        bufptr = IO_get_next_float( fp2, bufptr, buffer, &f_value);
		verts[itmp].z = f_value + z_trans;
	        bufptr = IO_get_next_float( fp2, bufptr, buffer, &f_value);
		if( bufptr != NULL )
		    verts[itmp].w = f_value;
		else
		    verts[itmp].w = 1.0;
		/*
		 * Find the min's and max's
		 */
		if( verts[itmp].x > global_max->x)
			global_max->x = verts[itmp].x;
		if( verts[itmp].x < global_min->x)
			global_min->x = verts[itmp].x;
		if( verts[itmp].y > global_max->y)
			global_max->y = verts[itmp].y;
		if( verts[itmp].y < global_min->y)
			global_min->y = verts[itmp].y;
		if( verts[itmp].z > global_max->z)
			global_max->z = verts[itmp].z;
		if( verts[itmp].z < global_min->z)
			global_min->z = verts[itmp].z;
	    (*vct)++;
	  } else if( *(buffer+1) == 'n' ){
	    sscanf(buffer, "vn %f %f %f\n",
		    &vertn[*vnct].x, &vertn[*vnct].y, &vertn[*vnct].z);
	    (*vnct)++;
	  } else if( *(buffer+1) == 't' ){
	    sscanf(buffer, "vt %f %f %f\n",
		    &vertt[*vtct].x, &vertt[*vtct].y, &vertt[*vtct].z);
	    (*vtct)++;
	  } else if( *(buffer+1) == 'p' ){
		itmp = *vpct;
		bufptr = buffer+3;
	        bufptr = IO_get_next_float( fp2, bufptr, buffer, &f_value);
		verts[itmp].x = f_value;
	        bufptr = IO_get_next_float( fp2, bufptr, buffer, &f_value);
		verts[itmp].y = f_value;
	        bufptr = IO_get_next_float( fp2, bufptr, buffer, &f_value);
		verts[itmp].z = f_value;

	    (*vpct)++;
	  }
	    break;
	case 'f' :		/* a polygon face */
	case 'l' :		/* a polyline */

	    if( *buffer == 'l' )
	        el_type = EL_TY_LINE;
	    else
	        el_type = EL_TY_POLY;

	    i=0;		/* count of num indicies for this primitive */
				/* clear out the newline '\n'. The */
				/* parse_buffer routine will keep parsing */
				/* until null is found */
	    bufptr = index( buffer, '\n'); /* find the newline (\n) */
	    if( bufptr == (char *) NULL ) {
		break;
	    }
	    *bufptr = '\000';

	    i = parse_buffer( fp2, buffer, Obj_data, el_type, *vct, &lcount);
	    if( i == 0 ) break;

		if( flip_flag ){	/* swap indices */
		    k = *numEdges - i;
		    for(j= k, itmp= *numEdges-1; 
			j < k + i/2; j++, itmp--){
			value = Obj_data->vert_indices[j];
			Obj_data->vert_indices[j]=Obj_data->vert_indices[itmp];
			Obj_data->vert_indices[itmp] = value;
	            }
		}

	        if( flag == 2 || flag == 3 ){
		    norm_Count[*poly_cnt]  = i;
		    /* printf("setting normal count to %d\n",i); */
		}
	     (*poly_cnt)++;
#ifdef USE_DB
	    DB_add_element(argv, el_type, Obj_data);
#endif
	  break;
	case 't' :		/* text */
	    i=0;
	    bufptr = index( buffer, '\n');
	    if( bufptr == NULL ) break;
	    *bufptr = '\000';
	    bufptr = buffer;
	    while( (*bufptr != '\0') && *bufptr != ' ' ) bufptr++;
	    if( *bufptr == '\0' ) break;
	    bufptr=strtok( bufptr, " ");
	    while( (*bufptr != '\0') && *bufptr == ' ' ) bufptr++;
	    if( *bufptr == '\0' ) break;

	    if( is_digit(*bufptr) )
		    value = atoi(bufptr);
	    else
		    break;
	    if( value == 0 ){ break;}
	    if( value > *vct+1 )
		    error_index("text", "vertex", lcount);
	    /* look for scale factor */
		bufptr= strtok(NULL," ");
	    /* while( bufptr && *bufptr == ' ' ) bufptr++; */
	    if( *bufptr == '\0' ) break;
		    f_value = atof(bufptr);
		bufptr= strtok(NULL," ");
	    /* while( bufptr && *bufptr != ' ' ) bufptr++; */
	    /* if( bufptr == NULL ) break; */
	    /* look for string */
	    /* while( bufptr && *bufptr == ' ' ) bufptr++; */
	    if( *bufptr == '\0' ) break;
	    strcpy(Obj_data->string, bufptr);
	    Obj_data->f_value = f_value;

		Obj_data->vert_indices[*numEdges] = value-1;
		(*numEdges)++;
		i++;

	    facet_Count[*poly_cnt] = i;
	   (*poly_cnt)++;
#ifdef USE_DB
	    DB_add_element(argv, EL_TY_TEXT, Obj_data);
#endif
	  break;
	case 'p' :		/* parm statement */

	    if( strncmp(buffer,"parm",4) != 0) break;
	    bufptr = index( buffer, '\n');
	    if( bufptr == NULL ) break;
	    *bufptr = '\000';
	    bufptr = buffer+5;
	    if( *bufptr == 'u' )	/* parm u or parm v */
		flag = 0;
	    else
		flag = 1;
	    bufptr = buffer+7;
	    if( *bufptr == '\\' )
      		    if(fgets(buffer, BUF_SIZE, fp2) == NULL)
	    		return;	/* error */
	    if( *bufptr == '\0' ) return ;

	    /* look past spaces */
	    while( (*bufptr != '\0') && isspace(*bufptr) ) bufptr++;

	    if( flag )
		f_ptr = Obj_data->parm_v;
	    else
		f_ptr = Obj_data->parm_u;
	    i = 0;
	    while( (*bufptr != '\0') ){
	        bufptr = IO_get_next_float( fp2, bufptr, buffer, f_ptr);
		if( bufptr == NULL ) break; /* check against NULL */
		i++; f_ptr++;
		/* Check to make sure that the parm buffer isn't full */
		if( flag ){  /* parm v */
		    if( i >= Obj_data->size_parm_v ){
			Obj_data->size_parm_v += DEF_PARM_SIZE;
		        if(( Obj_data->parm_v = (float *)realloc(
			    Obj_data->parm_v, Obj_data->size_parm_v)) == NULL)
			      return;
		    }
		} else {     /* parm u */
		    if( i >= Obj_data->size_parm_u ){
			Obj_data->size_parm_u += DEF_PARM_SIZE;
		        if(( Obj_data->parm_u = (float *)realloc(
			    Obj_data->parm_u, Obj_data->size_parm_u)) == NULL)
			      return;
		    }
		}
	    }
	    if( flag )
		Obj_data->nparm_v = i;
	    else
		Obj_data->nparm_u = i;
printf("parm read %d parameters\n",i);
	  break;
	case 's' :		/* a surf */
	    if( strncmp(buffer,"surf",4) != 0) break;
	    bufptr = index( buffer, '\n'); /* clear out the newline (\n) */
	    if( bufptr == NULL ) {   /* it makes it easier to parse */
		break;
	    }
	    *bufptr = '\000';
	    el_type = EL_TY_SURF;
	    bufptr = buffer;
		/* look for first space */
	    while( (*bufptr != '\0') && !isspace(*bufptr)) {
		if( *bufptr == '\\'){
      		    if(fgets(buffer, BUF_SIZE, fp2) == NULL)
	    		return;	/* error */
		    bufptr = buffer;
		} else
		    bufptr++;
	    }
	    /* look past spaces */
	    while( (*bufptr != '\0') &&  isspace(*bufptr) ) bufptr++;
	    if( *bufptr == '\0' ) break;

	    bufptr = IO_get_next_float( fp2, bufptr, buffer, &Obj_data->umin);
	    bufptr = IO_get_next_float( fp2, bufptr, buffer, &Obj_data->umax);
	    bufptr = IO_get_next_float( fp2, bufptr, buffer, &Obj_data->vmin);
	    bufptr = IO_get_next_float( fp2, bufptr, buffer, &Obj_data->vmax);
	    i = parse_buffer( fp2, bufptr, Obj_data, el_type, *vct, &lcount);
printf("surf read %d indices\n",i);
	    if( i == 0 ) break;
		/* facet_Count[*poly_cnt] = i; */
		(*poly_cnt)++;
	break;
	case 'c' :		/* a curv or cstype */
				/* curv umin umax {ind}* */
	    if( strncmp(buffer,"curv ",5) == 0) {
		bufptr = index( buffer, '\n');/* clear out the newline (\n) */
					      /* it makes it easier to parse */
		if( bufptr == NULL ) break;
		*bufptr = '\000';
		el_type = EL_TY_CURV;
		bufptr = buffer;
		/* look for first space */
		while( bufptr && (*bufptr != '\0') && !isspace(*bufptr)) {
		if( *bufptr == '\\'){
      		    if(fgets(buffer, BUF_SIZE, fp2) == NULL)
	    		return;	/* error */
		    bufptr = buffer;
		} else
		    bufptr++;
		}
		/* look past spaces */
		while(bufptr && (*bufptr != '\0') && isspace(*bufptr))bufptr++;
		if( *bufptr == '\0' ) break;

		bufptr=IO_get_next_float(fp2, bufptr, buffer, &Obj_data->umin);
		bufptr=IO_get_next_float(fp2, bufptr, buffer, &Obj_data->umax);
		i = parse_buffer( fp2, bufptr, Obj_data, el_type,*vct,&lcount);
		if( i == 0 ) break;
		/* facet_Count[*poly_cnt] = i; */
		(*poly_cnt)++;
	        break;
	    } else 
				/* curv2 {ind}* */
	    if( strncmp(buffer,"curv2",5) == 0) {
		bufptr = index( buffer, '\n');/* clear out the newline (\n) */
					      /* it makes it easier to parse */
		if( bufptr == NULL ) break;
		*bufptr = '\000';
		el_type = EL_TY_CURV;
		bufptr = buffer;
		/* look for first space */
		while( bufptr && (*bufptr != '\0') && !isspace(*bufptr)) {
		if( *bufptr == '\\'){
      		    if(fgets(buffer, BUF_SIZE, fp2) == NULL)
	    		return;	/* error */
		    bufptr = buffer;
		} else
		    bufptr++;
		}
		/* look past spaces */
		while(bufptr && (*bufptr != '\0') && isspace(*bufptr))bufptr++;
		if( *bufptr == '\0' ) break;

		bufptr=IO_get_next_float(fp2, bufptr, buffer, &Obj_data->umin);
		bufptr=IO_get_next_float(fp2, bufptr, buffer, &Obj_data->umax);
		i = parse_buffer( fp2, bufptr, Obj_data,el_type,*vpct,&lcount);
		if( i == 0 ) break;
		/* facet_Count[*poly_cnt] = i; */
		/* (*poly_cnt)++; */
	        break;
	    } else 
	    		/* cstype [rat] {bspline|bspline} */
	    if( strncmp(buffer,"cstype",6) == 0){
		bufptr = index( buffer, '\n');/* clear out the newline (\n) */
					      /* it makes it easier to parse */
		if( bufptr == NULL ) break;
		*bufptr = '\000';
		bufptr = buffer+6;
		/* look past spaces */
		while( (*bufptr != '\0') && isspace(*bufptr) ) bufptr++;
		if( bufptr == NULL ) break;
printf("cstype\n");
		if( strncmp(bufptr,"rat",3) == 0){
		    Obj_data->ratnl = 1;
		    bufptr += 3;
	            /* look past spaces */
	            while( (bufptr != NULL) && isspace(*bufptr) ) bufptr++;
		}
		if( strncmp(bufptr,"bsp",3) == 0) Obj_data->ptype = 1;
		if( strncmp(bufptr,"bez",3) == 0) Obj_data->ptype = 2;
printf(" type %s %s\n",(Obj_data->ratnl == 1) ? "rat" : "",
		(Obj_data->ptype ==1) ? "bspline" : "bezier");
		break;
	    }
	case 'd' :		/* deg */
	    if( strncmp(buffer,"deg",3) != 0) break;
	    bufptr = index( buffer, '\n'); /* clear out the newline (\n) */
	    if( bufptr == NULL ) {	   /* it makes it easier to parse */
		break;
	    }
	    *bufptr = '\000';
		bufptr = IO_parse_to_next_token( fp2, buffer);
		if( *bufptr == '\0' ) break;
		Obj_data->deg_u = atoi( bufptr);
		bufptr = IO_parse_to_next_token( fp2, bufptr);
		if( *bufptr == '\0' ) break;
		Obj_data->deg_v = atoi( bufptr);
	   break;
	case 'e' :		/* end */
	    if( *(buffer+1) == 'n' ){
		DB_add_element(argv, el_type, Obj_data);
	     }
	   break;
      } /* end switch */
    } /* end while */

    fclose(fp2);

    printf("%s contains %d vertices, %d vert norm's, %d polys, %d edges\n\n", 
		objname2, *vct, *vnct, *poly_cnt, *numEdges );

    global_pos->x = fabs(global_max->x);
    if( fabs(global_min->x) > global_pos->x )
	global_pos->x = fabs(global_min->x);

    global_pos->y = fabs(global_max->y);
    if( fabs(global_min->y) > global_pos->y )
	global_pos->y = fabs(global_min->y);

    global_pos->z = fabs(global_max->z);
    if( fabs(global_min->z) > global_pos->z )
	global_pos->z = fabs(global_min->z);

#ifdef PRINT_DATA
    {
	printf(" global_x %f, global_y %f, global_z %f\n",
		global_pos->x,global_pos->y,global_pos->z);
	printf(" global_x_max %f, global_y_max %f, global_z_max %f\n",
		global_max->x,global_max->y,global_max->z);
	printf(" global_x_min %f, global_y_min %f, global_z_min %f\n",
		global_min->x,global_min->y,global_min->z);
    }
#endif

        ctr->x = global_min->x + (global_max->x - global_min->x) / 2.0;
        ctr->y = global_min->y + (global_max->y - global_min->y) / 2.0;
        ctr->z = global_min->z + (global_max->z - global_min->z) / 2.0;

}

static char *
IO_parse_to_next_token( fp2, buffer)
FILE *fp2;
char *buffer;
{
	char *bufptr;
	bufptr = buffer;

		/* look for first space */
	    while( (*bufptr != '\0') && !isspace(*bufptr)) {
		if( *bufptr == '\\'){
      		    if(fgets(buffer, BUF_SIZE, fp2) == NULL)
	    		return NULL;	/* error */
		    bufptr = buffer;
		} else
		    bufptr++;
	    }
	    /* look past spaces */
	    while( (*bufptr != '\0') &&  isspace(*bufptr) ) bufptr++;
	    /* if( bufptr == '\0' ) return NULL; */

	return bufptr;
}

error_index( type, str, lcount)
int type;
char *str;
int  lcount;
{
    switch( type ){
	case EL_TY_LINE:
    	    printf("line %s index value is out of range at line %d\n",
		str,lcount);
	  break;
	case EL_TY_POLY:
    	    printf("polygon %s index value is out of range at line %d\n",
		str,lcount);
	  break;
	case EL_TY_MESH:
    	    printf("mesh %s index value is out of range at line %d\n",
		str,lcount);
	  break;
	case EL_TY_SURF:
    	    printf("nurb %s index value is out of range at line %d\n",
		str,lcount);
	  break;
	case EL_TY_TEXT:
    	    printf("text %s index value is out of range at line %d\n",
		str,lcount);
	  break;
	default:
    	    printf("unknown element type %s value is out of range at line %d\n",
		str,lcount);
	  break;
    }
    exit(-1);
}

#ifndef USE_DB
DB_add_element()
{
  return;
}
#endif

IO_init_data( obj_data)
ObjData *obj_data;
{

    obj_data->nparm_u = 0;
    obj_data->size_parm_u = DEF_PARM_SIZE;
    obj_data->parm_u = (float *) malloc(DEF_PARM_SIZE * sizeof(float));
    obj_data->nparm_v = 0;
    obj_data->size_parm_v = DEF_PARM_SIZE;
    obj_data->parm_v = (float *) malloc(DEF_PARM_SIZE * sizeof(float));
    obj_data->umin = obj_data->vmin = 0.0;
    obj_data->umax = obj_data->vmax = 1.0;
    obj_data->deg_u = obj_data->deg_v = 3;
    obj_data->ratnl = 0;
}


/*
 *  ROUTINE:	parse_buffer
 *  PURPOSE:	This routine parses the input buffer for the 'indicies'
 *		of an element.
 *  RETURNS:	number of 'items' in buffer
 */
int
parse_buffer( fp2, buffer, Obj_data, el_type, n_ind, lcount )
FILE	*fp2;
char	*buffer;
ObjData *Obj_data;
int	el_type;
int	n_ind;		/* max number of indices allowed for this element */
int	*lcount;
{
    char *bufptr, *ptr;
    int value;
    int i, j, k, itmp, max_index, max_cnt;
    int flag;
    int             *poly_cnt, *vct, *vtct, *vnct, *vpct, *numEdges;

	poly_cnt   = &Obj_data->poly_cnt;
	vct	   = &Obj_data->vct;
	vtct	   = &Obj_data->vtct;
	vnct	   = &Obj_data->vnct;
	vpct	   = &Obj_data->vpct;
	numEdges   = &Obj_data->numEdges;
	max_index  = Obj_data->max_index;
	max_cnt    = Obj_data->max_cnt;

	i = 0;
	    bufptr = buffer;		/* look for first whitespace */
	    while( (bufptr != NULL) && !isspace(*bufptr)) bufptr++;
	    if( bufptr == NULL ) return 0;

	    /* look past spaces */
	    while( (bufptr != NULL) &&  isspace(*bufptr) ) bufptr++;
	    while(1){	/* foreach v, v/vt, v/vt/vn, or v//vn in buffer */
		if( bufptr == NULL ) break;

		if( is_digit(*bufptr) )
		    value = atoi(bufptr);
		else 
		  if( *bufptr == '\\' ){
		     if(fgets(buffer, BUF_SIZE, fp2) == NULL)
			return 0;	/* error */
		     bufptr = buffer;
		     (*lcount)++;
		     while( (*bufptr != '\0') &&  isspace(*bufptr) ) bufptr++;
		     continue;
		} else {break;}

		if( value == 0 ){ break;}
		if( value < 0 )
		    value = n_ind + value + 1;
		if( value > n_ind+1 || value < 1 )
		    error_index(el_type,"vertex",*lcount);

		Obj_data->vert_indices[*numEdges] = value-1;
		i++; flag=0;
		if( *numEdges > max_index ){
		    fprintf(stderr,
			"ERROR... max number indicies reached at line %d\n",
			*lcount);
		    return 0;
		}
		(*numEdges)++;
		/* look for next non-digit */
	        while( (bufptr != NULL) && is_digit(*bufptr) ) bufptr++;
		if( bufptr == NULL ) break;

		/* check for a v//vn v/vt/vn */
		if( *bufptr == '/' ){
		   /* is it v/vt (case 1),  v//vn (case 2), v/vt/vn (case 3) */
		    bufptr++; 
		    flag = 1;
		    ptr = bufptr;
		    if( *bufptr == '/' ){
			flag = 2;  /* v//vn */
			bufptr++;
		    }
		    /* check for v/vt/vn */
		    while( ptr && flag != 3  && is_digit(*ptr) ) {
			if( *ptr == '/' ) flag = 3;
			ptr++;
		    }

		    switch( flag ){
			case 0:
			default:
			  break;
			case 1:		/* v/vt */
			    if( is_digit(*bufptr) )
				value   = atoi(bufptr);
			    else {break;}
			    if( value   > *vtct+1 )
				error_index(el_type,"texture",*lcount);
			    Obj_data->vt_indices[*numEdges-1] = value - 1;
			  break;
			case 3:		/* v/vt/vn */
			    if( is_digit(*bufptr) )
				value  = atoi(bufptr);
			    else {
				error_index( el_type, "texture", *lcount);
				break;
			    }

			    if( value   > *vtct+1 || value   < 1 )
				error_index(el_type,"texture",*lcount);
			    Obj_data->vt_indices[*numEdges-1] = value - 1;

			    if( is_digit(*ptr) )
				value   = atoi(ptr);
			    else {
				error_index( el_type, "normal", *lcount);
				break;
			    }

			    if( value   > *vnct+1 || value   < 1 )
				error_index(el_type,"normal",*lcount);
			    Obj_data->vn_indices[*numEdges-1] = value - 1;

			  break;
			case 2:		/* v//vn */
			    if( is_digit(*bufptr) )
				value   = atoi(bufptr);
			    else {
				error_index( el_type, "normal", *lcount);
				break;
			    }
			    if( value   > *vnct+1 )
				error_index(el_type,"normal",*lcount);
			    Obj_data->vn_indices[*numEdges-1] = value - 1;
			  break;
		     } /* end switch */
		    /* look forward to spaces */
	            while( (*bufptr != '\0') && !isspace(*bufptr)) bufptr++;
		}
		/* look past spaces */
	        while( (*bufptr != '\0') &&  isspace(*bufptr)) bufptr++;
		if( *bufptr == '\0' ) break;

	      } /* end while(1) */
	Obj_data->facet_Count[*poly_cnt] = i;
	if( flag == 2 || flag == 3 )
	    Obj_data->norm_Count[*poly_cnt]  = i;
	return i;
}

/*
 *  ROUTINE:	IO_get_next_float
 *  PURPOSE:	Gets the next floating point value in string 's', and
 *		stores it in 'value'.
 *		If it reaches a continuation char '\\', then another line
 *		is read from the file.
 *  RETURNS:	pointer to current position in buffer after getting
 *		the floating point value. The current position will be the
 *		trailing space or other non-digit.
 */
static char *
IO_get_next_float( fp2, s, buffer, value )
FILE	*fp2;
char *s;
char *buffer;
float *value;
{

	if( (s == NULL) )
	    return NULL;

	if( *s == '\n' || *s == '\0' )
	    return NULL;

	while( (s != (char *) NULL) && (*s != '\0') && isspace(*s)) s++;
	if( s == NULL ){
	    *value = 0.0;
	    return NULL;
	}

	if( *s == '\n' || (*s == '\0') ){
	    *value = 0.0;
	    return NULL;
	}

	if( *s == '\\' ){
	    if(fgets(buffer, 128, fp2) == NULL)
	   	return NULL;	/* error */
	    s = buffer;
	  while( (s != (char *) NULL) && isspace(*s)) s++;
	}

	*value = (float) atof(s);
	/* look past the 'digits' to trailing space */
	while((s != (char *)NULL) && (*s != '\0') 
		&& !isspace(*s) && *s != '\n' ) s++;
	return s;
}

int
is_digit( c )
int c;
{
    if( isdigit(c) || c == (int) '-' )
	return 1;
    else 
	return 0;
}
