/*
 ******************************************************************************
 ******************************************************************************
 *****                           stest.c                                  *****
 *****									  *****
 *****			 Simple PEXtk test program			  *****
 *****									  *****
 *****									  *****
 *                                                                            *
 *	Copyright 1992, 1993. All Rights Reserved                             *
 *			SHOgraphics					      *
 *			321 Soquel Way           			      *
 *			Sunnyvale, CA     94086   			      *
 *                                                                            *
 *****									  *****

    Permission to use, copy, modify, and distribute this software and
 its documentation for any purpose and without fee is hereby granted,
 provided that the above copyright notice appear in all copies and that
 both that copyright notice and this permission notice appear in
 supporting documentation, and that the name of SHOgraphics not be
 used in advertising or publicity pertaining to distribution of the 
 software without specific, written prior permission.

 SHOGRAPHICS DISCLAIMS ANY AND ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 INCLUDING ALL EXPRESSED AND IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS. IN NO EVENT SHALL SHOGRAPHICS BE LIABLE FOR ANY DAMAGES OF ANY
 KIND, INCLUDING BUT NOT LIMITED TO SPECIAL, INDIRECT OR CONSEQUENTIAL 
 DAMAGES RESULTING FROM LOSS OF USE, DATA OR PROSPECTIVE PROFITS, WHETHER 
 IN AN ACTION OF CONTRACT, NEGLIENCE OR OTHER TORTIOUS ACTION, ARISING OUT 
 OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

 ************************************************************************** */

/*
	Program:   stest.c

	Synopsis   This program is a simple test of PEXtk.

	Usage:	   stest [options] 

	    where:
		options are:
		    -d 			turns on double buffering
		    -i num		sets iteration count
		    -f 			uses Ppoly_fill instead of Ppoly
		    -w size		sets the size of the window

*/

#include <stdio.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <PEXtk.h>

Fcoord3D line[2] = {{ 0.0, 0.0, -5.0},{100.0, 100.0, -15.0}};
Fcoord3D poly[4] = {{ 5.0, 5.0, -15.0},{-5.0, 5.0, -15.0},
                    { -5.0, -5.0, -15.0},{ 5.0, -5.0, -15.0}};
Fcoord3D ply2[4] = {{ 6.0, 4.0, -16.0},{-4.0, 4.0, -16.0},
                    { -4.0, -6.0, -16.0},{ 6.0, -6.0, -16.0}};
    int			 win_width, win_height;

main(argc,argv)
int	argc;
char	*argv[];
{
	Display *dpy; 
	Window window;
	int i,iter=10,dbl_buf, fill_flag=0, sync_flag=0;
	PCOLOR	col;

	win_width=500; win_height=500;

	dbl_buf = 0;
	if( getenv("DBL_BUF") )
	    dbl_buf = 1;

	for(i=1;i<argc;i++){
	    if( strcmp(argv[i],"-i") == 0 ){
		i++;
		iter = atoi(argv[i]);
	    }
	    if( strcmp(argv[i],"-d") == 0 ){
		dbl_buf = 1;
	    }
	    if( strcmp(argv[i],"-f") == 0 ){
		fill_flag = 1;
	    }
	    if( strcmp(argv[i],"-s") == 0 ){
		sync_flag = 1;
	    }
	    if( strcmp(argv[i],"-w") == 0 ){
		i++;
		win_width = atoi(argv[i]);
		win_height = win_width;
	    }
	}

        if ((dpy = XOpenDisplay ("")) == NULL) {
            printf(" error opening display\n");
	    return 1;
        }

	if(Pinit(dpy)){
	    printf("ERROR... initializing PEX server\n");
	    exit(1);
	}

	if( (window=Pcreate_window(dpy, win_width, win_height)) == NULL){
	    printf("ERROR... creating window\n");
	    exit(1);
	}

	if( sync_flag ){
	    XSynchronize( dpy, True);
	}
	col.color.rgb.red=1.0;col.color.rgb.green=col.color.rgb.blue=0.;

	Pset_drawable(window);
	if( dbl_buf )
	    Pbuffer_mode(PEXtk_DOUBLE_BUFFER);
	else
	    Pbuffer_mode(PEXtk_SINGLE_BUFFER);
	Pzbuffer(1);
	Pset_view_index(0);
	Pviewport( 0, win_width-1, 0, win_height-1);
	Ppersp_view( 0, 90.0, 1.0, 1.0, 1000.0);
	Pclear_rgb( 0.0, 0.0, 0.0);

	for(i=0;i<iter;i++)
	{
	    Pset_view_index(0);
	    Pclear_rgb( 0.0, 0.0, 0.0);
	    Pline_color(&col);
	    Ppoly_line(2, line);
	    Pcolor(&col);
	    if( fill_flag ){
		Ppoly_fill(4, poly);
		Pcolor_rgb( 0.0, 1.0, 0.0);
		Ppoly_fill(4, ply2);
	    } else {
		Ppoly(4, (Fcoord3D *) poly);
		Pcolor_rgb( 0.0, 1.0, 0.0);
		Ppoly(4, (Fcoord3D *) ply2);
	    }
	    Protate_z( (Angle) 5.0, 1);
	    if( dbl_buf ){
		Pswap_buffers();
		Pflush(0);
	    } else
		Pflush(1);
	}
	Pflush(1);
	printf("Press return\n");scanf("%c",&iter);
	Pexit();
	XDestroyWindow( dpy, window);
	XCloseDisplay(dpy);
}
