/*
 ******************************************************************************
 ******************************************************************************
 *****                             PEXtk                                  *****
 *****									  *****
 *****		    An Immediate Mode Interface to PEX			  *****
 *****									  *****
 *****				Test Programs				  *****
 *****									  *****
 *                                                                            *
 *	Copyright 1991, 1993. All Rights Reserved                             *
 *			SHOgraphics					      *
 *			321 Soquel Way           			      *
 *			Sunnyvale, CA      94086  			      *
 *                                                                            *
 *****									  *****

    Permission to use, copy, modify, and distribute this software and
 its documentation for any purpose and without fee is hereby granted,
 provided that the above copyright notice appear in all copies and that
 both that copyright notice and this permission notice appear in
 supporting documentation, and that the name of SHOgraphics not be
 used in advertising or publicity pertaining to distribution of the 
 software without specific, written prior permission.

 SHOGRAPHICS DISCLAIMS ANY AND ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 INCLUDING ALL EXPRESSED AND IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS. IN NO EVENT SHALL SHOGRAPHICS BE LIABLE FOR ANY DAMAGES OF ANY
 KIND, INCLUDING BUT NOT LIMITED TO SPECIAL, INDIRECT OR CONSEQUENTIAL 
 DAMAGES RESULTING FROM LOSS OF USE, DATA OR PROSPECTIVE PROFITS, WHETHER 
 IN AN ACTION OF CONTRACT, NEGLIENCE OR OTHER TORTIOUS ACTION, ARISING OUT 
 OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

 ************************************************************************ 
 *		PEXtk Test Program for triangle mesh routines
 */
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <stdio.h>
#include <PEXtk.h>

#include "read_demo_obj.c"

    Fcoord3D	 	ctr;
    Fcoord3D	 	*vnorm;
    int			iter, verbose_flag, ModelType, str_flag;
    int			pause_flag, color_flag, fill_type;
    float		near, far;
    float		fov, aspect;
    int			win_width, win_height;


main(argc, argv)
int argc; char **argv;
{
    int 		i,j,k,m,n;
    Display 		*dpy;
    Window 		window;
    Fcoord3D	 	vt, vn, vtt[255];
    PCOLOR		colorObj;
    int		 	*infoData;

    ModelType = 0; pause_flag = verbose_flag = 0;
    iter = 100; str_flag = 0; color_flag = 0; fill_type = 1;
    win_width=500; win_height=500;
    near = 1.0; far = 1000.0; aspect = 1.0; fov = 90.0;
    ctr.x = ctr.y = 0.0; ctr.z = -5.0;

    parse_line(argc,argv);

    if ((dpy = XOpenDisplay ("")) == NULL) {
	printf(" Error opening display\n");
	return 1;
    }

    if(Pinit(dpy)){
	printf("ERROR... initializing PEX server\n");
	exit(1);
    }

    if( (window = Pcreate_window(dpy, win_width, win_height)) == NULL){
	printf("ERROR... creating window\n");
	exit(1);
    }

    Pset_drawable( window );
    Pzbuffer(1);
    Pbuffer_mode(  PEXtk_DOUBLE_BUFFER );
    Pbackface( PEXtk_CULL_NONE);

    colorObj.color.rgb.red   = 0.9; colorObj.color.rgb.green = 0.0;
    colorObj.color.rgb.blue  = 0.4;

	Set_view();

	Pcolor( &colorObj);
	Pline_color( &colorObj);

    i = 0;
	Create_str(str_flag);
    while( i < iter ){
	Pclear_rgb( 0.0, 0.0, 0.0);
	/* Protate_z( 3.0, 1 ); */
	Pload_identity();
	Ptranslate( 0.0, 0.0, -10.0, 1);
	Ptranslate( ctr.x, ctr.y, ctr.z, 1);
	Protate_y( ((float) i)*3.0, 1);
	Ptranslate( -ctr.x, -ctr.y, -ctr.z, 1);
	Exec_str( );
	Pswap_buffers(); 
	i++;
    }
    if( pause_flag ){
	printf("enter a char ");
	fscanf( stdin, "%c",&i);
	}
      Delete_str();
    Pexit();
    XDestroyWindow( dpy, window);
    XCloseDisplay( dpy);
    exit(0);
}

Setup_mode()
{
    PEXtk_LIGHT lgt;

    if(!( ModelType & PEXtk_SHADE_LIGHTS) ) return;
    Plight_color_rgb( 0.5, 0.5, 0.5);
    Plight_ambient( .6);
    lgt.type = PEXtk_LIGHT_POINT;
    /* lgt.dir.x = 0.0; lgt.dir.y = 0.0; lgt.dir.z = -1.0; */
    lgt.pos.x = 20.0; lgt.pos.y = 20.0; lgt.pos.z = 21.0;
    lgt.atten1 = .5;
    lgt.atten2 = 0.0;
    Plight( 1, &lgt);
    Pspecular_color_rgb( 0.0, 0.0, 0.0);            /* white specular */
}
	
Set_view()
{
    Ppersp_view( 0, fov, aspect, near, far);
}

Create_str( use_str)
int use_str;
{
    int i, j, k, m, n;
    Fcoord3D	 vn, vt;
    float pt[3];

  if( use_str )
    Pcreate_struct(1);
	Pbegin_tmesh(fill_type);
	    pt[0] = -10.0; pt[1] = -5.0; pt[2] = -10.0;
	    Pvertex( pt); /* pt 1 */
	    if( color_flag )
		Pvertex_color_rgb( 1.0, 0.0, 0.0);
	    pt[0] = -7.5; pt[1] = 0.0; pt[2] = -10.0;
	    Pvertex( pt); /* pt 2 */
	    if( color_flag )
		Pvertex_color_rgb( 0.8, 0.0, 0.2);
	    pt[0] = -5.0; pt[1] = -5.0; pt[2] = -10.0;
	    Pvertex( pt); /* pt 3 */
	    if( color_flag )
		Pvertex_color_rgb( 0.6, 0.0, 0.4);
	    pt[0] = -2.5; pt[1] = 0.0; pt[2] = -10.0;
	    Pvertex( pt); /* pt 4 */
	    if( color_flag )
		Pvertex_color_rgb( 0.4, 0.0, 0.6);
	    pt[0] = 0.0; pt[1] = -5.0; pt[2] = -10.0;
	    Pvertex( pt); /* pt 5 */
	    if( color_flag )
		Pvertex_color_rgb( 0.2, 0.0, 0.8);
	    pt[0] =  2.5; pt[1] = 0.0; pt[2] = -10.0;
	    Pvertex( pt); /* pt 6 */
	    if( color_flag )
		Pvertex_color_rgb( 0.0, 0.0, 1.0);
	    pt[0] = 5.0; pt[1] = -5.0; pt[2] = -10.0;
	    Pvertex( pt); /* pt 7 */
	    if( color_flag )
		Pvertex_color_rgb( 0.0, 0.6, 1.0);
	    Pswap_tmesh();
	    pt[0] = 7.5; pt[1] = 0.0; pt[2] = -10.0;
	    Pvertex( pt); /* pt 8 */
	    if( color_flag )
		Pvertex_color_rgb( 0.0, 0.8, 1.0);
	    Pswap_tmesh();
	    pt[0] = 5.0; pt[1] = 5.0; pt[2] = -10.0;
	    Pvertex( pt); /* pt 9 */
	    if( color_flag )
		Pvertex_color_rgb( 0.0, 0.6, 1.0);
	    Pswap_tmesh();
	    pt[0] = 0.0; pt[1] = 5.0; pt[2] = -10.0;
	    Pvertex( pt); /* pt 10 */
	    if( color_flag )
		Pvertex_color_rgb( 0.2, 0.0, 0.8);
	    pt[0] = -2.5; pt[1] = 0.0; pt[2] = -10.0;
	    Pvertex( pt); /* pt 4 */
	    if( color_flag )
		Pvertex_color_rgb( 0.4, 0.0, 0.6);
	    pt[0] = -5.0; pt[1] = 5.0; pt[2] = -10.0;
	    Pvertex( pt); /* pt 11 */
	    if( color_flag )
		Pvertex_color_rgb( 0.6, 0.0, 0.4);
	Pend_tmesh();
  if( use_str )
    Pclose_struct();
}

Delete_str()
{
    if( str_flag )
      if(Pdelete_struct(1))
	printf("ERROR... deleting structure 1\n");
}

Exec_str()
{
    if( str_flag )
	Pcall_struct( 1);
    else
	Create_str( str_flag);
}

/*
 *                      +---------------------------+
 * ---------------------+         parse_line        +------------------------
 *                      +---------------------------+
 *  FUNCTION:
 *      Parses the command line.
 *
 *  USAGE:
 *      parse_line(argc, argv,options)
 *	int	   argc 	   (in)  - number of arg's
 *	char   	   argv 	   (in)  - pointer to arg list
 *
 *  DESCRIPTION:
 *	This routine will parse the command line for the following options:
 *	   opt  argument	type	  comment
 *	   ---  ------------    ----      ----------------------------------
 *	    -f filename.obj	string	  .obj file to read
 *	    -i iter		int	  number of iterations
 *	    -p 			void	  sets pause flag
 *	    -v 			void	  sets verbose flag
 *	    -m mode		int	  shade mode, in hexadecimal
 *	    -w size		int	  window size
 *	    -z trans		float	  z translation value
 *	    -h near		float	  near (hither) plane
 *	    -y far		float	  far (yon) plane
 *	    -a aspect		float	  aspect ratio
 *                                
*/
int parse_line(argc,argv)
int argc;
char **argv;
{
     char *temp,ffile[32];
     extern char *optarg;
     extern int optind;
     int c;

    while((c=getopt(argc,argv,"a:t:i:h:y:m:f:t:w:cpsvz:")) != EOF)
    {
	switch(c){
	    case 'f':
		  read_demo_obj(optarg, &ctr, 1.0);
		break;
	    case 'i':
		  sscanf(optarg,"%d",&iter);
		break;
	    case 'a':
		  sscanf(optarg,"%f",&aspect);
		break;
	    case 'm':
		  sscanf(optarg,"%x",&ModelType);
		break;
	    case 't':
		  sscanf(optarg,"%x",&fill_type);
		break;
	    case 'c':
		  color_flag = 1;
		break;
	    case 's':
		  str_flag = 1;
		break;
	    case 'p':
		  pause_flag = 1;
		break;
	    case 'v':
		  verbose_flag = 1;
		break;
	    case 'w':
		  sscanf(optarg,"%d",&win_width);
		  win_height = win_width;
		break;
	    case 'z':
		  sscanf(optarg,"%f",&z_trans_factor);
		break;
	    case 'h':
		  sscanf(optarg,"%f",&near);
		break;
	    case 'y':
		  sscanf(optarg,"%f",&far);
		break;
	    default:
		Usage();
	}
    }

#ifdef READ_OBJ
    for(; optind < argc;optind++ ){
		  read_demo_obj(argv[optind], &ctr, 1.0);
    }
#endif
}

Usage()
{
    printf("Error...  Usage is:\n");
    printf(" prm_test [options] file.obj\n");
    printf(" options: \n");
    printf("\t -m mode {0-9}\t(shade mode)\n");
	printf("\t\t PEXtk_SHADE_NONE \t\t%d\n",PEXtk_SHADE_NONE);
	printf("\t\t PEXtk_SHADE_WIRE \t\t%d\n",PEXtk_SHADE_WIRE);
	printf("\t\t PEXtk_SHADE_FLAT \t\t%d\n",PEXtk_SHADE_FLAT);
	printf("\t\t PEXtk_SHADE_COLORS \t\t%d\n",PEXtk_SHADE_COLORS);
	printf("\t\t PEXtk_SHADE_NORMALS \t\t%d\n",PEXtk_SHADE_NORMALS);
	printf("\t\t PEXtk_SHADE_SPECULAR \t\t%d\n",PEXtk_SHADE_SPECULAR);
	printf("\t\t PEXtk_SHADE_LIGHTS \t\t%d\n",PEXtk_SHADE_LIGHTS);
	printf("\t\t PEXtk_SHADE_HIDDENLINE \t%d\n",PEXtk_SHADE_HIDDENLINE);
	printf("\t\t PEXtk_SHADE_ANTIALIAS \t\t%d\n",PEXtk_SHADE_ANTIALIAS);
	printf("\t\t PEXtk_SHADE_DOTPRODUCT \t%d\n",PEXtk_SHADE_DOTPRODUCT);
    printf("\t -i iter\t(number of iterations)\n");
    printf("\t -t fill_type {1,2}\t(fill type, 1=hollow,2=solid)\n");
    printf("\t -c\t\t(color flag)\n");
    printf("\t -p\t\t(pause flag)\n");
    printf("\t -s\t\t(use structure mode)\n");
    printf("\t -a aspect\t(aspect ratio, float)\n");
    printf("\t -h hither\t(set hither plane)\n");
    printf("\t -w size\t(set window width and height)\n");
    printf("\t -y yon\t\t(set yon plane)\n");
    printf("\t -v\t\t(verbose flag)\n");
    printf("\t -f file.obj\n");
}
