/*
 ******************************************************************************
 ******************************************************************************
 *****                             PEXtk                                  *****
 *****									  *****
 *****		    An Immediate Mode Interface to PEX			  *****
 *****									  *****
 *****									  *****
 *                                                                            *
 *	Copyright 1991, 1993. All Rights Reserved                             *
 *			SHOgraphics					      *
 *			321 Soquel Way           			      *
 *			Sunnyvale, CA      94086  			      *
 *                                                                            *
 *****									  *****

    Permission to use, copy, modify, and distribute this software and
 its documentation for any purpose and without fee is hereby granted,
 provided that the above copyright notice appear in all copies and that
 both that copyright notice and this permission notice appear in
 supporting documentation, and that the name of SHOgraphics not be
 used in advertising or publicity pertaining to distribution of the 
 software without specific, written prior permission.

 SHOGRAPHICS DISCLAIMS ANY AND ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 INCLUDING ALL EXPRESSED AND IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS. IN NO EVENT SHALL SHOGRAPHICS BE LIABLE FOR ANY DAMAGES OF ANY
 KIND, INCLUDING BUT NOT LIMITED TO SPECIAL, INDIRECT OR CONSEQUENTIAL 
 DAMAGES RESULTING FROM LOSS OF USE, DATA OR PROSPECTIVE PROFITS, WHETHER 
 IN AN ACTION OF CONTRACT, NEGLIENCE OR OTHER TORTIOUS ACTION, ARISING OUT 
 OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

 ************************************************************************ 
 *		PEXtk Test Program for begin/end routines
 */
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <stdio.h>
#include <PEXtk.h>

    Fcoord3D	 	ctr;
    int			iter, verbose_flag, ModelType, color_flag;
    float		start_angle, end_angle, x, y, z;
    float		radius;
    int			pause_flag;
    int			res;
    int 		win_width, win_height;

main(argc, argv)
int argc; char **argv;
{
    int 		i,j,k,m,n;
    Display 		*dpy;
    Window 		window;
    Fcoord3D	 	vt, vn, vtt[255];
    PCOLOR		colorObj, colorList[256];
    PEXtk_LIGHT 	lgt;
    PCOLOR 		col;

    ModelType = 0; pause_flag = verbose_flag = color_flag = 0;
    iter = 100;
    win_width=500; win_height=500;
    x = 0.0; y = 0.0; z = -10.0; radius = 5.0;
    start_angle = 0.0; end_angle = 45.0;
    res = 20;

    parse_line(argc,argv);

    if ((dpy = XOpenDisplay ("")) == NULL) {
	printf(" Error opening display\n");
	return 1;
    }

    if(Pinit(dpy)){
	printf("ERROR... initializing PEX server\n");
	exit(1);
    }

    if( (window = Pcreate_window(dpy, win_width, win_height)) == NULL){
	printf("ERROR... creating window\n");
	exit(1);
    }

    Pset_drawable( window );
    Pbuffer_mode(  PEXtk_DOUBLE_BUFFER );
    Pviewport( 0, win_width-1, 0, win_height-1);
    Pzbuffer(1);
    Ppersp_view( 0, 90.0, 1.0, 1.0, 1000.0);
    Pset_precision( res, res);

    colorObj.color.rgb.red   = 0.9; colorObj.color.rgb.green = 0.0;
    colorObj.color.rgb.blue  = 0.4;
    Pshade_mode( PEXtk_SHADE_LIGHTS );

    Pcreate_struct(1);

	switch( ModelType){
	    case 0:
		Parc( x,y,z, radius, start_angle, end_angle);
		break;
	    case 1:
		Parc_fill( x,y,z, radius, start_angle, end_angle);
		break;
	    case 2:
		Pcircle( x,y,z, radius);
		break;
	    case 3:
		Pcircle_fill( x,y,z, radius);
		break;
	    case 4:
		Psphere( x,y,z, radius);
		break;
	    case 5:
		Psphere_fill( x,y,z, radius);
		break;
	}
    Pclose_struct();

	Pcolor( &colorObj);
	Pline_color( &colorObj);
	Ptranslate( -ctr.x, -ctr.y, 0.0, 0);

    Plight_color_rgb( 0.5, 0.5, 0.5);
    Plight_ambient( .6);
    lgt.type = PEXtk_LIGHT_POINT;
    lgt.dir.x = 0.0; lgt.dir.y = 0.0; lgt.dir.z = -1.0;
    lgt.pos.x = 20.0; lgt.pos.y = 20.0; lgt.pos.z = 21.0;
    lgt.atten1 = .5;
    lgt.atten2 = 0.0;
    Plight( 1, &lgt);
    col.color.rgb.red = col.color.rgb.green = col.color.rgb.blue = 1.0;
    Pspecular_color( &col);		/* white specular */
    col.color.rgb.red = 1.0;
    col.color.rgb.green = col.color.rgb.blue = 0.0;
    Pcolor( &col);			/* polygon color */
    /* Pshade_mode( PEXtk_SHADE_LIGHTS ); */
    i = 0;
    while( i < iter ){
	Pclear_rgb( 0.0, 0.0, 0.0);
	Protate_z( 3.0, 1 );
	Pcall_struct( 1);
	Pswap_buffers(); 
	i++;
    }
    if( pause_flag ){
	printf("Press Enter Key\n");
	scanf("%c",&i);
    }
    Pexit();
    XDestroyWindow( dpy, window);
    XCloseDisplay( dpy);
    exit(0);
} 

/*
 *                      +---------------------------+
 * ---------------------+         parse_line        +------------------------
 *                      +---------------------------+
 *  FUNCTION:
 *      Parses the command line.
 *
 *  USAGE:
 *      parse_line(argc, argv,options)
 *	int	   argc 	   (in)  - number of arg's
 *	char   	   argv 	   (in)  - pointer to arg list
 *
 *  DESCRIPTION:
 *	This routine will parse the command line for the following options:
 *	   opt  argument	type	  comment
 *	   ---  ------------    ----      ----------------------------------
 *	    -i iter		int	  number of iterations
 *	    -t type		int	  type of begin sequence: Choice of
 *						0=arc, 1=arc_fill, 2=circle,
 *						3=circle_fill, 4=sphere,
 *						5=sphere_fill
 *	    -w size		int	  window size
 *	    -x x_pos		float	  sets the x position value
 *	    -y y_pos		float	  sets the y position value
 *	    -z z_pos		float	  sets the z position value
 *	    -n res		int	  sets the resolution
 *	    -r rad		float	  sets the radius
 *	    -s start		float	  sets the starting angle
 *	    -e end		float	  sets the ending angle
 *	    -p 			void	  sets pause flag
 *	    -v 			void	  sets verbose flag
 *                                
*/
int parse_line(argc,argv)
int argc;
char **argv;
{
     char *temp,ffile[32];
     extern char *optarg;
     extern int optind;
     int c;

    if( argc < 2 ){
	Usage();
	exit(1);
    }

    while((c=getopt(argc,argv,"pi:t:vs:n:e:w:x:y:z:")) != EOF)
    {
	switch(c){
	    case 'i':
		  sscanf(optarg,"%d",&iter);
		break;
	    case 't':
		  sscanf(optarg, "%d", &ModelType);
		  /* Check for valid type */
		  if( ModelType < 0 || ModelType > 5 ){
		    printf("ERROR... type must be 0 <= type <= 5:, %d\n",
			ModelType);
		    ModelType = 0;
		  }
		break;
	    case 'p':
		  pause_flag = 1;
		break;
	    case 'v':
		  verbose_flag = 1;
		break;
	    case 'e':
		  sscanf(optarg,"%f",&end_angle);
		break;
	    case 's':
		  sscanf(optarg,"%f",&start_angle);
		break;
	    case 'n':
		  sscanf(optarg,"%d",&res);
		break;
	    case 'r':
		  sscanf(optarg,"%f",&radius);
		break;
	    case 'w':
		  sscanf(optarg,"%d",&win_width);
		  win_height = win_width;
		break;
	    case 'x':
		  sscanf(optarg,"%f",&x);
		break;
	    case 'y':
		  sscanf(optarg,"%f",&y);
		break;
	    case 'z':
		  sscanf(optarg,"%f",&z);
		break;
	}
    }

}

usage()
{
    printf("Error...  Usage is:\n");
    printf(" utl_test -p -w size -i iter -n res -e end -s start -t type -v -x val -y val -z val\n");
    printf("\twhere res is resolution, p is pause, v is verbose\n");
    printf("\ttype may be:\n\t\t0 = arc\n");
    printf("\t\t1 = arc_fill\n\t\t2 = circle\n\t\t3 = circle_fill\n");
    printf("\t\t4 = sphere\n\t\t5 = sphere_fill\n");
}

Usage()
{
    printf("Error...  Usage is:\n");
    printf(" utl_test [options]\n");
    printf(" options: \n");
    printf("\t -t type {0-5}\t(primitive type)\n");
    printf("\t    type may be:\n\t\t0 = arc\n");
    printf("\t\t1 = arc_fill\n\t\t2 = circle\n\t\t3 = circle_fill\n");
    printf("\t\t4 = sphere\n\t\t5 = sphere_fill\n");
    printf("\t -i interations\n");
    printf("\t -e end angle\t(Parc)\n");
    printf("\t -n resolution\t(Pset_precision)\n");
    printf("\t -r radius\t(Pcircle)\n");
    printf("\t -s start angle\t(Parc)\n");
    printf("\t -w size\t(window size, int)\n");
    printf("\t -x x_value\t(Pcircle x position)\n");
    printf("\t -y y_value\t(Pcircle y position)\n");
    printf("\t -z z_value\t(Pcircle z position)\n");
    printf("\t -p\t\t(pause flag)\n");
    printf("\t -v\t\t(verbose flag)\n");
}
