/*
 *  $Header: /auto/nfs/sequoia/u/an/work/rws/pextk/src/pextk.c,v 1.5 1994/01/06 00:47:00 rws Exp $
 ******************************************************************************
 ******************************************************************************
 *****                             PEXtk                                  *****
 *****									  *****
 *****		    An Immediate Mode Interface to PEX			  *****
 *****									  *****
 *****									  *****
 *                                                                            *
 *	Copyright 1992, 1993. All Rights Reserved                             *
 *			SHOgraphics					      *
 *			321 Soquel Way           			      *
 *			Sunnyvale, CA     94086   			      *
 *                                                                            *
 *****									  *****

    Permission to use, copy, modify, and distribute this software and
 its documentation for any purpose and without fee is hereby granted,
 provided that the above copyright notice appear in all copies and that
 both that copyright notice and this permission notice appear in
 supporting documentation, and that the name of SHOgraphics not be
 used in advertising or publicity pertaining to distribution of the 
 software without specific, written prior permission.

 SHOGRAPHICS DISCLAIMS ANY AND ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 INCLUDING ALL EXPRESSED AND IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS. IN NO EVENT SHALL SHOGRAPHICS BE LIABLE FOR ANY DAMAGES OF ANY
 KIND, INCLUDING BUT NOT LIMITED TO SPECIAL, INDIRECT OR CONSEQUENTIAL 
 DAMAGES RESULTING FROM LOSS OF USE, DATA OR PROSPECTIVE PROFITS, WHETHER 
 IN AN ACTION OF CONTRACT, NEGLIENCE OR OTHER TORTIOUS ACTION, ARISING OUT 
 OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

 ************************************************************************ */
/*
 *
 *  Module Purpose:
 *	This is the module which contains the PEXtk API routines for
 *	initialization, basic line and polygon primitives, setting the
 *	line and polygon colors and other similar attributes, structure
 *	handling, inquiry routines, and light routines.
 *
 *  Module Notes:
 *
 *  Module Contents:
 *
 *    int   Pinit             Initializes PEXtk
 *    void  Pbackface         Turns on/off culling of backfacing polygons.
 *    void  Pborder           Sets the geometry outlineing attribute.
 *    void  Pbuffer_mode      Changes buffering mode to Single or Double Buffer
 *    void  Pcall_display_list    Calls a display list.
 *    void  Pcall_struct      Executes a structure and saves the state
 *    void  Pexec_struct      Executes a structure.
 *    void  Pclose_display_list   Closes an opened display list.
 *    void  Pclose_struct     Closes an opened structure for retained mode.
 *    int   Pcreate_struct    Opens a structure for retained mode.
 *    int   Pcreate_display_list  Creates a display list.
 *    void  Pclear            Clears viewport to specified color.
 *    void  Pclear_hsv        Clears viewport to specified hsv color.
 *    void  Pclear_ind        Clears viewport to specified color index.
 *    void  Pclear_rgb        Clears viewport to specified rgb color.
 *    void  Pcolor            Sets the polygon/fill color.
 *    void  Pcolor_hsv        Sets the polygon/fill color to an hsv triple.
 *    void  Pcolor_ind        Sets the polygon/fill color to an indexed value.
 *    void  Pcolor_rgb        Sets the polygon/fill color to an rgb triple.
 *    void  Pcurve_approx_method  Sets the curve approximation method.
 *    int   Pdelete_display_list  Deletes a display list.
 *    int   Pdelete_struct    Deletes a structure.
 *    void  Pdepth            Sets the near and far clipping planes to those 
 *    void  Pdepth_cue        Turns depth cue on or off.
 *    void  Pdither           Turns dithering on or off.
 *    int   Pdisplay_list_used Determines if a display list name is in use
 *    int   Pedit_multiply_matrix Gets a matrix from structure name 
 *    int   Pedit_struct      Opens an existing structure for editing.
 *    void  Pexit             Closes PEXtk and frees data.
 *    void  Pflush            Flushes the buffers.
 *    u long Pget_backbuffer      Returns the current back buffer ID.
 *    int   Pget_backface         Returns the current backfacing attribute
 *    int   Pget_border           Returns the current border attribute
 *    int   Pget_buffer_mode      Returns the current buffer_mode state
 *    void  Pget_color            Returns the polygon/fill color.
 *    int   Pget_depth_cue    Returns depthcue values, and whether its on/off
 *    u long Pget_display_buffer  Returns the current display buffer ID.
 *    u long Pget_frontbuffer     Returns the current front buffer ID.
 *    int   Pget_line_color   Returns the current line color.
 *    int   Pget_line_type    Returns the current line type.
 *    Fdata Pget_line_width   Returns the current line width.
 *    int   Pget_marker_color Returns the marker color.
 *    int   Pget_marker_type  Returns the marker type.
 *    int   Pget_shade_mode   Returns the current shading mode.
 *    int   Pget_structure    Returns the currently open structure name.
 *    int   Pget_structmatrix Returns a matrix from a struct at position label
 *    int   Pget_support_info Returns support information on PEXtk.
 *    void  Pget_zbuffer      Returns current zbuffer mode.
 *    int   Pinsert_struct    Sets edit pointer to a label for insert editing
 *    int   Plabel_used       Determines if a label is used in a structure 
 *    void  Plight            Sets the light data.
 *    void  Plight_ambient    Sets the ambient light intensity and color.
 *    void  Plight_color      Sets the light color.
 *    void  Plight_color_hsv  Sets the light color to h,s,v values.
 *    void  Plight_color_ind  Sets the light color to r,g,b values.
 *    void  Plight_color_rgb  Sets the light color to r,g,b values.
 *    void  Pline_color       Sets the line color.
 *    void  Pline_color_hsv   Sets the line color to h,s,v values.
 *    void  Pline_color_ind   Sets the line color to h,s,v values.
 *    void  Pline_color_rgb   Sets the line color to r,g,b values.
 *    void  Pline_width       Sets the line width.
 *    void  Pline_type        Sets the line type.
 *    void  Plist_fonts       Returns the available fonts
 *    void  Pload_attribute   Changes attribute attrib to location at *valueptr
 *    void  Plogical_function Sets the 3D pipeline logical operation.
 *    void  Pmake_label       Sets a label in a structure. The label may then
 *    void  Pmarker           Draws a 3D marker primitive.
 *    void  Pmarker_color     Sets the marker color.
 *    void  Pmarker_color_hsv Sets the marker color to h,s,v values.
 *    void  Pmarker_color_ind Sets the marker color to r,g,b values.
 *    void  Pmarker_color_rgb Sets the marker color to r,g,b values.
 *    void  Pmarker_scale     Sets the marker scale.
 *    void  Pmarker_type      Sets the marker type.
 *    void  Pmatrix_mode      TBD
 *    void  Pnurb_curve       Draws a PHIGS+ Nurb curve primitive.
 *    void  Pnurb_surface     Draws a PHIGS+ Nurb surface primitive.
 *    void  Ppick_id          Sets a pick id.
 *    void  Pplane_mask       Set the 3D pipeline bit plane mask.
 *    void  Ppop_attributes   Pops the attribute stack.
 *    void  Ppush_attributes  Pushes the attribute stack.
 *    void  Ppop_name         Pops the name stack stack. Only Immediate mode 
 *    void  Ppush_name        Pushes the name stack stack. Only Immediate mode
 *    void  Ppoly             Draws a 3D polygon primitive.
 *    void  Ppoly_fill        Draws a filled 3D polygon primitive.
 *    void  Ppoly_fill_area   Draws a PHIGS+ fill area 3D with data primitive.
 *    void  Ppoly_index       Draws a PHIGS+ Set of fill area sets primitive.
 *    void  Ppoly_line        Draws a 3D line primitive.
 *    void  Ppoly_shade       Draws a shaded 3D polygon primitive.
 *    void  Ppoly_shade_texture Draws a textured and shaded 3D polygon primitiv
 *    void  Ppoly_texture     Draws a textured 3D polygon primitive.
 *    void  Ppoly_with_data   Draws a shaded 3D polygon primitive.
 *    void  Prectangle        Draws a rectangle outline
 *    void  Prectangle_fill   Draws a filled rectangle
 *    void  Preflection_property Sets the specular reflection values 
 *    int   Preplace_struct   Sets edit pointer to a label for replacment edi
 *    void  Pset_color_mode   Sets the color mode to indexed or RGB.
 *    void  Pset_color_table     Sets the color table values.
 *    void  Pset_color_table_hsv Sets the color table h,s,v values.
 *    void  Pset_color_table_rgb Sets the color table r,g,b values.
 *    void  Pset_display_buffer  Sets the display buffer id.
 *    int   Pset_drawable        This routine will set the current window
 *    int   Pset_function        Sets the support function for PEXtk.
 *    void  Pset_update_buffer   Sets the update buffer id.
 *    void  Pshade_mode       Sets the shade mode.
 *    void  Pshape_flag       Sets the polygon shape flag.
 *    void  Pspecular_color   Sets the specular highlight color for surfaces.
 *    void  Pspecular_color_hsv Sets the specular highlight hsv for surfaces.
 *    void  Pspecular_color_ind Sets the specular highlight index for surfaces.
 *    void  Pspecular_color_rgb Sets the specular highlight rgb for surfaces.
 *    int   Pstruct_used      Determines if a specified structure name is in 
 *    void  Psurface_approx_method  Sets the surface approximation method.
 *    void  Psurface_parm_range     Sets the surface parametric range.
 *    void  Pswap_buffers     Swaps the buffers in double buffer mode.
 *    void  Ptext             Writes a text string at position x,y.
 *    void  Ptext_annotate    Writes an annotation text string at position x,y.
 *    void  Ptext_color       Sets the text color.
 *    void  Ptext_color_hsv   Sets the text color to h,s,v values.
 *    void  Ptext_color_ind   Sets the text color to r,g,b values.
 *    void  Ptext_color_rgb   Sets the text color to r,g,b values.
 *    void  Ptext_orientation Sets the text vector orientation
 *    void  Ptext_size        Sets the text size
 *    void  Ptext_font        Sets the text font
 *    void  Ptexture_info     Sets the texture mapping information.
 *    void  Ptexture_name     Sets the current texture map.
 *    void  Ptransparency     Turns transparency on or off.
 *    void  Pzbuffer          Turns the zbuffer on/off.
 *  Undocumented utilities:
 *    long  Pget_renderer_ID	     Get renderer XID
 *    long  Pget_view_ID	     Get view table XID
 *
 *    void  PEXtk_Get_Support_Info      Get PEX Server supported information.
 *    int   PEXtk_init_attributes       Initializes PEXtk default values.
 *    int   PEXtk_Get_Renderer_Dynamics Gets Renderer default values.
 *    int   PEXtk_Begin_Rendering       Begins immediate mode rendering
 *    int   PEXtk_End_Rendering         Ends immediate mode rendering
 *    int   PEXtk_check_pc              Checks the PC mask and update PC
 *    int   PEXtk_Get_X_Res   Get x resolution of current drawable
 *    int   PEXtk_Get_Y_Res   Get y resolution of current drawable
 *    int   PEXtk_check_poly_attrib  Checks current int style, interp meth
 *
 *
 *  Module Description:
 *
 *	The first PEXtk call to be made must be Pinit. This must be followed
 *	by the Pset_drawable routine to set the drawable and create the assoc-
 *	iated PEXLUTView id's, Renderer id's, PEXLUTLight id's. The window 
 *	depth, width, and height is then obtained.
 *
 *	PEXtk will then attempt to setup the colormap for an 8-bit system
 *	using Pset_color_equation. First, it will try to allocate a 6/6/6 
 *	pseudo colormap using the default colormap. If there are not enough
 *	colorcells available, it will create a new colormap, and attempt to 
 *	save as many of the existing default colormap cells as possible. The 
 *	default will be to try to use the first 16 colors from the default
 *	colormap.
 *	
 *	If Pbuffer_mode is called to do double buffering, then the back 
 *	buffer is created using one of the registered methods (that is,
 *	using a Pixmap back buffer or the Multibuffering Extension).
 *
 *	Shape Flag:
 *		The only shapes the reference manual states
 *		are supported are 'Convex' shapes. This allows the
 *		PEX server to use an efficient triangulation method.
 *		A new routine named Pshape_flag has been added so that
 *		users may set the shape flag to something else such as 
 *		'PEXtk_SHAPE_NONCONVEX', however the user would have to
 *		make sure to reset it to 'PEXtk_SHAPE_CONVEX' whenever
 *		possible for better performance.
 *
 *  General Notes:
 *
 *	Note that PEXtk considers the Geometry Transformations as:
 *		[x' y' z' w'] = [ x y z w ] T  
 *
 *	PEX, however, considers the Geometry Transformations as:
 *		    | x' | =     | x |
 *		    | y' | =  T  | y |
 *		    | z' | =     | z |
 *		    | w' | =     | w |
 *
 *	   or, alternately:
 *
 *			     t		     t
 *		[x' y' z' w'] = T [ x y z w ]    
 *
 *	The difference being PEXtk considers transformations for points
 *	as the points *row* array is postmultiplied by the transformation
 *	matrix T, and PEX considers the points *column* array is premultiplied 
 *	by the transformation matrix T.
 *	The transformation from one to the other is achieved by:
 *
 *			   t     t  t
 *			(AB) = (B)(A)
 *
 *
 *  State information
 *	An interior style is maintained as a "state attribute" for PEXtk.
 *	Since it is an attribute than can be set or changed inside a
 *	structure, two values are actually maintained. One is i_style for 
 *	immediate mode, and the other is i_style_str for structure mode.
 *	The i_style_str value is set to 0 when creating a structure so 
 *	that it will always be set "inside a structure".
 *	This value is also used for keeping track of the state for labels.
 *	Since a make_label may be called at any time (in structure mode, 
 *	basically), the state must be saved and reset when editing at a 
 *	label. This is due to the way PEXtk places "transparent" OC data 
 *	in the structure. 
 *	For example, for a sequence of 4 {make_label, poly_fill}, the 
 *	first poly_fill will have the interior style set, and the remaining 
 *	poly_fill's will not. 
 *	For editing (replace mode), the state *must* be set to the same
 *	as when in structure mode (creating the structure), otherwise the
 *	replacement of the first poly_fill could replace the Output-Command
 *	of the interior style, and not the poly_fill (i.e., the FillArea).
 *	The same is true for interp_meth and interp_meth_str.
 *	Note that the edgeFlag (edgeFlag_str) is also this type of attribute.
 *
 *  Known Bugs:
 *	When using the DEC multibuffering extension routines on a DEC PEX 5.0
 *	server, the z buffer must be turned on. If this is not done, then the 
 *	swap buffer routine will crash the X server if an expose event occurs 
 *	on that window/buffer.
 *
 *	If Pset_update_buffer and Pset_display_buffer are used, then
 *	the (undocumented) routines PEXtk_Get_X_Res and PEXtk_Get_Y_Res
 *	routines must NOT be called. They perform an XGetGeometry call,
 *	and if the buffers are MBX buffers, this will cause an error.
 *
 *	Using a backface culling other than NONE with Ppoly_index causes
 *	problems on DEC platforms.
 *
 *  PEXtk routines not implemented yet:
 *	The following routines are not implemented in this version of PEXtk:
 *		Pcreate_display_list, Pclose_display_list, Pcall_display_list,
 *
 *  PEXtk routines not fully verified yet:
 *		Plookat_view, Ppolar_view, Pmatrix_mode
 *
 *  PEXtk routines that are hardware dependent:
 *		Pdither, Ptransparency, Ptexture_name, Ptexture_info,
 *		Pvertex_texture, Ppoly_point_nv_uv, Ppoly_point_uv
 *
 *  Extended functionality:
 *
 *		*    Plane Mask and Logical Ops. PEX 5.0 does not 
 *			currently support the use of a plane mask in the 
 *			3D pipeline. There are the D.E.C. MPEX extensions 
 *			that can utilize this functionality.
 *			PEXtk supports these extensions with the Pplane_mask 
 *			and Plogical_function routines.
 *
 *  Open Issues:
 *	The following items still need to be addressed:
 *
 *		*    NURBS/Parametric Surfaces.
 *			Note that a PEX/PHIGS+ set of routines are
 *			available to support NURBS, which are:
 *				Pnurb_surface, Psurface_approx_method,
 *				Psurface_parm_range, Pnurb_curve,
 *				Pcurve_approx_method, and Pcurve_parm_range.
 *
 *		*    Error handling. This involves errors that occur when
 *			creating primitives, or other PEXtk errors.
 *
 *		*    Transport optimization. This involves implementing
 *			the DB_check_dirty_mask routine. This will allow
 *			items to be buffered (in immediate mode) and
 *			not be sent until the affected routine is invoked.
 *			For example, a the current matrix will not be sent
 *			(via SetGlobalTransform) every time a Protate or 
 *			Ptranslate routine is called until a Ppoly or other
 *			affected routine is called.
 *
 *
 *  Changes from V1.0 to V1.1
 *    1. PEXtk now saves any attribute changes in a Pipeline Context. This
 *	 allows for a much faster begin/end render sequence, since the
 *	 attributes don't have to be reset after the begin render.
 *    2. The PEXtk_reset_attributes is no longer needed due to 1. This 
 *	 was for PEXtk to reset attributes after a begin render call.
 *    3. Since the MIT PEXlib does not have any MPEX routines, the MPEX 
 *	 extensions are not supported, although the original calls are
 *	 left intact.
 *    4. The tranformation matrix is now cached in immediate mode. The
 *	 matrix is flushed on any geometry command, or view matrix
 *	 generating command.
 *	 This allows for a sequence of the translate, rotate, etc. 
 *	 routines to be called without sending down the matrix each time.
 *	 No caching is done in retained mode. 
 *    5. The ViewTable ID and the LightTable ID are now bound to the
 *	 renderer when it is created.
 *    6. Pbuffer_mode checks to see if the current buffer mode is the
 *	 same as the input mode, and does nothing if it is.
 *    7. Added function support for antialias. The user may set the antialias
 *	 function using type PEXtk_FUNC_ANTIALIAS passed to Pset_function. The
 *	 default function uses the HP PEXGSE.
 *    8. There is a problem with the PEXlib compiled on a SUN and displaying
 *	 on a DEC with the bitmask for PEXRAHLHSRMode set. If the PEXtk
 *	 library might be used with a SUN host and displayed on a DEC,
 *	 then include -DSUN_DEC_BUG in the CC options. 
 *    9. The Prectangle routine was corrected to use FillArea with Interior 
 *	 Style Hollow.
 *   10. The check for E&S double buffer escape also has a check for the
 *	 SunPEX server (SMI in vendor string). When the SunPEX server
 *	 supports PEX 5.1, this can be removed.
 *
 *  Changes from V1.1 to V1.2
 *    1. PEXtk will support the KPC Sphere OC primitive. To activate, add
 *	 -DKPC to the CFLAGS variable in the src/Makefile, and add -lKPEX 
 *	 to the linked libraries.
 *    2. Picking support was added.
 *    3. Type casts were added for PEXMarkers, PEXPolyline, PEXFillArea, and
 *	 PEXFillAreaSetWithData.
 */

#include <stdio.h>
#include <math.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <X11/PEX5/PEX.h>
#include <X11/PEX5/PEXproto.h>
#include <X11/PEX5/PEXlib.h>
#include <X11/PEX5/PEXHPlib.h>

#ifdef MPEX
#include <X11/MPEX.h>
#include <X11/MPEXproto.h>
#include <X11/MPEXlib.h>
#endif

#include <X11/extensions/multibuf.h>

#ifndef SYSV
#include <memory.h>
#endif

#include <PEXtk.h>

#include "pextk.h"
#include "pextk_def.h"
#include "pextk_ops.h"
#include "pextk_glyphs.h"

/*
 *
 *                       +--------------------------+
 * ----------------------+           Pinit          +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Initializes PEXtk
 *
 *  USAGE:
 *	Pinit( display )
 *	Display	    *display	(in)	pointer to connection id
 *
 *  RETURNS:
 *	0 if successful, non-zero if error occured:
 *		 1 if PEX is not supported
 *		 2 if Renderer is not supported
 *		 3 if memory allocation error
 *
 */
 
/*ARGSUSED*/
int
Pinit( displayptr)
Display	*displayptr;
{

    PEXExtensionInfo	*pex_info;
    PEXPCAttributes 	pcAttribs;
    int 		i, m_op, f_ev, f_err;
    Visual 		*win_visual;
    char 		error_string[128];

    dpy = displayptr;

    if(PEXInitialize( dpy, &pex_info, 128, error_string))
	return 1;

#ifdef DEBUG
    printf("\n  %s  Version %d\n\n",
	pex_info->vendor_name,pex_info->major_version);
#endif
    
    if( (pex_info->subset_info & 0xFFFF) == 0x02 )
	return 2;    /* Phigs Wks only.. Renderer not supported */

    /*
     * Save pointers to some various important data structures
     */
    p_object.state	= p_state   = &pstate;
    p_object.build	= p_build   = &pbuild;
    p_object.buffer	= p_buffer  = &pbuffer;
    p_object.support	= p_support = &psupport;
    p_object.dpy_list	= p_dl      = &p_dl_obj;
    p_object.clr_tbl	= p_clr     = &clr_tbl;
    p_object.x_data	= p_x_data  = &px_data;
    psupport.dbl_buf    = 0;
    px_data.red_shift   = px_data.green_shift = px_data.blue_shift  = 0;

    win_screen   = DefaultScreen (dpy);
    win_visual   = DefaultVisual(dpy,DefaultScreen(dpy));

    pstate.maj_ver    = pex_info->major_version;
    pstate.min_ver    = pex_info->minor_version;

    if( pstate.maj_ver == 5 && pstate.min_ver == 1)
	win_depth = PEXtk_find_target( &win_visual);
    else
	PEXtk_Get_Visuals( &win_visual);

    win_colormap = DefaultColormapOfScreen( ScreenOfDisplay( dpy, 
			win_screen));
    gc		 = DefaultGC( dpy, win_screen);

    /* Initialize the X data to the defaults */
    px_data.dpy		= displayptr;
    px_data.gc		= gc;
    px_data.win_screen	= win_screen;
    px_data.win_depth	= win_depth;
    px_data.win_visual	= win_visual;
    px_data.cmap	= 0;

    /*
     * Get PEX support information
     */
    PEXtk_Get_Support_Info(pex_info->vendor_name);

    /* Initialize the object pointers */

    p_object.surface_color		= PEXtk_surface_color;
    p_object.surface_color_ind		= PEXtk_default;
    p_object.concat			= PEXtk_concat;
    p_object.preconcat			= PEXtk_preconcat;
    p_object.dither			= PEXtk_dither;
    p_object.antialias			= PEXtk_antialias;
    p_object.transparency		= PEXtk_transparency;

    if( px_data.win_depth == 8 ){
	p_object.pixel_color_ind	= (int (*)()) PEXtk_pseudo_index;
	p_object.pixel_color_rgb	= (int (*)()) Ppseudo_color;
    } else {
	p_object.pixel_color_ind	= (int (*)()) PEXtk_pseudo_index;
	p_object.pixel_color_rgb	= (int (*)()) Ptrue_color;
    }

    pstate.color_type = PEXColorTypeRGB; /* The color type is fixed to RGB */
    pstate.req_type   = PEXOCRender;
    pstate.fp_format  = PEXIEEE_754_32;  /* UPDATE to get fp format */

    /* Initialize the stack location pointers */

    pstate.mtx_loc	 = 0;
    pstate.vpt_loc	 = 0;
    pstate.view_loc	 = 0;
    pstate.attr_loc	 = 0;
    pstate.name_loc	 = 0;

    pstate.cur_view	 	= 0;
    pstate.cur_w	 	= 0;
    pstate.cur_disp_buf		= 0;
    pstate.dither	 	= 0;
    pstate.transparency		= 0;

    for(i=0;i<PEXtk_MAX_BUFFERS;i++)
    {
	pstate.win[i].Rdr_ID	  = 0;
	pstate.win[i].View_ID	  = 0;
	pstate.win[i].Light_ID	  = 0;
	pstate.win[i].Light_ID0   = 0;
	pstate.win[i].Depthcue_ID = 0;
	pstate.win[i].Text_ID	  = 0;
	pstate.win[i].w_id    = pstate.win[i].back_w_id = 0;
	pstate.win[i].mbuf[0] = pstate.win[i].mbuf[1] = 0;
    }

    /*
     * Check if the Multi-Buffering extension is supported.
     * Note that a SunPEX server does not support PEX on any MBX buffer,
     * so do a check for 'SMI' in the vendor string.
     */
    if(XQueryExtension( dpy, "Multi-Buffering",
				&m_op, &f_ev, &f_err) &&
	!strstr (pex_info->vendor_name, "SMI") &&
	!strstr (pex_info->vendor_name, "Hewlett"))
	psupport.dbl_buf |= PEXtk_DBUF_MBUF;

    win_8bit_cmap = ( Colormap ) NULL;

    pstate.num_w_id	 = 0;	/* number of active windows/drawables */
    pstate.cur_struct	 = 0;
    pstate.cur_struct_loc = -1;	/* index of active struct in struct list */
    pstate.cur_str_name	 = PEXtk_STR_EMPTY_NAME;

    pstate.sys_mode 	 = PEXtk_RENDERING_MODE;
    pstate.edit_mode 	 = 0;
    pstate.shade_mode	 = PEXtk_SHADE_WIRE;
    pbuild.bld_mode	 = PEXtk_BLD_NONE;
    pstate.shape 	 = PEXShapeConvex;
    pstate.dirty_mask 	 = 0;

    for(i=0;i<=PEXtk_MAX_LIGHTS;i++)
    {
	LightOnTable[i] = 0;
    }
    active_light_mask = pstate.active_lgts = 0;
    pstate.num_lgts = PEXtk_MAX_LIGHTS;	/* max number of lights stored */

    pstate.attr = &p_attrib[0];	/* current attrib stack pointer */
    pstate.lut_approx  = 0;	/* Lookup table ID for color approx (8-bit) */
    pstate.lut_index   = 0;     /* Lookup table ID for color indexing */

    if(( name_stk = (PEXName *) malloc(psupport.max_names*sizeof(PEXName)))
	== NULL) return 3;

    name_stk[0] = 0;

    pstate.pc_mask[0] = pstate.pc_mask[1] = pstate.pc_mask[2] = 0;
    pstate.pc_ID = PEXCreatePipelineContext(dpy, pstate.pc_mask, &pcAttribs);
    pstate.rdrAttrib.pipeline_context = pstate.pc_ID;

    /* Create the picking nameset */
    pstate.nameSet = (PEXNameSet) PEXCreateNameSet( dpy);
    PEXChangeNameSet( dpy, pstate.nameSet, PEXNSAdd, 1, name_stk);
    pstate.names = name_stk;

    PEXtk_init_attributes();

    pstate.buffer_mode   = PEXtk_SINGLE_BUFFER;
    pstate.mat_mode	 = PEXtk_MATRIX_MODELING;
    pstate.win[0].width  = 1024;	/* set the default width/height */
    pstate.win[0].height = 768;

    pstate.debug_mode = 0;
    if( getenv("PEXTK_DEBUG") )
	pstate.debug_mode = 1;

    PEXtk_init_db(  &p_object);
    PEXtk_init_buf( &p_object);
    PEXtk_init_bld( &p_object);
    PEXtk_init_cmap( &p_object);
    PEXtk_init_utl( &p_object);
    PEXtk_init_dls( &p_object);
    PEXtk_init_pck( &p_object);
    PEXtk_init_mtx( &p_object);
    PEXtk_init_vue( &p_object);

    return 0;

}

/*
 *
 *                       +--------------------------+
 * ----------------------+         Pbackface        +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Turns on/off culling of backfacing polygons.
 *
 *  USAGE:
 *	Pbackface( flag )
 *	int	    flag	(in)	Turns on backface culling if True,
 *					turns off backface culling if False.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pbackface( flag )
int	flag;
{
    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_attrib( PEXtk_OP_BACKFACE, flag);
	return;
    }

    DB_CHECK_DIRTY_MASK( PEXtk_DIRTY_MTX);

    switch( flag ){
	case PEXtk_CULL_BACKFACE:
		PEXSetFacetCullingMode( dpy, pstate.cur_id, pstate.req_type,
			PEXBackFaces);
		/* Don't use backface attributes */
		PEXSetFacetDistinguishFlag( dpy, pstate.cur_id, 
			pstate.req_type, False);
		pstate.pcAttrib.culling_mode = PEXBackFaces;
		pstate.pcAttrib.distinguish  = False;
	    break;
	case PEXtk_CULL_FRONTFACE:
		PEXSetFacetCullingMode( dpy, pstate.cur_id, pstate.req_type,
			PEXFrontFaces);
		PEXSetFacetDistinguishFlag( dpy, pstate.cur_id, 
			pstate.req_type, False);
		pstate.pcAttrib.culling_mode = PEXFrontFaces;
		pstate.pcAttrib.distinguish  = False;
	    break;
	case PEXtk_CULL_NONE:
		PEXSetFacetCullingMode( dpy, pstate.cur_id, pstate.req_type,0);
		PEXSetFacetDistinguishFlag( dpy, pstate.cur_id, 
			pstate.req_type, False);
		pstate.pcAttrib.culling_mode = 0;
		pstate.pcAttrib.distinguish  = False;
	    break;
	default:
		flag = PEXtk_CULL_BACKFACE;
		PEXSetFacetCullingMode( dpy, pstate.cur_id, pstate.req_type,
			PEXBackFaces);
		PEXSetFacetDistinguishFlag( dpy, pstate.cur_id, 
			pstate.req_type, False);
		pstate.pcAttrib.culling_mode = PEXBackFaces;
		pstate.pcAttrib.distinguish  = False;
	    break;
    }

    pstate.attr->backcull = flag;
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCCullingMode);
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCDistinguishFlag);

}

/*
 *
 *                       +--------------------------+
 * ----------------------+          Pborder         +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the geometry outlineing attribute.
 *
 *  USAGE:
 *	Pborder( type )
 *	int	    type	(in)	Turns on border/polygon edge if not
 *					set to PEXtk_BORDER_NONE.
 *
 *  DESCRIPTION:
 *	This routine will set the PEX Surface Edge Type.
 *
 *  NOTE:
 *	This routine does not yet inquire the server on the edge types
 *	which are supported.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pborder( type )
int	type;
{
    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_attrib( PEXtk_OP_BORDER, type);
	return;
    }

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);

    switch( type ){
	case PEXtk_BORDER_NONE:
		pstate.attr->border = type;
		if( pstate.sys_mode == PEXtk_STRUCTURE_MODE )
		    pstate.attr->edgeFlag_str = PEXOff;
		else
		    pstate.attr->edgeFlag = PEXOff;
		PEXSetSurfaceEdgeFlag( dpy, pstate.cur_id, pstate.req_type, 
			PEXOff);
		pstate.attr->edgeType = PEXSurfaceEdgeSolid;
		/* If in struct mode, then don't set pipeline context */
		if( pstate.sys_mode == PEXtk_STRUCTURE_MODE ) return;
    		pstate.pcAttrib.surface_edges     = PEXOff;
    		pstate.pcAttrib.surface_edge_type = pstate.attr->edgeType;
    		PEXSetPCAttributeMask( pstate.pc_mask, PEXPCSurfaceEdgeFlag);
    		PEXSetPCAttributeMask( pstate.pc_mask, PEXPCSurfaceEdgeType);
		return;
	case PEXtk_BORDER_SOLID:
		pstate.attr->edgeType = PEXSurfaceEdgeSolid;
	    break;
	case PEXtk_BORDER_DASHED:
		pstate.attr->edgeType = PEXSurfaceEdgeDashed;
	    break;
	case PEXtk_BORDER_DOTTED:
		pstate.attr->edgeType = PEXSurfaceEdgeDotted;
	    break;
	case PEXtk_BORDER_DASHDOT:
		pstate.attr->edgeType = PEXSurfaceEdgeDashDot;
	    break;
	default:
		return;
    }

    if( pstate.sys_mode == PEXtk_STRUCTURE_MODE )
	pstate.attr->edgeFlag_str = PEXOn;
    else
	pstate.attr->edgeFlag = PEXOn;
    pstate.attr->border = type;
    PEXSetSurfaceEdgeType( dpy, pstate.cur_id, pstate.req_type, 
	pstate.attr->edgeType);
    PEXSetSurfaceEdgeFlag( dpy, pstate.cur_id, pstate.req_type, PEXOn);

    /* If in struct mode, then don't set pipeline context */
    if( pstate.sys_mode == PEXtk_STRUCTURE_MODE ) return;

    pstate.pcAttrib.surface_edges     = PEXOn;
    pstate.pcAttrib.surface_edge_type = pstate.attr->edgeType;
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCSurfaceEdgeFlag);
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCSurfaceEdgeType);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+        Pbuffer_mode      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets buffering mode to Single or Double Buffer.
 *
 *  USAGE:
 *	Pbuffer_mode( mode )
 *	int	    mode	(in)	Sets buffering mode to single buffer
 *					if mode = PEXtk_SINGLE_BUFFER, or 
 *					sets mode to double buffer if 
 *					mode = PEXtk_DOUBLE_BUFFER
 *
 *  NOTE:
 *	If the mode is set to double buffer, then a check is made to see
 *	if the current window has a back buffer associated with it. If it
 *	does not, then one is created. Otherwise, the current 'rendering
 *	drawable' is set to the back buffer.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pbuffer_mode( mode )
int	mode;
{
    int win_width, win_height;

    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_state( PEXtk_OP_BUFFER_MODE, mode);
	return;
    }

    if( pstate.buffer_mode == mode) return;

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
    XFlush( dpy);
    if( mode == PEXtk_DOUBLE_BUFFER )
    {
        if( ( pstate.win[pstate.cur_w].back_w_id == 0 )) {
	    if( ( pstate.win[pstate.cur_w].w_id != 0 ) ) {
		win_width = pstate.win[pstate.cur_w].width;
		win_height = pstate.win[pstate.cur_w].height;
		(*p_object.create_buffer)(win_width,win_height);
	    } else {  /* Error... don't know how this state happened */
		return; /* perhaps this was called before drawable set */
	    }
	}
	pstate.cur_draw = pstate.win[pstate.cur_w].back_w_id; 
	PEXtk_set_clear_function( mode);
    } else {
	pstate.cur_draw = pstate.win[pstate.cur_w].w_id; 
	PEXtk_set_clear_function( mode);
    }

    /* pstate.cur_rdr  = pstate.win[pstate.cur_w].Rdr_ID; */

    /* do we really need a reset_pex_state */
    /* VERIFY */
    if( pstate.cur_rdr != 0 && pstate.cur_draw != 0 ){
	    PEXtk_reset_pex_state(1,(long)0);
    }

    pstate.buffer_mode = mode;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+    Pcall_display_list    +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Executes a Display List.
 *
 *  USAGE:
 *	Pcall_display_list( name )
 *	int	    name	(in)	Name of display list to call/execute
 *
 *  DESCRIPTION:
 *	This routine will execute the display list with a id of 'name'. 
 *	The 'name' does NOT have to exist in the name_block table,
 *	although ONLY if this is put in a display list.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pcall_display_list( name )
int	name;
{

	int i=0;
	DLS_NAME *block_ptr;

	/* if( pstate.cur_dls_name == PEXtk_DLS_EMPTY_NAME ) */
	if( pstate.sys_mode != PEXtk_DLS_MODE )
	{
	    /* Find name in p_dl_obj.name_block */
#ifndef NO_DB_DLS
	    i = DB_find_dls_name( name, &block_ptr );
	    if( i == -1 )  /* error, name must exist */
		return;
	    DLI_call_dls( block_ptr->node[i] );
#endif
	    DLS_call_dpy_list( name );
	} else {	/* dpy list is being created, store this instruction */
	    DLS_call_dls( name );
	}
}

/*
 *
 *                       +--------------------------+
 * ----------------------+       Pcall_struct       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Executes a structure and saves the state
 *
 *  USAGE:
 *	Pcall_struct( name )
 *	int	    name	(in)	Name of structure to execute
 *
 *  DESCRIPTION:
 *	This routine will execute/call the structure with a id of 'name'.
 *	If 'name' does NOT exist in the str_list table, then an empty one 
 *	is created (i.e., one os created, and the 'empty' flag is set
 *	to 1). Since this routine may be called even when the structure
 *	'name' does not exist yet, this routines does not return anything.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pcall_struct( name )
int	name;
{
    int i=0;

    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_exec( PEXtk_OP_CALL_STRUCT, name);
	return;
    }

	DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
	/* Find name in pstate.str_list */
	i = DB_find_struct_name( name );

	if( i == -1 ){		/* doesn't exist yet; create an empty one */
	    if( pstate.sys_mode == PEXtk_RENDERING_MODE || 
	        pstate.sys_mode == PEXtk_PICKING_MODE )
		return ; /* in immed. mode, it MUST exist */

	    DB_check_struct_list();

	    i = DB_insert_struct_name(name);
	    pstate.str_list[i].empty = 0;

            if((pstate.str_list[i].id = PEXCreateStructure( dpy)) == 0)
		return ;		/* a Fatal error */
	}

	/* Push everything */

	PEXExecuteStructure( dpy, pstate.cur_id, pstate.req_type, 
		pstate.str_list[i].id);

}

/*
 *
 *                       +--------------------------+
 * ----------------------+    Pclose_display_list   +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Closes an opened display list.
 *
 *  USAGE:
 *	Pclose_display_list( )
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pclose_display_list( )
{

	if( pstate.sys_mode != PEXtk_DLS_MODE )
		return;
	XFlush( dpy);
	pstate.cur_dls_name = PEXtk_DLS_EMPTY_NAME;

	pstate.sys_mode 	 = PEXtk_RENDERING_MODE;

}

/*
 *
 *                       +--------------------------+
 * ----------------------+       Pclose_struct      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Closes an opened structure for retained mode.
 *
 *  USAGE:
 *	Pclose_struct( )
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pclose_struct( )
{

	XFlush( dpy);

	if( pstate.cur_struct )
	    pstate.cur_struct = 0;

	p_object.concat			= PEXtk_concat;
	p_object.preconcat		= PEXtk_preconcat;

	pstate.sys_mode 	 = PEXtk_RENDERING_MODE;
	pstate.cur_str_name	 = PEXtk_STR_EMPTY_NAME;
	pstate.cur_label	 = PEXtk_STR_EMPTY_LABEL;
	pstate.cur_struct_loc	 = -1;
	pstate.cur_label_loc	 = -1;
        pstate.attr->i_style_str = 0;	/* depends on 0 being not defined */
        pstate.attr->edgeFlag_str = PEXOff;
        pstate.attr->interp_meth_str = 0;/* depends on 0 being not defined */

	pstate.cur_id   = pstate.cur_rdr;
	pstate.req_type = PEXOCRender;

    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+   Pcreate_display_list   +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Creates a display list, which is simply a structure on the client
 *	side.
 *
 *  USAGE:
 *	Pcreate_display_list( name )
 *	int	    name	(in)	Name of display list to open/create
 *
 *  DESCRIPTION:
 *	This routine will check to see if a dpy_list is already open. If one
 *	is, a DLS_OPEN error is returned. It will also check to see if 'name'
 *	already exists, if it does, an DLS_EXISTS error is returned.
 *
 *	If the above two checks are passed, then the 'name' is placed in
 *	the name_block, and either a dls_node block is allocated, or a pointer
 *	is adjusted to the next available dls node.
 *
 *  RETURNS:
 *	0 if successful, non-zero if error
 */
 
/*ARGSUSED*/
int
Pcreate_display_list( name )
int	name;
{

	if( pstate.cur_dls_name != PEXtk_DLS_EMPTY_NAME ){
	    pstate.error = PEXtk_ERR_DLS_OPEN;
	    return pstate.error;	/* a dls is already open */
	}

	XFlush( dpy);
	DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);

	/* See if a display list of 'name' already exists */
	/* Should it be added to a 'Database' of DLS ? */
#define NO_DB_DLS
#ifndef NO_DB_DLS
	if(DB_add_dls_list(name))
	    return 1;
#endif

	if( DLS_create_dpy_list( name ) )
	    return 1;

	pstate.cur_dls_name    		= name;

	pstate.sys_mode 	 = PEXtk_DLS_MODE;

	return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      Pcreate_struct      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Opens a structure for retained mode.
 *
 *  USAGE:
 *	Pcreate_struct( name )
 *	int	    name	(in)	Name of structure to open/create
 *
 *  RETURNS:
 *	0 if successful, non-zero if error
 */
 
/*ARGSUSED*/
int
Pcreate_struct( name )
int	name;
{
    int i;

	if( pstate.cur_struct != 0 )	/* structure is already open */
	    return PEXtk_ERR_STR_OPEN;

	DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
	XFlush( dpy);

	i = DB_find_struct_name( name );

	if( i != -1 ){	/* does an empty structure exist ? */

	    /* make sure that the location returned is not larger
	     * than the number of existing structures (this error
	     * should NOT occur), and check that the empty flag
	     * is set to 1.
	     */
	    if( i > pstate.str_cnt-1 || pstate.str_list[i].empty == 1 )
		return 1;

	} else {

	    DB_check_struct_list();

	    i = DB_insert_struct_name(name);

            if((pstate.str_list[i].id = PEXCreateStructure( dpy)) == 0)
		return 1;
	}

	pstate.str_list[i].empty = 0;

	pstate.cur_struct 		= pstate.str_list[i].id;
	pstate.cur_str_name 		= name;
	p_object.concat			= PEXtk_concat_struct;
	p_object.preconcat		= PEXtk_preconcat_struct;

/* printf("create_struct %d, loc, %d %x\n",name,i,pstate.str_list[i].id); */

	pstate.sys_mode 	 = PEXtk_STRUCTURE_MODE;
	pstate.edit_mode 	 = 0;
	pstate.cur_id	 	 = pstate.cur_struct;
	pstate.req_type	 	 = PEXOCStore;
	pstate.cur_struct_loc	 = i;
	pstate.attr->i_style_str = 0;	/* set Interior Style to not defined */
        pstate.attr->edgeFlag_str = PEXOff;

	return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+       Pexec_struct       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Executes a structure.
 *
 *  USAGE:
 *	Pexec_struct( name )
 *	int	    name	(in)	Name of structure to execute
 *
 *  DESCRIPTION:
 *	This routine will execute the structure with a id of 'name'. If
 *	'name' does NOT exist in the str_list table, then an empty one is
 *	created (i.e., one is created, and the 'empty' flag is set to 1).
 *	Since this routine may be called even when the structure 'name'
 *	does not exist yet (although ONLY if this is put in a display 
 *	list or in a structure), this routines does not return anything.
 *
 *  NOTE:
 *	This routine is specified to not save any attributes, however,
 *	in the current version of PEX, the state is automatically saved.
 *	A future version of PEX will support push/pop control for this.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pexec_struct( name )
int	name;
{
	int i=0;

    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_exec( PEXtk_OP_EXEC_STRUCT, name);
	return;
    }

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);

	/* Find name in pstate.str_list */
	i = DB_find_struct_name( name );

	if( i == -1 ){		/* doesn't exist yet; create an empty one ? */
	    if( pstate.sys_mode == PEXtk_RENDERING_MODE ||
	        pstate.sys_mode == PEXtk_PICKING_MODE )
		return ; /* in immed. mode, it MUST exist */
	    DB_check_struct_list();

	    i = DB_insert_struct_name(name);
	    pstate.str_list[i].empty = 0;

            if((pstate.str_list[i].id = PEXCreateStructure( dpy)) == 0)
		return ;		/* a Fatal error */
	}

	PEXExecuteStructure( dpy, pstate.cur_id, pstate.req_type, 
		pstate.str_list[i].id);

    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+           Pclear         +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Clears viewport to specified color.
 *
 *  USAGE:
 *	Pclear( col )
 *	PCOLOR	    *col	(in)	color to clear to
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pclear( col )
PCOLOR	*col;
{
#ifdef MPEX
    long pix;
#endif

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
    if( col->type == PEXtk_COLOR_INDEXED ){
	pstate.attr->colorBg.indexed.index = col->color.index.value;
	PEXtk_rgb_index( col );
#ifdef MPEX
	pix = Ppixel_value_rgb( col->color.rgb.red,
			col->color.rgb.green, col->color.rgb.blue);
#else
#endif
    } else {
#ifdef MPEX
	pix = (*p_object.pixel_color_rgb)( col->color.rgb.red, 
				col->color.rgb.green, col->color.rgb.blue);
#endif
	pstate.attr->colorBg.rgb.red  = col->color.rgb.red;
	pstate.attr->colorBg.rgb.green= col->color.rgb.green;
	pstate.attr->colorBg.rgb.blue = col->color.rgb.blue;
    }

#ifdef MPEX
    (*p_object.clear_buffer)(pix);
#else
    (*p_object.clear_buffer)( (float) col->color.rgb.red, (float) col->color.rgb.green,
	(float) col->color.rgb.blue);
#endif
    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+         Pclear_hsv       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Clears viewport to specified hsv color.
 *
 *  USAGE:
 *	Pclear_hsv( h, s, v )
 *	Fdata	    h		(in)	Value for red
 *	Fdata	    s		(in)	Value for green
 *	Fdata	    v		(in)	Value for blue
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pclear_hsv( h, s, v )
Fdata	h;
Fdata	s;
Fdata	v;
{
    float r,g,b;

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
    if( pstate.color_type == PEXColorTypeRGB ){
	PEXtk_convert_hsv_to_rgb( h, s, v, &r, &g, &b);
	pstate.attr->colorBg.rgb.red   = r;
	pstate.attr->colorBg.rgb.green = g;
	pstate.attr->colorBg.rgb.blue  = b;
    } else {
	pstate.attr->colorBg.hsv.hue   = h;
	pstate.attr->colorBg.hsv.saturation = s;
	pstate.attr->colorBg.hsv.value  = v;
    }
	
#ifdef MPEX
    { long pix;
    pix = (*p_object.pixel_color_rgb)( r, g, b);
    (*p_object.clear_buffer)(pix);
    }
#else
    (*p_object.clear_buffer)(r, g, b);
#endif
    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+         Pclear_ind       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Clears viewport to specified color index.
 *
 *  USAGE:
 *	Pclear_ind( index )
 *	int	    index	(in)	Index of color to clear to
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pclear_ind( index )
int	index;
{
    PCOLOR col;

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
    col.color.index.value = index;
    PEXtk_rgb_index( &col );
#ifdef MPEX
    { long pix;
	pix = Ppixel_value_rgb( col.color.rgb.red,
			col.color.rgb.green, col.color.rgb.blue);
	(*p_object.clear_buffer)( pix );
    }
#else
    { float r,g,b;
    r= col.color.rgb.red;
	 g= col.color.rgb.green;  b= col.color.rgb.blue;
/*
    (*p_object.clear_buffer)( (float) r,
	 (float) g,  (float) b);
*/
    (*p_object.clear_buffer)( (float) col.color.rgb.red, 
	 (float) col.color.rgb.green,  (float) col.color.rgb.blue);
    }
#endif
    pstate.attr->colorBg.rgb.red  = col.color.rgb.red;
    pstate.attr->colorBg.rgb.green= col.color.rgb.green;
    pstate.attr->colorBg.rgb.blue = col.color.rgb.blue;

    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+         Pclear_rgb       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Clears viewport to specified rgb color.
 *
 *  USAGE:
 *	Pclear_rgb( r, g, b )
 *	Fdata	    r		(in)	Value for red
 *	Fdata	    g		(in)	Value for green
 *	Fdata	    b		(in)	Value for blue
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pclear_rgb( r, g, b )
Fdata	r;
Fdata	g;
Fdata	b;
{

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
    pstate.attr->colorBg.rgb.red  = r;
    pstate.attr->colorBg.rgb.green= g;
    pstate.attr->colorBg.rgb.blue = b;

#ifdef MPEX
    { long pix;
    pix = (*p_object.pixel_color_rgb)( r, g, b);
    (*p_object.clear_buffer)(pix);
    }
#else
    (*p_object.clear_buffer)(r, g, b);
#endif
    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+         Pcolor           +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the polygon/fill color.
 *
 *  USAGE:
 *	Pcolor( col )
 *	PCOLOR	    *col	(in)	color of poly
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pcolor( col )
PCOLOR	*col;
{

    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_attrib_clr( PEXtk_OP_COLOR, col );
	return;
    }
    if( col->type == PEXtk_COLOR_HSV && pstate.color_type == PEXColorTypeRGB)
    {
	PEXtk_convert_hsv_to_rgb(  col->color.hsv.hue,
			col->color.hsv.saturation,
			col->color.hsv.value,
			&pstate.attr->colorPoly.rgb.red,
			&pstate.attr->colorPoly.rgb.green,
			&pstate.attr->colorPoly.rgb.blue);
    } else {
	pstate.attr->colorPoly.rgb.red   = col->color.rgb.red;
	pstate.attr->colorPoly.rgb.green = col->color.rgb.green;
	pstate.attr->colorPoly.rgb.blue  = col->color.rgb.blue;
    }
	
    (*p_object.surface_color)( &pstate.attr->colorPoly);
    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+         Pcolor_hsv       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the polygon/fill color.
 *
 *  USAGE:
 *	Pcolor_hsv( h, s, v )
 *	Fdata	    h		(in)	Value for hue
 *	Fdata	    s		(in)	Value for saturation
 *	Fdata	    v		(in)	Value for value
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pcolor_hsv( h, s, v )
Fdata	h;
Fdata	s;
Fdata	v;
{
    float r,g,b;

    if( pstate.color_type == PEXColorTypeRGB ){
	PEXtk_convert_hsv_to_rgb( h, s, v, &r, &g, &b);
	pstate.attr->colorPoly.rgb.red   = r;
	pstate.attr->colorPoly.rgb.green = g;
	pstate.attr->colorPoly.rgb.blue  = b;
    } else {
	pstate.attr->colorPoly.hsv.hue   = h;
	pstate.attr->colorPoly.hsv.saturation = s;
	pstate.attr->colorPoly.hsv.value  = v;
    }
	
    (*p_object.surface_color)( &pstate.attr->colorPoly);
    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+         Pcolor_ind       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the polygon/fill color to an indexed value.
 *
 *  USAGE:
 *	Pcolor_ind( index )
 *	int	    index	(in)	index color of poly
 *
 *  NOTE:
 *	The 'index' represents the position in the color table (set by
 *	the set_color_table routine) to get the color values.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pcolor_ind( index )
int	index;
{

    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_attrib_clr_ind( PEXtk_OP_COLOR_IND, index );
	return;
    }

    pstate.attr->colorPolyInd = index;
    (*p_object.surface_color_ind)( pstate.attr->colorPolyInd);
    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+         Pcolor_rgb       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the polygon/fill color.
 *
 *  USAGE:
 *	Pcolor_rgb( r, g, b )
 *	Fdata	    r		(in)	Value for red
 *	Fdata	    g		(in)	Value for green
 *	Fdata	    b		(in)	Value for blue
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pcolor_rgb( r, g, b )
Fdata	r;
Fdata	g;
Fdata	b;
{
	
    pstate.attr->colorPoly.rgb.red   = r;
    pstate.attr->colorPoly.rgb.green = g;
    pstate.attr->colorPoly.rgb.blue  = b;
	
    (*p_object.surface_color)( &pstate.attr->colorPoly);
    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+   Pcurve_approx_method   +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the curve approximation method.
 *
 *  USAGE:
 *	Pcurve_approx_method( meth, tol )
 *	int	    meth	(in)	Curve approximation method to use
 *	Fdata	    tol		(in)	Curve tolerance in u-dir
 *
 *  RETURNS:
 *      Nothing
 */
                                                           
/*ARGSUSED*/
void
Pcurve_approx_method( meth, tol )
int     meth;
Fdata   tol;
{
    switch( meth ){
	case PEXtk_APPROX_IMP_DEP:
		pstate.pcAttrib.curve_approx.method = PEXApproxImpDep;
	    break;
	case PEXtk_APPROX_CONSTANT_KNOTS:
		pstate.pcAttrib.curve_approx.method = PEXApproxConstantBetweenKnots;
	    break;
	case PEXtk_APPROX_WC_CHORD_LEN:
		pstate.pcAttrib.curve_approx.method = PEXApproxWCChordalSize;
	    break;
	case PEXtk_APPROX_NC_CHORD_LEN:
		pstate.pcAttrib.curve_approx.method = PEXApproxNPCChordalSize;
	    break;
	case PEXtk_APPROX_DC_CHORD_LEN:
		pstate.pcAttrib.curve_approx.method = PEXApproxDCChordalSize;
	    break;
	case PEXtk_APPROX_WC_CHORD_DEV:
		pstate.pcAttrib.curve_approx.method = PEXCurveApproxWCChordalDev;
	    break;
	case PEXtk_APPROX_NC_CHORD_DEV:
		pstate.pcAttrib.curve_approx.method = PEXCurveApproxNPCChordalDev;
	    break;
	case PEXtk_APPROX_DC_CHORD_DEV:
		pstate.pcAttrib.curve_approx.method = PEXCurveApproxDCChordalDev;
	    break;
	case PEXtk_APPROX_WC_RELATIVE:
		pstate.pcAttrib.curve_approx.method = PEXApproxWCRelative;
	    break;
	case PEXtk_APPROX_NC_RELATIVE:
		pstate.pcAttrib.curve_approx.method = PEXApproxNPCRelative;
	    break;
	case PEXtk_APPROX_DC_RELATIVE:
		pstate.pcAttrib.curve_approx.method = PEXApproxDCRelative;
	    break;
	default:
	    break;
    }
    pstate.attr->curv_meth = meth;
    pstate.attr->curv_utol = tol;
    PEXSetCurveApprox( dpy, pstate.cur_id, pstate.req_type, 
	pstate.pcAttrib.curve_approx.method, (double) tol);

    if( ! IMMEDIATE_MODE(pstate.sys_mode)) return;
    pstate.pcAttrib.curve_approx.tolerance = tol;
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCCurveApprox);
    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+    Pcurve_parm_range     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the curve parametric range.
 *
 *  USAGE:
 *	Pcurve_parm_range( umin, umax )
 *	Fdata	    umin	(in)	Curve min parametric value
 *	Fdata	    umax	(in)	Curve max parametric value
 *
 *  RETURNS:
 *	Nothing
 */

/*ARGSUSED*/
void
Pcurve_parm_range( umin, umax )
Fdata   umin;
Fdata   umax;
{
	pstate.attr->curv_umin = umin;
	pstate.attr->curv_umax = umax;
    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+   Pdelete_display_list   +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Deletes a display list.
 *
 *  USAGE:
 *	Pdelete_display_list( name )
 *	int	    name	(in)	Name of display list to delete
 *
 *  RETURNS:
 *	0 if successful, non-zero if error
 */
 
/*ARGSUSED*/
int
Pdelete_display_list( name )
int	name;
{
    int i;
    DLS_NAME *block_ptr;

	XFlush( dpy);

	i = DB_find_dls_name( name, &block_ptr );

	if( i == -1 )	/* does not exist */
	    return 1;

#ifndef NO_DB_DLS
	DB_delete_dls_name( i, block_ptr, name);
#endif

	return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      Pdelete_struct      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Deletes a structure.
 *
 *  USAGE:
 *	Pdelete_struct( name )
 *	int	    name	(in)	Name of structure to delete
 *
 *  RETURNS:
 *	0 if successful, 1 if error
 */
 
/*ARGSUSED*/
int
Pdelete_struct( name )
int	name;
{
	XFlush( dpy);

	if( (DB_delete_struct_name(name)) == -1 ) return 1;

	return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+         Pdepth           +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the near and far clipping planes to those given in near
 *	and far.
 *
 *  USAGE:
 *	Pdepth( near, far )
 *	Fcoord	    near	(in)	near clipping plane
 *	Fcoord	    far		(in)	far clipping plane
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pdepth( near, far)
Fcoord	near;
Fcoord	far;
{

    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_view_mode( PEXtk_OP_DEPTH, 0, near, far, 0);
	return;
    }

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
    switch( pstate.view->type ){

	case PEXtk_VIEW_PERSP:
		Ppersp_view( pstate.view_loc, pstate.view->fov, 
			pstate.view->aspect, near, far);
		break;
	case PEXtk_VIEW_ORTHO:
		Portho_view( pstate.view_loc, pstate.view->left, 
			pstate.view->right,
			pstate.view->bottom, pstate.view->top,
			near, far);
		break;
	case PEXtk_VIEW_ORTHO_2D:
		break;
	case PEXtk_VIEW_LOOKAT:
	case PEXtk_VIEW_POLAR:
		break;
    }
}

/*
 *
 *                       +--------------------------+
 * ----------------------+        Pdepth_cue        +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Turns depth cue on or off.
 *
 *  USAGE:
 *	Pdepth_cue( flag, rmin, rmax )
 *	int	    flag	(in)	Mode for depth cue, 1=on, 0=off
 *	Fcoord	    rmin	(in)	Minimum range for depth cue
 *	Fcoord	    rmax	(in)	Maximum range for depth cue
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pdepth_cue( flag, rmin, rmax )
int	flag;
Fcoord	rmin;
Fcoord	rmax;
{
    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_view_mode( PEXtk_OP_DEPTH_CUE, flag, rmin, rmax, 0);
	return;
    }

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
    if( pstate.win[pstate.cur_w].Depthcue_ID == 0 )
	pstate.win[pstate.cur_w].Depthcue_ID = PEXCreateLookupTable( dpy, 
	    pstate.win[pstate.cur_w].w_id, PEXLUTDepthCue);

    if( flag )
	Depthcue_table.mode = PEXOn;  
    else
	Depthcue_table.mode = PEXOff;  

    Depthcue_table.front_plane = rmax; Depthcue_table.back_plane = rmin; 
    Depthcue_table.front_scaling = 1.0; Depthcue_table.back_scaling = 0.0;
    Depthcue_table.color.type = PEXColorTypeRGB;
   /* 
    * This should be the background color
    */
    Depthcue_table.color.value.rgb.red = 
				pstate.attr->colorBg.rgb.red;
    Depthcue_table.color.value.rgb.green = 
				pstate.attr->colorBg.rgb.green;
    Depthcue_table.color.value.rgb.blue =
				pstate.attr->colorBg.rgb.blue;

    PEXSetTableEntries( dpy, pstate.win[pstate.cur_w].Depthcue_ID, 1, 1,
		PEXLUTDepthCue, &Depthcue_table);

    PEXtk_reset_pex_state( 0, PEXRADepthCueTable);

    if( flag ){
	PEXSetDepthCueIndex( dpy, pstate.cur_id, pstate.req_type, 1);
	pstate.pcAttrib.depth_cue_index = 1;
    } else {
	PEXSetDepthCueIndex( dpy, pstate.cur_id, pstate.req_type, 0);
	pstate.pcAttrib.depth_cue_index = 0;
    }
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCDepthCueIndex);

}

/*
 *
 *                       +--------------------------+
 * ----------------------+          Pdither         +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Turns dithering on or off.
 *
 *  USAGE:
 *	Pdither( mode, pattern )
 *	int	    mode	(in)	Mode for dithering, 1=on, 0=off
 *	int	    pattern	(in)	Pattern to use for dithering
 *
 *  NOTE:
 *	Support for dithering is platform specific.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pdither( mode, pattern )
int	mode;
int	pattern;
{
    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_dither( PEXtk_OP_DITHER, mode, pattern );
	return;
    }
    pstate.dither = mode;
    pstate.dither_pattern = pattern;
    (*p_object.dither)( mode, pattern );
}

/*
 *
 *                       +--------------------------+
 * ----------------------+    Pdisplay_list_used    +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Determines if a specified display list name is in use
 *
 *  USAGE:
 *	Pdisplay_list_used( name )
 *	int	    name	(in)	Name of display list to check
 *
 *  RETURNS:
 *	0 if not in use, 1 if it is in use
 */
 
/*ARGSUSED*/
int
Pdisplay_list_used( name )
int	name;
{
    int i;
    DLS_NAME *block_ptr;

	i = DB_find_dls_name(name, &block_ptr);

	if( i == -1 )
	    return 0;
	else
	    return 1;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+   Pedit_multiply_matrix  +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Gets a matrix from structure name at position label, and 
 *	multiplies it by mat.
 *
 *  USAGE:
 *	Pedit_multiply_matrix( name, label, mat, flag )
 *	int	    name	(in)	The structure id to apply matrix to
 *	int	    label	(in)	The label at which matrix is located
 *	Matrix	    mat		(out)	The matrix to pre-concat
 *	int	    flag	(in)	Multiply flag, FALSE=preconcat,
 *					TRUE=postconcat
 *
 *  RETURNS:
 *	0 if successful, 1 if error
 */
 
/*ARGSUSED*/
int
Pedit_multiply_matrix( name, label, mat, flag )
int	name;
int	label;
Matrix	mat;
int	flag;
{
    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
#ifdef UPDATE
	Matrix prod, cur_mat;
	pxlLocalTransform pex_lxform;
	char *pxinfo;

	char *addr;
	unsigned short inclList, exclList;
	int i;
	unsigned long offset, numEl, sizeRet;

	XFlush( dpy);
	i = DB_find_struct_name( name );
	if( i == -1 ) return 1;

	PEXSetElementPtr( dpy, pstate.str_list[i].id, PEXBeginning, 0);
	PEXSetElementPtrAtLabel( dpy, pstate.str_list[i].id, label,1);

	inclList = PEXOCLocalTransform;
	exclList=0;
	if( PEXElementSearch( dpy, pstate.str_list[i].id, PEXCurrent, 0,
		PEXForward, 1, &inclList, 0, &exclList, &offset))
	    return 1;
	if( PEXFetchElements( dpy, pstate.str_list[i].id, PEXCurrent, 0, 
		PEXCurrent, 0, pstate.fp_format, &numEl, &sizeRet, &pxinfo))
	    return 1;
	if( sizeRet < (sizeof(pxlLocalTransform) >> 2) )
	    return 1;

	addr = (char *) pxinfo;

	memcpy( (char *) &pex_lxform, addr, sizeof(pxlLocalTransform));

	/* make sure it is a LocalTransform command */
	if( pex_lxform.head.elementType != PEXOCLocalTransform ){
	    /* PEXFree( (char *) pxinfo ); */
	    return 1;
	}

	/* Don't forget to transpose the input matrix before we multiply it */
	/* VERIFY!!!! */
	PEXtk_matrix_tcopy( cur_mat, mat);

	if( POSTFLAG_IS_TRUE(flag) ) /* post multiply the matrix */
	    PEXtk_matrix_mult( cur_mat, pex_lxform.matrix, prod);
	else /* pre multiply the matrix */
	    PEXtk_matrix_mult( pex_lxform.matrix, cur_mat, prod);

	/* Should the composition type be changed ?? */
	/* VERIFY */
	/* pex_lxform.compType = PEXReplace; */
	PEXtk_matrix_copy( pex_lxform.matrix, prod);
	PEXSetEditingMode( dpy, pstate.str_list[i].id, 
		PEXStructureReplace);
	/* replace the matrix */
	/* PEXSetLocalTransform( OCBuf, &local_mtx); */
	/* local_mtx.composition = pex_lxform.compType; */
	memcpy( (char *) local_mtx.matrix, (char *) pex_lxform.matrix, 
		sizeof(pexMatrix));
	PEXSetLocalTransform( dpy, pstate.cur_id, pstate.req_type, 
		pex_lxform.compType, local_mtx);
	XFlush( dpy);
	XFree( (char *) pxinfo );

#endif
	return 0;

}

/*
 *
 *                       +--------------------------+
 * ----------------------+       Pedit_struct       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Opens an existing structure for editing.
 *
 *  USAGE:
 *	Pedit_struct( name )
 *	int	    name	(in)	Name of structure to edit
 *
 *  RETURNS:
 *	0 if successful, 1 if error
 */
 
/*ARGSUSED*/
int
Pedit_struct( name )
int	name;
{
    int i, attrib;

	DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
	XFlush( dpy);

	i = DB_find_struct_name(name);
	if( i == -1 ) return 1;
	if( pstate.str_list[i].id == 0 ) return 1;
	pstate.cur_struct	 = pstate.str_list[i].id;
	pstate.sys_mode 	 = PEXtk_STRUCTURE_MODE;
	pstate.edit_mode 	 = PEXtk_EDITING_MODE;
	pstate.cur_id	 	 = pstate.cur_struct;
	pstate.req_type	 	 = PEXOCStoreSingle;
	pstate.cur_struct_loc	 = i;
	pstate.cur_str_name 	 = name;
	pstate.cur_label	 = 0;
	pstate.cur_label_loc	 = 0;

	PEXSetEditingMode( dpy, pstate.cur_struct, PEXStructureInsert);
	PEXSetElementPtr( dpy, pstate.cur_struct, PEXEnd, 0);
	attrib = pstate.str_list[pstate.cur_struct_loc].attr;
	pstate.attr->i_style_str     = attrib & 0x0f;
	pstate.attr->edgeFlag_str    = (( attrib & 0xf0 ) >> 4 );
	pstate.attr->interp_meth_str = (( attrib & 0xf00 ) >> 8 );

	p_object.concat			= PEXtk_concat_struct;
	p_object.preconcat		= PEXtk_preconcat_struct;

	return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+           Pexit          +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Closes PEXtk and frees data.
 *
 *  USAGE:
 *	Pexit()
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pexit()
{
    int i;

    PEXEndRendering(dpy, pstate.win[pstate.cur_w].Rdr_ID, True);

    /* Free the PC and the LUT used for the colormap */
    if( px_data.win_depth == 8 ){ 
	if( pstate.lut_approx != 0 ) 
	    PEXFreeLookupTable( dpy, pstate.lut_approx);
	if( pstate.lut_index != 0 ) 
	    PEXFreeLookupTable( dpy, pstate.lut_index);
	pstate.lut_approx = 0;
	pstate.lut_index = 0;
	PEXtk_FreeColormap();
    }
    if( pstate.pc_ID != 0 ){
	PEXFreePipelineContext( dpy, pstate.pc_ID);
	pstate.pc_ID = 0;
    }
    PEXFreeNameSet( dpy, pstate.nameSet);
    pstate.nameSet = 0;
    if( pstate.attr->font )
	PEXUnloadFont( dpy, pstate.attr->font);

    DB_free_struct_list();	/* free struct labels and struct list */
    PEXtk_close_dls();
    PEXtk_close_db();
    PEXtk_close_utl();
    if( name_stk != NULL )
	free( (char *) name_stk );
    name_stk = NULL;
    pstate.names = NULL;

	i = 0;
	if( pstate.win[i].Rdr_ID )
	    PEXFreeRenderer( dpy, pstate.win[i].Rdr_ID);
	if( pstate.win[i].View_ID )
	    PEXFreeLookupTable( dpy, pstate.win[i].View_ID);
	if( pstate.win[i].Light_ID )
	    PEXFreeLookupTable( dpy, pstate.win[i].Light_ID);
#ifdef LIGHT_BUG
	if( pstate.win[i].Light_ID0 )
	    PEXFreeLookupTable( dpy, pstate.win[i].Light_ID0);
#endif
	if( pstate.win[i].Text_ID )
	    PEXFreeLookupTable( dpy, pstate.win[i].Text_ID);

    for(i=0;i<PEXtk_MAX_BUFFERS;i++){
	pstate.cur_w = i;
	(*p_object.delete_buffer)();
	pstate.win[i].Rdr_ID    = 0;
	pstate.win[i].View_ID   = 0;
	pstate.win[i].Light_ID  = 0;
	pstate.win[i].Light_ID0 = 0;
	pstate.win[i].Text_ID   = 0;
    }
}

/*
 *
 *                       +--------------------------+
 * ----------------------+          Pflush          +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Flushes the buffers.
 *
 *  USAGE:
 *	Pflush( flag )
 *	int	    flag	(in)	Flag set to True to do an End/Begin
 *					Rendering sequence.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pflush( flag )
int	flag;
{
	DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
	XFlush(dpy);
	/*
	 *   Note: an End/Begin Render sequence is needed 
	 *   mostly for single buffer mode.
	 */
	if( flag )
	    PEXtk_reset_pex_state(1,(long)0);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+     Pget_backbuffer      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Returns the current back buffer ID.
 *
 *  USAGE:
 *	Pget_backbuffer( )
 *
 *  RETURNS:
 *	The ID of the back buffer. This may be and XID or an
 *	implementation dependent ID.
 */
 
/*ARGSUSED*/
unsigned long
Pget_backbuffer()
{
    return ((unsigned long) pstate.win[pstate.cur_w].back_w_id);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      Pget_backface       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Returns the current backfacing attribute.
 *
 *  USAGE:
 *	Pget_backface( )
 *
 *  RETURNS:
 *	The value of the backfacing attribute
 */
 
/*ARGSUSED*/
int
Pget_backface()
{
    return pstate.attr->backcull;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+       Pget_border        +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Returns the current border attribute.
 *
 *  USAGE:
 *	Pget_border( )
 *
 *  RETURNS:
 *	The value of the border attribute
 */
 
/*ARGSUSED*/
int
Pget_border()
{
    return pstate.attr->border;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+     Pget_buffer_mode     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Returns the current buffering mode.
 *
 *  USAGE:
 *	Pget_buffer_mode( )
 *
 *  RETURNS:
 *	The value of the buffer_mode state
 */
 
/*ARGSUSED*/
int
Pget_buffer_mode()
{
    return pstate.buffer_mode;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+         Pget_color       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Returns the polygon/fill color.
 *
 *  USAGE:
 *	Pget_color( *col )
 *	PCOLOR	    *col	(in)	pointer to color for poly
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pget_color( col )
PCOLOR	*col;
{

    col->color.rgb.red   = pstate.attr->colorPoly.rgb.red;
    col->color.rgb.green = pstate.attr->colorPoly.rgb.green;
    col->color.rgb.blue  = pstate.attr->colorPoly.rgb.blue;
	
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      Pget_depth_cue      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Returns depthcue values, and whether its on/off.
 *
 *  USAGE:
 *	Pget_depth_cue( rmin, rmax )
 *	Fcoord	    *rmin	(out)	Minimum range for depth cue
 *	Fcoord	    *rmax	(out)	Maximum range for depth cue
 *
 *  RETURNS:
 *	0 if depth cueing is off, 1 if it is on
 */
 
/*ARGSUSED*/
int
Pget_depth_cue( rmin, rmax )
Fcoord	*rmin;
Fcoord	*rmax;
{
    if( Depthcue_table.mode == PEXOff )
	return 0;

    *rmin = Depthcue_table.back_plane; 
    *rmax = Depthcue_table.front_plane;
    return 1;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+    Pget_display_buffer   +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Returns the current display buffer ID.
 *
 *  USAGE:
 *	Pget_display_buffer( )
 *
 *  RETURNS:
 *	The ID of the display buffer. This might be an XID or an
 *	implementation dependent ID.
 */
 
/*ARGSUSED*/
unsigned long
Pget_display_buffer()
{
    if( pstate.buffer_mode == PEXtk_DOUBLE_BUFFER )
	return ((unsigned long) pstate.win[pstate.cur_w].back_w_id);
    else
	return ((unsigned long) pstate.win[pstate.cur_w].w_id);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+     Pget_frontbuffer     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Returns the current front buffer ID.
 *
 *  USAGE:
 *	Pget_frontbuffer( )
 *
 *  RETURNS:
 *	The ID of the front buffer. This might be an XID or an
 *	implementation dependent ID.
 */
 
/*ARGSUSED*/
unsigned long
Pget_frontbuffer()
{
    return ((unsigned long) pstate.win[pstate.cur_w].w_id);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+     Pget_light_color     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Returns the light color.
 *
 *  USAGE:
 *	Pget_light_color( *col )
 *	PCOLOR	    *col	(in)	pointer to color for light 
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pget_light_color( col )
PCOLOR	*col;
{
    col->color.rgb.red   = pstate.attr->colorLight.value.rgb.red;
    col->color.rgb.green = pstate.attr->colorLight.value.rgb.green;
    col->color.rgb.blue  = pstate.attr->colorLight.value.rgb.blue;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      Pget_line_color     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Returns the line color.
 *
 *  USAGE:
 *	Pget_line_color( *col )
 *	PCOLOR	    *col	(in)	pointer to color for line 
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pget_line_color( col )
PCOLOR	*col;
{
    col->color.rgb.red   = pstate.attr->colorLine.rgb.red;
    col->color.rgb.green = pstate.attr->colorLine.rgb.green;
    col->color.rgb.blue  = pstate.attr->colorLine.rgb.blue;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      Pget_line_type      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Returns the current line type.
 *
 *  USAGE:
 *	Pget_line_type( )
 *
 *  RETURNS:
 *	The current line type.
 */
 
/*ARGSUSED*/
int
Pget_line_type()
{
	return (pstate.attr->line_type);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      Pget_line_width     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Returns the current line width.
 *
 *  USAGE:
 *	Pget_line_width( )
 *
 *  RETURNS:
 *	The current line width.
 */
 
/*ARGSUSED*/
Fdata
Pget_line_width()
{
	return (pstate.attr->line_width);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+    Pget_marker_color     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Returns the marker color.
 *
 *  USAGE:
 *	Pget_marker_color( *col )
 *	PCOLOR	    *col	(in)	pointer to color for marker
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pget_marker_color( col )
PCOLOR	*col;
{
    col->color.rgb.red   = pstate.attr->colorMrkr.rgb.red;
    col->color.rgb.green = pstate.attr->colorMrkr.rgb.green;
    col->color.rgb.blue  = pstate.attr->colorMrkr.rgb.blue;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+     Pget_marker_type     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Returns the marker type.
 *
 *  USAGE:
 *	Pget_marker_type( )
 *
 *  RETURNS:
 *	The current marker type.
 */
 
/*ARGSUSED*/
int
Pget_marker_type( )
{
	return (pstate.attr->mrkr_type);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+     Pget_shade_mode      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Returns the current shade mode.
 *
 *  USAGE:
 *	Pget_shade_mode( )
 *
 *  RETURNS:
 *	The bitmask of the current shading mode.
 */
 
/*ARGSUSED*/
int
Pget_shade_mode()
{
    return ((unsigned long) pstate.shade_mode);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+    Pget_specular_color   +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Returns the specular color.
 *
 *  USAGE:
 *	Pget_specular_color( *col )
 *	PCOLOR	    *col	(in)	pointer to color for specular hilite
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pget_specular_color( col )
PCOLOR	*col;
{
    col->color.rgb.red   = 
	pstate.attr->poly_mtl.specular_color.value.rgb.red;
    col->color.rgb.green = 
	pstate.attr->poly_mtl.specular_color.value.rgb.green;
    col->color.rgb.blue  = 
	pstate.attr->poly_mtl.specular_color.value.rgb.blue;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+     Pget_structmatrix    +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Returns a matrix from a structure at position label.
 *
 *  USAGE:
 *	Pget_structmatrix( name, label, mat )
 *	int	    name	(in)	The structure id to apply matrix to
 *	int	    label	(in)	The label at which matrix is located
 *	Matrix	    mat		(out)	The matrix to retrieve
 *
 *  RETURNS:
 *	0 if successful, 1 if error
 */
 
/*ARGSUSED*/
int
Pget_structmatrix( name, label, mat )
int	name;
int	label;
Matrix	mat;
{
    char *pxinfo;

	pexLocalTransform pex_lxform;
	char *addr;
	unsigned short inclList, exclList;
	int i;
	unsigned long numEl, sizeRet, offset;

	DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
	XFlush(dpy);
	i = DB_find_struct_name( name );
	if( i == -1 ) return 1;
	PEXSetElementPtr( dpy, pstate.str_list[i].id, PEXBeginning, 0);
	PEXSetElementPtrAtLabel( dpy, pstate.str_list[i].id,label,0);

	inclList = PEXOCLocalTransform;
	exclList=0;
	if( PEXElementSearch( dpy, pstate.str_list[i].id, PEXCurrent, 0,
		PEXForward, 1, &inclList, 0, &exclList, &offset))
	    return 1;
	if( PEXFetchElements( dpy, pstate.str_list[i].id, PEXCurrent, 0, 
		PEXCurrent, 0, pstate.fp_format, &numEl, &sizeRet, &pxinfo))
/*
printf("Pget ret Local %d, Search offset %d\n",j,offset);
*/

	/* if( sizeRet < sizeof(pexLocalTransform)) */
	if( sizeRet < (sizeof(pexLocalTransform) >> 2) )
	    return 1;
	addr = (char *) pxinfo;

	memcpy((char *) &pex_lxform,(char *) addr, sizeof(pexLocalTransform));
	XFree( (char *) pxinfo );

	/* make sure it is a LocalTransform command */
	if( pex_lxform.head.elementType != PEXOCLocalTransform )
	    return 1;

	/* Don't forget to transpose the matrix before we return it */
	/* VERIFY!!!!*/
	PEXtk_matrix_tcopy( mat, pex_lxform.matrix);

	return 0;

}

/*
 *
 *                       +--------------------------+
 * ----------------------+     Pget_structure       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Return the currently open structure name.
 *
 *  USAGE:
 *	Pget_structure( )
 *
 *  RETURNS:
 *	structure name if one is open, -1 if not
 */
 
/*ARGSUSED*/
int
Pget_structure( )
{
	if( pstate.cur_struct == 0 )
	    return -1;
	else if( pstate.cur_struct_loc == -1 )
	    return -1;
	else
	    return (pstate.str_list[pstate.cur_struct_loc].name);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+     Pget_support_info    +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Returns support information on PEXtk.
 *
 *  USAGE:
 *	Pget_saupport_info( type, data )
 *	int	    type	(in)	The type of information to return
 *	int	    **data 	(out)	The returned data
 *
 *  CAVEATS:
 *	free should be invoked by the caller to free the allocated memory.
 *
 *  RETURNS:
 *	0 if successful, non-zero if error:
 *					     1 if memory allocation
 *					     2 if incorrect 'type'
 */
 
/*ARGSUSED*/
int
Pget_support_info( type, data )
int	type;
int	**data;
{
    int i, count;
    int *ptr;
    switch( type ){
	case PEXtk_INFO_MAX_LIGHTS:
		if((*data = (int *)malloc( sizeof(int) )) == NULL)
		    return 1;
		*((int *)*data) = psupport.max_lgts;
	    break;
	case PEXtk_INFO_MAX_VIEWS:
		if((*data = (int *)malloc( sizeof(int) )) == NULL)
		    return 1;
		*((int *)*data) = psupport.max_views;
	    break;
	case PEXtk_INFO_ZBUFFER:
		if((*data = (int *)malloc( sizeof(int) )) == NULL)
		    return 1;
		*((int *)*data) = psupport.allow_zbuffer;
	    break;
	case PEXtk_INFO_ANTIALIAS:
		if((*data = (int *)malloc( sizeof(int) )) == NULL)
		    return 1;
		*((int *)*data) = psupport.allow_antialias;
	    break;
	case PEXtk_INFO_DITHER:
		if((*data = (int *)malloc( sizeof(int) )) == NULL)
		    return 1;
		*((int *)*data) = psupport.allow_dither;
	    break;
	case PEXtk_INFO_LIGHT_TYPES:
		/*
		 * Max number of items currently supported are 4,
		 * but allocate 8+1 (size of light_type buffer)
		 */
		if((*data = (int *)malloc( sizeof(int) * 9)) == NULL)
		    return 1;
		ptr = (int *) *data;
		for(i=0,count=0;i<PEXLightWCSpot; i++)
		    if( psupport.light_type[i] ){
			switch( i+1 ){
			    case PEXLightAmbient:
				ptr[i+1] = PEXtk_LIGHT_AMBIENT;
				break;
			    case PEXLightWCVector:
				ptr[i+1] = PEXtk_LIGHT_INFINITE;
				break;
			    case PEXLightWCPoint:
				ptr[i+1] = PEXtk_LIGHT_POINT;
				break;
			    case PEXLightWCSpot:
				ptr[i+1] = PEXtk_LIGHT_SPOT;
				break;
			}
			count++;
		    }
		ptr[0] = count;
	    break;
	case PEXtk_INFO_SHADE_MODES:
		/*
		 * Shades modes are currently fixed.
		 * Max number of items currently supported are 10,
		 * so allocate 10+1 (PEXtk_SHADE_NONE is always supported).
		 */
		if((*data = (int *)malloc( sizeof(int) * 11)) == NULL)
		    return 1;
		ptr = (int *) *data;
		for(i=0,count=0;i<10; i++)
		    switch(i){
			case 0:
			    ptr[i+1] = PEXtk_SHADE_NONE;
			    count++;
			    break;
			case 1:
			    ptr[i+1] = PEXtk_SHADE_WIRE;
			    count++;
			    break;
			case 2:
			    ptr[i+1] = PEXtk_SHADE_FLAT;
			    count++;
			    break;
			case 3:
			    ptr[i+1] = PEXtk_SHADE_COLORS;
			    count++;
			    break;
			case 4:
			    ptr[i+1] = PEXtk_SHADE_NORMALS;
			    count++;
			    break;
			case 5:
			    ptr[i+1] = PEXtk_SHADE_SPECULAR;
			    count++;
			    break;
			case 6:
			    ptr[i+1] = PEXtk_SHADE_LIGHTS;
			    count++;
			    break;
			case 7:
			    ptr[i+1] = PEXtk_SHADE_HIDDENLINE;
			    count++;
			    break;
			case 8:
			    if( psupport.allow_antialias ){
			        ptr[i+1] = PEXtk_SHADE_ANTIALIAS;
				count++;
			    }
			    break;
			case 9:
			    ptr[i+1] = PEXtk_SHADE_DOTPRODUCT;
			    ptr[i+1] = 0;
			    /* not supported */
			    break;
		    }
		ptr[0] = count;
	    break;
	default:
		return 2;
    }
    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      Pget_text_color     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Returns the text color.
 *
 *  USAGE:
 *	Pget_text_color( *col )
 *	PCOLOR	    *col	(in)	pointer to color for text
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pget_text_color( col )
PCOLOR	*col;
{
    col->color.rgb.red   = pstate.attr->colorText.rgb.red;
    col->color.rgb.green = pstate.attr->colorText.rgb.green;
    col->color.rgb.blue  = pstate.attr->colorText.rgb.blue;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+       Pget_zbuffer       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Returns the current Z Buffer Mode.
 *
 *  USAGE:
 *	Pget_zbuffer( )
 *
 *  RETURNS:
 *	The current zbuffer mode.
 */
 
/*ARGSUSED*/
int
Pget_zbuffer()
{
	return (pstate.zbuffer);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      Pinsert_struct      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets edit pointer to a label for insert editing.
 *
 *  USAGE:
 *	Pinsert_struct( label )
 *	int	    label	(in)	Label at which elements are inserted
 *
 *  NOTE:
 *	Requires Pedit_struct to precede this call.
 *
 *  RETURNS:
 *	0 if successful, non-zero if error
 *			 1 - a structure is not open
 *			 2 - label not found in current structure
 */
 
/*ARGSUSED*/
int
Pinsert_struct( label )
int	label;
{
    int i, attrib;


	if(( pstate.sys_mode != PEXtk_STRUCTURE_MODE ) &&
	  (( pstate.edit_mode != PEXtk_EDITING_MODE  ) ||
	   ( pstate.edit_mode != PEXtk_REPLACE_MODE  ) ||
	   ( pstate.edit_mode != PEXtk_INSERT_MODE   )))
	    return 1;

	XFlush(dpy);

	i = DB_find_label_name( pstate.cur_str_name, label);
	if( i == -1 )
	    return 2;	/* label not found */

	pstate.cur_id	 	 = pstate.cur_struct;
	pstate.req_type	 	 = PEXOCStoreSingle;

	PEXSetEditingMode( dpy, pstate.cur_struct, PEXStructureInsert);
	PEXSetElementPtr( dpy, pstate.cur_struct, PEXBeginning, 0);
	PEXSetElementPtrAtLabel( dpy, pstate.cur_struct, label, 1);
	attrib = pstate.str_list[pstate.cur_struct_loc].cur_lbl->attr[i];
	pstate.attr->i_style_str     = attrib & 0x0f;
	pstate.attr->edgeFlag_str    = (( attrib & 0xf0 ) >> 4 );
	pstate.attr->interp_meth_str = (( attrib & 0xf00 ) >> 8 );
/*
printf(" Label %d, style %d\n",
		pstate.str_list[pstate.cur_struct_loc].cur_lbl->label[i],
		pstate.attr->i_style_str);
*/

	p_object.concat			= PEXtk_concat_struct;
	p_object.preconcat		= PEXtk_preconcat_struct;

	pstate.edit_mode 	 = PEXtk_INSERT_MODE;
	pstate.cur_label	 = label;
	pstate.cur_label_loc	 = i;

	return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+        Plabel_used       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Determines if a label is used in a structure 
 *
 *  USAGE:
 *	Plabel_used( name, label )
 *	int	    name	(in)	Structure name
 *	int	    label	(in)	Label value
 *
 *  RETURNS:
 *	0 if not in use, 1 if it is in use
 */
 
/*ARGSUSED*/
int
Plabel_used( name, label )
int	name;
int	label;
{
    int i;
	i = DB_find_label_name(name,label);
	if( i == -1 )
	    return 0;

	return 1;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+           Plight         +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the light data.
 *
 *  USAGE:
 *	Plight( n, lgt )
 *	int	    n		(in)	ID number of light, 0 turns off
 *					all lights. n starts at 1.
 *	PEXtk_LIGHT *lgt	(in)	Light structure. If lgt is NULL,
 *					the light is turned off.
 *  NOTE:
 *	    The lights are maintained by keeping a mask of the active lights
 *	named active_light_mask, and a table of the lights which are turned
 *	on, LightOnTable. When a light is to be activated, the LightOnTable
 *	is searched to see if 'n' is already in the array. If it is, no action
 *	is taken. If it is not, it is placed at the first non-zero entry.
 *	    When a light is to be removed, the LightOnTable is searched for
 *	the specified light id 'n'. If is is not found, no action is taken.
 *	If it is found, it is removed from the table, and the last item in
 *	the array is moved to that location.
 *	The bit range for active_light_mask in this routine is from bit 2 to
 *	bit 32. Note that bit 2 is reserved for the ambient light.
 *
 *	The #define for LIGHT_BUG is due to problems with turning off
 *	all the lights, and not having the server take the correct action
 *	when turning another one back on. It usually happens in single
 *	buffer mode. If this problem happens on a specific server, then
 *	put a #define LIGHT_BUG at the beginning of this file.
 *
 *	This routine has been updated to support structure mode for
 *	enabling or disabling a light.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Plight( n, lgt )
int		n;
PEXtk_LIGHT	*lgt;
{
    int i, lnum;
    pexTableIndex del_light;

    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_light( n, lgt);
	return;
    }

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
    if( n == 0 ){	/* Turn off all lights */
	active_light_mask = 0;
	PEXSetLightSourceState( dpy, pstate.cur_id, pstate.req_type, 0, 0, 
		pstate.active_lgts, LightOnTable);
	pstate.pcAttrib.light_state.count    = 0;
	PEXSetPCAttributeMask( pstate.pc_mask, PEXPCLightState);
	for(i=0;i<pstate.active_lgts;i++) {
	    LightOnTable[i] = 0;
	}
#ifdef LIGHT_BUG
	    pstate.rdrAttrib.light_table = pstate.win[pstate.cur_w].Light_ID0;
	    PEXChangeRenderer( dpy, pstate.cur_rdr, PEXRALightTable, 
		&pstate.rdrAttrib);
	    XFlush(dpy);
	    PEXtk_End_Rendering();
	    PEXtk_Begin_Rendering();
	    PEXtk_reset_pex_state(0,(long)0);
#endif
	return;
    }

	lnum=n+1;	/* Base number is 1, but that is saved for ambient */
    if( lgt == NULL ){
	if( active_light_mask & (1L << n))
	{
	    active_light_mask &= ~( 1L << n );
	    pstate.active_lgts  = PEXtk_del_light( lnum );

	    if( pstate.active_lgts  != 0 || 
	        pstate.buffer_mode == PEXtk_DOUBLE_BUFFER )
	    {
		del_light = lnum;
		PEXSetLightSourceState( dpy, pstate.cur_id, pstate.req_type, 
			pstate.active_lgts, LightOnTable, 1, &del_light);
	    } else {
		/*
		 * In single buffer mode, a situation existed on the DEC server
		 * where deactivating lights may not take effect until an
		 * End/Begin Render sequence, even if they are dynamic.
		 */
#ifdef LIGHT_BUG
		pstate.rdrAttrib.light_table = 
					pstate.win[pstate.cur_w].Light_ID0;
		PEXChangeRenderer( dpy, pstate.cur_rdr, PEXRALightTable, 
			&pstate.rdrAttrib);
	        XFlush(dpy);
		PEXtk_End_Rendering();
		PEXtk_Begin_Rendering();
		PEXtk_reset_pex_state(0,(long)0);
#else
		del_light = lnum;
		PEXSetLightSourceState( dpy, pstate.cur_id, pstate.req_type, 
			pstate.active_lgts, LightOnTable, 1, &del_light);
		PEXSetLightSourceState( dpy, pstate.cur_id, pstate.req_type, 
			0, 0, 0, 0);
#endif
	    }
	pstate.pcAttrib.light_state.count    = pstate.active_lgts;
	PEXSetPCAttributeMask( pstate.pc_mask, PEXPCLightState);

	} /* else the light is not active */
	return;
    }

#ifdef LIGHT_BUG
        XFlush(dpy);
	pstate.rdrAttrib.light_table = pstate.win[pstate.cur_w].Light_ID;
	PEXChangeRenderer( dpy, pstate.cur_rdr, PEXRALightTable, 
		&pstate.rdrAttrib);
	XFlush(dpy);
#endif

      switch( lgt->type ){

       case PEXtk_LIGHT_POINT:

	  light_table.type = PEXLightWCPoint;
   	  light_table.concentration = 1.0;
	  if( pstate.color_type == PEXColorTypeRGB ){
	      light_table.color.type = PEXColorTypeRGB;

	      light_table.color.value.rgb.red  = 
	  	  pstate.attr->colorLight.value.rgb.red;
	      light_table.color.value.rgb.green= 
	  	  pstate.attr->colorLight.value.rgb.green;
	      light_table.color.value.rgb.blue = 
	  	  pstate.attr->colorLight.value.rgb.blue;

	  } else {
	      light_table.color.type = PEXColorTypeIndexed;
	      light_table.color.value.indexed.index =
	  	  pstate.attr->colorLightInd;
	  }

	  light_table.point.x = lgt->pos.x;
	  light_table.point.y = lgt->pos.y;
	  light_table.point.z = lgt->pos.z;
	  if( lgt->atten1 == 0.0 && lgt->atten2 == 0.0 ){
		light_table.attenuation1 = ATTENUATION1;
		light_table.attenuation2 = ATTENUATION2;
	  } else {
		light_table.attenuation1 = lgt->atten1;
		light_table.attenuation2 = lgt->atten2;
	  }

	  PEXSetTableEntries( dpy, pstate.win[pstate.cur_w].Light_ID, 
		lnum, 1, PEXLUTLight,
		&light_table);

	  active_light_mask |= ( 1L << (n));     /* turn bit on  */
	  pstate.active_lgts = PEXtk_add_light(lnum);

          PEXSetLightSourceState( dpy, pstate.cur_id, pstate.req_type, 
		pstate.active_lgts, LightOnTable, 0, 0);

/*
printf(" lgt pos %f, %f, %f, lightcolor %f %f %f\n",
	lgt->pos.x,lgt->pos.y,lgt->pos.z,
	  light_table.color.value.rgb.red  ,
	  light_table.color.value.rgb.green,
	  light_table.color.value.rgb.blue );
*/

	  break;

       case PEXtk_LIGHT_INFINITE:
	  light_table.type = PEXLightWCVector;
   	  light_table.concentration = 1.0;
	  if( pstate.color_type == PEXColorTypeRGB ){
	      light_table.color.type = PEXColorTypeRGB;

	      light_table.color.value.rgb.red  = 
	  	  pstate.attr->colorLight.value.rgb.red;
	      light_table.color.value.rgb.green= 
	  	  pstate.attr->colorLight.value.rgb.green;
	      light_table.color.value.rgb.blue = 
	  	  pstate.attr->colorLight.value.rgb.blue;

	  } else {
	      light_table.color.type = PEXColorTypeIndexed;
	      light_table.color.value.indexed.index =
	  	  pstate.attr->colorLightInd;
	  }

	  light_table.direction.x = lgt->dir.x;
	  light_table.direction.y = lgt->dir.y;
	  light_table.direction.z = lgt->dir.z;

	  PEXSetTableEntries( dpy, pstate.win[pstate.cur_w].Light_ID, 
	  			lnum, 1, PEXLUTLight, 
				&light_table);

	  active_light_mask |= ( 1L << (n));     /* turn bit on  */
	  pstate.active_lgts = PEXtk_add_light(lnum);

          PEXSetLightSourceState( dpy, pstate.cur_id, pstate.req_type, 
			pstate.active_lgts, LightOnTable, 0, 0);
	  break;
       case PEXtk_LIGHT_SPOT:
	  light_table.type = PEXLightWCSpot;
	  if( lgt->conc == 0.0 )
   		light_table.concentration = 1.0;
	  else
   		light_table.concentration = lgt->conc;
	  if( lgt->angle == 0.0 )
   	      light_table.spread_angle = 90.0*3.141592/360.0;/* default val */
	  else
   	      light_table.spread_angle = lgt->angle;
	  if( pstate.color_type == PEXColorTypeRGB ){
	      light_table.color.type = PEXColorTypeRGB;

	      light_table.color.value.rgb.red  = 
	  	  pstate.attr->colorLight.value.rgb.red;
	      light_table.color.value.rgb.green= 
	  	  pstate.attr->colorLight.value.rgb.green;
	      light_table.color.value.rgb.blue = 
	  	  pstate.attr->colorLight.value.rgb.blue;

	  } else {
	      light_table.color.type = PEXColorTypeIndexed;
	      light_table.color.value.indexed.index =
	  	  pstate.attr->colorLightInd;
	  }

	  light_table.point.x = lgt->pos.x;
	  light_table.point.y = lgt->pos.y;
	  light_table.point.z = lgt->pos.z;
	  light_table.direction.x = lgt->dir.x;
	  light_table.direction.y = lgt->dir.y;
	  light_table.direction.z = lgt->dir.z;

	  if( lgt->atten1 == 0.0 && lgt->atten2 == 0.0 ){
		light_table.attenuation1 = ATTENUATION1;
		light_table.attenuation2 = ATTENUATION2;
	  } else {
		light_table.attenuation1 = lgt->atten1;
		light_table.attenuation2 = lgt->atten2;
	  }

	  PEXSetTableEntries( dpy, pstate.win[pstate.cur_w].Light_ID, 
	  			lnum, 1, PEXLUTLight, 
				&light_table);

	  active_light_mask |= ( 1L << (n));     /* turn bit on  */
	  pstate.active_lgts = PEXtk_add_light(lnum);

	  PEXSetLightSourceState( dpy, pstate.cur_id, pstate.req_type, 
		pstate.active_lgts, LightOnTable, 0, 0);
	  break;
      } /* end switch */

    pstate.pcAttrib.light_state.count    = pstate.active_lgts;
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCLightState);

    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+       Plight_ambient     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the ambient light intensity and color.
 *
 *  USAGE:
 *	Plight_ambient( amb )
 *	Fdata	    amb		(in)	ambient light factor
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Plight_ambient( amb )
Fdata	amb;
{
    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_light_ambient( amb );
	return;
    }
	DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
        XFlush(dpy);

	/* Is a change rdr request needed here ? */
	/* VERIFY !!!! */
/*
	pstate.rdrAttrib.light_table = pstate.win[pstate.cur_w].Light_ID;
	PEXChangeRenderer( dpy, pstate.cur_rdr, PEXRALightTable, 
		&pstate.rdrAttrib);
	XFlush(dpy);
*/

	light_table.type = PEXLightAmbient;
	  if( pstate.color_type == PEXColorTypeRGB ){
	      light_table.color.type = PEXColorTypeRGB;

	      light_table.color.value.rgb.red  = amb *
	  	  pstate.attr->colorLight.value.rgb.red;
	      light_table.color.value.rgb.green= amb *
	  	  pstate.attr->colorLight.value.rgb.green;
	      light_table.color.value.rgb.blue = amb *
	  	  pstate.attr->colorLight.value.rgb.blue;

	  } else {
	      light_table.color.type = PEXColorTypeIndexed;
	      light_table.color.value.indexed.index =
	  	  pstate.attr->colorLightInd;
	  }

	/* pstate.attr->poly_mtl.ambient = amb; */

	/* LightLUT number for ambient light is 1 */
	PEXSetTableEntries( dpy, pstate.win[pstate.cur_w].Light_ID, 
				1, 1, PEXLUTLight,
				&light_table);

	active_light_mask |=  1;     /* turn first bit on  */
	pstate.active_lgts = PEXtk_add_light(1);
    	PEXSetLightSourceState( dpy, pstate.cur_id, pstate.req_type, 
		pstate.active_lgts, LightOnTable, 0, 0);

	if( ! IMMEDIATE_MODE(pstate.sys_mode)) return;
	pstate.pcAttrib.light_state.count    = pstate.active_lgts;
	PEXSetPCAttributeMask( pstate.pc_mask, PEXPCLightState);
	return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+        Plight_color      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the light color.
 *
 *  USAGE:
 *	Plight_color( col )
 *	PCOLOR	    *col	(in)	light color
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Plight_color( col )
PCOLOR	*col;
{
    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_attrib_clr( PEXtk_OP_LIGHT_CLR, col );
	return;
    }
	pstate.attr->colorLight.value.rgb.red   = col->color.rgb.red;
	pstate.attr->colorLight.value.rgb.green = col->color.rgb.green;
	pstate.attr->colorLight.value.rgb.blue  = col->color.rgb.blue;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+     Plight_color_hsv     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the light color to h,s,v values.
 *
 *  USAGE:
 *	Plight_color_hsv( h, s, v )
 *	Fdata	    h		(in)	light color hue
 *	Fdata	    s		(in)	light color saturation
 *	Fdata	    v		(in)	light color value
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Plight_color_hsv( h, s, v )
Fdata	h;
Fdata	s;
Fdata	v;
{
    float r,g,b;

    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_attrib_clr_3f( PEXtk_OP_LIGHT_CLR_HSV, h, s, v );
	return;
    }
    PEXtk_convert_hsv_to_rgb( h, s, v, &r, &g, &b);
	pstate.attr->colorLight.value.rgb.red   = r;
	pstate.attr->colorLight.value.rgb.green = g;
	pstate.attr->colorLight.value.rgb.blue  = b;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+     Plight_color_ind     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the light color to r,g,b values.
 *
 *  USAGE:
 *	Plight_color_ind( index )
 *	int	    index	(in)	light color index
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Plight_color_ind( index )
int	index;
{

    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_attrib_clr_ind( PEXtk_OP_LIGHT_CLR_IND, index );
	return;
    }

	pstate.attr->colorLight.value.indexed.index  = index;
	pstate.attr->colorLightInd = index;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+     Plight_color_rgb     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the light color to r,g,b values.
 *
 *  USAGE:
 *	Plight_color_rgb( r, g, b )
 *	Fdata	    r		(in)	light color red
 *	Fdata	    g		(in)	light color green
 *	Fdata	    b		(in)	light color blue
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Plight_color_rgb( r, g, b )
Fdata	r;
Fdata	g;
Fdata	b;
{
    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_attrib_clr_3f( PEXtk_OP_LIGHT_CLR_RGB, r, g, b );
	return;
    }
	pstate.attr->colorLight.value.rgb.red   = r;
	pstate.attr->colorLight.value.rgb.green = g;
	pstate.attr->colorLight.value.rgb.blue  = b;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+        Pline_color       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the line color.
 *
 *  USAGE:
 *	Pline_color( col )
 *	PCOLOR	    *col	(in)	color of line
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pline_color( col )
PCOLOR	*col;
{
    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_attrib_clr( PEXtk_OP_LINE_CLR, col );
	return;
    }
    if( col->type == PEXtk_COLOR_HSV && pstate.color_type == PEXColorTypeRGB)
    {
	PEXtk_convert_hsv_to_rgb(  col->color.hsv.hue,
			col->color.hsv.saturation,
			col->color.hsv.value,
			&pstate.attr->colorLine.rgb.red,
			&pstate.attr->colorLine.rgb.green,
			&pstate.attr->colorLine.rgb.blue);
    } else {
	pstate.attr->colorLine.rgb.red   = col->color.rgb.red;
	pstate.attr->colorLine.rgb.green = col->color.rgb.green;
	pstate.attr->colorLine.rgb.blue  = col->color.rgb.blue;
    }
	
    PEXSetLineColor( dpy, pstate.cur_id, pstate.req_type, 
	pstate.color_type, &pstate.attr->colorLine);

    /* If in struct mode, then don't set pipeline context */
    if( ! IMMEDIATE_MODE(pstate.sys_mode)) return;

    memcpy( &pstate.pcAttrib.line_color.value, &pstate.attr->colorLine,
		sizeof(PEXColor));
    pstate.pcAttrib.line_color.type 	     = pstate.color_type;
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCLineColor);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      Pline_color_hsv     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the line color to h,s,v values.
 *
 *  USAGE:
 *	Pline_color_hsv( h, s, v )
 *	Fdata	    h		(in)	line color hue
 *	Fdata	    s		(in)	line color saturation
 *	Fdata	    v		(in)	line color value
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pline_color_hsv( h, s, v )
Fdata	h;
Fdata	s;
Fdata	v;
{
    float r,g,b;

    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_attrib_clr_3f( PEXtk_OP_LINE_CLR_HSV, h, s, v );
	return;
    }
    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
    if( pstate.color_type == PEXColorTypeRGB ){
	PEXtk_convert_hsv_to_rgb( h, s, v, &r, &g, &b);
	pstate.attr->colorLine.rgb.red   = r;
	pstate.attr->colorLine.rgb.green = g;
	pstate.attr->colorLine.rgb.blue  = b;
    } else {
	pstate.attr->colorLine.hsv.hue   = h;
	pstate.attr->colorLine.hsv.saturation = s;
	pstate.attr->colorLine.hsv.value  = v;
    }
	
    PEXSetLineColor( dpy, pstate.cur_id, pstate.req_type, 
	pstate.color_type, &pstate.attr->colorLine);

    /* If in struct mode, then don't set pipeline context */
    if( ! IMMEDIATE_MODE(pstate.sys_mode)) return;

    memcpy( &pstate.pcAttrib.line_color.value, &pstate.attr->colorLine,
		sizeof(PEXColor));
    pstate.pcAttrib.line_color.type 	     = pstate.color_type;
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCLineColor);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      Pline_color_ind     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the line color to h,s,v values.
 *
 *  USAGE:
 *	Pline_color_ind( index )
 *	int	    index	(in)	color of line
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pline_color_ind( index )
int	index;
{

    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_attrib_clr_ind( PEXtk_OP_LINE_CLR_IND, index );
	return;
    }

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
    pstate.attr->colorLineInd = index;
    pstate.attr->colorLine.indexed.index  = index;
	
    PEXSetLineColorIndex( dpy, pstate.cur_id, pstate.req_type, index);

    /* If in struct mode, then don't set pipeline context */
    if( ! IMMEDIATE_MODE(pstate.sys_mode)) return;

    pstate.pcAttrib.line_color.type 	   	   = PEXColorTypeIndexed;
    pstate.pcAttrib.line_color.value.indexed.index = index;
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCLineColor);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      Pline_color_rgb     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the line color to r,g,b values.
 *
 *  USAGE:
 *	Pline_color_rgb( r, g, b )
 *	Fdata	    r		(in)	line color red
 *	Fdata	    g		(in)	line color green
 *	Fdata	    b		(in)	line color blue
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pline_color_rgb( r, g, b )
Fdata	r;
Fdata	g;
Fdata	b;
{

    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_attrib_clr_3f( PEXtk_OP_LINE_CLR_RGB, r, g, b );
	return;
    }
    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
    pstate.attr->colorLine.rgb.red   = r;
    pstate.attr->colorLine.rgb.green = g;
    pstate.attr->colorLine.rgb.blue  = b;
	
    PEXSetLineColor( dpy, pstate.cur_id, pstate.req_type, 
	pstate.color_type, &pstate.attr->colorLine);

    /* If in struct mode, then don't set pipeline context */
    if( ! IMMEDIATE_MODE(pstate.sys_mode)) return;

    memcpy( &pstate.pcAttrib.line_color.value, &pstate.attr->colorLine,
		sizeof(PEXColor));
    pstate.pcAttrib.line_color.type 	       = pstate.color_type;
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCLineColor);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+        Pline_type        +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the line type.
 *
 *  USAGE:
 *	Pline_type( type )
 *	int	    type	(in)	The type of polyline.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pline_type( type )
int	type;
{
    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_attrib( PEXtk_OP_LINE_TYPE, type );
	return;
    }
	DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
	switch( type ){
	    case PEXtk_LINE_SOLID:
		pstate.pcAttrib.line_type = PEXLineTypeSolid;
		break;
	    case PEXtk_LINE_DASHED:
		pstate.pcAttrib.line_type = PEXLineTypeDashed;
		break;
	    case PEXtk_LINE_DOTTED:
		pstate.pcAttrib.line_type = PEXLineTypeDotted;
		break;
	    case PEXtk_LINE_DASHDOT:
		pstate.pcAttrib.line_type = PEXLineTypeDashDot;
		break;
	    default:
		pstate.pcAttrib.line_type = PEXLineTypeSolid;
		break;
	}
	pstate.attr->line_type = type;
	PEXSetLineType( dpy, pstate.cur_id, pstate.req_type, 
		pstate.pcAttrib.line_type);
	if( ! IMMEDIATE_MODE(pstate.sys_mode)) return;
	PEXSetPCAttributeMask( pstate.pc_mask, PEXPCLineType);
    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+        Pline_width       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the line width.
 *
 *  USAGE:
 *	Pline_width( width )
 *	Fdata	    width	(in)	The width of a polyline.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pline_width( width )
Fdata	width;
{
    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_attrib_f( PEXtk_OP_LINE_WIDTH, width );
	return;
    }
	DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
	pstate.attr->line_width = width;
	PEXSetLineWidth( dpy, pstate.cur_id, pstate.req_type, 
		(double) pstate.attr->line_width);
	if( ! IMMEDIATE_MODE(pstate.sys_mode)) return;
	pstate.pcAttrib.line_width		= width;
	PEXSetPCAttributeMask( pstate.pc_mask, PEXPCLineWidth);
    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+        Plist_fonts       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Returns the supported fonts.
 *
 *  USAGE:
 *	Plist_fonts( max_fonts, pattern, fn )
 *	int	    max_fonts 	(in)	max number of fonts to search for
 *	char	    *pattern   	(in)	font pattern to search for
 *	char	    **fn   	(in)	number of fonts found
 *
 *  CAVEATS:
 *	The font name array returned must be freed by caller when done
 *
 *  RETURNS:
 *	The array of font names if successfull, 
 *	NULL if a memory error occurred
 */
 
/*ARGSUSED*/
char **
Plist_fonts( max_fonts, pattern, num_found)
int  	max_fonts;
char 	*pattern;
int 	*num_found;
{
  int i, j;
  char **pex_fonts; /* font names returned by PEXlib */
  char **fn;	    /* font names to be returned to caller */
  unsigned long cnt;
  
    pex_fonts = PEXListFonts( dpy, pattern, max_fonts, &cnt);
    *num_found = cnt;

    if( pex_fonts == NULL || cnt == 0 ){
	return NULL;
    }

    if(( fn = (char **) malloc( cnt * sizeof(char *))) == NULL){
	PEXFreeFontNames( cnt, pex_fonts);
	return NULL;
    }
    for( i=0; i < cnt; i++){
	j = strlen(pex_fonts[i]);	
	if( j == 0 ){
	    *num_found = i;
	    PEXFreeFontNames( (unsigned long) i, pex_fonts);
	    return fn;
	}
	if(( fn[i] = (char *) malloc(j*sizeof(char))) == NULL){
	    *num_found = i;
	    PEXFreeFontNames( (unsigned long) i, pex_fonts);
	    return fn;
	}
	strcpy(fn[i],pex_fonts[i]);
    }

    PEXFreeFontNames( cnt, pex_fonts);
    return fn;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      Pload_attribute     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Changes the attribute attrib to location at addr valueptr
 *
 *  USAGE:
 *	Pload_attribute( attrib, *valueptr )
 *	int	    attrib	(in)	The type of attribute to load
 *	void	    *valueptr	(in)	The pointer to the address stack
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pload_attribute( attrib, valueptr)
int	attrib;
void    *valueptr;
{
    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+     Plogical_function    +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the 3D pipeline logical function/operation.
 *
 *  USAGE:
 *	Plogical_function( func )
 *	u long	    func	(in)	Logical function to use for pixels.
 *
 *  NOTE:
 *	This routine is platform specific.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Plogical_function( func )
unsigned long	func;
{
#ifdef MPEX
    unsigned long mask;
    mpxlRendererAttributes rdrNewAttribs;

    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_ext( PEXtk_OP_LOGICAL, func );
	return;
    }
        XFlush(dpy);
	PEXtk_End_Rendering();

	rdrNewAttribs.planeMask = func;

	mask = MPEXNRALogicalOp;

	MPEXChangeRenderer( dpy, pstate.cur_rdr, 
			mask, &rdrNewAttribs);
	PEXtk_Begin_Rendering();

	PEXtk_reset_pex_state(0,(long)0);
#endif
}

/*
 *
 *                       +--------------------------+
 * ----------------------+        Pmake_label       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets a label in a structure. The label may then be used as a
 *	reference position for various editing commands.
 *
 *  USAGE:
 *	Pmake_label( label )
 *	int	    label	(in)	Label value
 *
 *  NOTE:
 *	This routine may only be used inside a structure (at least until
 *	the dpy list editing is functional).
 *
 *  RETURNS:
 *	0 if successful,  non-zero if an error occurs.
 *			  1 - memory alloc error
 *			  2 - a structure is not open
 *			  3 - label exists
 *			  4 - structure location error
 */
 
/*ARGSUSED*/
int
Pmake_label( label )
int	label;
{
    int i;

    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_attrib( PEXtk_OP_MAKE_LABEL, label);
	return 0;
    }

    if( pstate.cur_struct == 0 )
	return 2;		/* error... structure not open */

	i = DB_find_label_name( pstate.cur_str_name, label);
	if( i != -1 )
	    return 3;		/* error... label exists */

	DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
       /*
	* Add this label to the current structure name label list
        */
	if(pstate.cur_struct_loc == -1 )
	    return 4;	/* structure location error */

	PEXLabel( dpy, pstate.cur_id, pstate.req_type, (long) label);

    return (DB_add_label_list( pstate.cur_struct_loc, label));

}

/*
 *
 *                       +--------------------------+
 * ----------------------+          Pmarker         +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Draws a 3D marker primitive.
 *
 *  USAGE:
 *	Pmarker( n, points )
 *	int	    n		(in)	Number of points in array
 *	Fcoord3D    *points	(in)	Array of 3D vertices
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pmarker( n, points )
int	 n;
Fcoord3D *points;
{
	DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
	if( pstate.dirty_mask & PEXtk_DIRTY_MRKR_TYPE ){
		PEXSetMarkerType( dpy, pstate.cur_id, pstate.req_type, 
			pstate.pcAttrib.marker_type);
		pstate.dirty_mask &= ~PEXtk_DIRTY_MRKR_TYPE;
	}
	PEXMarkers( dpy, pstate.cur_id, pstate.req_type, n, (PEXCoord *)points);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      Pmarker_color       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the marker color.
 *
 *  USAGE:
 *	Pmarker_color( col )
 *	PCOLOR	    *col	(in)	color of marker
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pmarker_color( col )
PCOLOR	*col;
{
    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_attrib_clr( PEXtk_OP_MRKR_CLR, col );
	return;
    }
    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
    if( col->type == PEXtk_COLOR_HSV && pstate.color_type == PEXColorTypeRGB)
    {
	PEXtk_convert_hsv_to_rgb(  col->color.hsv.hue,
			col->color.hsv.saturation,
			col->color.hsv.value,
			&pstate.attr->colorMrkr.rgb.red,
			&pstate.attr->colorMrkr.rgb.green,
			&pstate.attr->colorMrkr.rgb.blue);
    } else {
	pstate.attr->colorMrkr.rgb.red   = col->color.rgb.red;
	pstate.attr->colorMrkr.rgb.green = col->color.rgb.green;
	pstate.attr->colorMrkr.rgb.blue  = col->color.rgb.blue;
    }
	
    PEXSetMarkerColor( dpy, pstate.cur_id, pstate.req_type, 
	pstate.color_type, &pstate.attr->colorMrkr);

    /* If in struct mode, then don't set pipeline context */
    if( ! IMMEDIATE_MODE(pstate.sys_mode)) return;

    memcpy( &pstate.pcAttrib.marker_color.value, &pstate.attr->colorMrkr,
		sizeof(PEXColor));
    pstate.pcAttrib.marker_color.type    = pstate.color_type;
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCMarkerColor);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+    Pmarker_color_hsv     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the marker color to h,s,v values.
 *
 *  USAGE:
 *	Pmarker_color_hsv( h, s, v )
 *	Fdata	    h		(in)	marker color hue
 *	Fdata	    s		(in)	marker color saturation
 *	Fdata	    v		(in)	marker color value
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pmarker_color_hsv( h, s, v )
Fdata	h;
Fdata	s;
Fdata	v;
{
    float r,g,b;

    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_attrib_clr_3f( PEXtk_OP_MRKR_CLR_HSV, h, s, v );
	return;
    }
    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
    if( pstate.color_type == PEXColorTypeRGB ){
	PEXtk_convert_hsv_to_rgb( h, s, v, &r, &g, &b);
	pstate.attr->colorMrkr.rgb.red   = r;
	pstate.attr->colorMrkr.rgb.green = g;
	pstate.attr->colorMrkr.rgb.blue  = b;
    } else {
	pstate.attr->colorMrkr.hsv.hue   = h;
	pstate.attr->colorMrkr.hsv.saturation = s;
	pstate.attr->colorMrkr.hsv.value  = v;
    }
    PEXSetMarkerColor( dpy, pstate.cur_id, pstate.req_type, 
	pstate.color_type, &pstate.attr->colorMrkr);

    /* If in struct mode, then don't set pipeline context */
    if( ! IMMEDIATE_MODE(pstate.sys_mode)) return;

    memcpy( &pstate.pcAttrib.marker_color.value, &pstate.attr->colorMrkr,
		sizeof(PEXColor));
    pstate.pcAttrib.marker_color.type    = pstate.color_type;
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCMarkerColor);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+    Pmarker_color_ind     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the marker color to r,g,b values.
 *
 *  USAGE:
 *	Pmarker_color_ind( index )
 *	int	    index	(in)	marker color index
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pmarker_color_ind( index )
int	index;
{

    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_attrib_clr_ind( PEXtk_OP_MRKR_CLR_IND, index );
	return;
    }

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
    pstate.attr->colorMrkr.indexed.index      = index;
    pstate.attr->colorMrkrInd = index;
    PEXSetMarkerColorIndex( dpy, pstate.cur_id, pstate.req_type, index);

    /* If in struct mode, then don't set pipeline context */
    if( ! IMMEDIATE_MODE(pstate.sys_mode)) return;

    pstate.pcAttrib.marker_color.type 	     	     = PEXColorTypeIndexed;
    pstate.pcAttrib.marker_color.value.indexed.index = index;
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCMarkerColor);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+    Pmarker_color_rgb     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the marker color to r,g,b values.
 *
 *  USAGE:
 *	Pmarker_color_rgb( r, g, b )
 *	Fdata	    r		(in)	marker color red
 *	Fdata	    g		(in)	marker color green
 *	Fdata	    b		(in)	marker color blue
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pmarker_color_rgb( r, g, b )
Fdata	r;
Fdata	g;
Fdata	b;
{
    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_attrib_clr_3f( PEXtk_OP_MRKR_CLR_RGB, r, g, b );
	return;
    }

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);

    pstate.attr->colorMrkr.rgb.red   = r;
    pstate.attr->colorMrkr.rgb.green = g;
    pstate.attr->colorMrkr.rgb.blue  = b;
    PEXSetMarkerColor( dpy, pstate.cur_id, pstate.req_type, 
	pstate.color_type, &pstate.attr->colorMrkr);

    /* If in struct mode, then don't set pipeline context */
    if( ! IMMEDIATE_MODE(pstate.sys_mode)) return;

    memcpy( &pstate.pcAttrib.marker_color.value, &pstate.attr->colorMrkr,
		sizeof(PEXColor));
    pstate.pcAttrib.marker_color.type    = pstate.color_type;
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCMarkerColor);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      Pmarker_scale       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the marker scale.
 *
 *  USAGE:
 *	Pmarker_scale( scale )
 *	Fdata	    scale	(in)	The scale of marker.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pmarker_scale( scale )
Fdata	scale;
{
    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_attrib_f( PEXtk_OP_MRKR_SCALE, scale );
	return;
    }
	DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
	pstate.attr->mrkr_scale = scale;
	PEXSetMarkerScale( dpy, pstate.cur_id, pstate.req_type, 
		(double) pstate.attr->mrkr_scale);

	/* If in struct mode, then don't set pipeline context */
	if( ! IMMEDIATE_MODE(pstate.sys_mode)) return;
	pstate.pcAttrib.marker_scale = scale;
	PEXSetPCAttributeMask( pstate.pc_mask, PEXPCMarkerScale);
    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+       Pmarker_type       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the marker type.
 *
 *  USAGE:
 *	Pmarker_type( type )
 *	int	    type	(in)	The type of marker.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pmarker_type( type )
int	type;
{
    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_attrib( PEXtk_OP_MRKR_TYPE, type );
	return;
    }
    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
	switch( type ){
	    case PEXtk_MRKR_POINT:
		pstate.pcAttrib.marker_type = PEXMarkerDot;
		break;
	    case PEXtk_MRKR_CIRCLE:
		pstate.pcAttrib.marker_type = PEXMarkerCircle;
		break;
	    case PEXtk_MRKR_CROSS:
		pstate.pcAttrib.marker_type = PEXMarkerCross;
		break;
	    case PEXtk_MRKR_X:
		pstate.pcAttrib.marker_type = PEXMarkerX;
		break;
	    case PEXtk_MRKR_ASTERISK:
		pstate.pcAttrib.marker_type = PEXMarkerAsterisk;
		break;
#ifdef MPEX
	    case PEXtk_MRKR_BOX:
		pstate.pcAttrib.marker_type = MPEXMarkerGlyph1;
		break;
#endif
	    default:
		pstate.pcAttrib.marker_type = PEXMarkerX;
		break;
	}
	pstate.attr->mrkr_type = type;
	PEXSetMarkerType( dpy, pstate.cur_id, pstate.req_type, 
		pstate.pcAttrib.marker_type);
	pstate.dirty_mask &= ~PEXtk_DIRTY_MRKR_TYPE;

	/* If in struct mode, then don't set pipeline context */
	if( ! IMMEDIATE_MODE(pstate.sys_mode)) return;

	PEXSetPCAttributeMask( pstate.pc_mask, PEXPCMarkerType);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+       Pmatrix_mode       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the current matrix mode, modeling/viewing or projection.
 *
 *  USAGE:
 *	Pmatrix_mode( mode )
 *	int	    mode	(in)	Matrix mode
 *
 *  NOTE:
 *	This routine is for immediate mode only.
 *	This routine is not yet fully operational.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pmatrix_mode( mode )
int	mode;
{
    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_state( PEXtk_OP_MATRIX_MODE, mode);
	return;
    }

    switch( mode ){
	case PEXtk_MATRIX_MODELING:
	case PEXtk_MATRIX_VIEWING:
	case PEXtk_MATRIX_PROJ:
		pstate.mat_mode = mode;
	    break;
    }

}

/*
 *
 *                       +--------------------------+
 * ----------------------+        Pnurb_curve       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *      Draws a PHIGS+ Nurb curve primitive.
 *
 *  USAGE:
 *	Pnurb_curve( type, order, numKnots, knots, numPoints, points )
 *	int	    type	(in)	Type PEXtkRational or PEXtkNonRational
 *	int	    order	(in)	order in u-dir, degree+1
 *	int	    numKnots   (in)	Number of Knots in uKnots list
 *	float       *knots	(in)	Knot list
 *	int	    numPoints	(in)	Number of control points
 *	Fcoord	    *points	(in)	Control points
 *
 *  DESCRIPTION:
 *	This routine is intended as a direct interface to the PEX
 *	routine 'Nurb Curve'.
 *
 *  RETURNS:
 *	Nothing
 */

/*ARGSUSED*/
void
Pnurb_curve( type, order, numKnots, knots, numPoints, points)
int             type;		/* PEXtk_RATIONAL or PEXtk_NONRATIONAL */
int             order;		/* order = degree - 1 */
int             numKnots;
float           *knots;
int             numPoints;
Fcoord          *points;
{
    int pex_type;
    PEXArrayOfCoord pts;

    if( type == PEXtk_RATIONAL )
	pex_type = PEXRational;
    else
	pex_type = PEXNonRational;

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
    pts.point = (PEXCoord *) points;

	PEXNURBCurve( dpy, pstate.cur_id, pstate.req_type, 
			pex_type, order, knots, numPoints, pts,
			(double) pstate.attr->curv_umin, 
			(double) pstate.attr->curv_umax);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+       Pnurb_surface      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *      Draws a PHIGS+ Nurb surface primitive.
 *
 *  USAGE:
 *	Pnurb_surface( type, uOrder, vOrder, numUknots, uKnots, numVknots,
 *		       vknots, uPts, vPts, points, trimCurveLists, numLists)
 *	int	    type	(in)	Type PEXtkRational or PEXtkNonRational
 *	int	    uOrder	(in)	order in u-dir, degree+1
 *	int	    vOrder	(in)	order in v-dir, degree+1
 *	int	    numUknots   (in)	Number of Knots in uKnots list
 *	float       *uKnots	(in)	Knot list
 *	int	    numVknots   (in)	Number of Knots in vKnots list
 *	float       *vKnots	(in)	Knot list
 *	int	    mPts	(in)	Number of points in u-dir
 *	int	    nPts	(in)	Number of points in v-dir
 *	Fcoord	    *points	(in)	Control points
 *	PEXtkTrimLoop
 *              *trimCurveLoops (in)    List of trimming loops
 *	int	    numLoops    (in)    Number of trimming loops in list
 *
 *  DESCRIPTION:
 *	This routine is intended as a direct interface to the PEX
 *	routine 'Nurb Surface'.
 *
 *  RETURNS:
 *	Nothing
 */

/*ARGSUSED*/
void
Pnurb_surface( type, uOrder, vOrder, numUknots, uKnots, numVknots,
        vKnots, uPts, vPts, points, trimCurveLoops, numLoops)
int             type;                   /* PEXtkRational or PEXtkNonRational */
int             uOrder;                 /* order = degree - 1 */
int             vOrder;                 /* order = degree - 1 */
int             numUknots;
float           *uKnots;
int             numVknots;
float           *vKnots;
int             uPts;
int             vPts;
Fcoord          *points;
PEXtkTrimLoop   *trimCurveLoops;
int             numLoops;
{
#ifdef UPDATE
    pxlNurbSurfaceData nurb;

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);

	if( type == PEXtk_NONRATIONAL )
	    nurb.type       = PEXNonRational;
	else
	    nurb.type       = PEXRational;

	nurb.uorder     = uOrder;
	nurb.vorder     = vOrder;
#ifdef PEX4_0
	nurb.umin       = pstate.attr->surf_umin;
	nurb.umax       = pstate.attr->surf_umax;
	nurb.vmin       = pstate.attr->surf_vmin;
	nurb.vmax       = pstate.attr->surf_vmax;
#endif
#ifdef DEC_NURB_BUG
	nurb.numMPoints = vPts;
	nurb.numNPoints = uPts;
	nurb.uknots     = vKnots;
	nurb.vknots     = uKnots;
#else
	nurb.numMPoints = uPts;
	nurb.numNPoints = vPts;
	nurb.uknots     = uKnots;
	nurb.vknots     = vKnots;
#endif
	nurb.points     = (char *) points;
	nurb.trimLoop   = (pxlTrimLoopData *) trimCurveLoops;
	nurb.numTrimLoops = numLoops;

	PEXNURBSurface( dpy, pstate.cur_id, pstate.req_type, &nurb);
#endif
}

/*
 *
 *                       +--------------------------+
 * ----------------------+         Ppick_id         +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets a pick id.
 *
 *  USAGE:
 *	Ppick_id()
 *	int    id		(in)	Pick id
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ppick_id( id )
int	id;
{
    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_attrib( PEXtk_OP_PICK_ID, id );
	return;
    }
    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
    PEXSetPickID( dpy, pstate.cur_id, pstate.req_type, (unsigned long) id);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+       Pplane_mask        +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the 3D pipeline plane mask.
 *
 *  USAGE:
 *	Pplane_mask( bit_mask )
 *	u long	    bit_mask	(in)	plane bitmask
 *
 *  NOTE:
 *	This routine is platform specific.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pplane_mask( bit_mask )
unsigned long	bit_mask;
{
#ifdef MPEX
    unsigned long mask;
    mpxlRendererAttributes rdrNewAttribs;

    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_ext( PEXtk_OP_PLANE_MASK, mask );
	return;
    }

        XFlush(dpy);
	PEXtk_End_Rendering();

	rdrNewAttribs.planeMask = bit_mask;

	mask = MPEXNRAPlaneMask;

	MPEXChangeRenderer( dpy, pstate.cur_rdr, 
			mask, &rdrNewAttribs);
	PEXtk_Begin_Rendering();

	PEXtk_reset_pex_state(0,(long)0);
#endif
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      Ppop_attributes     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Pops the attribute stack.
 *
 *  USAGE:
 *	Ppop_attributes( )
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ppop_attributes( )
{
    if( pstate.sys_mode == PEXtk_RENDERING_MODE ||
        pstate.sys_mode == PEXtk_PICKING_MODE ){
	if( pstate.attr_loc > 0 )
	{
	    pstate.attr_loc--;
	    pstate.attr = &p_attrib[pstate.attr_loc];
	}
	return;
    } else
    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_stack( PEXtk_OP_POP_ATTR);
	return;
    }
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      Ppush_attributes    +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Pushes the attribute stack.
 *
 *  USAGE:
 *	Ppush_attributes( )
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ppush_attributes( )
{
    if( pstate.sys_mode == PEXtk_RENDERING_MODE ||
        pstate.sys_mode == PEXtk_PICKING_MODE ){
	if( pstate.attr_loc < PEXtk_ATTRIB_SIZE-1 )
	{
	    pstate.attr_loc++;
	    pstate.attr = &p_attrib[pstate.attr_loc];
	    memcpy( (char *) &p_attrib[pstate.attr_loc],
		(char *) &p_attrib[pstate.attr_loc-1], sizeof(PEXtk_ATTRIB));
	}
	return;
    } else
    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_stack( PEXtk_OP_PUSH_ATTR );
	return;
    }
}

/*
 *
 *                       +--------------------------+
 * ----------------------+         Ppop_name        +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Pops the name stack.
 *
 *  USAGE:
 *	Ppop_name( )
 *
 *  NOTE:
 *	This routine removes the current name from the name stack.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ppop_name( )
{
    if( pstate.sys_mode == PEXtk_RENDERING_MODE ||
        pstate.sys_mode == PEXtk_STRUCTURE_MODE ||
        pstate.sys_mode == PEXtk_PICKING_MODE ){
	if( pstate.name_loc > 0 ){
	    PEXRemoveFromNameSet( dpy, pstate.cur_id, pstate.req_type, 
		1, &(name_stk[pstate.name_loc-1]));
	    pstate.name_loc--;
	}
	return;
    } else
    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_name_stack( PEXtk_OP_POP_NAME, 0);
	return;
    }
    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+         Ppush_name       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Pushes a value on the name stack.
 *
 *  USAGE:
 *	Ppush_name( )
 *	int	    n		(in)	Name to push onto the stack
 *
 *  NOTE:
 *	This routine adds the name 'n' to the name stack.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ppush_name( n )
int	n;
{
    if( pstate.sys_mode == PEXtk_RENDERING_MODE ||
        pstate.sys_mode == PEXtk_STRUCTURE_MODE ||
        pstate.sys_mode == PEXtk_PICKING_MODE ){
	if( pstate.name_loc < psupport.max_names-1 ){
	    name_stk[pstate.name_loc] = n;
	    PEXAddToNameSet( dpy, pstate.cur_id, pstate.req_type, 
		1, &(name_stk[pstate.name_loc]));
	    pstate.name_loc++;
	}
	return;
    } else
    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_name_stack( PEXtk_OP_PUSH_NAME, n);
	return;
    }
    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+            Ppoly         +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Draws a 3D polygon primitive.
 *
 *  USAGE:
 *	Ppoly( n, points )
 *	int	    n		(in)	Number of points in array
 *	Fcoord3D    *points	(in)	Array of 3D vertices
 *
 *  ATTRIBUTES:
 *	Facet Normal, Colors per Vertex.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ppoly( n, points )
int	 	n;
Fcoord3D	*points;
{
  unsigned long fa, va;
  int i;
  char *ptr;

    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_poly( PEXtk_OP_POLY, n, points);
	return;
    }

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
    i = (pbuild.color_cnt) ? PEXSurfaceInterpColor : 0;
    PEXtk_check_poly_attrib( PEXInteriorStyleHollow, i);

    if( pbuild.polynorm_cnt == 0 && pbuild.color_cnt == 0 ){
	PEXFillArea( dpy, pstate.cur_id, pstate.req_type, 
		pstate.shape, False, n, (PEXCoord *)points);
	return;
    }

    fa = va = 0;
    Ppolygon.vertices.no_data = (PEXCoord *) points;
    Ppolygon.count = n;
    if( pbuild.polynorm_cnt ){
	fa = PEXGANormal;
	/* Ppolygon.facetData = (char *) &pbuild.bld_polynorm[0]; */
    }
    if( pbuild.color_cnt ){
	va = PEXGAColor;
	ptr = pbuild.buffer;

	/* check buffer size */
	i = 2 * n * 3 * sizeof(Fcoord);
	if( DB_check_build_buffer_size(i))
	    return; /* memory alloc error */
	for(i=0;i<n;i++){
	    memcpy( ptr, (char *) &points[i], 3*sizeof(Fcoord));
	    ptr += 3*sizeof(Fcoord);
	    memcpy( ptr, 
		(char *) &(pbuild.bld_color[i].color.rgb.red), 
		3*sizeof(Fcoord));
	    ptr += 3*sizeof(Fcoord);
	}
	Ppolygon.vertices.no_data = (PEXCoord *) pbuild.buffer;
    }

    PEXFillAreaSetWithData( dpy, pstate.cur_id, pstate.req_type, pstate.shape, 
	False, PEXContourDisjoint, fa, va, pstate.color_type, 1,
	(PEXFacetData *)&pbuild.bld_polynorm[0], &Ppolygon);

	pbuild.polynorm_cnt = 0;
	pbuild.color_cnt = 0;

}

/*
 *
 *                       +--------------------------+
 * ----------------------+         Ppoly_fill       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Draws a filled 3D polygon primitive.
 *
 *  USAGE:
 *	Ppoly_fill( n, points )
 *	int	    n		(in)	Number of points in array
 *	Fcoord3D    *points	(in)	Array of 3D vertices
 *
 *  ATTRIBUTES:
 *	Facet Normal, Colors per Vertex.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ppoly_fill( n, points )
int		n;
Fcoord3D	*points;
{
  unsigned long fa, va;
  int i;
  char *ptr;

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
    i = (pbuild.color_cnt) ? PEXSurfaceInterpColor : 0;
    PEXtk_check_poly_attrib( PEXInteriorStyleSolid, i);

    fa = va = 0;
    Ppolygon.count = n;

    if( pbuild.polynorm_cnt ){
	/* Ppolygon.facetData = (char *) &pbuild.bld_polynorm[0]; */
	fa = PEXGANormal;
    }
    if( pbuild.color_cnt == 0 ){
	Ppolygon.vertices.no_data = (PEXCoord *) points;
    } else {
	va = PEXGAColor;
	ptr = pbuild.buffer;

	/* check buffer size */
	i = 2 * n * 3 * sizeof(Fcoord);
	if( DB_check_build_buffer_size(i))
	    return; /* memory alloc error */
	for(i=0;i<n;i++){
	    memcpy( ptr, (char *) &points[i], 3*sizeof(Fcoord));
	    ptr += 3*sizeof(Fcoord);
	    memcpy( ptr, 
		(char *) &(pbuild.bld_color[i].color.rgb.red), 
		3*sizeof(Fcoord));
	    ptr += 3*sizeof(Fcoord);
	}
	Ppolygon.vertices.no_data = (PEXCoord *) pbuild.buffer;
    }
    PEXFillAreaSetWithData( dpy, pstate.cur_id, pstate.req_type, pstate.shape, 
	False, PEXContourDisjoint, fa, va, pstate.color_type, 1,
	(PEXFacetData *)&pbuild.bld_polynorm[0], &Ppolygon);

    pbuild.color_cnt    = 0;
    pbuild.polynorm_cnt = 0;
    pbuild.bld_mask 	= 0;
    pbuild.bld_mode 	= 0;
    pbuild.vert_cnt 	= 0;
    pbuild.normal_cnt 	= 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      Ppoly_fill_area     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Draws a PEX/PHIGS+ fill area with data 3D primitive.
 *
 *  USAGE:
 *	Ppoly_fill_area( polyMask, vertmask, nvert, ctype, polyData, vertData )
 *	int	    polyMask	(in)	Bitmask indicating polygon data given 
 *	int	    vertMask	(in)	Bitmask indicating vertex data given
 *	int	    nvert	(in)	Number of points in array
 *	int	    ctype	(in)	Color type data given
 *	Fcoord	    *polyData	(in)	Array containing the polygon data
 *	Fcoord	    *vertData	(in)	Array containing the vertex data
 *
 *  DESCRIPTION:
 *	This routine is intended as a direct interface to the PEX
 *	routine 'Fill Area 3D with Data'.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ppoly_fill_area( polyMask, vertMask, nvert, ctype, polyData, vertData )
long	polyMask;
long	vertMask;
int	nvert;
int	ctype;
Fcoord	*polyData;
Fcoord	*vertData;
{
    int fa=0, va=0;

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);

	/* Ppolygon.facetData = (char *) polyData; */
	Ppolygon.vertices.no_data = (PEXCoord *) vertData;
	Ppolygon.count = nvert;
	if( polyMask & GACOLOR ) fa |= PEXGAColor;
	if( vertMask & GACOLOR ) va |= PEXGAColor;
	if( polyMask & GANORMAL ) fa |= PEXGANormal;
	if( vertMask & GANORMAL ) va |= PEXGANormal;

    PEXFillAreaSetWithData( dpy, pstate.cur_id, pstate.req_type, pstate.shape, 
	False, PEXContourDisjoint, fa, va, pstate.color_type, 1,
	(PEXFacetData *)polyData, &Ppolygon);

    pbuild.color_cnt    = 0;
    pbuild.polynorm_cnt = 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+         Ppoly_index      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Draws an indexed 3D polygon primitive.
 *
 *  USAGE:
 *	Ppoly_index( polyMask, vertMask, npoly, nvert, numEdges, ctype, 
 *		     polyCounts, polyData, edges, vertData)
 *	long	    polyMask	(in)	Bitmask indicating the facet data given 
 *	long	    vertMask	(in)	Bitmask indicating the vertex data 
 *	int	    npoly	(in)	Number of polygons
 *	int	    nvert	(in)	Number of points in 'vertData' array
 *	int	    numEdges	(in)	Total number of edges in 'edges' array
 *	int	    ctype	(in)	Color type data given
 *	short	    *polyCounts	(in)	Array with no. of verts in each polygon
 *	Fcoord	    *polyData	(in)	Array containing the polygon data
 *	short	    *edges	(in)	Array containing the vertex connectivity
 *	Fcoord	    *vertData	(in)	Array containing the vertex data
 *
 *  DESCRIPTION:
 *	This routine is intended as a direct interface to the PEX
 *	routine 'Set of Fill Area Sets'.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ppoly_index( polyMask, vertMask, npoly, nvert, numEdges, ctype, polyCounts, 
	polyData, edges, vertData )
long	polyMask;
long	vertMask;
int	npoly;
int	nvert;
int	numEdges;
int	ctype;
short	*polyCounts;
Fcoord	*polyData;
unsigned short	*edges;
Fcoord	*vertData;
{
    int i,j,k,c_type;
    PEXArrayOfVertex vertList;
    PEXArrayOfFacetData facetList;

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
    if( numEdges == 0 ){
	for(i=0;i<npoly;i++)
	    numEdges += polyCounts[i];
    }
    i = numEdges;
    /* DB_check_indicies_size( i ); */
    i = npoly;
    DB_check_connectivity_size( i );

    j=0;
    for(i=0,j=0;i<npoly;i++){
	/* pbuffer.cdata[i].count = 1; */
	/* this should already be done */
	/* pbuffer.cdata[i].lists = &pbuffer.shorts_list[i]; */
	pbuffer.shorts_list[i].count = k = polyCounts[i];
	pbuffer.shorts_list[i].shorts = &edges[j];
	j += k;
    }

    vertList.no_data = (PEXCoord *)  vertData;
    facetList.normal = (PEXVector *) polyData;

    switch( ctype ){
	case PEXtk_COLOR_RGB:
		c_type = PEXColorTypeRGB;
		break;
	default:
		c_type = PEXColorTypeRGB;
		break;
    }

#ifdef OLD
	PEXSetOfFillAreaSets( pstate.shape, PEXContourDisjoint, True,
		polyMask, vertMask, 0,
		npoly, nvert, numEdges, npoly,
		(char *) polyData, 
		(char *) vertData, (char *) NULL, 
		(char *) pbuffer.indicies);
#endif
	PEXSetOfFillAreaSets( dpy, pstate.cur_id, pstate.req_type,
		pstate.shape, polyMask, vertMask, 0,
		PEXContourDisjoint, True, c_type,
		npoly, facetList, nvert, vertList, numEdges,
		(PEXSwitch *) NULL, pbuffer.cdata);

    pbuild.color_cnt    = 0;
    pbuild.polynorm_cnt = 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+         Ppoly_line       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Draws a 3D line primitive.
 *
 *  USAGE:
 *	Ppoly_line( n, points )
 *	int	    n		(in)	Number of points in array
 *	Fcoord3D    *points	(in)	Array of 3D vertices
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ppoly_line( n, points )
int		n;
Fcoord3D	*points;
{
    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
    PEXPolyline( dpy, pstate.cur_id, pstate.req_type, n, (PEXCoord *)points);
    pbuild.polynorm_cnt = 0;
    pbuild.color_cnt    = 0;
    pbuild.bld_mask 	= 0;
    pbuild.bld_mode 	= 0;
    pbuild.vert_cnt 	= 0;
    pbuild.normal_cnt 	= 0;
}
 
/*
 *
 *                       +--------------------------+
 * ----------------------+         Ppoly_shade      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Draws a shaded 3D polygon primitive.
 *
 *  USAGE:
 *	Ppoly_shade( n, points, vnarray )
 *	int	    n		(in)	Number of points in array
 *	Fcoord3D    *points	(in)	Array of 3D vertices
 *	Fcoord3D    *vnarray	(in)	Array of vertex normals
 *
 *  ATTRIBUTES:
 *	Facet Normal, Colors per vertex, Normals per vertex. 
 *
 *  NOTE:
 *	Colors per vertex must be in RGB format.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ppoly_shade( n, points, vnarray )
int		n;
Fcoord3D	*points;
Fcoord3D 	*vnarray;
{
  unsigned long fa, va;
  int i, nitems;
  char *ptr;
  float *p_ptr, *vn_ptr, *cv_ptr;;

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
    PEXtk_check_poly_attrib( PEXInteriorStyleSolid, PEXSurfaceInterpColor);

    fa = 0;
    Ppolygon.count = n;

    if( pbuild.polynorm_cnt ){
	/* Ppolygon.facetData = (char *) &pbuild.bld_polynorm[0]; */
	fa = PEXGANormal;
    }

    va = 0; nitems = 1;
    if( pbuild.normal_cnt == n ){
	vn_ptr = (float *) pbuild.bld_normal;
	va = PEXGANormal;
	nitems = 1;
    }
    if( vnarray == (Fcoord3D *) NULL ){
	Ppolygon.vertices.no_data = (PEXCoord *) points;
    } else {
	vn_ptr = (float *) vnarray;
	va = PEXGANormal;
	nitems = 1;
    }
    if( pbuild.color_cnt == n ){
	cv_ptr = (float *) pbuild.bld_color;
	va |= PEXGAColor;
	nitems++;
    }
    
    if( va != 0 ){
	ptr = pbuild.buffer;

	/* check buffer size */
	i = (nitems+1) * n * 3 * sizeof(Fcoord);
	if( DB_check_build_buffer_size(i))
	    return; /* memory alloc error */
	p_ptr = (float *) points;
	for(i=0;i<n;i++){
	    memcpy( ptr, (char *) p_ptr, 3*sizeof(Fcoord));
	    ptr += 3*sizeof(Fcoord); p_ptr += 3;
	    if( va & PEXGAColor ){
		memcpy( ptr, (char *) cv_ptr, 3*sizeof(Fcoord));
		ptr += 3*sizeof(Fcoord); cv_ptr += 3;
	    }
	    if( va & PEXGANormal ){
		memcpy( ptr, (char *) vn_ptr, 3*sizeof(Fcoord));
		ptr += 3*sizeof(Fcoord); vn_ptr += 3;
	    }
	}
	Ppolygon.vertices.no_data = (PEXCoord *) pbuild.buffer;
    }

    PEXFillAreaSetWithData( dpy, pstate.cur_id, pstate.req_type, pstate.shape, 
	False, PEXContourDisjoint, fa, va, pstate.color_type, 1,
	(PEXFacetData *)&pbuild.bld_polynorm[0], &Ppolygon);

    pbuild.polynorm_cnt = 0;
    pbuild.color_cnt    = 0;
    pbuild.bld_mask 	= 0;
    pbuild.bld_mode 	= 0;
    pbuild.vert_cnt 	= 0;
    pbuild.normal_cnt 	= 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+    Ppoly_shade_texture   +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Draws a textured and shaded 3D polygon primitive.
 *
 *  USAGE:
 *	Ppoly_shade_texture( n, points, vn, txvert )
 *	int	    n		(in)	Number of points in array
 *	Fcoord3D    *points	(in)	Array of 3D vertices
 *	Fcoord3D    *vn		(in)	Array of vertex normals
 *	Tvar        *txvert	(in)	Array of texture vertices
 *
 *  NOTE:
 *	Support for texture mapping is platform specific.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ppoly_shade_texture( n, points, vn, txvert )
int		n;
Fcoord3D	*points;
Fcoord3D	*vn;
Tvar		*txvert;
{
  unsigned long fa, va;
  int i;
  char *ptr;

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
    PEXtk_check_poly_attrib( PEXInteriorStyleSolid, PEXSurfaceInterpColor);

    fa = va = 0;
    Ppolygon.count = n;

    if( pbuild.polynorm_cnt ){
	/* Ppolygon.facetData = (char *) &pbuild.bld_polynorm[0]; */
	fa = PEXGANormal;
    }

    if( vn == (Fcoord3D *) NULL ){
	Ppolygon.vertices.no_data = (PEXCoord *) points;
    } else {
	va = PEXGANormal;
	ptr = pbuild.buffer;
	/* check buffer size */
	i = 2 * n * 3 * sizeof(Fcoord);
	if( DB_check_build_buffer_size(i))
	    return; /* memory alloc error */
	for(i=0;i<n;i++){
	    memcpy( ptr, (char *) &points[i], 3*sizeof(Fcoord));
	    ptr += 3*sizeof(Fcoord);
	    memcpy( ptr, (char *) &vn[i], 3*sizeof(Fcoord));
	    ptr += 3*sizeof(Fcoord);
	}
	Ppolygon.vertices.no_data = (PEXCoord *) pbuild.buffer;
    }

    PEXFillAreaSetWithData( dpy, pstate.cur_id, pstate.req_type, pstate.shape, 
	False, PEXContourDisjoint, fa, va, pstate.color_type, 1,
	(PEXFacetData *)&pbuild.bld_polynorm[0], &Ppolygon);

    pbuild.polynorm_cnt = 0;
    pbuild.color_cnt    = 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+       Ppoly_texture      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Draws a textured 3D polygon primitive.
 *
 *  USAGE:
 *	Ppoly_texture( n, points, txvert )
 *	int	    n		(in)	Number of points in array
 *	Fcoord3D    *points	(in)	Array of 3D vertices
 *	Tvar        *txvert	(in)	Array of texture vertices
 *
 *  NOTE:
 *	Support for texture mapping is platform specific.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ppoly_texture( n, points, txvert )
int		n;
Fcoord3D	*points;
Tvar		*txvert;
{
    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
    PEXtk_check_poly_attrib( PEXInteriorStyleSolid, 0);

    PEXFillArea( dpy, pstate.cur_id, pstate.req_type, pstate.shape, 
	False, n, (PEXCoord *) points);
    pbuild.polynorm_cnt = 0;
    pbuild.color_cnt    = 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+       Ppoly_with_data    +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Draws a shaded 3D polygon primitive.
 *
 *  USAGE:
 *	Ppoly_with_data( type, n, points, ndata, poly_data)
 *	int	    type	(in)	type of polygon primitive
 *	int	    n		(in)	Number of points in array
 *	Fcoord3D    *points	(in)	Array of 3D vertices
 *	int	    ndata	(in)	Number of items in poly_data
 *	void	    *poly_data	(in)	polygon data
 *
 *  NOTE:
 *	This routine is used for extensions to PEXtk and it uses
 *	the PEX state the same as Ppoly_fill_area and Ppoly_index
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ppoly_with_data( type, n, points, ndata, poly_data )
int	  type;
int	  n;
Fcoord3D  *points;
int	  ndata;
void	  *poly_data;
{

/*
 short facet_Count = n;
    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
    DB_check_index_list(n);
    switch( type ){
	case PEXtk_POLY_INDEXED:
	    PEXSetOfFillAreaSets( dpy, pstate.cur_id, pstate.req_type, 
		pstate.shape, 0, 0, 0,
		PEXContourDisjoint, True,
		n, NULL, n, vertData, n,
		(char *) NULL, points);
#ifdef OLD
		pstate.shape, 0, 0, 0, 
		1, n, n, &facet_Count, NULL, pbuffer.vert_index, points );
#endif
	break;
    }
*/
    pbuild.polynorm_cnt = 0;
    pbuild.color_cnt    = 0;
    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+        Prectangle        +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Draws a rectangle outline
 *
 *  USAGE:
 *	Prectangle ( x1, y1, x2, y2)
 *	Fcoord	    x1		(in)	upper left coordinate x position
 *	Fcoord	    y1		(in)	upper left coordinate y position
 *	Fcoord	    x2		(in)	lower right coordinate x position
 *	Fcoord	    y2		(in)	lower right coordinate y position
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Prectangle( x1, y1, x2, y2)
Fcoord	x1;
Fcoord	y1;
Fcoord	x2;
Fcoord	y2;
{
    static PEXCoord norm = { 0.0, 0.0, 1.0};
    float coords[5][3];

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);

	coords[0][0] = x1; coords[0][1] = y1; coords[0][2] = 0.0; /* up left */
	coords[1][0] = x1; coords[1][1] = y2; coords[1][2] = 0.0; /* btm lft */
	coords[2][0] = x2; coords[2][1] = y2; coords[2][2] = 0.0; /* btm rt */
	coords[3][0] = x2; coords[3][1] = y1; coords[3][2] = 0.0; /* up rt */
	coords[4][0] = x1; coords[4][1] = y1; coords[4][2] = 0.0; /* up left */
	
    PEXtk_check_poly_attrib( PEXInteriorStyleHollow, 0);

#ifdef OLD
	PEXPolyline( dpy, pstate.cur_id, pstate.req_type, 5, coords);
#endif

    Ppolygon.vertices.no_data = (PEXCoord *) coords;
    Ppolygon.count = 4;
    if( pbuild.polynorm_cnt ){
	PEXFillAreaSetWithData( dpy, pstate.cur_id, pstate.req_type, PEXShapeConvex,
	    False, PEXContourDisjoint, PEXGANormal, 0, pstate.color_type, 1,
	    (PEXFacetData *)&pbuild.bld_polynorm[0], &Ppolygon);
    } else {
	PEXFillAreaSetWithData( dpy, pstate.cur_id, pstate.req_type, PEXShapeConvex,
	    False, PEXContourDisjoint, PEXGANormal, 0, pstate.color_type, 1,
	    (PEXFacetData *)&norm, &Ppolygon);
    }
    pbuild.polynorm_cnt = 0;
    pbuild.color_cnt    = 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      Prectangle_fill     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Draws a filled rectangle
 *
 *  USAGE:
 *	Prectangle_fill( x1, y1, x2, y2)
 *	Fcoord	    x1		(in)	upper left coordinate x position
 *	Fcoord	    y1		(in)	upper left coordinate y position
 *	Fcoord	    x2		(in)	lower right coordinate x position
 *	Fcoord	    y2		(in)	lower right coordinate y position
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Prectangle_fill( x1, y1, x2, y2)
Fcoord	x1;
Fcoord	y1;
Fcoord	x2;
Fcoord	y2;
{
    static PEXCoord norm = { 0.0, 0.0, 1.0};
    PEXCoord coords[4];

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);

	coords[0].x = x1; coords[0].y = y1; coords[0].z = 0.0; /* up left */
	coords[1].x = x1; coords[1].y = y2; coords[1].z = 0.0; /* btm lft */
	coords[2].x = x2; coords[2].y = y2; coords[2].z = 0.0; /* btm rt  */
	coords[3].x = x2; coords[3].y = y1; coords[3].z = 0.0; /* up rt   */
	
    PEXtk_check_poly_attrib( PEXInteriorStyleSolid, 0);

#ifdef OLD
    PEXFillArea( dpy, pstate.cur_id, pstate.req_type, PEXShapeConvex, False, 
	4, coords);
#endif

    Ppolygon.vertices.no_data = (PEXCoord *) coords;
    Ppolygon.count = 4;
    if( pbuild.polynorm_cnt ){
	PEXFillAreaSetWithData( dpy, pstate.cur_id, pstate.req_type, PEXShapeConvex,
	    False, PEXContourDisjoint, PEXGANormal, 0, pstate.color_type, 1,
	    (PEXFacetData *)&pbuild.bld_polynorm[0], &Ppolygon);
    } else {
	PEXFillAreaSetWithData( dpy, pstate.cur_id, pstate.req_type, PEXShapeConvex,
	    False, PEXContourDisjoint, PEXGANormal, 0, pstate.color_type, 1,
	    (PEXFacetData *)&norm, &Ppolygon);
    }
    pbuild.polynorm_cnt = 0;
    pbuild.color_cnt    = 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+   Preflection_property   +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the specular reflection values, such as the ambient
 *	coefficient, diffuse coefficient, specular coefficient, and
 *	concentration/exponent.
 *
 *  USAGE:
 *	Preflection_property( type, prop )
 *	int	    type	(in)	Material property type
 *	Fdata	    prop	(in)	Material property value
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Preflection_property( type, prop )
int	type;
Fdata	prop;
{

    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_refl_prop( PEXtk_OP_REFL_PROP, type, prop );
	return;
    }
    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
    switch( type ){
      case PEXtk_REFL_AMBIENT:
		pstate.attr->poly_mtl.ambient = prop;
		break;
      case PEXtk_REFL_DIFFUSE:
		pstate.attr->poly_mtl.diffuse = prop;
		break;
      case PEXtk_REFL_SPECULAR:
		pstate.attr->poly_mtl.specular = prop;
		break;
      case PEXtk_REFL_EXPONENT:
		pstate.attr->poly_mtl.specular_conc = prop;
		break;
    }
	
	/* turn on in immediate mode ? */
	PEXSetReflectionAttributes( dpy, pstate.cur_id, pstate.req_type, 
		&pstate.attr->poly_mtl);

	/* If in struct mode, then don't set pipeline context */
	if( ! IMMEDIATE_MODE(pstate.sys_mode)) return;

	memcpy( &pstate.pcAttrib.reflection_attr, &pstate.attr->poly_mtl,
		sizeof(PEXReflectionAttributes));
    	PEXSetPCAttributeMask( pstate.pc_mask, PEXPCReflectionAttr);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      Preplace_struct     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets edit pointer to a label for replacment editing.
 *
 *  USAGE:
 *	Preplace_struct( label )
 *	int	    label	(in)	Label at which elements are replaced
 *
 *  NOTE:
 *	Requires Pedit_struct to precede this call.
 *
 *  RETURNS:
 *	0 if successful, non-zero if error
 *			 1 - a structure is not open
 *			 2 - label not found in current structure
 */
 
/*ARGSUSED*/
int
Preplace_struct( label )
int	label;
{
    int i, attrib;

	if(( pstate.sys_mode != PEXtk_STRUCTURE_MODE ) &&
	  (( pstate.edit_mode != PEXtk_EDITING_MODE  ) ||
	   ( pstate.edit_mode != PEXtk_REPLACE_MODE  ) ||
	   ( pstate.edit_mode != PEXtk_INSERT_MODE   )))
	    return 1;

	XFlush(dpy);

	i = DB_find_label_name( pstate.cur_str_name, label);
	if( i == -1 )
	    return 2;	/* label not found */

	pstate.cur_id	 	 = pstate.cur_struct;
	pstate.req_type	 	 = PEXOCStoreSingle;

	PEXSetEditingMode( dpy, pstate.cur_struct,PEXStructureReplace);
	PEXSetElementPtr( dpy, pstate.cur_struct, PEXBeginning, 0);
	PEXSetElementPtrAtLabel( dpy, pstate.cur_struct, label, 1);
	attrib = pstate.str_list[pstate.cur_struct_loc].cur_lbl->attr[i];
	pstate.attr->i_style_str     = attrib & 0x0f;
	pstate.attr->edgeFlag_str    = (( attrib & 0xf0 ) >> 4 );
	pstate.attr->interp_meth_str = (( attrib & 0xf00 ) >> 8 );
/*
printf(" cur_str %x, label %d, style %d\n", pstate.cur_struct,
		pstate.str_list[pstate.cur_struct_loc].cur_lbl->label[i],
		pstate.attr->i_style_str);
*/

	p_object.concat			= PEXtk_concat_struct;
	p_object.preconcat		= PEXtk_preconcat_struct;

	pstate.edit_mode 	 = PEXtk_REPLACE_MODE;
	pstate.cur_label	 = label;
	pstate.cur_label_loc	 = i;

	return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      Pset_color_mode     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the color mode to indexed or RGB.
 *
 *  USAGE:
 *	Pset_color_mode( mode )
 *	int	    mode	(in)	Mode, either PEXtk_COLOR_INDEXED or
 *					PEXtk_COLOR_RGB
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pset_color_mode( mode )
int	mode;
{
    pstate.color_mode	= mode;

    switch( mode ){
	case PEXtk_COLOR_INDEXED:
		pstate.color_type = PEXColorTypeIndexed;
		if( pstate.shade_mode & PEXtk_SHADE_HIDDENLINE ){
		    p_object.surface_color	= PEXtk_default;
		    p_object.surface_color_ind  = PEXtk_hiddenline_color_ind;
		} else {
		    p_object.surface_color	= PEXtk_default;
		    p_object.surface_color_ind  = PEXtk_surface_color_ind;
		}
	    break;
		/*
		 * Even for color type HSV, use PEX RGB color type.
		 */
	default:
		pstate.color_type = PEXColorTypeRGB;
		if( pstate.shade_mode & PEXtk_SHADE_HIDDENLINE ){
		    p_object.surface_color	= PEXtk_hiddenline_color;
		    p_object.surface_color_ind  = PEXtk_default;
		} else {
		    p_object.surface_color	= PEXtk_surface_color_ind;
		    p_object.surface_color_ind  = PEXtk_default;
		}
	    break;
    }
}

/*
 *
 *                       +--------------------------+
 * ----------------------+     Pset_color_table     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the color table values.
 *
 *  USAGE:
 *	Pset_color_table( index, col )
 *	int	    index	(in)	Index of slot to set
 *	PCOLOR	    *col	(in)	Pointer to color structure
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pset_color_table( index, col )
int	index;
PCOLOR	*col;
{
	/* check color table size */
	/* TBD... */

	DB_insert_color_values( index, col);
	PEXtk_SetColorTable( pstate.win[pstate.cur_w].Rdr_ID, index, col);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+   Pset_color_table_hsv   +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the color table h,s,v values.
 *
 *  USAGE:
 *	Pset_color_table_hsv( index, h, s, v )
 *	int	    index	(in)	Index of slot to set
 *	Fdata	    h		(in)	hue   (0.0 - 1.0)
 *	Fdata	    s		(in)	saturation (0.0 - 1.0)
 *	Fdata	    v		(in)	value  (0.0 - 1.0)
 *
 *  NOTE:
 *	No check is made on the color mode in effect.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pset_color_table_hsv( index, h, s, v )
int	index;
Fdata	h;
Fdata	s;
Fdata	v;
{
    PCOLOR col;

	col.type 	 	 = PEXtk_COLOR_HSV;
	col.color.hsv.hue	 = h;
	col.color.hsv.saturation = s;
	col.color.hsv.value	 = v;

	DB_insert_color_values( index, &col);
	PEXtk_SetColorTable( pstate.win[pstate.cur_w].Rdr_ID, index, &col);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+   Pset_color_table_rgb   +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the color table r,g,b values.
 *
 *  USAGE:
 *	Pset_color_table_rgb( index, r, g, b )
 *	int	    index	(in)	Index of slot to set
 *	Fdata	    r		(in)	Red   color value (0.0 - 1.0)
 *	Fdata	    g		(in)	Green color value (0.0 - 1.0)
 *	Fdata	    b		(in)	Blue  color value (0.0 - 1.0)
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pset_color_table_rgb( index, r, g, b )
int	index;
Fdata	r;
Fdata	g;
Fdata	b;
{
    PCOLOR col;

	col.type	    = PEXtk_COLOR_RGB;
	col.color.rgb.red   = r;
	col.color.rgb.green = g;
	col.color.rgb.blue  = b;

	DB_insert_color_values( index, &col);
	PEXtk_SetColorTable( pstate.win[pstate.cur_w].Rdr_ID, index, &col);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+    Pset_display_buffer   +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the display buffer id.
 *
 *  USAGE:
 *	Pset_display_buffer( buf )
 *	unsigned long	    buf		(in)	buffer id
 *
 *  DESCRIPTION:
 *	This routine is used for double buffering by setting
 *	the display/update buffer to be used.
 *
 *  NOTE:
 *	This routine, along with Pset_update_buffer, should be used
 *	together. Pbuffer_mode should NOT be called.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pset_display_buffer( buf )
unsigned long	buf;
{
    static int first_time = 1;

	XmbufDisplayBuffers( dpy, 1, &buf, 0, 0);
        pstate.win[pstate.cur_w].back_w_id = buf;
	pstate.cur_disp_buf = buf;
    if( first_time ){
	if( pstate.cur_disp_buf != pstate.cur_draw )
	    pstate.buffer_mode = PEXtk_DOUBLE_BUFFER;
	first_time = 0;
    }

#ifdef OLD
 /* TBD .... */
  /* The definition for this module still needs to be defined.
   * Should this be more of a low level interface, or
   *   should PEXtk maintain what the 'buf' is?
   * Also, should this routine only be activated with the
   *   multibuffering extension?
   */
  if( pstate.cur_draw != buf ){
    /*
     * Either this is a new buffer id, or first time called 
     *   Look for 'buf' in array of window id's 
     *   If buf is not there, add it, and change pstate.cur_w 
     */
    /*
     * get the index into the window table pstate.win[index] for
     * the buffer 'buf' if it exists 
     */
      pstate.cur_w = PEXtk_find_window_id( 1, buf );

	/*
	 * Get renderer dynamics infomation
	 */
	PEXtk_Get_Renderer_Dynamics( pstate.cur_w, 
			pstate.win[pstate.cur_w].Rdr_ID );
	pstate.cur_draw = pstate.win[pstate.cur_w].back_w_id; 
    }
#endif
}

/*
 *
 *                       +--------------------------+
 * ----------------------+       Pset_drawable      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	This routine will set the current window/drawable.
 *
 *  USAGE:
 *	Pset_drawable( buf )
 *	Window	    buf		(in)	buffer id
 *
 *  DESCRIPTION:
 *	This module sets the current drawable. 
 *	This routine is called to initially set the destination drawable,
 *	to change the destination drawable to another drawable, or when
 *	the current drawable has been resized.
 *
 *	A list is maintained for all drawables used, and their associated 
 *	back buffer id's. The default back buffer id is NULL (i.e., single 
 *	buffer). Each window id has associated with it:
 *
 *		Item			variable
 *		--------------		-------------
 *		Renderer id		Rdr_ID
 *		Back buffer id		back_w_id
 *		Light id		Light_ID
 *		Light id (null)		Light_ID0 (LIGHT_BUG)
 *		Text id			Text_ID
 *
 *  NOTE:
 *	When the current drawable is changed, the current window,
 *	pstate.cur_w, and the current renderer pstate.cur_rdr are
 *	changed to the current values. These are used simply for
 *	convenience.
 *
 *	If buf is equal to the the current drawable, then an
 *	XGetGeometry call is made to get the current size, since a
 *	resize could have occurred.
 *
 *  RETURNS:
 *	0 if successful, non-zero if an error occurred
 */
 
/*ARGSUSED*/
int
Pset_drawable( buf )
Window	buf;
{
    int 	i, j;
    Window	root;
    unsigned int utemp, win_width, win_height, depth;

    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_set_drawable( PEXtk_OP_SET_DRAWABLE, (long) buf );
	return 0;
    }

    if( pstate.win[pstate.cur_w].w_id != buf ){
    /*
     * Either this is a new window id, or first time called 
     *   Look for 'buf' in array of window id's 
     *   If buf is not there, add it, and change pstate.cur_w 
     */

    /*
     * get the index into the window table pstate.win[index] for
     * the buffer 'buf' if it exists 
     */
      XGetGeometry( dpy, buf, &root,
		&j, &j, &win_width, &win_height, &utemp, &depth);
	win_depth = depth;

	/*
	 * See if this 'buf' has already been specified as a target
	 * drawable before. If it has, then all pertinent information
	 * has already been obtained about it. If it hasn't, then gather
	 * some information about this drawable.
	 */
	pstate.cur_w = PEXtk_find_window_id( 0, buf);/* find/create index */

	pstate.win[pstate.cur_w].w_id = buf; /* not done in find_id */

	/*
	 * If in double buffer mode, make sure the back buffer exists.
	 */
	PEXtk_Check_Back_Buffer( pstate.cur_w);

	/*
	 * Save information about this drawable in the window info structure
	 */
	pstate.win[pstate.cur_w].width  = win_width;
	pstate.win[pstate.cur_w].height = win_height;
	pstate.win[pstate.cur_w].depth  = depth;

	/*
	 * If this is the first drawable specified, then we may need to
	 * create a color approx table if it is an 8-bit drawable.
	 */
	if( pstate.num_w_id <= 1 ){
	    if(win_depth == 8 && pstate.lut_approx == 0 ){
		/* set up a 6,6,6 colormap */
		i = 15;
		if( px_data.cmap == 0 )
		    Pset_color_equation( NULL, &i, 6, 6, 6, 6*6, 6, 1, 2.3);
	    }
	}

	pstate.cur_rdr  = pstate.win[pstate.cur_w].Rdr_ID;
#ifdef OLD
	Light_ID 	= pstate.win[pstate.cur_w].Light_ID;
	Light_ID0	= pstate.win[pstate.cur_w].Light_ID0;
#endif
	pstate.cur_id   = pstate.cur_rdr; /* set up immed mode */
	pstate.req_type = PEXOCRender; /* set up immed mode */

	if( pstate.buffer_mode == PEXtk_DOUBLE_BUFFER ){
	    pstate.cur_draw = pstate.win[pstate.cur_w].back_w_id; 
	} else {
	    pstate.cur_draw = pstate.win[pstate.cur_w].w_id; 
	}

	PEXtk_set_viewport(); /* set viewport to default settings */
	PEXtk_reset_pex_state( 1, (long) (PEXRAViewport|PEXRALightTable));

    } else { /* w_id == buf, get current geometry... could have had a resize */

	    XGetGeometry( dpy, pstate.win[pstate.cur_w].w_id, &root,
		&j, &j, &win_width, &win_height, &utemp, &depth);

	    pstate.win[pstate.cur_w].width  = win_width;
	    pstate.win[pstate.cur_w].height = win_height;
    }

    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+       Pset_function      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the support function for PEXtk.
 *
 *  USAGE:
 *	Pset_function( int (*func)() )
 *	int	    type	(in)	type of function to be set
 *	int	    (*func)()	(in)	pointer to function
 *
 *  NOTE:
 *	These routines are necessary for intrnal use to PEXtk. If
 *	an application wishes to to have PEXtk use its own routines
 *	for these items, then it must pass PEXtk the function to use.
 *
 *  RETURNS:
 *	0 if successful, non-zero if error
 */
 
/*ARGSUSED*/
int
Pset_function( type, func )
int	type;
int	(*func)();
{
    switch( type ){
	case PEXtk_FUNC_CLEAR_BUFFER:
		p_object.clear_buffer	 = func;
		p_object.push_clear_buffer=func;
	    break;
	case PEXtk_FUNC_CREATE_BUFFER:
		p_object.create_buffer	 = func;
	    break;
	case PEXtk_FUNC_DELETE_BUFFER:
		p_object.delete_buffer	 = func;
	    break;
	case PEXtk_FUNC_PIXEL_RGB:
		p_object.pixel_color_rgb = func;
	    break;
	case PEXtk_FUNC_PIXEL_INDEX:
		p_object.pixel_color_ind = func;
	    break;
	case PEXtk_FUNC_SWAP_BUFFER:
		p_object.swap_buffer	 = func;
	    break;
	case PEXtk_FUNC_CLEAR_ZBUFFER:
		p_object.clear_zbuffer	 = func;
	    break;
	case PEXtk_FUNC_DITHER:
		p_object.dither		 = func;
	    break;
	case PEXtk_FUNC_TRANSPARENCY:
		p_object.transparency	 = func;
	    break;
	case PEXtk_FUNC_ANTIALIAS:
		p_object.antialias	 = func;
	    break;
	default:
		return 1;
    }
    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+    Pset_update_buffer    +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the update buffer id. This is the buffer that
 *	primitives are rendered to.
 *
 *  USAGE:
 *	Pset_update_buffer( buf )
 *	unsigned long	    buf		(in)	buffer id
 *
 *  NOTE:
 *	This routine, along with Pset_display_buffer, should be used
 *	together. Pbuffer_mode should NOT be called.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pset_update_buffer( buf )
unsigned long	buf;
{
    static int first_time = 1;

    pstate.cur_draw = buf;
    if( first_time ){
	if( pstate.cur_disp_buf != pstate.cur_draw )
	    pstate.buffer_mode = PEXtk_DOUBLE_BUFFER;
	first_time = 0;
    }
}

/*
 *
 *                       +--------------------------+
 * ----------------------+       Pshade_mode        +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the shade mode.
 *
 *  USAGE:
 *	Pshade_mode( mode )
 *	int	    mode	(in)	Shade mode
 *
 *  NOTE:
 *	This routine is for Immediate Mode Only.
 *
 *  DESCRIPTION:
 *	This routine controls the Reflection Model, the Interior Style,
 *	and the Surface Interp Method. For hiddenline mode, it also
 *	controls setting the Edge Flags, Edge Color and Back Culling.
 *
 *	The 'mode' controls the following:
 *
 *	      mode      Reflection   InterpMethod    Interior Style
 *	      FLAT	  Diffuse	 None		NA
 *	      WIRE          NA           NA  		Hollow
 *	      !WIRE         NA           NA  		Solid
 *	      LIGHTS      Diffuse	 NA   		NA
 *	      COLORS      NA     	 Color		NA
 *	      SPECULAR    Specular	 NA    		NA
 *
 *  CAVEATS:
 *	Lights have to be activated to enable SPECULAR.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pshade_mode( mode )
int	mode;
{
  static old_backcull = PEXtk_CULL_BACKFACE;

    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_state( PEXtk_OP_SHADE_MODE, mode );
	return;
    }
    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);

    if( mode & PEXtk_SHADE_FLAT ){
	   /* pstate.attr->refl_model  = PEXReflectionDiffuse; */
	   pstate.attr->interp_meth = PEXSurfaceInterpNone;
	   pstate.attr->line_interp = PEXPolylineInterpNone;
	   PEXSetPCAttributeMask( pstate.pc_mask, PEXPCPolylineInterp);
    	   PEXSetPCAttributeMask( pstate.pc_mask, PEXPCSurfaceInterp);
    	   PEXSetPCAttributeMask( pstate.pc_mask, PEXPCReflectionModel);
    }

    if( mode & PEXtk_SHADE_WIRE ){
	   /* pstate.attr->refl_model  = PEXReflectionDiffuse; */
	   /* pstate.attr->interp_meth = PEXSurfaceInterpNone; */
	   pstate.attr->i_style     = PEXInteriorStyleHollow;
    	   /* PEXSetPCAttributeMask( pstate.pc_mask, PEXPCSurfaceInterp); */
    	   /* PEXSetPCAttributeMask( pstate.pc_mask, PEXPCReflectionModel); */
    	   PEXSetPCAttributeMask( pstate.pc_mask, PEXPCInteriorStyle);
    } else {
	   pstate.attr->i_style     = PEXInteriorStyleSolid;
    	   PEXSetPCAttributeMask( pstate.pc_mask, PEXPCInteriorStyle);
    }

    if( psupport.allow_antialias ){ /* check if mode is supported */
	if( mode & PEXtk_SHADE_ANTIALIAS ){
	    (*p_object.antialias)( 1 );
	} else {
	    if( pstate.shade_mode & PEXtk_SHADE_ANTIALIAS )
		(*p_object.antialias)( 1 ); /* turn off only if it was on */
	}
    } else
	mode &= ~PEXtk_SHADE_ANTIALIAS; /* turn mode off */

    if( mode & PEXtk_SHADE_HIDDENLINE ){
	/*
	 * Set fill color to bg, and edge color to previous fill color
	 */
	if( mode & PEXtk_SHADE_WIRE ){	/* wireframe not allowed */
	    mode &= ~PEXtk_SHADE_WIRE;
	   pstate.attr->i_style     = PEXInteriorStyleSolid;
    	   PEXSetPCAttributeMask( pstate.pc_mask, PEXPCInteriorStyle);
	}
	if( pstate.sys_mode == PEXtk_STRUCTURE_MODE )
	    pstate.attr->edgeFlag_str = PEXOn;
	else
	    pstate.attr->edgeFlag = PEXOn;
	old_backcull = pstate.attr->backcull;
	pstate.attr->backcull = PEXtk_CULL_BACKFACE;
	/* setup surface color routines for edges */
	if( pstate.color_type == PEXColorTypeRGB ){
	    p_object.surface_color	= PEXtk_hiddenline_color;
	    p_object.surface_color_ind  = PEXtk_default;
	} else {
	    p_object.surface_color	= PEXtk_default;
	    p_object.surface_color_ind  = PEXtk_hiddenline_color_ind;
	}
    } else {
	if( pstate.shade_mode & PEXtk_SHADE_HIDDENLINE ) /* toggling mode ? */
	    pstate.attr->backcull = old_backcull;
	if( pstate.sys_mode == PEXtk_STRUCTURE_MODE )
	    pstate.attr->edgeFlag_str = PEXOff;
	else
	    pstate.attr->edgeFlag = PEXOff;
	/* setup surface color routines */
	if( pstate.color_type == PEXColorTypeRGB ){
	    p_object.surface_color	= PEXtk_surface_color;
	    p_object.surface_color_ind  = PEXtk_default;
	} else {
	    p_object.surface_color	= PEXtk_default;
	    p_object.surface_color_ind  = PEXtk_surface_color_ind;
	}
    }

    if( mode & PEXtk_SHADE_LIGHTS ){
	pstate.attr->refl_model = PEXReflectionDiffuse;
	PEXSetPCAttributeMask( pstate.pc_mask, PEXPCReflectionModel);
#ifdef LIGHT_BUG
        pstate.rdrAttrib.light_table = Light_ID;
        PEXChangeRenderer( dpy, pstate.cur_rdr, PEXRALightTable, 
		&pstate.rdrAttrib);
#endif
	/* UPDATE - Is this OC needed here */
        PEXSetLightSourceState( dpy, pstate.cur_id, pstate.req_type, 
		pstate.active_lgts, LightOnTable, 0, 0);
	pstate.pcAttrib.light_state.count    = pstate.active_lgts;
	PEXSetPCAttributeMask( pstate.pc_mask, PEXPCLightState);
    } else {
	pstate.attr->refl_model = PEXReflectionNone;
	PEXSetPCAttributeMask( pstate.pc_mask, PEXPCReflectionModel);
    }

    if( mode & PEXtk_SHADE_COLORS ){
	pstate.attr->line_interp = PEXPolylineInterpColor;
	pstate.attr->interp_meth = PEXSurfaceInterpColor;
	PEXSetPCAttributeMask( pstate.pc_mask, PEXPCSurfaceInterp);
	PEXSetPCAttributeMask( pstate.pc_mask, PEXPCPolylineInterp);
    }

    /*
     * Reflection Model for specular highlights has a higher priority
     * that that for 'lights' only.
     */
    if( mode & PEXtk_SHADE_SPECULAR ){
	pstate.attr->refl_model = PEXReflectionSpecular;
	PEXSetPCAttributeMask( pstate.pc_mask, PEXPCReflectionModel);
    }

    if( mode & PEXtk_SHADE_NORMALS ){
	pstate.attr->line_interp = PEXPolylineInterpColor;
	pstate.attr->interp_meth = PEXSurfaceInterpNormal;
	PEXSetPCAttributeMask( pstate.pc_mask, PEXPCSurfaceInterp);
	PEXSetPCAttributeMask( pstate.pc_mask, PEXPCPolylineInterp);
    }

    pstate.pcAttrib.surface_interp     = pstate.attr->interp_meth;
    pstate.pcAttrib.reflection_model   = pstate.attr->refl_model;
    pstate.pcAttrib.interior_style     = pstate.attr->i_style;
    pstate.pcAttrib.polyline_interp    = pstate.attr->line_interp;

	/* VERIFY */
    if( pstate.shade_mode != mode && IMMEDIATE_MODE(pstate.sys_mode))
	PEXtk_reset_pex_state(1,0);

    pstate.shade_mode = mode;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+       Pshape_flag        +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the polygon shape flag.
 *
 *  USAGE:
 *	Pshape_flag( flag )
 *	int	    flag	(in)	Shape flag: 
 *					    PEXtk_SHAPE_CONVEX
 *					    PEXtk_SHAPE_COMPLEX
 *					    PEXtk__SHAPE_NONCONVEX
 *					    PEXtk_SHAPE_UNKNOWN
 *
 *  NOTE:
 *	The shape flag is not an attribute and will not be push/popped.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pshape_flag( flag )
int	flag;
{
    switch( flag ){
	case PEXtk_SHAPE_CONVEX:
	case PEXtk_SHAPE_NONCONVEX:
	case PEXtk_SHAPE_COMPLEX:
	case PEXtk_SHAPE_UNKNOWN:
		pstate.shape = flag;
	    break;
	default:
	    break;
    }
}

/*
 *
 *                       +--------------------------+
 * ----------------------+     Pspecular_color      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the specular highlight color for surfaces.
 *
 *  USAGE:
 *	Pspecular_color( col )
 *	PCOLOR	    *col	(in)	Specular Highlight color 
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pspecular_color( col )
PCOLOR	*col;
{

    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_attrib_clr( PEXtk_OP_COLOR_SPEC, col );
	return;
    }
	pstate.attr->poly_mtl.specular_color.value.rgb.red  = 
			col->color.rgb.red;
	pstate.attr->poly_mtl.specular_color.value.rgb.green= 
			col->color.rgb.green;
	pstate.attr->poly_mtl.specular_color.value.rgb.blue = 
			col->color.rgb.blue;

	PEXSetReflectionAttributes( dpy, pstate.cur_id, pstate.req_type, 
		&(pstate.attr->poly_mtl));

	/* If in struct mode, then don't set pipeline context */
	if( ! IMMEDIATE_MODE(pstate.sys_mode)) return;

	memcpy( &pstate.pcAttrib.reflection_attr, &pstate.attr->poly_mtl,
		sizeof(PEXReflectionAttributes));
    	PEXSetPCAttributeMask( pstate.pc_mask, PEXPCReflectionAttr);

}

/*
 *
 *                       +--------------------------+
 * ----------------------+    Pspecular_color_hsv   +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the specular highlight color hsv for surfaces.
 *
 *  USAGE:
 *	Pspecular_color_hsv( h, s, v )
 *	Fdata	    h		(in)	specular color red
 *	Fdata	    s		(in)	specular color green
 *	Fdata	    v		(in)	specular color blue
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pspecular_color_hsv( h, s, v )
Fdata	h;
Fdata	s;
Fdata	v;
{
    float r,g,b;

    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_attrib_clr_3f( PEXtk_OP_COLOR_SPEC_HSV, h, s, v );
	return;
    }
    if( pstate.color_type == PEXColorTypeRGB ){
	PEXtk_convert_hsv_to_rgb( h, s, v, &r, &g, &b);
	pstate.attr->poly_mtl.specular_color.value.rgb.red  = r;
	pstate.attr->poly_mtl.specular_color.value.rgb.green= g;
	pstate.attr->poly_mtl.specular_color.value.rgb.blue = b;
    } else {
	pstate.attr->poly_mtl.specular_color.value.hsv.hue  = h;
	pstate.attr->poly_mtl.specular_color.value.hsv.saturation=
					s;
	pstate.attr->poly_mtl.specular_color.value.hsv.value = v;
    }

	PEXSetReflectionAttributes( dpy, pstate.cur_id, pstate.req_type, 
		&(pstate.attr->poly_mtl));

	/* If in struct mode, then don't set pipeline context */
	if( ! IMMEDIATE_MODE(pstate.sys_mode)) return;

	memcpy( &pstate.pcAttrib.reflection_attr, &pstate.attr->poly_mtl,
		sizeof(PEXReflectionAttributes));
    	PEXSetPCAttributeMask( pstate.pc_mask, PEXPCReflectionAttr);

}

/*
 *
 *                       +--------------------------+
 * ----------------------+   Pspecular_color_ind    +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the specular highlight color index for surfaces.
 *
 *  USAGE:
 *	Pspecular_color_ind( index )
 *	int	    index	(in)	Specular Highlight color index
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pspecular_color_ind( index )
int	index;
{

    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_attrib_clr_ind( PEXtk_OP_COLOR_SPEC_IND, index );
	return;
    }
    pstate.attr->poly_mtl.specular_color.value.indexed.index = index;

    PEXSetReflectionAttributes( dpy, pstate.cur_id, pstate.req_type, 
	&(pstate.attr->poly_mtl));

    /* If in struct mode, then don't set pipeline context */
    if( ! IMMEDIATE_MODE(pstate.sys_mode)) return;

    memcpy( &pstate.pcAttrib.reflection_attr, &pstate.attr->poly_mtl,
	sizeof(PEXReflectionAttributes));
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCReflectionAttr);

}

/*
 *
 *                       +--------------------------+
 * ----------------------+    Pspecular_color_rgb   +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the specular highlight color rgb for surfaces.
 *
 *  USAGE:
 *	Pspecular_color_rgb( r, g, b )
 *	Fdata	    r		(in)	specular color red
 *	Fdata	    g		(in)	specular color green
 *	Fdata	    b		(in)	specular color blue
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pspecular_color_rgb( r, g, b )
Fdata	r;
Fdata	g;
Fdata	b;
{

    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_attrib_clr_3f( PEXtk_OP_COLOR_SPEC_RGB, r, g, b );
	return;
    }
	pstate.attr->poly_mtl.specular_color.value.rgb.red  = r;
	pstate.attr->poly_mtl.specular_color.value.rgb.green= g;
	pstate.attr->poly_mtl.specular_color.value.rgb.blue = b;

	PEXSetReflectionAttributes( dpy, pstate.cur_id, pstate.req_type, 
		&(pstate.attr->poly_mtl));

	/* If in struct mode, then don't set pipeline context */
	if( pstate.sys_mode == PEXtk_STRUCTURE_MODE ) return;

	memcpy( &pstate.pcAttrib.reflection_attr, &pstate.attr->poly_mtl,
		sizeof(PEXReflectionAttributes));
    	PEXSetPCAttributeMask( pstate.pc_mask, PEXPCReflectionAttr);

}

/*
 *
 *                       +--------------------------+
 * ----------------------+      Pstruct_used        +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Determines if a specified structure name is in use
 *
 *  USAGE:
 *	Pstruct_used( name )
 *	int	    name	(in)	Name of structure to check
 *
 *  RETURNS:
 *	0 if not in use, 1 if it is in use
 */
 
/*ARGSUSED*/
int
Pstruct_used( name )
int	name;
{
    int i;
	XFlush(dpy);

	i = DB_find_struct_name(name);
	if( i == -1 )
	    return 0;
	else
	    return 1;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+  Psurface_approx_method  +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the surface approximation method.
 *
 *  USAGE:
 *	Psurface_approx_method( meth, u_tol, v_tol )
 *	int	    meth	(in)	Surface approximation method to use
 *	Fdata	    u_tol	(in)	Surface tolerance in u-dir
 *	Fdata	    v_tol	(in)	Surface tolerance in v-dir
 *
 *  RETURNS:
 *      Nothing
 */
                                                           
/*ARGSUSED*/
void
Psurface_approx_method( meth, u_tol, v_tol )
int     meth;
Fdata   u_tol;
Fdata   v_tol;
{
    switch( meth ){
	case PEXtk_APPROX_IMP_DEP:
		pstate.pcAttrib.surface_approx.method = PEXApproxImpDep;
	    break;
	case PEXtk_APPROX_CONSTANT_KNOTS:
		pstate.pcAttrib.surface_approx.method =PEXApproxConstantBetweenKnots;
	    break;
	case PEXtk_APPROX_WC_CHORD_LEN:
		pstate.pcAttrib.surface_approx.method = PEXApproxWCChordalSize;
	    break;
	case PEXtk_APPROX_NC_CHORD_LEN:
		pstate.pcAttrib.surface_approx.method = PEXApproxNPCChordalSize;
	    break;
	case PEXtk_APPROX_DC_CHORD_LEN:
		pstate.pcAttrib.surface_approx.method = PEXApproxDCChordalSize;
	    break;
	case PEXtk_APPROX_WC_CHORD_DEV:
		pstate.pcAttrib.surface_approx.method = PEXSurfaceApproxWCPlanarDev;
	    break;
	case PEXtk_APPROX_NC_CHORD_DEV:
		pstate.pcAttrib.surface_approx.method = PEXSurfaceApproxNPCPlanarDev;
	    break;
	case PEXtk_APPROX_DC_CHORD_DEV:
		pstate.pcAttrib.surface_approx.method = PEXSurfaceApproxDCPlanarDev;
	    break;
	case PEXtk_APPROX_WC_RELATIVE:
		pstate.pcAttrib.surface_approx.method = PEXApproxWCRelative;
	    break;
	case PEXtk_APPROX_NC_RELATIVE:
		pstate.pcAttrib.surface_approx.method = PEXApproxNPCRelative;
	    break;
	case PEXtk_APPROX_DC_RELATIVE:
		pstate.pcAttrib.surface_approx.method = PEXApproxDCRelative;
	    break;
	default:
	    break;
    }
    pstate.attr->surf_meth = meth;
    pstate.attr->surf_utol = u_tol;
    pstate.attr->surf_vtol = v_tol;
    PEXSetSurfaceApprox( dpy, pstate.cur_id, pstate.req_type, 
	pstate.pcAttrib.surface_approx.method,
	(double) pstate.attr->surf_utol, (double) pstate.attr->surf_vtol);

    /* If in struct mode, then don't set pipeline context */
    if( ! IMMEDIATE_MODE(pstate.sys_mode)) return;
    pstate.pcAttrib.surface_approx.u_tolerance = u_tol;
    pstate.pcAttrib.surface_approx.v_tolerance = v_tol;
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCSurfaceApprox);
    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+   Psurface_parm_range    +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the surface parametric range.
 *
 *  USAGE:
 *	Psurface_parm_range( umin, umax, vmin, vmax )
 *	Fdata	    umin	(in)	Surface min value in u-dir
 *	Fdata	    umax	(in)	Surface max value in u-dir
 *	Fdata	    vmin	(in)	Surface min value in v-dir
 *	Fdata	    vmax	(in)	Surface max value in v-dir
 *
 *  RETURNS:
 *	Nothing
 */

/*ARGSUSED*/
void
Psurface_parm_range( umin, umax, vmin, vmax )
Fdata   umin;
Fdata   umax;
Fdata   vmin;
Fdata   vmax;
{
	pstate.attr->surf_umin = umin;
	pstate.attr->surf_umax = umax;
	pstate.attr->surf_vmin = vmin;
	pstate.attr->surf_vmax = vmax;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+       Pswap_buffers       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Swaps the buffers in double buffer mode.
 *
 *  USAGE:
 *	Pswap_buffers( )
 *
 *  NOTE:
 *	This routine cannot be called when in Pick mode.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pswap_buffers()
{

    if( pstate.buffer_mode != PEXtk_DOUBLE_BUFFER )
	return;

	XFlush(dpy);
	PEXtk_End_Rendering();

	(*p_object.swap_buffer)();
	XFlush(dpy);
	PEXtk_Begin_Rendering();
    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+          Ptext           +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Writes a text string at position x,y,z.
 *
 *  USAGE:
 *	Ptext( x, y, z, string )
 *	Fcoord	    x		(in)	Text x position
 *	Fcoord	    y		(in)	Text y position
 *	Fcoord	    z		(in)	Text z position
 *	char	    *string	(in)	Text string
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ptext( x, y, z, string )
Fcoord	x;
Fcoord	y;
Fcoord	z;
char	*string;
{

	PEXCoord origin;
	PEXVector vec1,vec2;  

	if( pstate.sys_mode == PEXtk_DLS_MODE ){
	    DLS_text( PEXtk_OP_TEXT, x, y, z, string );
	    return;
	}
	DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
	origin.x = x; origin.y = y;
	origin.z = z;

	vec1.x = vec2.y = 1.0; vec1.y = vec1.z = vec2.x = vec2.z = 0.;

        PEXText( dpy, pstate.cur_id, pstate.req_type, &origin, &vec1, &vec2,
		(int) strlen(string), string);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      Ptext_annotate      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Writes an annotation text string at position x,y.
 *
 *  USAGE:
 *	Ptext_annotate( x, y, z, string )
 *	Fcoord	    x		(in)	Text x position
 *	Fcoord	    y		(in)	Text y position
 *	Fcoord	    z		(in)	Text z position
 *	char	    *string	(in)	Text string
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ptext_annotate( x, y, z, string )
Fcoord	x;
Fcoord	y;
Fcoord	z;
char	*string;
{
	PEXCoord origin,offset;

	if( pstate.sys_mode == PEXtk_DLS_MODE ){
	    DLS_text( PEXtk_OP_TEXT_ANNO, x, y, z, string );
	    return;
	}

	DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_MTX);
	origin.x = x; origin.y = y;
	origin.z = z;
	offset.x = 0.0; offset.y = 0.0; offset.z = 0.0;

	PEXAnnotationText( dpy, pstate.cur_id, pstate.req_type, &origin, 
		&offset, (int) strlen(string), string);
    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+    Ptext_annotate_size   +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the annotation text size (or scaling factor).
 *
 *  USAGE:
 *	Ptext_annotate_size( xsize, ysize )
 *	Fdata	    xsize  	(in)	Text string horizontal size
 *	Fdata	    ysize  	(in)	Text string vertical size
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ptext_annotate_size( xsize, ysize )
Fdata	xsize;
Fdata	ysize;
{
	if( pstate.sys_mode == PEXtk_DLS_MODE ){
	    DLS_attrib_text( PEXtk_OP_TEXTA_SZ, xsize, ysize );
	    return;
	}
	pstate.attr->texta_xsize = xsize;
	pstate.attr->texta_ysize = ysize;
	PEXSetATextHeight( dpy,pstate.cur_rdr, PEXOCRender, xsize);

	/* If in struct mode, then don't set pipeline context */
	if( ! IMMEDIATE_MODE(pstate.sys_mode)) return;
	pstate.pcAttrib.atext_height 	= xsize;
	PEXSetPCAttributeMask( pstate.pc_mask, PEXPCATextHeight);
    return;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+       Ptext_color        +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the text color.
 *
 *  USAGE:
 *	Ptext_color( col )
 *	PCOLOR	    *col	(in)	color of poly
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ptext_color( col )
PCOLOR	*col;
{

    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_attrib_clr( PEXtk_OP_TEXT_CLR, col );
	return;
    }
    if( col->type == PEXtk_COLOR_HSV && pstate.color_type == PEXColorTypeRGB)
    {
	PEXtk_convert_hsv_to_rgb(  col->color.hsv.hue,
			col->color.hsv.saturation,
			col->color.hsv.value,
			&pstate.attr->colorText.rgb.red,
			&pstate.attr->colorText.rgb.green,
			&pstate.attr->colorText.rgb.blue);
    } else {
	pstate.attr->colorText.rgb.red   = col->color.rgb.red;
	pstate.attr->colorText.rgb.green = col->color.rgb.green;
	pstate.attr->colorText.rgb.blue  = col->color.rgb.blue;
    }
	
    PEXSetTextColor( dpy, pstate.cur_id, pstate.req_type, pstate.color_type, 
	&pstate.attr->colorText);

    /* If in struct mode, then don't set pipeline context */
    if( ! IMMEDIATE_MODE(pstate.sys_mode)) return;

    memcpy( &pstate.pcAttrib.text_color.value, &pstate.attr->colorText,
		sizeof(PEXColor));
    pstate.pcAttrib.text_color.type 	       = pstate.color_type;
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCTextColor);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+     Ptext_color_hsv      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the text color to h,s,v values.
 *
 *  USAGE:
 *	Ptext_color_hsv( h, s, v )
 *	Fdata	    h		(in)	text color hue
 *	Fdata	    s		(in)	text color saturation
 *	Fdata	    v		(in)	text color value
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ptext_color_hsv( h, s, v )
Fdata	h;
Fdata	s;
Fdata	v;
{
    float r,g,b;
    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_attrib_clr_3f( PEXtk_OP_TEXT_CLR_HSV, h, s, v );
	return;
    }
    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_COLOR);
    if( pstate.color_type == PEXColorTypeRGB ){
	PEXtk_convert_hsv_to_rgb( h, s, v, &r, &g, &b);
	pstate.attr->colorText.rgb.red   = r;
	pstate.attr->colorText.rgb.green = g;
	pstate.attr->colorText.rgb.blue  = b;
    } else {
	pstate.attr->colorText.hsv.hue   = h;
	pstate.attr->colorText.hsv.saturation = s;
	pstate.attr->colorText.hsv.value  = v;
    }
    PEXSetTextColor( dpy, pstate.cur_id, pstate.req_type, pstate.color_type, 
	&pstate.attr->colorText);

    /* If in struct mode, then don't set pipeline context */
    if( ! IMMEDIATE_MODE(pstate.sys_mode)) return;

    memcpy( &pstate.pcAttrib.text_color.value, &pstate.attr->colorText,
		sizeof(PEXColor));
    pstate.pcAttrib.text_color.type 	       = pstate.color_type;
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCTextColor);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+     Ptext_color_ind      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the text color to r,g,b values.
 *
 *  USAGE:
 *	Ptext_color_ind( index )
 *	int	    index	(in)	text color index
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ptext_color_ind( index )
int	index;
{

    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_attrib_clr_ind( PEXtk_OP_TEXT_CLR_IND, index );
	return;
    }

    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_COLOR);

    pstate.attr->colorText.indexed.index   = index;
    pstate.attr->colorTextInd 		   = index;
    PEXSetTextColorIndex( dpy, pstate.cur_id, pstate.req_type, index);

    /* If in struct mode, then don't set pipeline context */
    if( ! IMMEDIATE_MODE(pstate.sys_mode)) return;

    pstate.pcAttrib.text_color.type 	   	   = PEXColorTypeIndexed;
    pstate.pcAttrib.text_color.value.indexed.index = index;
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCTextColor);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+     Ptext_color_rgb      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the text color to r,g,b values.
 *
 *  USAGE:
 *	Ptext_color_rgb( r, g, b )
 *	Fdata	    r		(in)	text color red
 *	Fdata	    g		(in)	text color green
 *	Fdata	    b		(in)	text color blue
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ptext_color_rgb( r, g, b )
Fdata	r;
Fdata	g;
Fdata	b;
{
    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_attrib_clr_3f( PEXtk_OP_TEXT_CLR_RGB, r, g, b );
	return;
    }
    DB_CHECK_DIRTY_MASK(PEXtk_DIRTY_COLOR);
	pstate.attr->colorText.rgb.red   = r;
	pstate.attr->colorText.rgb.green = g;
	pstate.attr->colorText.rgb.blue  = b;
    PEXSetTextColor( dpy, pstate.cur_id, pstate.req_type, pstate.color_type, 
	&pstate.attr->colorText);

    /* If in struct mode, then don't set pipeline context */
    if( ! IMMEDIATE_MODE(pstate.sys_mode)) return;

    memcpy( &pstate.pcAttrib.text_color.value, &pstate.attr->colorText,
		sizeof(PEXColor));
    pstate.pcAttrib.text_color.type 	       = pstate.color_type;
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCTextColor);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      Ptext_extents       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Returns information about the specified font.
 *
 *  USAGE:
 *	Ptext_extents( expan, spacing, ht, str, up_right, low_left, next_pt )
 *	Fdata	    expan   	(in)	text expansion
 *	Fdata	    spacing   	(in)	text spacing
 *	Fdata	    ht   	(in)	text height
 *	char	    *str  	(in)	character string
 *	Fcoord2D    *up_right  	(in)	upper right position
 *	Fcoord2D    *low_left  	(in)	lower left position
 *	Fcoord2D    *concat_pt 	(in)	concat position
 *
 *  CAVEATS:
 *	The text extents returned must be freed by caller when done
 *
 *  RETURNS:
 *	True if successful, False if an error occurred
 */
 
/*ARGSUSED*/
int
Ptext_extents( expan, spacing, height, str, up_right, low_left, concat_pt )
Fdata 		expan;
Fdata 		spacing;
Fdata 		height;
char		*str;
Fcoord2D 	*up_right;
Fcoord2D 	*low_left;
Fcoord2D 	*concat_pt;
{
    int			f_path;
    char 		**pex_fonts;	/* font names returned by PEXlib */
    char 		**fn;		/* font names to be returned     */
    PEXTextExtent	*extents;
    PEXStringData 	phont_str;
    unsigned int 	font;
  
    phont_str.length = strlen(str);
    phont_str.ch     = str;

    f_path = pstate.pcAttrib.text_path;
    font = (unsigned int) pstate.pcAttrib.text_font;

    if( font == 0 ) return 0;
    extents = PEXQueryTextExtents( dpy, pstate.cur_rdr, font, f_path, 
				     (double) expan, (double) spacing, 
				     (double) height, PEXHAlignLeft,
				     PEXVAlignTop, 1, &phont_str);

    if( extents == NULL ){
	return 0;
    }

    up_right->x  = extents[0].upper_right.x;
    up_right->y  = extents[0].upper_right.y;
    low_left->x  = extents[0].lower_left.x;
    low_left->y  = extents[0].lower_left.y;
    concat_pt->x = extents[0].concat_point.x;
    concat_pt->y = extents[0].concat_point.y;

    XFree( (char *) extents);
    return 1;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+         Ptext_font       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the text font
 *
 *  USAGE:
 *	Ptext_font( font )
 *	char	    *font  	(in)	Text string font
 *
 *  CAVEATS:
 *	The string passed as 'font' must be one returned by Plist_fonts
 *
 *  DESCRIPTION:
 *	This routine will load the font specified, and then set the table
 *	entry for this font.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ptext_font( font )
char	*font;
{
    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_attrib( PEXtk_OP_TEXT_FNT, font );
	return;
    }

    pstate.attr->font = PEXLoadFont( dpy, font);
    pstate.attr->text_entry.count = 1;
    pstate.attr->text_entry.fonts = &pstate.attr->font;
    PEXSetTableEntries( dpy, pstate.win[pstate.cur_w].Text_ID, 1, 1,
		PEXLUTTextFont, (PEXPointer) &pstate.attr->text_entry);

    if( ! pstate.attr->font ) return;
    /* If in struct mode, then don't set pipeline context */
    if( ! IMMEDIATE_MODE(pstate.sys_mode)) return;

    pstate.pcAttrib.text_font 	   = 1;
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCTextFont);

}

/*
 *
 *                       +--------------------------+
 * ----------------------+    Ptext_orientation     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the text vector orientation
 *
 *  USAGE:
 *	Ptext_orientation( vec )
 *	Fcoord	    x		(in)	Text string up x dir vector orientation
 *	Fcoord	    y		(in)	Text string up y dir vector orientation
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ptext_orientation( x, y )
Fcoord	x;
Fcoord	y;
{
    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_attrib_text( PEXtk_OP_TEXT_OR, x, y );
	return;
    }
	pstate.attr->text_vec.x = x;
	pstate.attr->text_vec.y = y;
        PEXSetCharUpVector( dpy, pstate.cur_id, pstate.req_type, 
		&pstate.attr->text_vec);

	/* If in struct mode, then don't set pipeline context */
	if( ! IMMEDIATE_MODE(pstate.sys_mode)) return;

	pstate.pcAttrib.char_up_vector.x  = x;
	pstate.pcAttrib.char_up_vector.y  = y;
	pstate.pcAttrib.atext_up_vector.x  = x;
	pstate.pcAttrib.atext_up_vector.y  = y;
	PEXSetPCAttributeMask( pstate.pc_mask, PEXPCCharUpVector);
	PEXSetPCAttributeMask( pstate.pc_mask, PEXPCATextUpVector);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+         Ptext_path       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the text path
 *
 *  USAGE:
 *	Ptext_path( path )
 *	int	    path  	(in)	Text string path
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ptext_path( path )
int	path;
{
    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_attrib( PEXtk_OP_TEXT_PATH, path );
	return;
    }
    switch( path ){
	case PEXtk_TEXT_PATH_UP:
	    pstate.attr->path   = path;
	    pstate.pcAttrib.text_path = PEXPathUp;
	  break;
	case PEXtk_TEXT_PATH_DOWN:
	    pstate.attr->path   = path;
	    pstate.pcAttrib.text_path = PEXPathDown;
	  break;
	case PEXtk_TEXT_PATH_LEFT:
	    pstate.attr->path   = path;
	    pstate.pcAttrib.text_path = PEXPathLeft;
	  break;
	case PEXtk_TEXT_PATH_RIGHT:
	    pstate.attr->path   = path;
	    pstate.pcAttrib.text_path = PEXPathRight;
	  break;
	default:
	    pstate.attr->path   = PEXtk_TEXT_PATH_RIGHT;
	    pstate.pcAttrib.text_path = PEXPathRight;
	  break;
    }

	PEXSetTextPath( dpy, pstate.cur_id, pstate.req_type,
		pstate.pcAttrib.text_path);

	/* If in struct mode, then don't set pipeline context */
	if( ! IMMEDIATE_MODE(pstate.sys_mode)) return;

	pstate.pcAttrib.atext_path  = pstate.pcAttrib.text_path;
	PEXSetPCAttributeMask( pstate.pc_mask, PEXPCTextPath);
	PEXSetPCAttributeMask( pstate.pc_mask, PEXPCATextPath);

}

/*
 *
 *                       +--------------------------+
 * ----------------------+         Ptext_size       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the text size (or scaling factor).
 *
 *  USAGE:
 *	Ptext_size( xsize, ysize )
 *	Fdata	    xsize  	(in)	Text string horizontal size
 *	Fdata	    ysize  	(in)	Text string vertical size
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ptext_size( xsize, ysize )
Fdata	xsize;
Fdata	ysize;
{
    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_attrib_text( PEXtk_OP_TEXT_SZ, xsize, ysize );
	return;
    }
	pstate.attr->text_xsize = xsize;
	pstate.attr->text_ysize = ysize;
        PEXSetCharHeight( dpy, pstate.cur_id, pstate.req_type, 
		pstate.attr->text_xsize);

	/* If in struct mode, then don't set pipeline context */
	if( ! IMMEDIATE_MODE(pstate.sys_mode)) return;

	pstate.pcAttrib.char_height  = xsize;
	PEXSetPCAttributeMask( pstate.pc_mask, PEXPCCharHeight);

}

/*
 *
 *                       +--------------------------+
 * ----------------------+       Ptexture_info      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the texture mapping information.
 *
 *  USAGE:
 *	Ptexture_info( tb, tc, tn )
 *	long	    tb		(in)	The texture back id
 *	long	    tc		(in)	The id for the colormap map
 *	int	    tn		(in)	The name for the texture map
 *
 *  NOTE:
 *	Support for texture mapping is platform specific.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ptexture_info( tb, tc, tn )
long	tb;
long	tc;
int	tn;
{
}

/*
 *
 *                       +--------------------------+
 * ----------------------+       Ptexture_name      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Sets the current texture map.
 *
 *  USAGE:
 *	Ptexture_name( tn )
 *	int	    tn		(in)	The name of the texture map to use
 *
 *  NOTE:
 *	Support for texture mapping is platform specific.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ptexture_name( tn )
int	tn;
{
}
 
/*
 *
 *                       +--------------------------+
 * ----------------------+        Ptransparency     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Turns transparency on or off.
 *
 *  USAGE:
 *	Ptransparency( mode, value )
 *	int	    mode	(in)	Mode for transparency, 1=on, 0=off
 *	Fdata	    value	(in)	MPEXTransparency value, 0.0 = opaque,
 *						1.0 = transparent
 *  NOTE:
 *	Support for transparency is platform specific.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Ptransparency( mode, value )
int	mode;
Fdata	value;
{
    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_transparency( mode, value);
	return;
    }
    (*p_object.transparency)( mode, value );
}

/*
 *
 *                       +--------------------------+
 * ----------------------+         Pzbuffer         +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Turns the zbuffer on/off.
 *
 *  USAGE:
 *	Pzbuffer( mode )
 *	int	    mode	(in)	Z Buffer mode. Available modes:
 *				  PEXtk_ZBUFFER_OFF	  = Not activated
 *				  PEXtk_ZBUFFER_ON	  = activated
 *				  PEXtk_ZBUFFER_RESET	  = reset z values
 *
 *  NOTE:
 *	On some platforms, when Antialias is in use, the Z Buffer may 
 *	be deactivated.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
void
Pzbuffer( mode )
int	mode;
{
    if( pstate.sys_mode == PEXtk_DLS_MODE ){
	DLS_state( PEXtk_OP_Z_BUFFER, mode );
	return;
    }
    switch( mode ){
	case PEXtk_ZBUFFER_ON:
		if( psupport.allow_zbuffer )
		    pstate.zbuffer = mode;
		else
		    return;
	    break;
	case PEXtk_ZBUFFER_OFF:
		pstate.zbuffer = mode;
	    break;
	case PEXtk_ZBUFFER_RESET:
		if(pstate.zbuffer == PEXtk_ZBUFFER_ON )
			(*p_object.clear_zbuffer)();
		return;
	default:
		pstate.zbuffer = PEXtk_ZBUFFER_OFF;
		return;
    }
    PEXtk_reset_pex_state(1,PEXRAHLHSRMode);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+  PEXtk_Begin_Rendering   +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Begins a rendering session.
 *
 *  USAGE:
 *	PEXtk_Begin_Rendering( )
 *
 *  DESCRIPTION:
 *	This routine is called to begin (or continue) immediate mode
 *	rendering. The Pipeline Context is checked to see if something
 *	was update. The current mode is checked to make sure PEXtk is
 *	in PEXtk_RENDERING_MODE.
 *
 *  RETURNS:
 *	0 if successful, non-zero if incorrect mode
 */
 
/*ARGSUSED*/
int
PEXtk_Begin_Rendering( )
{
    PEXtk_check_pc();
    if( pstate.sys_mode != PEXtk_RENDERING_MODE ) return 1;/* not immed mode */
    PEXBeginRendering( dpy, pstate.cur_draw, pstate.cur_rdr);
    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+    PEXtk_End_Rendering   +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Ends a rendering session, depending on mode.
 *
 *  USAGE:
 *	PEXtk_End_Rendering( )
 *
 *  RETURNS:
 *	0 if successful, non-zero if incorrect mode
 */
 
/*ARGSUSED*/
int
PEXtk_End_Rendering( )
{
    if( pstate.sys_mode == PEXtk_RENDERING_MODE ){ /* immediate mode */
	PEXEndRendering( dpy, pstate.cur_rdr, True);
	return 0;
    } else return 1;
}

unsigned long
Pget_renderer_ID()
{
	return ((unsigned long) pstate.win[pstate.cur_w].Rdr_ID);
}

unsigned long
Pget_view_ID()
{
	return ((unsigned long) pstate.win[pstate.cur_w].View_ID);
}

/*
 *
 *                       +--------------------------+
 * ----------------------+  PEXtk_hiddenline_color  +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Object module for surface_color. Sets the hiddenline color
 *	(i.e., SurfaceColor is set to background/Pclear color, 
 *	polygon edge color is set to Pcolor.
 *
 *  USAGE:
 *	PEXtk_hiddenline_color( clr_poly )
 *	PEXColor     *clr_poly	(in)	Pointer to Polygon color
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_hiddenline_color( clr_poly )
PEXColor  *clr_poly;
{
    PEXSetSurfaceColor( dpy, pstate.cur_id, pstate.req_type, 
	pstate.color_type, &pstate.attr->colorBg);
    PEXSetSurfaceEdgeColor( dpy, pstate.cur_id, pstate.req_type, 
	pstate.color_type, clr_poly);

    /* If not in immediate mode, then don't set pipeline context */
    if( ! IMMEDIATE_MODE(pstate.sys_mode)) return 0;

    memcpy( &pstate.pcAttrib.surface_color.value, &pstate.attr->colorBg,
		sizeof(PEXColor));
    pstate.pcAttrib.surface_color.type 	       = pstate.color_type;
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCSurfaceColor);
    memcpy( &pstate.pcAttrib.surface_edge_color.value, clr_poly,
		sizeof(PEXColor));
    pstate.pcAttrib.surface_edge_color.type    = pstate.color_type;
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCSurfaceEdgeColor);
    return 0;
}

/*
 *
 *                       +----------------------------+
 * ----------------------+ PEXtk_hiddenline_color_ind +------------------------
 *                       +----------------------------+
 *
 *  PURPOSE:
 *	Object module for surface_color_ind. Sets the hiddenline color
 *	(i.e., SurfaceColor is set to background/Pclear_ind color, 
 *	polygon edge color is set to Pcolor_ind.
 *
 *  USAGE:
 *	PEXtk_hiddenline_color( clr_poly )
 *	int	     clr_poly	(in)	Integer represening Polygon color ind
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_hiddenline_color_ind( clr_poly )
int clr_poly;
{
    PEXSetSurfaceColorIndex( dpy, pstate.cur_id, pstate.req_type, 
	pstate.attr->colorBgInd);

    PEXSetSurfaceEdgeColorIndex( dpy, pstate.cur_id, pstate.req_type,clr_poly);

    /* If not in immediate mode, then don't set pipeline context */
    if( ! IMMEDIATE_MODE(pstate.sys_mode)) return 0;

    pstate.pcAttrib.surface_color.type 	   	   = PEXColorTypeIndexed;
    pstate.pcAttrib.surface_color.value.indexed.index=pstate.attr->colorBgInd;
    pstate.pcAttrib.surface_edge_color.type 	   = PEXColorTypeIndexed;
    pstate.pcAttrib.surface_color.value.indexed.index = clr_poly;
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCSurfaceColor);
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCSurfaceEdgeColor);

    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+    PEXtk_surface_color   +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Object module for surface_color. Sets the polygon color
 *	(i.e., SurfaceColor is set to Pcolor color, 
 *	polygon edges are not typically drawn.
 *
 *  USAGE:
 *	PEXtk_surface_color( clr_poly )
 *	PEXColor     *clr_poly	(in)	Pointer to Polygon color
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_surface_color( clr_poly )
PEXColor  *clr_poly;
{
    PEXSetSurfaceColor( dpy, pstate.cur_id, pstate.req_type, 
	pstate.color_type, clr_poly);

    /* If not in immediate mode, then don't set pipeline context */
    if( IMMEDIATE_MODE(pstate.sys_mode)){
	memcpy( &pstate.pcAttrib.surface_color.value, clr_poly,
	    sizeof(PEXColor));
	pstate.pcAttrib.surface_color.type     = pstate.color_type;
	PEXSetPCAttributeMask( pstate.pc_mask, PEXPCSurfaceColor);
    }

    if( pstate.sys_mode == PEXtk_STRUCTURE_MODE ){
	if( pstate.attr->edgeFlag_str == PEXOn )
	    PEXSetSurfaceEdgeColor( dpy, pstate.cur_id, pstate.req_type, 
		pstate.color_type, clr_poly);
    } else if( pstate.attr->edgeFlag == PEXOn ){
	PEXSetSurfaceEdgeColor( dpy, pstate.cur_id, pstate.req_type, 
		pstate.color_type, clr_poly);
	memcpy( &pstate.pcAttrib.surface_edge_color.value, clr_poly,
	    sizeof(PEXColor));
	pstate.pcAttrib.surface_edge_color.type = pstate.color_type;
	PEXSetPCAttributeMask( pstate.pc_mask, PEXPCSurfaceEdgeColor);
    }
    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+  PEXtk_surface_color_ind +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Object module for surface_color with index. Sets the polygon color
 *	(i.e., SurfaceColor is set to Pcolor_ind color, 
 *	polygon edges are not typically drawn.
 *
 *  USAGE:
 *	PEXtk_surface_color_ind( clr_poly )
 *	int	     clr_poly	(in)	Integer representing Polygon color ind
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_surface_color_ind( clr_poly )
int clr_poly;
{
    PEXSetSurfaceColorIndex( dpy, pstate.cur_id, pstate.req_type, clr_poly);

    /* If not in immediate mode, then don't set pipeline context */
    if( IMMEDIATE_MODE(pstate.sys_mode)) {
	pstate.pcAttrib.surface_color.type 	  = PEXColorTypeIndexed;
	pstate.pcAttrib.surface_color.value.indexed.index = clr_poly;
	PEXSetPCAttributeMask( pstate.pc_mask, PEXPCSurfaceColor);
    }

    if( pstate.sys_mode == PEXtk_STRUCTURE_MODE ){
	if( pstate.attr->edgeFlag_str == PEXOn )
	    PEXSetSurfaceEdgeColorIndex( dpy, pstate.cur_id,pstate.req_type, 
		clr_poly);
    } else if( pstate.attr->edgeFlag == PEXOn ){
	PEXSetSurfaceEdgeColorIndex( dpy, pstate.cur_id, pstate.req_type, 
		clr_poly);
	pstate.pcAttrib.surface_edge_color.type   = PEXColorTypeIndexed;
	pstate.pcAttrib.surface_color.value.indexed.index = clr_poly;
	PEXSetPCAttributeMask( pstate.pc_mask, PEXPCSurfaceEdgeColor);
    }
    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+ PEXtk_check_poly_attrib  +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Checks the current interior style against i_style, and the
 *	current surface interpolation method against interp_meth.
 *
 *  USAGE:
 *	PEXtk_check_poly_attrib( i_style, interp_meth )
 *	int	     i_style	 (in)	required interior style
 *	int	     interp_meth (in)	required interpolation method
 *
 *  NOTE:
 *	The method used in this routine to check for the interpolation
 *	method REQUIRES that 0 is not a valid value for interp_meth.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_check_poly_attrib( i_style, interp_meth )
int	i_style;
int	interp_meth;
{
    if( pstate.sys_mode == PEXtk_STRUCTURE_MODE ){
        if( pstate.attr->i_style_str != i_style ){
	    PEXSetInteriorStyle( dpy, pstate.cur_id, pstate.req_type, i_style);
	    pstate.attr->i_style_str = i_style;
	}
	if( interp_meth )
	{
	    if( pstate.attr->interp_meth_str != interp_meth )
	    {
		pstate.attr->interp_meth_str = interp_meth;
		PEXSetSurfaceInterpMethod( dpy, pstate.cur_id, pstate.req_type,
			interp_meth);
	    }
	}
    } else {	/* otherwise immediate mode */
	if( pstate.attr->i_style != i_style )
	{
	    PEXSetInteriorStyle( dpy, pstate.cur_id, pstate.req_type, i_style);
	    pstate.attr->i_style = i_style;
    	    pstate.pcAttrib.interior_style     = i_style;
    	    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCInteriorStyle);
	}
	if( interp_meth )
	{
	    if( pstate.attr->interp_meth != interp_meth )
	    {
		pstate.attr->interp_meth = interp_meth;
		PEXSetSurfaceInterpMethod( dpy, pstate.cur_id, pstate.req_type,
			interp_meth);
    		pstate.pcAttrib.surface_interp     = interp_meth;
    		PEXSetPCAttributeMask( pstate.pc_mask, PEXPCSurfaceInterp);
	    }
	}
    }
    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      PEXtk_antialias     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Object module for anti-alias support. Turns on/off anti-alias.
 *
 *  USAGE:
 *	PEXtk_antialias( mode )
 *	int	    mode	(in)	Mode for antialiasing,0=off,1=on
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_antialias( mode )
int    mode;
{
    hpGSESetAntialiasMode   gse_aa;
    /*
     * Check to make sure antialias is supported
     */
    if( ! psupport.allow_antialias ) return 1;
    if( mode )
	gse_aa.antialias_mode = HP_ANTIALIAS_MODE_BEST;
    else
	gse_aa.antialias_mode = HP_ANTIALIAS_MODE_OFF;
    PEXGSE( dpy, pstate.cur_id, pstate.req_type, HP_GSE_SET_ANTIALIAS_MODE,
	sizeof(hpGSESetAntialiasMode), (char *) &gse_aa);
    pstate.antialias = mode;
    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+       PEXtk_dither       +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Object module for dithering. Sets the current dithering
 *	mode and value.
 *
 *  USAGE:
 *	PEXtk_dither( mode, pattern )
 *	int	    mode	(in)	Mode for dithering,0=off,1=on
 *	int	    pattern	(in)	Dithering pattern to use
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_dither( mode, pattern )
int    mode;
int    pattern;
{
    pstate.dither = mode;
    pstate.dither_pattern = pattern;
    /* Need to change the color approx entry to turn on dithering */
    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+    PEXtk_transparency    +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Object module for transparency. Sets the current transparency
 *	mode and value.
 *
 *  USAGE:
 *	PEXtk_transparency( mode, value )
 *	int	    mode	(in)	Mode for transparency,0=off,1=on
 *	Fdata	    value	(in)	value for transparency
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_transparency( mode, value )
int    mode;
Fdata  value;
{

    pstate.transparency = mode;
    pstate.attr->poly_mtl.transmission = value;
    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+   PEXtk_reset_pex_state  +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Resets the PEX state.
 *
 *  USAGE:
 *	PEXtk_reset_pex_state( i, mask )
 *	int	    i		(in)	Flag set to 1 will End/Begin Rendering
 *	long	    mask	(in)	bitmask specifying which Renderer
 *					attributes are to be reset.
 *					Supported masks are:
 *						PEXRAViewTable
 *						PEXRALightTable
 *						PEXRAHLHSRMode
 *						PEXRANPCSubVolume
 *						PEXRAViewport
 *
 *  NOTE:
 *	Since this routine may be called when in structure mode, the
 *	attribute state is only reset in Immediate mode.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_reset_pex_state(i,mask)
int	i;
long	mask;
{
	XFlush( dpy);

	/*
	 * If not a full reset, then check if the renderer item is
	 * dynamic. If it is, no need to reset.
	 */
#define CHECK_RESET
#ifdef CHECK_RESET
	if( !i && !( pstate.win[pstate.cur_w].dyn_state & mask )) return 0;
#endif

#ifdef MPEX
	(*p_object.mpex_reset_state)(i,mask);
#endif

	if( mask ){
	    pstate.rdrAttrib.view_table  = 
		pstate.win[pstate.cur_w].View_ID;
	    pstate.rdrAttrib.light_table = 
		pstate.win[pstate.cur_w].Light_ID;
	    pstate.rdrAttrib.text_font_table = 
		pstate.win[pstate.cur_w].Text_ID;
	    pstate.rdrAttrib.depth_cue_table = 
		pstate.win[pstate.cur_w].Depthcue_ID;
	    pstate.rdrAttrib.hlhsr_mode = 
	    (pstate.zbuffer == PEXtk_ZBUFFER_ON) ? PEXHLHSRZBuffer:PEXHLHSROff;
	/*
	 * For some reason, when PEXtk/PEXlib is compiled on the SUN 
	 * and displaying on a DEC, having the Zbuffer turned on causes
	 * a problem. So if the mask has ZbufferMode in it, turn it off.
	 */
#ifdef SUN_DEC_BUG
	   { char *ptr;
	    ptr = ServerVendor(dpy);
	     
	    if( *ptr == 'D' && *(ptr+1) == 'E' ) /* Check for DEC */
		mask = mask & ~PEXRAHLHSRMode;
	   }
#endif

	    PEXChangeRenderer( dpy, pstate.win[pstate.cur_w].Rdr_ID,
		mask, &pstate.rdrAttrib);
	}

	if( i ){
	    PEXtk_End_Rendering();
	    PEXtk_Begin_Rendering();
	}

	XFlush(dpy);

	return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+      PEXtk_check_pc      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Checks the pipeline context mask and updates the pipeline context
 *	if necessary. This routine is called between a End/Begin rendering.
 *
 *  USAGE:
 *	PEXtk_check_pc( )
 *
 *  NOTE:
 *	This routine is only called when the renderer is Idle.
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_check_pc()
{

    if( pstate.dirty_mask ) DB_check_all_dirty_mask();
    if( pstate.pc_mask[0] == 0 && pstate.pc_mask[1] == 0 ) return 0;
    PEXChangePipelineContext( dpy, pstate.pc_ID, pstate.pc_mask, 
	&pstate.pcAttrib);
    pstate.pc_mask[0] = pstate.pc_mask[1] = pstate.pc_mask[2] = 0;
    return 0;
}

/*
 *                       +--------------------------+
 * ----------------------+      PEXtk_add_light     +------------------------
 *                       +--------------------------+
 *  FUNCTION:
 *      Add a light id to the Light Table
 *
 *  USAGE:
 *      PEXtk_add_light(num)
 *      int	    num		(in)	The id number of light to add. The
 *					base number is 1.
 *
 *  RETURNS:
 *      Returns number of lights in buffer
 */

/*ARGSUSED*/
int
PEXtk_add_light(num)
int num;
{
    int i;

    for( i = 0 ; i <= pstate.num_lgts; i++)
    {
	if( LightOnTable[i] == num )
	{
	    return (pstate.active_lgts);
	} else
	if( LightOnTable[i] == 0 )
	{
	    LightOnTable[i] = num;
	    return (i+1);
	}
    }
    return (pstate.active_lgts);
}

/*
 *                       +--------------------------+
 * ----------------------+      PEXtk_del_light     +------------------------
 *                       +--------------------------+
 *  FUNCTION:
 *      Delete a light id from the Light Table
 *
 *  USAGE:
 *      PEXtk_del_light(num)
 *      int	    num		(in)	id number of light to del
 *
 *  DESCRIPTION:
 *	To delete light number num, the LightOnTable array is searched
 *	until num is found. When it is, then the last item in the array
 *	is put in its slot, and the last item is zeroed out. If the light
 *	is not found, then return.
 *	If the light is the last item, it is just zeroed out.
 *	Note that PEXtk_MAX_LIGHTS does NOT include the ambient light.
 *
 *  RETURNS:
 *      Returns number of lights in buffer
 */

/*ARGSUSED*/
int
PEXtk_del_light(num)
int num;
{
    int i,j;

    for( i = 0 ; i <= PEXtk_MAX_LIGHTS; i++)
    {
        if( LightOnTable[i] == num )
        {
    	/* The light is found... look for last item to swap with
    	 */
    	for( j = pstate.num_lgts+1; j > i; j--)
    	    if(LightOnTable[j] != 0 )
    		{
    		   LightOnTable[i] = LightOnTable[j];
    		   LightOnTable[j] = 0;
    		   return (j);
    		}
    	/* This light is the last item in the list.
    	 */
    	LightOnTable[i] = 0;
    	return (i);
        }
    }
    return (pstate.active_lgts);
}

/*
 * NOT USED
 */
PEXtk_get_rdr_for_window( w, rdr )
Window		w;
PEXRenderer	*rdr;
{
    int i, j;
    unsigned int utemp, win_width, win_height, win_depth;
    Window	root;

    /* Find render ID */
    for(i=0;i<pstate.num_rdr;i++)
	if( w == pstate.win[i].w_id )
	  break;

    if( i >= pstate.num_rdr )
    {
	if( pstate.num_rdr < PEXtk_MAX_BUFFERS )
	{
	    /* Create a Renderer for this ID */
	} else {
	    /* Remove one of the Renderers */
	    /* Create a Renderer for this ID */
	}
    } else {
	/* Set current renderer */
    }
    XGetGeometry( dpy, pstate.win[i].w_id, &root,
		&j, &j, &win_width, &win_height, &utemp, &win_depth);
	    pstate.win[pstate.cur_w].width  = win_width;
	    pstate.win[pstate.cur_w].height = win_height;
	    pstate.win[pstate.cur_w].depth  = win_depth;

}

/*
 *                       +--------------------------+
 * ----------------------+   PEXtk_find_window_id   +------------------------
 *                       +--------------------------+
 *  FUNCTION:
 *      Searches the window id list and returns an index to that id. If
 *	id is not in list, a new slot is created for it, and all associated
 *	id's are created for it.
 *
 *  USAGE:
 *      PEXtk_find_window_id( flag, buf )
 *      int	    flag	(in)	Flag set to False (0) to search front 
 *					buffer only, set to True (1) to search 
 *					both buffers.
 *      Window	    buf		(in)	Window id 
 *
 *  DESCRIPTION:
 *	If the id is in the pstate.win.w_id list, then return the index to it.
 *	If the id is not in the list (i.e., search pstate.num_w_id times and
 *	it is not found), then create a Renderer ID, ViewLUT ID, TextLUT ID, 
 *	LightLUT ID, and an empty LightLUT ID (Light_ID0) if LIGHT_BUG is 
 *	defined.
 *
 *  NOTE:
 *	It is up to the calling routine to set the window id in the
 *	'win' array to 'buf'. That is, to set 'win[i].w_id = buf'.
 *      Only one renderer is created, since only when the num_w_id value
 *	is zero is the create_all_id routine called.
 *
 *  RETURNS:
 *      Index to location in list for buf
 */

/*ARGSUSED*/
int
PEXtk_find_window_id( flag, buf )
int     flag;
Window	buf;
{
    int i, j;

	for(i=0;i<pstate.num_w_id;i++){
	    if( buf == pstate.win[i].w_id )
		return i;
	    if( flag && buf == pstate.win[i].back_w_id )
		return i;
	    if( buf < pstate.win[i].w_id ){ 
		for(j=pstate.num_w_id;j>i;j-- ){ /* move everyone down */
		    memcpy( (char *) &pstate.win[j], (char *)&pstate.win[j-1], 
			sizeof(PEXtk_WINDOW));
		}
		/* pstate.win[i].w_id = buf; */
		if( pstate.num_w_id == 0 ){
		    if( PEXtk_create_all_id( i, buf ))
			return -1;
		} else {
		    if( PEXtk_copy_all_id( 0, i ))
			return -1;
		}
		pstate.num_w_id++;
		return i;
	    }
	}

	/*
	 * 'buf' id is not in list, so add to list, and create
	 *  associated id's
	 */

	/* pstate.win[pstate.num_w_id].w_id = buf; */
	if( pstate.num_w_id == 0 ){
	    if( PEXtk_create_all_id( pstate.num_w_id, buf ))
		return -1;
	} else {
	    if( PEXtk_copy_all_id( 0, pstate.num_w_id ))
		return -1;
	}
	pstate.num_w_id++;
	return (pstate.num_w_id-1); /* return the 'index', (i.e., n-1) */
}

/*
 *                       +--------------------------+
 * ----------------------+    PEXtk_create_all_id   +------------------------
 *                       +--------------------------+
 *  PURPOSE:
 *      Creates the id's necessary for rendering.
 *
 *  USAGE:
 *      PEXtk_create_all_id( loc, buf )
 *      int	    loc		(in)	Location in state list to store id's
 *      Window	    buf		(in)	Window id to allocate LUT's, Rdr on
 *
 *  DESCRIPTION:
 *	This routine will create the Renderer ID, a Light ID,
 *	a view ID, and will allocate a back buffer if in double 
 *	buffer mode. The view ID and light ID will be bound to
 *	the renderer.
 *
 *  CAVEATS:
 *	This routine requires that 'buf' be a valid Window id
 *	to bind the ID's to.
 *
 *  RETURNS:
 *      0 if successful, non-zero if error
 */

PEXtk_create_all_id( i, buf )
int	i;
Window  buf;
{
	unsigned long rmask;
	/*
	 * If zbuffer is supported, use it as the default
	 */
	if( psupport.allow_zbuffer ){
	    pstate.zbuffer = PEXtk_ZBUFFER_ON;
	    pstate.rdrAttrib.hlhsr_mode = PEXHLHSRZBuffer;
	} else {
	    pstate.zbuffer = PEXtk_ZBUFFER_OFF;
	    pstate.rdrAttrib.hlhsr_mode = PEXHLHSROff;
	}
	/*
	 *  Create the Renderer, PEXLUTView Table, and PEXLUTLight Table to
	 *	be associated with this drawable.
	 */
	pstate.win[i].View_ID = PEXCreateLookupTable( dpy, 
		buf, PEXLUTView);
	if( pstate.win[i].View_ID == 0 )
		return 1;
	pstate.rdrAttrib.view_table = pstate.win[i].View_ID;

	pstate.win[i].Light_ID = PEXCreateLookupTable( dpy, 
		buf, PEXLUTLight);
	if( pstate.win[i].Light_ID == 0 )
		return 1;
	pstate.rdrAttrib.light_table = pstate.win[i].Light_ID;

	pstate.win[i].Text_ID = PEXCreateLookupTable( dpy, 
		buf, PEXLUTTextFont);
	if( pstate.win[i].Text_ID == 0 )
		return 1;
	pstate.rdrAttrib.text_font_table = pstate.win[i].Text_ID;

#ifdef LIGHT_BUG
	pstate.win[i].Light_ID0 = PEXCreateLookupTable( dpy, 
		buf, PEXLUTLight);
	if( pstate.win[i].Light_ID0 == 0 )
		return 1;
#endif

	rmask = PEXRAViewTable|PEXRALightTable|PEXRAPipelineContext|
				PEXRATextFontTable|
				PEXRAPickIncl|PEXRAHLHSRMode;

	/*
	 * For some reason, when PEXtk/PEXlib is compiled on the SUN 
	 * and displaying on a DEC, having the Zbuffer turned on causes
	 * a problem.
	 */
#ifdef SUN_DEC_BUG
	{ char *ptr;
	  ptr = ServerVendor(dpy);
	     
	    if( *ptr == 'D' && *(ptr+1) == 'E' ) /* Check for DEC */
		rmask = PEXRAViewTable|PEXRALightTable|PEXRAPipelineContext|
			PEXRAPickIncl;
	}
#endif

	pstate.rdrAttrib.pick_incl = pstate.nameSet;
	pstate.win[i].Rdr_ID = PEXCreateRenderer( dpy, 
		buf, rmask, 
		&pstate.rdrAttrib);
	if( pstate.win[i].Rdr_ID == 0 ){
		fprintf(stderr, "Error... Could not create Renderer\n");
		return 1;
	} 
	pstate.cur_rdr  = pstate.win[i].Rdr_ID;
	pstate.cur_id   = pstate.cur_rdr;
	/*
	 * Create items associated with the renderer, namely marker glyphs
	 */
#ifdef MPEX
	/* Setup the box marker type */
	MPEX_Set_Marker_Glyph( dpy, pstate.win[i].Rdr_ID, 
		MPEXMarkerGlyph1, mrkr_glyph1, 8, 8);
#endif

	/*
	 * Get renderer dynamics infomation
	 */
	PEXtk_Get_Renderer_Dynamics( i, pstate.win[i].Rdr_ID );

	return 0;
}

/*
 *                       +--------------------------+
 * ----------------------+     PEXtk_copy_all_id    +------------------------
 *                       +--------------------------+
 *  PURPOSE:
 *      Copies the id's necessary for rendering.
 *
 *  USAGE:
 *      PEXtk_copy_all_id( src, dst )
 *      int	    src		(in)	Location in state list to get id's
 *      int	    dst		(in)	Location in state list to store id's
 *
 *  DESCRIPTION:
 *	This routine will copy the Renderer ID, two Light ID's,
 *	a view ID, and will allocate a back buffer if in double 
 *	buffer mode.
 *
 *  CAVEATS:
 *	This routine assumes that pstate.win[loc].w_id has been
 *	set to the buffer with which the id's are bound to, and
 *	that the win[loc].width and win[loc].height have been set.
 *
 *  RETURNS:
 *      0 if successful, non-zero if error
 */

int
PEXtk_copy_all_id( src, i )
int	src;
int	i;
{
	/*
	 *  Create the Renderer, PEXLUTView Table, and PEXLUTLight Table to
	 *	be associated with this drawable.
	 */
	pstate.win[i].View_ID = pstate.win[src].View_ID;
	if( pstate.win[i].View_ID == 0 )
		return 1;

	pstate.win[i].Light_ID = pstate.win[src].Light_ID ;
	if( pstate.win[i].Light_ID == 0 )
		return 1;

#ifdef LIGHT_BUG
	pstate.win[i].Light_ID0 = pstate.win[src].Light_ID0;
	if( pstate.win[i].Light_ID0 == 0 )
		return 1;
#endif

	pstate.win[i].Rdr_ID = pstate.win[src].Rdr_ID;
	if( pstate.win[i].Rdr_ID == 0 )
		return 1;

	return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+  PEXtk_Check_Back_Buffer +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	This routine will check for the existence of the back buffer when
 *	in double buffer mode.
 *
 *  USAGE:
 *	PEXtk_Check_Back_Buffer( i )
 *	int	    i		(in)	index into 'win' array on where
 *					to check for back buffer.
 *
 *  DESCRIPTION:
 *	This module checks the current buffer mode. If it is in double
 *	buffer mode, then the back buffer is checked to see if it exists.
 *	If it does not, then the 'create_buffer' routine is called.
 *	The current width and height of the Window at index 'i' must
 *	have been set already.
 *
 *  RETURNS:
 *	0 if successful, non-zero if an error occurred
 */
 
/*ARGSUSED*/
int
PEXtk_Check_Back_Buffer( i )	/* Check for existance of back buffer */
int	i;
{
    int save_w;
        if(( pstate.buffer_mode == PEXtk_DOUBLE_BUFFER ) &&
           ( pstate.win[i].back_w_id == 0 ) &&
	   ( pstate.win[i].w_id != 0 ) )
	{
	    save_w = pstate.cur_w;	/* save the current window index */
	    pstate.cur_w = i;
	    (*p_object.create_buffer)( pstate.win[i].width,
					  pstate.win[i].height);
	    pstate.cur_w = save_w;	/* restore the current window index */
	}
	return 0;
}

int
PEXtk_Get_X_Res()	/* Get x resolution of current drawable */
{
    int j;
    Window	root;
    unsigned int utemp, win_width, win_height, win_depth;
    XGetGeometry( dpy, pstate.win[pstate.cur_w].w_id, &root,
		&j, &j, &win_width, &win_height, &utemp, &win_depth);
	return win_width;
}

int
PEXtk_Get_Y_Res()	/* Get y resolution of current drawable */
{
    int j;
    Window	root;
    unsigned int utemp, win_width, win_height, win_depth;
    XGetGeometry( dpy, pstate.win[pstate.cur_w].w_id, &root,
		&j, &j, &win_width, &win_height, &utemp, &win_depth);
	return win_height;
}

int
find_lsbit(n)
register unsigned long n;
{
    register int count;

    if (n == 0) return 0;
    for( count = 0; !(n & 1); n = n >> 1)
	count++;
    return count;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+       PEXtk_default      +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Default routine to simply return for unsupported items. 
 *
 *  USAGE:
 *	PEXtk_default( )
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_default( )
{
    return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+    PEXtk_Get_Visuals     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Find any 24-bit X Visuals that are supported.
 *
 *  USAGE:
 *	PEXtk_Get_Visuals( visual )
 *	Visual	    *visual	(in)	pointer to visual
 *
 *  NOTE:
 *	This routine is for PEX 5.0 server
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_Get_Visuals( win_visual)
Visual **win_visual;
{
    int i, j, c_class;
    XVisualInfo vTemplate, *visualList;

    win_depth    = 0;
    px_data.red_shift = px_data.green_shift = px_data.blue_shift = 0;
    /*
     *	See if any 24-bit visuals are available
     *  If the default is 24-bit, then get the rgb masks for true color
     *	if it exists.
     */
	vTemplate.screen = DefaultScreen( dpy);
	vTemplate.depth  = 24;
	visualList = XGetVisualInfo( dpy, VisualScreenMask |
			VisualDepthMask, &vTemplate, &i);
	for(j=0;j<i;j++){
#if defined(__cplusplus) || defined(c_plusplus)
	    if( (c_class = visualList[j].c_class) == TrueColor )
#else
	    if( (c_class = visualList[j].class  ) == TrueColor )
#endif
	    {
		*win_visual = visualList[j].visual;
		win_depth = 24;
		px_data.c_class  = c_class;
		/*
		 * Calculate the shift amount for each rgb mask
		 */
		px_data.red_shift   = find_lsbit(visualList[j].red_mask);
		px_data.green_shift = find_lsbit(visualList[j].green_mask);
		px_data.blue_shift  = find_lsbit(visualList[j].blue_mask);
	    }
	}

	if( win_depth == 0 ){
	    /*
	     * Check if there were 24-bit visuals, but no TrueColor visual
	     */
	    if( i != 0 ){
		*win_visual = visualList[0].visual;/* use first one returned */
		win_depth    = 24;
		px_data.c_class  = c_class;
	        if( (c_class == TrueColor) || (c_class == DirectColor )){
		    px_data.red_shift   = find_lsbit(visualList[0].red_mask);
		    px_data.green_shift = find_lsbit(visualList[0].green_mask);
		    px_data.blue_shift  = find_lsbit(visualList[0].blue_mask);
		}
	    } else {
		win_depth = DefaultDepth(dpy, win_screen);
		px_data.c_class  = c_class;
	    }
	}
    return 0;
}
	    
/*
 *
 *                       +--------------------------+
 * ----------------------+    PEXtk_find_target     +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Finds the best "PEX" visual to use. This routine will first
 *	look for a 24 bit BufferDrawable, and then a 24 bit AnyDrawable,
 *	and then any drawable that supports PEX, with TrueColor and
 *	larger depth having a higher priority.
 *
 *  USAGE:
 *	PEXtk_find_target( visual )
 *	Visual	    **visual	(out)	visual found matching needs
 *
 *  RETURNS:
 *	0 if no visual found, non-zero value is depth of the visual
 */
 
/*ARGSUSED*/
int
PEXtk_find_target( visual)
Visual	**visual;
{
#define N_MRT 9
    int i, j, f_err, depth, c_class = -1;
    unsigned long max_targets=100, count;
    PEXRenderingTarget use_target, *targets, *sav_target;
    static PEXTK_MRT_OBJECT bufs[N_MRT] = {{ 24, PEXBufferDrawable, 1}, /* MBX */
    				           { 24, PEXWindowDrawable, 2}, /* E&S */
    				           { 12, PEXBufferDrawable, 1},
    				           { 12, PEXWindowDrawable, 2},
    				           { 24, PEXPixmapDrawable, 3}, /* Pixmap */
    				           { 12, PEXPixmapDrawable, 3},
    				           {  8, PEXBufferDrawable, 1},
    				           {  8, PEXWindowDrawable, 2},
    				           {  8, PEXPixmapDrawable, 3}};

    px_data.red_shift = px_data.green_shift = px_data.blue_shift = 0;
    depth = 0;

    for( j=0; j < N_MRT ; j++ ) {
      f_err = PEXMatchRenderingTargets( dpy, 
	        RootWindow(dpy,DefaultScreen(dpy)), 
	        bufs[j].depth, bufs[j].type,
	        NULL, max_targets, &count, &targets);
      if (f_err && count != 0){
	sav_target = targets;
	for (i=0; i<count; i++,targets++) {
#if defined(__cplusplus) || defined(c_plusplus)
	    c_class = targets->visual->c_class;
#else
	    c_class = targets->visual->class;
#endif
	    switch( c_class ) {
		case TrueColor:
			if( PEXtk_check_visual_db( targets->depth, targets->visual)){
			    *visual = targets->visual;
			    PEXtk_get_visual_mask( targets->visual);
			    depth = targets->depth;
			    px_data.c_class  = c_class;
			    XFree(sav_target);
			    return depth; /* this is what we want */
			}
		    break;
		case DirectColor:
			if( depth < targets->depth && 
			    PEXtk_check_visual_db( targets->depth, targets->visual)){
			    depth = targets->depth;
			    px_data.c_class  = c_class;
			    *visual = targets->visual;
			}
		    break;
		default:
			if( depth == 0 && 
			    PEXtk_check_visual_db( targets->depth, targets->visual)){
			    depth = targets->depth;
			    px_data.c_class  = c_class;
			    *visual = targets->visual;
			}
		    break;
	    }
	}
	XFree(sav_target);
      }
    }
    /* PEXtk_get_visual_mask( *visual); */
    return depth;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+  PEXtk_check_visual_db   +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Check the visual to make sure it can be double buffered.
 *	This is only needed on the HP.
 *
 *  USAGE:
 *	PEXtk_check_visual_db( )
 *
 *  RETURNS:
 *	0 if visual is not double bufferable,
 *	1 if MBX double bufferable
 *	2 if ES Escape double bufferable
 */
 
/*ARGSUSED*/
int
PEXtk_check_visual_db( depth, visual )
int    depth;
Visual *visual;
{
    PEXEnumTypeDesc *enumInfo,*ptr;
    int 	i, j, ret, win_create, enum_types;
    unsigned long  mask, *count;
    XSetWindowAttributes win_attrs;
    Colormap  cmap;
    Window    win;
    Multibuffer mbuf[2];

    /*
     * Get DBL Buffer Escape information
     *     Do not check this for 5.0 servers.
     */
    ret = 0;
#ifdef OLD
    if( visual == DefaultVisual( dpy, win_screen)){
	win = RootWindow( dpy, win_screen);
	win_create = 0;
    } else {
#endif
    {
	cmap = XCreateColormap( dpy, RootWindow( dpy, win_screen),
				visual, AllocNone);
	if( cmap == 0 )
	    return 0;
	win_attrs.override_redirect = True;
	win_attrs.background_pixel  = 0;
	win_attrs.border_pixel      = 0;
	win_attrs.colormap	    = cmap;
	mask =  CWOverrideRedirect | CWColormap | CWBackPixel |
		CWBorderPixel;
	win = XCreateWindow( dpy, RootWindow( dpy, win_screen), 
		10, 10, 1, 1, 0, depth, InputOutput,
		visual, mask, &win_attrs);
	if( win == 0 ){
	    XFreeColormap( dpy, cmap);
	    return 0;
	}
	win_create = 1;
    }
    if(XmbufCreateBuffers( dpy, win, 2,
        MultibufferUpdateActionUntouched, MultibufferUpdateHintFrequent,
        mbuf) != 2)
    {
	ret = 1;
    }

    if( ret == 0 ){
	enum_types = PEXETEscape;
	PEXGetEnumTypeInfo( dpy, win, 1,
	    &enum_types, PEXETIndex, &count, &enumInfo);

	/*
	 * For each entry, see if the ES DBL Buffer is supported
	 */
	for(i=0,ptr=enumInfo; i<count[0] ;i++,ptr++) {
	    if( (short) ptr->index == (short) ES_ESCAPE_ET_DBLBUFFER )
		ret = 1;
	}
	PEXFreeEnumInfo( 1, count, enumInfo);
    }

    if( win_create ){
	XFreeColormap( dpy, cmap);
	XDestroyWindow( dpy, win);
    }
    return ret;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+  PEXtk_Get_Best_Visual   +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Get best PEX supported visual. 
 *
 *  USAGE:
 *	PEXtk_Get_Best_Visual( )
 *
 *  RETURNS:
 *	0 if no visual available, non-zero is depth of the visual found
 */
 
/*ARGSUSED*/
int
PEXtk_Get_Best_Visual( disp, visual)
Display *disp;
Visual  **visual;
{
    if( px_data.win_visual == NULL )
	return 0;
    *visual = px_data.win_visual;
    return win_depth;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+  PEXtk_get_visual_mask   +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Get best PEX supported visual. 
 *
 *  USAGE:
 *	PEXtk_get_visual_mask( )
 *	Visual	    *visual	(out)	visual found matching needs
 *
 *  CAVEATS:
 *	This routine should only be called with a true or direct color visual.
 *
 *  RETURNS:
 *	0 if visual direct or true, non-zero otherwise 
 */
 
/*ARGSUSED*/
int
PEXtk_get_visual_mask( visual )
Visual *visual;
{
    int i, c_class;
    XVisualInfo vTemplate, *visualList;

	vTemplate.visualid = XVisualIDFromVisual( visual);
	visualList = XGetVisualInfo( dpy, VisualIDMask, &vTemplate, &i);

#if defined(__cplusplus) || defined(c_plusplus)
	    c_class = visual->c_class;
#else
	    c_class = visual->class;
#endif
        if( (c_class == TrueColor) || (c_class == DirectColor )){
	    px_data.red_shift   = find_lsbit(visualList[0].red_mask);
	    px_data.green_shift = find_lsbit(visualList[0].green_mask);
	    px_data.blue_shift  = find_lsbit(visualList[0].blue_mask);

	} else return 1;
	return 0;
}

/*
 *
 *                       +--------------------------+
 * ----------------------+  PEXtk_Get_Support_Info  +------------------------
 *                       +--------------------------+
 *
 *  PURPOSE:
 *	Get PEX Server supported information. This information
 *	describes what services the server provides.
 *
 *  USAGE:
 *	PEXtk_Get_Support_Info( vendor_str )
 *	char	    *vendor_str	(in)	vendor string
 *
 *  CAVEATS:
 *	The data is inquired on the RootWindow.
 *	The argument 'vendor_str' is a temporary fix to get SunPEX double
 *	buffering to work.
 *
 *  DESCRIPTION:
 *	The routine sets up the arrays that describe Enumerated Type Info
 *	that the server supports for:
 *		HLHSRMode, LineType, EdgeType, LightType,
 *		CurveApproxMethod, SurfaceApproxMethod,
 *		Escapes, GSE.
 *	For each supported item Type 'm', store True at location 'm-1' in
 *	the array describing the support. So for Hlhsr mode PEXHLHSRModeZBuffer
 *	(currently the value 2), the value one is stored in psupport.hlhsr[1].
 *
 *	As for the Escapes and GSE's, this routine will specifically check for 
 *	the E&S/HP Double Buffer escapes, and for the HP Anti-Alias GSE.
 *
 *	This routine also determines the Implementation Dependant constants for:
 *
 *		 PEXIDDitheringSupported
 *		 PEXIDTransparencySupported
 *		 PEXIDMaxNonAmbientLights
 *		 PEXIDMaxLineWidth
 *		 PEXIDMaxMarkerSize
 *		 PEXIDMaxLineWidth
 *		 PEXIDMaxNameSetNames
 *
 *  RETURNS:
 *	Nothing
 */
 
/*ARGSUSED*/
int
PEXtk_Get_Support_Info( vendor_str )
char	*vendor_str;
{
    PEXEnumTypeDesc 	*enumInfo,*ptr;
    int			i, j, enum_types;
    unsigned long	*count;
    unsigned short	namesID[10];
    PEXImpDepConstant   *valuesReturn;
    PEXTableInfo	tbl_info;

    /* 
     * Get HLHSR/Z_Buffer information
     */
	enum_types = PEXETHLHSRMode;
	psupport.allow_antialias = False;
	for(i=0;i<5;i++) psupport.hlhsr[i] = 0;

	PEXGetEnumTypeInfo( dpy, RootWindow(dpy,win_screen), 1, &enum_types, 
		PEXETIndex, &count, &enumInfo);

	for(i=0,ptr=enumInfo;i<count[0] && i < 5;i++,ptr++)
	{
            /* printf(" HLHSR Mode %d is supported\n",ptr->index); */
	    if( ptr->index > 0 ){	/* some are < 0 */
		j = (ptr->index > 5) ? 0 : (ptr->index - 1);
		psupport.hlhsr[j] = 1;
		if( ptr->index == PEXHLHSRZBuffer )
		    psupport.allow_zbuffer = True;
	    }
	}

	PEXFreeEnumInfo( 1, count, enumInfo);

    /*
     * Get line type information
     */
	enum_types = PEXETLineType;
	for(i=0;i<10;i++) psupport.line_type[i] = 0; 

	PEXGetEnumTypeInfo( dpy, RootWindow(dpy,win_screen), 1, &enum_types, 
		PEXETIndex, &count, &enumInfo);

	for(i=0,ptr=enumInfo;i<count[0] && i < 10;i++,ptr++)
	{
	    j = (ptr->index > 10) ? 0 : (ptr->index - 1);
	    if( j >= 0 )	/* make sure it's a positive offset */
		psupport.line_type[j] = 1;
	}

	PEXFreeEnumInfo( 1, count, enumInfo);

    /*
     * Get surface edge type information
     */
	enum_types = PEXETSurfaceEdgeType;
	for(i=0;i<10;i++) psupport.edge_type[i] = 0;

	PEXGetEnumTypeInfo( dpy, RootWindow(dpy,win_screen), 1, &enum_types, 
		PEXETIndex, &count, &enumInfo);

	for(i=0,ptr=enumInfo;i<count[0] && i < 10;i++,ptr++)
	{
	    j = (ptr->index > 10) ? 0 : (ptr->index - 1);
	    if( j >= 0 )	/* make sure it's a positive offset */
		psupport.edge_type[j] = 1;
	}

	PEXFreeEnumInfo( 1, count, enumInfo);

    /*
     * Get supported light type information
     */
	enum_types = PEXETLightType;
	for(i=0;i<8;i++) psupport.light_type[i] = 0;

	PEXGetEnumTypeInfo( dpy, RootWindow(dpy,win_screen), 1, &enum_types, 
		PEXETIndex, &count, &enumInfo);

	for(i=0,ptr=enumInfo;i<count[0] && i < 10;i++,ptr++)
	{
	    j = (ptr->index > 8) ? 0 : (ptr->index - 1);
	    if( j >= 0 )	/* make sure it's a positive offset */
		psupport.light_type[j] = 1;
	}

	PEXFreeEnumInfo( 1, count, enumInfo);

    /*
     * Get supported curve approx method information
     */
	enum_types = PEXETCurveApproxMethod;
	for(i=0;i<12;i++) psupport.curve_approx[i] = 0;

	PEXGetEnumTypeInfo( dpy, RootWindow(dpy,win_screen), 1, &enum_types, 
		PEXETIndex, &count, &enumInfo);

	for(i=0,ptr=enumInfo;i<count[0] && i < 10;i++,ptr++)
	{
	    j = (ptr->index > 12) ? 0 : (ptr->index - 1);
	    if( j >= 0 )	/* make sure it's a positive offset */
		psupport.curve_approx[j] = 1;
	}

	PEXFreeEnumInfo( 1, count, enumInfo);

    /*
     * Get supported surface approx method information
     */
	enum_types = PEXETSurfaceApproxMethod;
	for(i=0;i<12;i++) psupport.surf_approx[i] = 0;

	PEXGetEnumTypeInfo( dpy, RootWindow(dpy,win_screen), 1, &enum_types, 
		PEXETIndex, &count, &enumInfo);

	for(i=0,ptr=enumInfo;i<count[0] && i < 10;i++,ptr++)
	{
	    j = (ptr->index > 12) ? 0 : (ptr->index - 1);
	    if( j >= 0 )	/* make sure it's a positive offset */
		psupport.surf_approx[j] = 1;
	}

	PEXFreeEnumInfo( 1, count, enumInfo);

    /*
     * Get DBL Buffer Escape information, and GSE information (anti-alias)
     *     Do not check this for 5.0 servers, or for the SunPEX server.
     */
      if(!( (pstate.maj_ver == 5) && (pstate.min_ver == 0) &&
	     !strstr(vendor_str, "SMI") )){
	enum_types = PEXETEscape;
	PEXGetEnumTypeInfo( dpy, RootWindow(dpy,win_screen), 1,
        		&enum_types, PEXETIndex, &count, &enumInfo);

	/*
	 * For each entry, see if the ES DBL Buffer is supported
	 */
	for(i=0,ptr=enumInfo; i<count[0] ;i++,ptr++)
	{
	    if( (short) ptr->index == (short) ES_ESCAPE_ET_DBLBUFFER )
		psupport.dbl_buf |= PEXtk_DBUF_ES_ESCAPE;
	}

	PEXFreeEnumInfo( 1, count, enumInfo);

	enum_types = PEXETGSE;
	PEXGetEnumTypeInfo( dpy, RootWindow(dpy,win_screen), 1,
        		&enum_types, PEXETIndex, &count, &enumInfo);

	/*
	 * For each entry, see if the HP GSE Anti-alias is supported
	 */
	for(i=0,ptr=enumInfo; i<count[0] ;i++,ptr++)
	{
	    if( (short) ptr->index == (short) HP_GSE_ET_SET_ANTIALIAS_MODE )
		psupport.allow_antialias = True;
	}

	PEXFreeEnumInfo( 1, count, enumInfo);

	psupport.gdp_sphere = 0;
#ifdef KPC
	enum_types = PEXETGDP;
	PEXGetEnumTypeInfo( dpy, RootWindow(dpy,win_screen), 1,
        		&enum_types, PEXETAll, &count, &enumInfo);

	/*
	 * For each entry, see if the KPC GDP Sphere is supported
	 */
	for(i=0,ptr=enumInfo; i<count[0] ;i++,ptr++)
	{
	    if( strcmp( ptr->descriptor, "KPC_PolySphere") == 0)
		psupport.gdp_sphere = True;
	}
	PEXFreeEnumInfo( 1, count, enumInfo);
#endif

      }

    /*
     * Get Implementation Dependant information
     */

	namesID[0] = PEXIDDitheringSupported;
	namesID[1] = PEXIDTransparencySupported;
	namesID[2] = PEXIDMaxNonAmbientLights;
	namesID[3] = PEXIDMaxLineWidth;
	namesID[4] = PEXIDMaxMarkerSize;
	namesID[5] = PEXIDMaxLineWidth;
	namesID[6] = PEXIDMaxNameSetNames;

	if(PEXGetImpDepConstants( dpy, RootWindow(dpy, win_screen),
               		7, namesID, (PEXImpDepConstant **)&valuesReturn)){
	    psupport.allow_dither 	= valuesReturn[0].integer;
	    psupport.allow_transparency = valuesReturn[1].integer;
	    psupport.max_mrkr_sz	= valuesReturn[4].integer;
	    psupport.max_line_sz	= valuesReturn[5].integer;
	    psupport.max_names		= valuesReturn[6].integer;
	    psupport.max_lgts	= valuesReturn[2].integer < PEXtk_MAX_LIGHTS ? 
				  valuesReturn[2].integer : PEXtk_MAX_LIGHTS;
	    XFree( valuesReturn);
	} else {
	    psupport.allow_dither = psupport.allow_transparency = False;
	    psupport.max_mrkr_sz  = psupport.max_line_sz = 1;
	    psupport.max_lgts     = 6; 
	    psupport.max_names    = 16;
	}

	if(PEXGetTableInfo( dpy, RootWindow(dpy,win_screen), PEXLUTView,
		&tbl_info))
	    psupport.max_views	= tbl_info.definable_entries;
	else
	    psupport.max_views	= 2;

    return 0;
}

/*
 *  ROUTINE:	PEXtk_Get_Renderer_Dynamics
 *  PURPOSE:	Get Renderer Dynamics information. This information describes
 *		which attributes may be modified at any time and have that
 *		change take place immediately. A bit value of 0 indicates a
 *		dynamically modifiable attribute. A bit value of 1 indicates
 *		an attribute that may not be modified dynamically, so the
 *		change is "pending" and will take place at the next
 *		"begin rendering" sequence.
 */
int
PEXtk_Get_Renderer_Dynamics( i, rdr)
int	    i;
PEXRenderer rdr;
{
    unsigned long tRet, nRet, aRet;

    pstate.win[i].dyn_state =   PEXRAViewTable | PEXRAColorTable | 
				PEXRADepthCueTable | PEXRALightTable | 
				PEXRAColorApproxTable | PEXRAHLHSRMode |
			        PEXRANPCSubVolume | PEXRAViewport | 
				PEXRAClipList;

    if( PEXGetRendererDynamics( dpy, rdr, &tRet, &nRet, &aRet)){
	if( pstate.debug_mode )
	    fprintf(stderr, "ERROR... GetRendererDynamics\n");
	return 1;
    }

    if( tRet != 0 )
    {
	/* Ignore the Bundle entries */
	if( tRet & PEXRDTViewTable )
		pstate.win[i].dyn_state &= ~PEXRAViewTable;
	if( tRet & PEXRDTColorTable )
		pstate.win[i].dyn_state &= ~PEXRAColorTable;
	if( tRet & PEXRDTDepthCueTable )
		pstate.win[i].dyn_state &= ~PEXRADepthCueTable;
	if( tRet & PEXRDTLightTable )
		pstate.win[i].dyn_state &= ~PEXRALightTable;
	if( tRet & PEXRDTColorApproxTable )
			pstate.win[i].dyn_state &= ~PEXRAColorApproxTable;
    }

    if( aRet != 0 )
    {
	if( aRet & PEXRDAHLHSRMode )
		pstate.win[i].dyn_state &= ~PEXRAHLHSRMode;
	if( aRet & PEXRDANPCSubVolume )
		pstate.win[i].dyn_state &= ~PEXRANPCSubVolume;
	if( aRet & PEXRDAViewport )
		pstate.win[i].dyn_state &= ~PEXRAViewport;
	if( aRet & PEXRDAClipList )
		pstate.win[i].dyn_state &= ~PEXRAClipList;
    }
    return 0;
}

/*
 *  ROUTINE:	PEXtk_init_attributes
 *  PURPOSE:	Initializes the PEXtk attribute state.
 */
int
PEXtk_init_attributes()
{
    pstate.pc_mask[0] = pstate.pc_mask[1] = pstate.pc_mask[2] = 0;

    pstate.attr->interp_meth 		= PEXSurfaceInterpNone;
    pstate.pcAttrib.surface_interp	= PEXSurfaceInterpNone;
    pstate.attr->interp_meth_str= 0; /* depends on 0 being not defined */
    pstate.attr->i_style		= PEXInteriorStyleHollow;
    pstate.pcAttrib.interior_style	= PEXInteriorStyleHollow;
    pstate.attr->i_style_str	= 0; /* depends on 0 being not defined */
    pstate.attr->edgeFlag   		= PEXOff;
    pstate.pcAttrib.surface_edges	= PEXOff;
    pstate.attr->edgeFlag_str		= PEXOff;
    pstate.attr->border			= PEXtk_BORDER_NONE;
    pstate.attr->edgeType		= PEXSurfaceEdgeSolid;
    pstate.pcAttrib.surface_edge_type	= PEXSurfaceEdgeSolid;
    pstate.attr->backcull		= PEXtk_CULL_BACKFACE;
    pstate.pcAttrib.culling_mode 	= PEXBackFaces;
    pstate.attr->line_interp		= PEXPolylineInterpNone;
    pstate.pcAttrib.polyline_interp	= PEXPolylineInterpNone;
    pstate.attr->line_width		= 1.0;
    pstate.pcAttrib.line_width		= 1.0;
    pstate.attr->line_type		= PEXtk_LINE_SOLID;
    pstate.pcAttrib.line_type		= PEXLineTypeSolid;
    pstate.attr->mrkr_scale		= 1;
    pstate.pcAttrib.marker_scale	= 1.0;
    pstate.attr->mrkr_type		= PEXtk_MRKR_POINT;
    pstate.pcAttrib.marker_type		= PEXMarkerDot;

    pstate.pcAttrib.distinguish  	= False;

    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCCullingMode);
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCDistinguishFlag);
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCInteriorStyle);
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCLineType);
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCLineWidth);
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCMarkerScale);
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCMarkerType);
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCPolylineInterp);
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCSurfaceEdgeFlag);
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCSurfaceEdgeType);
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCSurfaceInterp);

    pstate.attr->refl_model  		= PEXReflectionNone;
    pstate.pcAttrib.reflection_model 	= PEXReflectionNone;

    pstate.attr->poly_mtl.ambient = SURFACE_AMBIENT;
    pstate.attr->poly_mtl.diffuse = SURFACE_DIFFUSE;
    pstate.attr->poly_mtl.specular = SURFACE_SPEC;
    pstate.attr->poly_mtl.specular_conc = 8.0;
    pstate.attr->poly_mtl.transmission = 0.0;

    pstate.attr->poly_mtl.specular_color.type  	= PEXColorTypeRGB;
    pstate.attr->poly_mtl.specular_color.value.rgb.red  = 0.0;
    pstate.attr->poly_mtl.specular_color.value.rgb.green= 0.0;
    pstate.attr->poly_mtl.specular_color.value.rgb.blue = 0.0;

    memcpy( &pstate.pcAttrib.reflection_attr, &pstate.attr->poly_mtl,
	sizeof(PEXReflectionAttributes));
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCReflectionAttr);
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCReflectionModel);

    light_table.attenuation1 = ATTENUATION1;
    light_table.attenuation2 = ATTENUATION2;
    pstate.pcAttrib.light_state.indices = LightOnTable;

    pstate.attr->colorBgInd		= 0;
    pstate.attr->colorBg.rgb.red	= 0.0;
    pstate.attr->colorBg.rgb.green      = 0.0;
    pstate.attr->colorBg.rgb.blue	= 0.0;

    pstate.attr->colorLineInd		= 0;
    pstate.attr->colorLine.rgb.red	= 1.0;
    pstate.attr->colorLine.rgb.green	= 1.0;
    pstate.attr->colorLine.rgb.blue	= 1.0;
    memcpy( &pstate.pcAttrib.surface_color.value, &pstate.attr->colorLine,
		sizeof(PEXColor));
    pstate.pcAttrib.line_color.type  = PEXColorTypeRGB;
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCLineColor);

    pstate.attr->colorPolyInd		= 0;
    pstate.attr->colorPoly.rgb.red  	= 1.0;
    pstate.attr->colorPoly.rgb.green	= 1.0;
    pstate.attr->colorPoly.rgb.blue 	= 1.0;
    memcpy( &pstate.pcAttrib.surface_color.value, &pstate.attr->colorPoly,
		sizeof(PEXColor));
    pstate.pcAttrib.surface_color.type  = PEXColorTypeRGB;
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCSurfaceColor);

    pstate.attr->colorMrkrInd		= 0;
    pstate.attr->colorMrkr.rgb.red  	= 1.0;
    pstate.attr->colorMrkr.rgb.green	= 1.0;
    pstate.attr->colorMrkr.rgb.blue 	= 1.0;
    memcpy( &pstate.pcAttrib.marker_color.value, &pstate.attr->colorMrkr,
		sizeof(PEXColor));
    pstate.pcAttrib.marker_color.type 	       = PEXColorTypeRGB;
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCMarkerColor);

    pstate.attr->colorTextInd		= 0;
    pstate.attr->colorText.rgb.red  	= 1.0;
    pstate.attr->colorText.rgb.green	= 1.0;
    pstate.attr->colorText.rgb.blue 	= 1.0;
    memcpy( &pstate.pcAttrib.text_color.value, &pstate.attr->colorText,
		sizeof(PEXColor));
    pstate.pcAttrib.text_color.type 	       = PEXColorTypeRGB;
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCTextColor);

    pstate.attr->colorLightInd		   = 0;
    pstate.attr->colorLight.type	   = PEXColorTypeRGB;
    pstate.attr->colorLight.value.rgb.red  = 1.0;
    pstate.attr->colorLight.value.rgb.green= 1.0;
    pstate.attr->colorLight.value.rgb.blue = 1.0;

    pstate.attr->text_vec.x 		= 0.0;
    pstate.attr->text_vec.y 		= 1.0;
    pstate.pcAttrib.char_up_vector.x 	= 0.0;
    pstate.pcAttrib.char_up_vector.y 	= 1.0;
    pstate.pcAttrib.atext_up_vector.x 	= 0.0;
    pstate.pcAttrib.atext_up_vector.y 	= 1.0;
    pstate.attr->font        		= 0;
    pstate.pcAttrib.text_font 	   	= 1;
    pstate.attr->path        		= PEXtk_TEXT_PATH_RIGHT;
    pstate.pcAttrib.text_path		= PEXPathRight;
    pstate.attr->text_xsize  		= 0.25;
    pstate.attr->text_ysize  		= 0.25;
    pstate.pcAttrib.char_height 	= 0.25;
    pstate.attr->texta_xsize 		= 0.015;
    pstate.attr->texta_ysize 		= 0.015;
    pstate.pcAttrib.atext_height 	= 0.015;

    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCCharUpVector);
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCCharHeight);
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCTextPath);
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCATextUpVector);
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCATextPath);
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCATextHeight);

    pstate.attr->u_prec		= 20;
    pstate.attr->v_prec		= 20;
    if( psupport.curve_approx[PEXApproxConstantBetweenKnots] ){
	pstate.attr->curv_meth	= PEXtk_APPROX_CONSTANT_KNOTS;
	pstate.pcAttrib.curve_approx.method	   = PEXApproxConstantBetweenKnots;
    } else {
	pstate.attr->curv_meth	= PEXtk_APPROX_IMP_DEP;
	pstate.pcAttrib.curve_approx.method	   = PEXApproxImpDep;
    }
    pstate.attr->curv_utol	= 0.1;
    pstate.attr->curv_umin	= 0.0;
    pstate.attr->curv_umax	= 1.0;
    pstate.pcAttrib.curve_approx.tolerance = pstate.attr->curv_utol;
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCCurveApprox);

    if( psupport.curve_approx[PEXApproxConstantBetweenKnots] ){
	pstate.attr->surf_meth   	= PEXtk_APPROX_CONSTANT_KNOTS;
	pstate.pcAttrib.surface_approx.method      = PEXApproxConstantBetweenKnots;
    } else {
	pstate.attr->surf_meth   	= PEXtk_APPROX_IMP_DEP;
	pstate.pcAttrib.surface_approx.method      = PEXApproxImpDep;
    }
    pstate.attr->surf_utol 	= 0.1;
    pstate.attr->surf_vtol 	= 0.1;
    pstate.pcAttrib.surface_approx.u_tolerance = 0.1;
    pstate.pcAttrib.surface_approx.v_tolerance = 0.1;
    PEXSetPCAttributeMask( pstate.pc_mask, PEXPCSurfaceApprox);

    pstate.attr->surf_umin = 0.0;
    pstate.attr->surf_umax = 1.0;
    pstate.attr->surf_vmin = 0.0;
    pstate.attr->surf_vmax = 1.0;

    Depthcue_table.mode = PEXOff;  

    pstate.rdrAttrib.npc_subvolume.min.x = 0.0;
    pstate.rdrAttrib.npc_subvolume.max.x = 1.0;
    pstate.rdrAttrib.npc_subvolume.min.y = 0.0;
    pstate.rdrAttrib.npc_subvolume.max.y = 1.0;
    pstate.rdrAttrib.npc_subvolume.min.z = 0.0;
    pstate.rdrAttrib.npc_subvolume.max.z = 1.0;

    PEXChangePipelineContext( dpy, pstate.pc_ID, pstate.pc_mask, 
	&pstate.pcAttrib);
    pstate.pc_mask[0] = pstate.pc_mask[1] = pstate.pc_mask[2] = 0;

    return 0;
}
